# -*- coding: utf-8 -*-
"""
DO NOT EDIT THIS FILE!
It is automatically generated from opcfoundation.org schemas.
Model Uri:http://opcfoundation.org/UA/
Version:1.04.9
Publication date:2021-01-21T00:00:00Z

File creation Date:2021-04-14 12:39:17.011969
"""
import datetime
from dateutil.tz import tzutc
from asyncua import ua
from asyncua.ua import NodeId, QualifiedName, NumericNodeId, StringNodeId, GuidNodeId
from asyncua.ua import NodeClass, LocalizedText


def create_standard_address_space_Services(server):

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3062, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3063, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BaseDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(24, 0),
        BrowseName=QualifiedName("BaseDataType", 0),
        NodeClass_=NodeClass.DataType,
        NodeAttributes=attrs,
    )
    server.add_nodes([node])

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Number"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(26, 0),
        BrowseName=QualifiedName("Number", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(26, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Integer"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(27, 0),
        BrowseName=QualifiedName("Integer", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(26, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(27, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(26, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UInteger"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(28, 0),
        BrowseName=QualifiedName("UInteger", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(26, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(28, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(26, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Enumeration"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(29, 0),
        BrowseName=QualifiedName("Enumeration", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(29, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Boolean"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(1, 0),
        BrowseName=QualifiedName("Boolean", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(1, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SByte"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2, 0),
        BrowseName=QualifiedName("SByte", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(27, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(27, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Byte"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3, 0),
        BrowseName=QualifiedName("Byte", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(28, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(28, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Int16"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(4, 0),
        BrowseName=QualifiedName("Int16", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(27, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(4, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(27, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UInt16"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(5, 0),
        BrowseName=QualifiedName("UInt16", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(28, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(5, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(28, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Int32"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(6, 0),
        BrowseName=QualifiedName("Int32", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(27, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(6, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(27, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UInt32"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7, 0),
        BrowseName=QualifiedName("UInt32", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(28, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(7, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(28, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Int64"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8, 0),
        BrowseName=QualifiedName("Int64", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(27, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(8, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(27, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UInt64"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9, 0),
        BrowseName=QualifiedName("UInt64", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(28, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(28, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Float"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10, 0),
        BrowseName=QualifiedName("Float", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(26, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(10, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(26, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Double"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11, 0),
        BrowseName=QualifiedName("Double", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(26, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(26, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("String"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12, 0),
        BrowseName=QualifiedName("String", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DateTime"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13, 0),
        BrowseName=QualifiedName("DateTime", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(13, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Guid"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14, 0),
        BrowseName=QualifiedName("Guid", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ByteString"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15, 0),
        BrowseName=QualifiedName("ByteString", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("XmlElement"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16, 0),
        BrowseName=QualifiedName("XmlElement", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(16, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NodeId"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17, 0),
        BrowseName=QualifiedName("NodeId", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ExpandedNodeId"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18, 0),
        BrowseName=QualifiedName("ExpandedNodeId", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("StatusCode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19, 0),
        BrowseName=QualifiedName("StatusCode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("QualifiedName"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20, 0),
        BrowseName=QualifiedName("QualifiedName", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(20, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("LocalizedText"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21, 0),
        BrowseName=QualifiedName("LocalizedText", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Structure"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(22, 0),
        BrowseName=QualifiedName("Structure", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(22, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataValue"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23, 0),
        BrowseName=QualifiedName("DataValue", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DiagnosticInfo"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(25, 0),
        BrowseName=QualifiedName("DiagnosticInfo", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(25, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Image"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(30, 0),
        BrowseName=QualifiedName("Image", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(30, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Decimal"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(50, 0),
        BrowseName=QualifiedName("Decimal", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(26, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(50, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(26, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("References"),
        IsAbstract=True,
        Symmetric=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(31, 0),
        BrowseName=QualifiedName("References", 0),
        NodeClass_=NodeClass.ReferenceType,
        NodeAttributes=attrs,
    )
    server.add_nodes([node])

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("NonHierarchicalReferences"),
        IsAbstract=True,
        Symmetric=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(32, 0),
        BrowseName=QualifiedName("NonHierarchicalReferences", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(31, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(32, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(31, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HierarchicalReferences"),
        InverseName=LocalizedText("HierarchicalReferences"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(33, 0),
        BrowseName=QualifiedName("HierarchicalReferences", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(31, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(33, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(31, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasChild"),
        InverseName=LocalizedText("ChildOf"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(34, 0),
        BrowseName=QualifiedName("HasChild", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(33, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(34, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(33, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("Organizes"),
        InverseName=LocalizedText("OrganizedBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(35, 0),
        BrowseName=QualifiedName("Organizes", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(33, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(35, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(33, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasEventSource"),
        InverseName=LocalizedText("EventSourceOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(36, 0),
        BrowseName=QualifiedName("HasEventSource", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(33, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(36, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(33, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasModellingRule"),
        InverseName=LocalizedText("ModellingRuleOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(37, 0),
        BrowseName=QualifiedName("HasModellingRule", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(37, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasEncoding"),
        InverseName=LocalizedText("EncodingOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(38, 0),
        BrowseName=QualifiedName("HasEncoding", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(38, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasDescription"),
        InverseName=LocalizedText("DescriptionOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(39, 0),
        BrowseName=QualifiedName("HasDescription", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(39, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasTypeDefinition"),
        InverseName=LocalizedText("TypeDefinitionOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(40, 0),
        BrowseName=QualifiedName("HasTypeDefinition", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(40, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("GeneratesEvent"),
        InverseName=LocalizedText("GeneratedBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(41, 0),
        BrowseName=QualifiedName("GeneratesEvent", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(41, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("AlwaysGeneratesEvent"),
        InverseName=LocalizedText("AlwaysGeneratedBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3065, 0),
        BrowseName=QualifiedName("AlwaysGeneratesEvent", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(41, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(41, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("Aggregates"),
        InverseName=LocalizedText("AggregatedBy"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(44, 0),
        BrowseName=QualifiedName("Aggregates", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(34, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(44, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(34, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasSubtype"),
        InverseName=LocalizedText("SubtypeOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(45, 0),
        BrowseName=QualifiedName("HasSubtype", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(34, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(45, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(34, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasProperty"),
        InverseName=LocalizedText("PropertyOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(46, 0),
        BrowseName=QualifiedName("HasProperty", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(44, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(46, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(44, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasComponent"),
        InverseName=LocalizedText("ComponentOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(47, 0),
        BrowseName=QualifiedName("HasComponent", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(44, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(47, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(44, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasNotifier"),
        InverseName=LocalizedText("NotifierOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(48, 0),
        BrowseName=QualifiedName("HasNotifier", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(36, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(48, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(36, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasOrderedComponent"),
        InverseName=LocalizedText("OrderedComponentOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(49, 0),
        BrowseName=QualifiedName("HasOrderedComponent", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(49, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("FromState"),
        InverseName=LocalizedText("ToTransition"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(51, 0),
        BrowseName=QualifiedName("FromState", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(51, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("ToState"),
        InverseName=LocalizedText("FromTransition"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(52, 0),
        BrowseName=QualifiedName("ToState", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(52, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasCause"),
        InverseName=LocalizedText("MayBeCausedBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(53, 0),
        BrowseName=QualifiedName("HasCause", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(53, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasEffect"),
        InverseName=LocalizedText("MayBeEffectedBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(54, 0),
        BrowseName=QualifiedName("HasEffect", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(54, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasSubStateMachine"),
        InverseName=LocalizedText("SubStateMachineOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(117, 0),
        BrowseName=QualifiedName("HasSubStateMachine", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasHistoricalConfiguration"),
        InverseName=LocalizedText("HistoricalConfigurationOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(56, 0),
        BrowseName=QualifiedName("HasHistoricalConfiguration", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(44, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(56, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(44, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BaseObjectType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(58, 0),
        BrowseName=QualifiedName("BaseObjectType", 0),
        NodeClass_=NodeClass.ObjectType,
        NodeAttributes=attrs,
    )
    server.add_nodes([node])

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("FolderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(61, 0),
        BrowseName=QualifiedName("FolderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(61, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("BaseVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(62, 0),
        BrowseName=QualifiedName("BaseVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        NodeAttributes=attrs,
    )
    server.add_nodes([node])

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("BaseDataVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(63, 0),
        BrowseName=QualifiedName("BaseDataVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(62, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(63, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(62, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("PropertyType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(68, 0),
        BrowseName=QualifiedName("PropertyType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(62, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(68, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(62, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("DataTypeDescriptionType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(69, 0),
        BrowseName=QualifiedName("DataTypeDescriptionType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(69, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(104, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(69, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(105, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(69, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(104, 0),
        BrowseName=QualifiedName("DataTypeVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(69, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DictionaryFragment"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(105, 0),
        BrowseName=QualifiedName("DictionaryFragment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(69, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("DataTypeDictionaryType"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(72, 0),
        BrowseName=QualifiedName("DataTypeDictionaryType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(72, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(106, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(72, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(72, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15001, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(72, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(106, 0),
        BrowseName=QualifiedName("DataTypeVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(72, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(72, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(107, 0),
        BrowseName=QualifiedName("NamespaceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(72, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(72, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Deprecated"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15001, 0),
        BrowseName=QualifiedName("Deprecated", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(72, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(72, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataTypeSystemType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(75, 0),
        BrowseName=QualifiedName("DataTypeSystemType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(75, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataTypeEncodingType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(76, 0),
        BrowseName=QualifiedName("DataTypeEncodingType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(76, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NamingRuleType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(120, 0),
        BrowseName=QualifiedName("NamingRuleType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='Mandatory'),
                Description=LocalizedText(Text='The BrowseName must appear in all instances of the type.'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='Optional'),
                Description=LocalizedText(Text='The BrowseName may appear in an instance of the type.'),
                ),
            ua.EnumValueType(
                Value=3,
                DisplayName=LocalizedText(Text='Constraint'),
                Description=LocalizedText(Text='The modelling rule defines a constraint and the BrowseName is not used in an instance of the type.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12169, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(120, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(120, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ModellingRuleType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(77, 0),
        BrowseName=QualifiedName("ModellingRuleType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(77, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(111, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(77, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamingRule"),
        DataType=NumericNodeId(120, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(111, 0),
        BrowseName=QualifiedName("NamingRule", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(77, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(77, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Mandatory"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(78, 0),
        BrowseName=QualifiedName("Mandatory", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(77, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(78, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(112, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(78, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(77, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamingRule"),
        DataType=NumericNodeId(120, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(112, 0),
        BrowseName=QualifiedName("NamingRule", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(78, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Optional"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(80, 0),
        BrowseName=QualifiedName("Optional", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(77, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(80, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(80, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(77, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamingRule"),
        DataType=NumericNodeId(120, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(113, 0),
        BrowseName=QualifiedName("NamingRule", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(80, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ExposesItsArray"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(83, 0),
        BrowseName=QualifiedName("ExposesItsArray", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(77, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(83, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(114, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(83, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(77, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamingRule"),
        DataType=NumericNodeId(120, 0),
        Value=ua.Variant(3, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(114, 0),
        BrowseName=QualifiedName("NamingRule", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(83, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(83, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("OptionalPlaceholder"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11508, 0),
        BrowseName=QualifiedName("OptionalPlaceholder", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(77, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11509, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(77, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamingRule"),
        DataType=NumericNodeId(120, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11509, 0),
        BrowseName=QualifiedName("NamingRule", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11508, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MandatoryPlaceholder"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11510, 0),
        BrowseName=QualifiedName("MandatoryPlaceholder", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(77, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11511, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(77, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamingRule"),
        DataType=NumericNodeId(120, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11511, 0),
        BrowseName=QualifiedName("NamingRule", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11510, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11510, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Root"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(84, 0),
        BrowseName=QualifiedName("Root", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(84, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Objects"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(85, 0),
        BrowseName=QualifiedName("Objects", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(84, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(85, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(84, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(85, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Types"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(86, 0),
        BrowseName=QualifiedName("Types", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(84, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(86, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(84, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(86, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Views"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(87, 0),
        BrowseName=QualifiedName("Views", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(84, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(87, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(84, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(87, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ObjectTypes"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(88, 0),
        BrowseName=QualifiedName("ObjectTypes", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(86, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(88, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(86, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(88, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(88, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("VariableTypes"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(89, 0),
        BrowseName=QualifiedName("VariableTypes", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(86, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(89, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(86, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(89, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(62, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(89, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DataTypes"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(90, 0),
        BrowseName=QualifiedName("DataTypes", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(86, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(90, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(86, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(90, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(90, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReferenceTypes"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(91, 0),
        BrowseName=QualifiedName("ReferenceTypes", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(86, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(91, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(86, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(91, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(31, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(91, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("XML Schema"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(92, 0),
        BrowseName=QualifiedName("XML Schema", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(90, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(75, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(92, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(90, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(92, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(75, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("OPC Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(93, 0),
        BrowseName=QualifiedName("OPC Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(90, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(75, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(93, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(90, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(93, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(75, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasArgumentDescription"),
        InverseName=LocalizedText("ArgumentDescriptionOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(129, 0),
        BrowseName=QualifiedName("HasArgumentDescription", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasOptionalInputArgumentDescription"),
        InverseName=LocalizedText("OptionalInputArgumentDescriptionOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(131, 0),
        BrowseName=QualifiedName("HasOptionalInputArgumentDescription", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(129, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(129, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("http://opcfoundation.org/UA/"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15957, 0),
        BrowseName=QualifiedName("http://opcfoundation.org/UA/", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11616, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15958, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15959, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15960, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15961, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15962, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15963, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15964, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16134, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16135, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16136, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11715, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('http://opcfoundation.org/UA/', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15958, 0),
        BrowseName=QualifiedName("NamespaceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('1.04.9', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15959, 0),
        BrowseName=QualifiedName("NamespaceVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespacePublicationDate"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        Value=ua.Variant(datetime.datetime(2021, 1, 21, 0, 0, tzinfo=tzutc()), ua.VariantType.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15960, 0),
        BrowseName=QualifiedName("NamespacePublicationDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IsNamespaceSubset"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        Value=ua.Variant(False, ua.VariantType.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15961, 0),
        BrowseName=QualifiedName("IsNamespaceSubset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticNodeIdTypes"),
        DataType=NumericNodeId(256, 0),
        Value=ua.Variant([0], ua.VariantType.Int32),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15962, 0),
        BrowseName=QualifiedName("StaticNodeIdTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticNumericNodeIdRange"),
        DataType=NumericNodeId(291, 0),
        Value=ua.Variant(['1:2252', '2270', '2273', '2280', '2283', '2291:2293', '2297:2734', '2738:2991', '2995', '2998:3703', '3710:11191', '11194:11195', '11202:11241', '11243:11272', '11276:11280', '11284:11311', '11315:11491', '11495:11501', '11503:11701', '11706', '11708', '11716:12164', '12169:12748', '12752:12872', '12875:12884', '12888:12910', '12912:14414', '14416:15003', '15005:15605', '15607:16300', '16306:17633', '17635:19090', '19092:2147483647'], ua.VariantType.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15963, 0),
        BrowseName=QualifiedName("StaticNumericNodeIdRange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticStringNodeIdPattern"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15964, 0),
        BrowseName=QualifiedName("StaticStringNodeIdPattern", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultRolePermissions"),
        DataType=NumericNodeId(96, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16134, 0),
        BrowseName=QualifiedName("DefaultRolePermissions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultUserRolePermissions"),
        DataType=NumericNodeId(96, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16135, 0),
        BrowseName=QualifiedName("DefaultUserRolePermissions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultAccessRestrictions"),
        DataType=NumericNodeId(95, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16136, 0),
        BrowseName=QualifiedName("DefaultAccessRestrictions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15957, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15957, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NodeVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3068, 0),
        BrowseName=QualifiedName("NodeVersion", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ViewVersion"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12170, 0),
        BrowseName=QualifiedName("ViewVersion", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Icon"),
        DataType=NumericNodeId(30, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3067, 0),
        BrowseName=QualifiedName("Icon", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocalTime"),
        DataType=NumericNodeId(8912, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3069, 0),
        BrowseName=QualifiedName("LocalTime", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AllowNulls"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3070, 0),
        BrowseName=QualifiedName("AllowNulls", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ValueAsText"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11433, 0),
        BrowseName=QualifiedName("ValueAsText", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxStringLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11498, 0),
        BrowseName=QualifiedName("MaxStringLength", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11498, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxCharacters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15002, 0),
        BrowseName=QualifiedName("MaxCharacters", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15002, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxByteStringLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12908, 0),
        BrowseName=QualifiedName("MaxByteStringLength", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12908, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxArrayLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11512, 0),
        BrowseName=QualifiedName("MaxArrayLength", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EngineeringUnits"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11513, 0),
        BrowseName=QualifiedName("EngineeringUnits", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11432, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3071, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12745, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3072, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3073, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultInstanceBrowseName"),
        DataType=ua.NodeId(ua.ObjectIds.QualifiedName),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17605, 0),
        BrowseName=QualifiedName("DefaultInstanceBrowseName", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ImageBMP"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2000, 0),
        BrowseName=QualifiedName("ImageBMP", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(30, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(30, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ImageGIF"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2001, 0),
        BrowseName=QualifiedName("ImageGIF", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(30, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(30, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ImageJPG"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2002, 0),
        BrowseName=QualifiedName("ImageJPG", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(30, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2002, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(30, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ImagePNG"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2003, 0),
        BrowseName=QualifiedName("ImagePNG", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(30, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(30, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AudioDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16307, 0),
        BrowseName=QualifiedName("AudioDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(16307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ServerType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2004, 0),
        BrowseName=QualifiedName("ServerType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2005, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2006, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15003, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2007, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2008, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2742, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12882, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17612, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2010, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2011, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2012, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11527, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11489, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12871, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12883, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ServerArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2005, 0),
        BrowseName=QualifiedName("ServerArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("NamespaceArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2006, 0),
        BrowseName=QualifiedName("NamespaceArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("UrisVersion"),
        DataType=NumericNodeId(20998, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15003, 0),
        BrowseName=QualifiedName("UrisVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ServerStatus"),
        DataType=NumericNodeId(862, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2007, 0),
        BrowseName=QualifiedName("ServerStatus", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2138, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3074, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3075, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3077, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3084, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3085, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StartTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3074, 0),
        BrowseName=QualifiedName("StartTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2007, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2007, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3075, 0),
        BrowseName=QualifiedName("CurrentTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2007, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2007, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(852, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3076, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2007, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2007, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BuildInfo"),
        DataType=NumericNodeId(338, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3077, 0),
        BrowseName=QualifiedName("BuildInfo", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2007, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(3051, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3078, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3081, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3082, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2007, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3078, 0),
        BrowseName=QualifiedName("ProductUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3077, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ManufacturerName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3079, 0),
        BrowseName=QualifiedName("ManufacturerName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3077, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3080, 0),
        BrowseName=QualifiedName("ProductName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3077, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("SoftwareVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3081, 0),
        BrowseName=QualifiedName("SoftwareVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3077, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildNumber"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3082, 0),
        BrowseName=QualifiedName("BuildNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3077, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildDate"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3083, 0),
        BrowseName=QualifiedName("BuildDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3077, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecondsTillShutdown"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3084, 0),
        BrowseName=QualifiedName("SecondsTillShutdown", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2007, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2007, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ShutdownReason"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3085, 0),
        BrowseName=QualifiedName("ShutdownReason", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2007, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2007, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ServiceLevel"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2008, 0),
        BrowseName=QualifiedName("ServiceLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("Auditing"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2742, 0),
        BrowseName=QualifiedName("Auditing", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("EstimatedReturnTime"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12882, 0),
        BrowseName=QualifiedName("EstimatedReturnTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("LocalTime"),
        DataType=NumericNodeId(8912, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17612, 0),
        BrowseName=QualifiedName("LocalTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ServerCapabilities"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2009, 0),
        BrowseName=QualifiedName("ServerCapabilities", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2013, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3086, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3087, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3088, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3089, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3090, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3091, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerProfileArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3086, 0),
        BrowseName=QualifiedName("ServerProfileArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIdArray"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3087, 0),
        BrowseName=QualifiedName("LocaleIdArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MinSupportedSampleRate"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3088, 0),
        BrowseName=QualifiedName("MinSupportedSampleRate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxBrowseContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3089, 0),
        BrowseName=QualifiedName("MaxBrowseContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxQueryContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3090, 0),
        BrowseName=QualifiedName("MaxQueryContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxHistoryContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3091, 0),
        BrowseName=QualifiedName("MaxHistoryContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SoftwareCertificates"),
        DataType=NumericNodeId(344, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3092, 0),
        BrowseName=QualifiedName("SoftwareCertificates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ModellingRules"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3093, 0),
        BrowseName=QualifiedName("ModellingRules", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateFunctions"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3094, 0),
        BrowseName=QualifiedName("AggregateFunctions", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2009, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2009, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ServerDiagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2010, 0),
        BrowseName=QualifiedName("ServerDiagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2020, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3111, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3114, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsSummary"),
        DataType=NumericNodeId(859, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3095, 0),
        BrowseName=QualifiedName("ServerDiagnosticsSummary", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2010, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2150, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3096, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3097, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3099, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3104, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3105, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3106, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3108, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2010, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerViewCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3096, 0),
        BrowseName=QualifiedName("ServerViewCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3097, 0),
        BrowseName=QualifiedName("CurrentSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3098, 0),
        BrowseName=QualifiedName("CumulatedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3099, 0),
        BrowseName=QualifiedName("SecurityRejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3100, 0),
        BrowseName=QualifiedName("RejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionTimeoutCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3101, 0),
        BrowseName=QualifiedName("SessionTimeoutCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionAbortCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3102, 0),
        BrowseName=QualifiedName("SessionAbortCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingIntervalCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3104, 0),
        BrowseName=QualifiedName("PublishingIntervalCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3105, 0),
        BrowseName=QualifiedName("CurrentSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3106, 0),
        BrowseName=QualifiedName("CumulatedSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3107, 0),
        BrowseName=QualifiedName("SecurityRejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3108, 0),
        BrowseName=QualifiedName("RejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3095, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsArray"),
        DataType=NumericNodeId(874, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3110, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2010, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2171, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2010, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SessionsDiagnosticsSummary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3111, 0),
        BrowseName=QualifiedName("SessionsDiagnosticsSummary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2010, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2026, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3112, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2026, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2010, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsArray"),
        DataType=NumericNodeId(865, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3112, 0),
        BrowseName=QualifiedName("SessionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3111, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2196, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2196, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsArray"),
        DataType=NumericNodeId(868, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3113, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3111, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2243, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2243, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledFlag"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        AccessLevel=3,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3114, 0),
        BrowseName=QualifiedName("EnabledFlag", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2010, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2010, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("VendorServerInfo"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2011, 0),
        BrowseName=QualifiedName("VendorServerInfo", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2033, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2033, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ServerRedundancy"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2012, 0),
        BrowseName=QualifiedName("ServerRedundancy", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2034, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3115, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2034, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RedundancySupport"),
        DataType=NumericNodeId(851, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3115, 0),
        BrowseName=QualifiedName("RedundancySupport", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2012, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2012, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Namespaces"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11527, 0),
        BrowseName=QualifiedName("Namespaces", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11645, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11645, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetMonitoredItems"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11489, 0),
        BrowseName=QualifiedName("GetMonitoredItems", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11490, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11491, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11490, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11489, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11489, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ServerHandles',
                DataType=NumericNodeId(7, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='ClientHandles',
                DataType=NumericNodeId(7, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11491, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11489, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11489, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ResendData"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12871, 0),
        BrowseName=QualifiedName("ResendData", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12872, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12872, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12871, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12871, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetSubscriptionDurable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12746, 0),
        BrowseName=QualifiedName("SetSubscriptionDurable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12747, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12748, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='LifetimeInHours',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12747, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12746, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12746, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RevisedLifetimeInHours',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12748, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12746, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12746, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RequestServerStateChange"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12883, 0),
        BrowseName=QualifiedName("RequestServerStateChange", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2004, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12884, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='State',
                DataType=NumericNodeId(852, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='EstimatedReturnTime',
                DataType=NumericNodeId(13, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecondsTillShutdown',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Reason',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Restart',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12884, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12883, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ServerCapabilitiesType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2013, 0),
        BrowseName=QualifiedName("ServerCapabilitiesType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2014, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2016, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2017, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2732, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2733, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3049, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11549, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11550, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12910, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11551, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2019, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2754, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11562, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16295, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerProfileArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2014, 0),
        BrowseName=QualifiedName("ServerProfileArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIdArray"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2016, 0),
        BrowseName=QualifiedName("LocaleIdArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MinSupportedSampleRate"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2017, 0),
        BrowseName=QualifiedName("MinSupportedSampleRate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxBrowseContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2732, 0),
        BrowseName=QualifiedName("MaxBrowseContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxQueryContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2733, 0),
        BrowseName=QualifiedName("MaxQueryContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxHistoryContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2734, 0),
        BrowseName=QualifiedName("MaxHistoryContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SoftwareCertificates"),
        DataType=NumericNodeId(344, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3049, 0),
        BrowseName=QualifiedName("SoftwareCertificates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3049, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3049, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3049, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxArrayLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11549, 0),
        BrowseName=QualifiedName("MaxArrayLength", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxStringLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11550, 0),
        BrowseName=QualifiedName("MaxStringLength", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxByteStringLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12910, 0),
        BrowseName=QualifiedName("MaxByteStringLength", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12910, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12910, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12910, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("OperationLimits"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11551, 0),
        BrowseName=QualifiedName("OperationLimits", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11564, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ModellingRules"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2019, 0),
        BrowseName=QualifiedName("ModellingRules", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateFunctions"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2754, 0),
        BrowseName=QualifiedName("AggregateFunctions", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("<VendorCapability>"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11562, 0),
        BrowseName=QualifiedName("<VendorCapability>", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2137, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2137, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("RoleSet"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16295, 0),
        BrowseName=QualifiedName("RoleSet", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15607, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16296, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16299, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddRole"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16296, 0),
        BrowseName=QualifiedName("AddRole", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16295, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16297, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16298, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16295, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='NamespaceUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16297, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16296, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16296, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16298, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16296, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16296, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveRole"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16299, 0),
        BrowseName=QualifiedName("RemoveRole", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16295, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16300, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16295, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16300, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16299, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16299, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2020, 0),
        BrowseName=QualifiedName("ServerDiagnosticsType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2022, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2023, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2744, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2025, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsSummary"),
        DataType=NumericNodeId(859, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2021, 0),
        BrowseName=QualifiedName("ServerDiagnosticsSummary", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2020, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2150, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3116, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3117, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3120, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3121, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3122, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3124, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3125, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3126, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3128, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerViewCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3116, 0),
        BrowseName=QualifiedName("ServerViewCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3117, 0),
        BrowseName=QualifiedName("CurrentSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3118, 0),
        BrowseName=QualifiedName("CumulatedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3119, 0),
        BrowseName=QualifiedName("SecurityRejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3120, 0),
        BrowseName=QualifiedName("RejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionTimeoutCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3121, 0),
        BrowseName=QualifiedName("SessionTimeoutCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionAbortCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3122, 0),
        BrowseName=QualifiedName("SessionAbortCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingIntervalCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3124, 0),
        BrowseName=QualifiedName("PublishingIntervalCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3125, 0),
        BrowseName=QualifiedName("CurrentSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3126, 0),
        BrowseName=QualifiedName("CumulatedSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3127, 0),
        BrowseName=QualifiedName("SecurityRejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3128, 0),
        BrowseName=QualifiedName("RejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2021, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnosticsArray"),
        DataType=NumericNodeId(856, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2022, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2020, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2164, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2164, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsArray"),
        DataType=NumericNodeId(874, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2023, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2020, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2171, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SessionsDiagnosticsSummary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2744, 0),
        BrowseName=QualifiedName("SessionsDiagnosticsSummary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2020, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2026, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3129, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3130, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2026, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsArray"),
        DataType=NumericNodeId(865, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3129, 0),
        BrowseName=QualifiedName("SessionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2744, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2196, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2196, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2744, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsArray"),
        DataType=NumericNodeId(868, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3130, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2744, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2243, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2243, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2744, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledFlag"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        AccessLevel=3,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2025, 0),
        BrowseName=QualifiedName("EnabledFlag", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2020, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SessionsDiagnosticsSummaryType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2026, 0),
        BrowseName=QualifiedName("SessionsDiagnosticsSummaryType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2028, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12097, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsArray"),
        DataType=NumericNodeId(865, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2027, 0),
        BrowseName=QualifiedName("SessionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2026, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2196, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2196, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2026, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsArray"),
        DataType=NumericNodeId(868, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2028, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2026, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2243, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2243, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2026, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<ClientName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12097, 0),
        BrowseName=QualifiedName("<ClientName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2026, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2029, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12152, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2029, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2026, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnostics"),
        DataType=NumericNodeId(865, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12098, 0),
        BrowseName=QualifiedName("SessionDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12097, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2197, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12099, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12103, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12104, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12105, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12106, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12108, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12109, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12111, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12112, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12114, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12115, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12116, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12117, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12120, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12121, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12122, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12123, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12124, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12125, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12126, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12128, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12129, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12130, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12131, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12132, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12133, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12134, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12135, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12136, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12137, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12138, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12139, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12140, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12141, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12097, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12099, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12100, 0),
        BrowseName=QualifiedName("SessionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientDescription"),
        DataType=NumericNodeId(308, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12101, 0),
        BrowseName=QualifiedName("ClientDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12102, 0),
        BrowseName=QualifiedName("ServerUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrl"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12103, 0),
        BrowseName=QualifiedName("EndpointUrl", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIds"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12104, 0),
        BrowseName=QualifiedName("LocaleIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActualSessionTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12105, 0),
        BrowseName=QualifiedName("ActualSessionTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxResponseMessageSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12106, 0),
        BrowseName=QualifiedName("MaxResponseMessageSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientConnectionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12107, 0),
        BrowseName=QualifiedName("ClientConnectionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientLastContactTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12108, 0),
        BrowseName=QualifiedName("ClientLastContactTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12109, 0),
        BrowseName=QualifiedName("CurrentSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12110, 0),
        BrowseName=QualifiedName("CurrentMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentPublishRequestsInQueue"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12111, 0),
        BrowseName=QualifiedName("CurrentPublishRequestsInQueue", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalRequestCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12112, 0),
        BrowseName=QualifiedName("TotalRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnauthorizedRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12113, 0),
        BrowseName=QualifiedName("UnauthorizedRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12114, 0),
        BrowseName=QualifiedName("ReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12115, 0),
        BrowseName=QualifiedName("HistoryReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriteCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12116, 0),
        BrowseName=QualifiedName("WriteCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryUpdateCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12117, 0),
        BrowseName=QualifiedName("HistoryUpdateCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CallCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12118, 0),
        BrowseName=QualifiedName("CallCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12119, 0),
        BrowseName=QualifiedName("CreateMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifyMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12120, 0),
        BrowseName=QualifiedName("ModifyMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetMonitoringModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12121, 0),
        BrowseName=QualifiedName("SetMonitoringModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetTriggeringCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12122, 0),
        BrowseName=QualifiedName("SetTriggeringCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12123, 0),
        BrowseName=QualifiedName("DeleteMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateSubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12124, 0),
        BrowseName=QualifiedName("CreateSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifySubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12125, 0),
        BrowseName=QualifiedName("ModifySubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetPublishingModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12126, 0),
        BrowseName=QualifiedName("SetPublishingModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12127, 0),
        BrowseName=QualifiedName("PublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12128, 0),
        BrowseName=QualifiedName("RepublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12129, 0),
        BrowseName=QualifiedName("TransferSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12130, 0),
        BrowseName=QualifiedName("DeleteSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12131, 0),
        BrowseName=QualifiedName("AddNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12132, 0),
        BrowseName=QualifiedName("AddReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12133, 0),
        BrowseName=QualifiedName("DeleteNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12134, 0),
        BrowseName=QualifiedName("DeleteReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12135, 0),
        BrowseName=QualifiedName("BrowseCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12136, 0),
        BrowseName=QualifiedName("BrowseNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TranslateBrowsePathsToNodeIdsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12137, 0),
        BrowseName=QualifiedName("TranslateBrowsePathsToNodeIdsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryFirstCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12138, 0),
        BrowseName=QualifiedName("QueryFirstCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12139, 0),
        BrowseName=QualifiedName("QueryNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RegisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12140, 0),
        BrowseName=QualifiedName("RegisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnregisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12141, 0),
        BrowseName=QualifiedName("UnregisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12098, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnostics"),
        DataType=NumericNodeId(868, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12142, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12097, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2244, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12143, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12144, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12145, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12146, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12147, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12148, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12149, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12150, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12151, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12097, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12143, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdOfSession"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12144, 0),
        BrowseName=QualifiedName("ClientUserIdOfSession", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdHistory"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12145, 0),
        BrowseName=QualifiedName("ClientUserIdHistory", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationMechanism"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12146, 0),
        BrowseName=QualifiedName("AuthenticationMechanism", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Encoding"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12147, 0),
        BrowseName=QualifiedName("Encoding", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransportProtocol"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12148, 0),
        BrowseName=QualifiedName("TransportProtocol", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12149, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityPolicyUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12150, 0),
        BrowseName=QualifiedName("SecurityPolicyUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12151, 0),
        BrowseName=QualifiedName("ClientCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsArray"),
        DataType=NumericNodeId(874, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12152, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12097, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2171, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12097, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsObjectType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2029, 0),
        BrowseName=QualifiedName("SessionDiagnosticsObjectType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2032, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnostics"),
        DataType=NumericNodeId(865, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2030, 0),
        BrowseName=QualifiedName("SessionDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2197, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3131, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3132, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3133, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3134, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3135, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3136, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3137, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3138, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3139, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3140, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3141, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3142, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3143, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8898, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3151, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3152, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3153, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3154, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3155, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3156, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3157, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3158, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3159, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3160, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3161, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3162, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3163, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3164, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3165, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3166, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3167, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3168, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3170, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3173, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3174, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3176, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3177, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3131, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3132, 0),
        BrowseName=QualifiedName("SessionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientDescription"),
        DataType=NumericNodeId(308, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3133, 0),
        BrowseName=QualifiedName("ClientDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3134, 0),
        BrowseName=QualifiedName("ServerUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrl"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3135, 0),
        BrowseName=QualifiedName("EndpointUrl", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIds"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3136, 0),
        BrowseName=QualifiedName("LocaleIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActualSessionTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3137, 0),
        BrowseName=QualifiedName("ActualSessionTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxResponseMessageSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3138, 0),
        BrowseName=QualifiedName("MaxResponseMessageSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientConnectionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3139, 0),
        BrowseName=QualifiedName("ClientConnectionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientLastContactTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3140, 0),
        BrowseName=QualifiedName("ClientLastContactTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3141, 0),
        BrowseName=QualifiedName("CurrentSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3142, 0),
        BrowseName=QualifiedName("CurrentMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentPublishRequestsInQueue"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3143, 0),
        BrowseName=QualifiedName("CurrentPublishRequestsInQueue", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalRequestCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8898, 0),
        BrowseName=QualifiedName("TotalRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnauthorizedRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11891, 0),
        BrowseName=QualifiedName("UnauthorizedRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3151, 0),
        BrowseName=QualifiedName("ReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3152, 0),
        BrowseName=QualifiedName("HistoryReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriteCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3153, 0),
        BrowseName=QualifiedName("WriteCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryUpdateCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3154, 0),
        BrowseName=QualifiedName("HistoryUpdateCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CallCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3155, 0),
        BrowseName=QualifiedName("CallCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3156, 0),
        BrowseName=QualifiedName("CreateMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifyMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3157, 0),
        BrowseName=QualifiedName("ModifyMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetMonitoringModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3158, 0),
        BrowseName=QualifiedName("SetMonitoringModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetTriggeringCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3159, 0),
        BrowseName=QualifiedName("SetTriggeringCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3160, 0),
        BrowseName=QualifiedName("DeleteMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateSubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3161, 0),
        BrowseName=QualifiedName("CreateSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifySubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3162, 0),
        BrowseName=QualifiedName("ModifySubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetPublishingModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3163, 0),
        BrowseName=QualifiedName("SetPublishingModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3164, 0),
        BrowseName=QualifiedName("PublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3165, 0),
        BrowseName=QualifiedName("RepublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3166, 0),
        BrowseName=QualifiedName("TransferSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3167, 0),
        BrowseName=QualifiedName("DeleteSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3168, 0),
        BrowseName=QualifiedName("AddNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3169, 0),
        BrowseName=QualifiedName("AddReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3170, 0),
        BrowseName=QualifiedName("DeleteNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3171, 0),
        BrowseName=QualifiedName("DeleteReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3172, 0),
        BrowseName=QualifiedName("BrowseCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3173, 0),
        BrowseName=QualifiedName("BrowseNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TranslateBrowsePathsToNodeIdsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3174, 0),
        BrowseName=QualifiedName("TranslateBrowsePathsToNodeIdsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryFirstCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3175, 0),
        BrowseName=QualifiedName("QueryFirstCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3176, 0),
        BrowseName=QualifiedName("QueryNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RegisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3177, 0),
        BrowseName=QualifiedName("RegisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnregisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3178, 0),
        BrowseName=QualifiedName("UnregisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2030, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnostics"),
        DataType=NumericNodeId(868, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2031, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2244, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3179, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3180, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3181, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3182, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3183, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3184, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3185, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3186, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3187, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3179, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdOfSession"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3180, 0),
        BrowseName=QualifiedName("ClientUserIdOfSession", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdHistory"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3181, 0),
        BrowseName=QualifiedName("ClientUserIdHistory", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationMechanism"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3182, 0),
        BrowseName=QualifiedName("AuthenticationMechanism", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Encoding"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3183, 0),
        BrowseName=QualifiedName("Encoding", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransportProtocol"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3184, 0),
        BrowseName=QualifiedName("TransportProtocol", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3185, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityPolicyUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3186, 0),
        BrowseName=QualifiedName("SecurityPolicyUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3187, 0),
        BrowseName=QualifiedName("ClientCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2031, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsArray"),
        DataType=NumericNodeId(874, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2032, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2171, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("VendorServerInfoType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2033, 0),
        BrowseName=QualifiedName("VendorServerInfoType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ServerRedundancyType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2034, 0),
        BrowseName=QualifiedName("ServerRedundancyType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2035, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RedundancySupport"),
        DataType=NumericNodeId(851, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2035, 0),
        BrowseName=QualifiedName("RedundancySupport", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2034, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2034, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TransparentRedundancyType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2036, 0),
        BrowseName=QualifiedName("TransparentRedundancyType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2034, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2037, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2038, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2034, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentServerId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2037, 0),
        BrowseName=QualifiedName("CurrentServerId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2036, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RedundantServerArray"),
        DataType=NumericNodeId(853, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2038, 0),
        BrowseName=QualifiedName("RedundantServerArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2036, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NonTransparentRedundancyType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2039, 0),
        BrowseName=QualifiedName("NonTransparentRedundancyType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2034, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2039, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2040, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2039, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2034, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerUriArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2040, 0),
        BrowseName=QualifiedName("ServerUriArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2039, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2039, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NonTransparentNetworkRedundancyType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11945, 0),
        BrowseName=QualifiedName("NonTransparentNetworkRedundancyType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2039, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2039, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerNetworkGroups"),
        DataType=NumericNodeId(11944, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11948, 0),
        BrowseName=QualifiedName("ServerNetworkGroups", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11945, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11945, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("OperationLimitsType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11564, 0),
        BrowseName=QualifiedName("OperationLimitsType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11565, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12161, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12162, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12163, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12164, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11569, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11570, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11571, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11572, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11574, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerRead"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11565, 0),
        BrowseName=QualifiedName("MaxNodesPerRead", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryReadData"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12161, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryReadData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryReadEvents"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12162, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryReadEvents", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerWrite"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11567, 0),
        BrowseName=QualifiedName("MaxNodesPerWrite", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryUpdateData"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12163, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryUpdateData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryUpdateEvents"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12164, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryUpdateEvents", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerMethodCall"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11569, 0),
        BrowseName=QualifiedName("MaxNodesPerMethodCall", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerBrowse"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11570, 0),
        BrowseName=QualifiedName("MaxNodesPerBrowse", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerRegisterNodes"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11571, 0),
        BrowseName=QualifiedName("MaxNodesPerRegisterNodes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerTranslateBrowsePathsToNodeIds"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11572, 0),
        BrowseName=QualifiedName("MaxNodesPerTranslateBrowsePathsToNodeIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerNodeManagement"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11573, 0),
        BrowseName=QualifiedName("MaxNodesPerNodeManagement", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxMonitoredItemsPerCall"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11574, 0),
        BrowseName=QualifiedName("MaxMonitoredItemsPerCall", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11564, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("FileType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11575, 0),
        BrowseName=QualifiedName("FileType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11576, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12687, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11579, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13341, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11583, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11585, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11588, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11590, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11593, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11576, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12686, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12687, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11579, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MimeType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13341, 0),
        BrowseName=QualifiedName("MimeType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11580, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11581, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11581, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11580, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11580, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11582, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11580, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11580, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11583, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11584, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11584, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11583, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11583, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11585, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11586, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11587, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11586, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11585, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11585, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11587, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11585, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11585, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11588, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11589, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11589, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11588, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11588, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11590, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11591, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11592, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11591, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11590, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11590, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11592, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11590, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11590, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11593, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11594, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11594, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11593, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11593, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AddressSpaceFileType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11595, 0),
        BrowseName=QualifiedName("AddressSpaceFileType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11615, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ExportNamespace"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11615, 0),
        BrowseName=QualifiedName("ExportNamespace", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11595, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11595, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NamespaceMetadataType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11616, 0),
        BrowseName=QualifiedName("NamespaceMetadataType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11617, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11618, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11619, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11620, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11621, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11623, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16137, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16138, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16139, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11617, 0),
        BrowseName=QualifiedName("NamespaceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11618, 0),
        BrowseName=QualifiedName("NamespaceVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespacePublicationDate"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11619, 0),
        BrowseName=QualifiedName("NamespacePublicationDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IsNamespaceSubset"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11620, 0),
        BrowseName=QualifiedName("IsNamespaceSubset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticNodeIdTypes"),
        DataType=NumericNodeId(256, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11621, 0),
        BrowseName=QualifiedName("StaticNodeIdTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticNumericNodeIdRange"),
        DataType=NumericNodeId(291, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11622, 0),
        BrowseName=QualifiedName("StaticNumericNodeIdRange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticStringNodeIdPattern"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11623, 0),
        BrowseName=QualifiedName("StaticStringNodeIdPattern", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("NamespaceFile"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11624, 0),
        BrowseName=QualifiedName("NamespaceFile", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11595, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11625, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12690, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12691, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11628, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11632, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11634, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11637, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11639, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11642, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11595, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11625, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12690, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12691, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11628, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11628, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11628, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11628, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11629, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11630, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11631, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11630, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11629, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11629, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11631, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11629, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11629, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11632, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11633, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11633, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11632, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11632, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11634, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11636, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11635, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11634, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11634, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11636, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11634, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11634, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11637, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11638, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11638, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11637, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11638, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11638, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11638, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11639, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11640, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11641, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11640, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11639, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11639, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11641, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11639, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11639, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11642, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(11624, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11643, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11642, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultRolePermissions"),
        DataType=NumericNodeId(96, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16137, 0),
        BrowseName=QualifiedName("DefaultRolePermissions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultUserRolePermissions"),
        DataType=NumericNodeId(96, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16138, 0),
        BrowseName=QualifiedName("DefaultUserRolePermissions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultAccessRestrictions"),
        DataType=NumericNodeId(95, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16139, 0),
        BrowseName=QualifiedName("DefaultAccessRestrictions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NamespacesType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11645, 0),
        BrowseName=QualifiedName("NamespacesType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<NamespaceIdentifier>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11646, 0),
        BrowseName=QualifiedName("<NamespaceIdentifier>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11645, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11616, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11647, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11648, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11649, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11651, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11652, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11645, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11647, 0),
        BrowseName=QualifiedName("NamespaceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11648, 0),
        BrowseName=QualifiedName("NamespaceVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespacePublicationDate"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11649, 0),
        BrowseName=QualifiedName("NamespacePublicationDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IsNamespaceSubset"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11650, 0),
        BrowseName=QualifiedName("IsNamespaceSubset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticNodeIdTypes"),
        DataType=NumericNodeId(256, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11651, 0),
        BrowseName=QualifiedName("StaticNodeIdTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticNumericNodeIdRange"),
        DataType=NumericNodeId(291, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11652, 0),
        BrowseName=QualifiedName("StaticNumericNodeIdRange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StaticStringNodeIdPattern"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11653, 0),
        BrowseName=QualifiedName("StaticStringNodeIdPattern", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11653, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11653, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11653, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BaseEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2041, 0),
        BrowseName=QualifiedName("BaseEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2042, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2043, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2044, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2045, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2046, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2047, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3190, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2050, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2051, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventId"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2042, 0),
        BrowseName=QualifiedName("EventId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventType"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2043, 0),
        BrowseName=QualifiedName("EventType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2044, 0),
        BrowseName=QualifiedName("SourceNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2044, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2044, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2044, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2045, 0),
        BrowseName=QualifiedName("SourceName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Time"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2046, 0),
        BrowseName=QualifiedName("Time", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceiveTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2047, 0),
        BrowseName=QualifiedName("ReceiveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocalTime"),
        DataType=NumericNodeId(8912, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3190, 0),
        BrowseName=QualifiedName("LocalTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Message"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2050, 0),
        BrowseName=QualifiedName("Message", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Severity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2051, 0),
        BrowseName=QualifiedName("Severity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2052, 0),
        BrowseName=QualifiedName("AuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2053, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2054, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2057, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActionTimeStamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2053, 0),
        BrowseName=QualifiedName("ActionTimeStamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Status"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2054, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2055, 0),
        BrowseName=QualifiedName("ServerId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientAuditEntryId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2056, 0),
        BrowseName=QualifiedName("ClientAuditEntryId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2057, 0),
        BrowseName=QualifiedName("ClientUserId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditSecurityEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2058, 0),
        BrowseName=QualifiedName("AuditSecurityEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17615, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatusCodeId"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17615, 0),
        BrowseName=QualifiedName("StatusCodeId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2058, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2058, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditChannelEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2059, 0),
        BrowseName=QualifiedName("AuditChannelEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2058, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2745, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2058, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecureChannelId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2745, 0),
        BrowseName=QualifiedName("SecureChannelId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2059, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2059, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditOpenSecureChannelEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2060, 0),
        BrowseName=QualifiedName("AuditOpenSecureChannelEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2059, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2061, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2062, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2063, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2065, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2066, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2059, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2061, 0),
        BrowseName=QualifiedName("ClientCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2060, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificateThumbprint"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2746, 0),
        BrowseName=QualifiedName("ClientCertificateThumbprint", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2060, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RequestType"),
        DataType=NumericNodeId(315, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2062, 0),
        BrowseName=QualifiedName("RequestType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2060, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityPolicyUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2063, 0),
        BrowseName=QualifiedName("SecurityPolicyUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2060, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2065, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2060, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RequestedLifetime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2066, 0),
        BrowseName=QualifiedName("RequestedLifetime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2060, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditSessionEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2069, 0),
        BrowseName=QualifiedName("AuditSessionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2058, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2070, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2058, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2070, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2069, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2069, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCreateSessionEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2071, 0),
        BrowseName=QualifiedName("AuditCreateSessionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2069, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2072, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2747, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2074, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2069, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecureChannelId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2072, 0),
        BrowseName=QualifiedName("SecureChannelId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2071, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2071, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2073, 0),
        BrowseName=QualifiedName("ClientCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2071, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2071, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificateThumbprint"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2747, 0),
        BrowseName=QualifiedName("ClientCertificateThumbprint", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2071, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2071, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RevisedSessionTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2074, 0),
        BrowseName=QualifiedName("RevisedSessionTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2071, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2071, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditUrlMismatchEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2748, 0),
        BrowseName=QualifiedName("AuditUrlMismatchEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2071, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2749, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2071, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrl"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2749, 0),
        BrowseName=QualifiedName("EndpointUrl", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2748, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2748, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditActivateSessionEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2075, 0),
        BrowseName=QualifiedName("AuditActivateSessionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2069, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2077, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11485, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2069, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientSoftwareCertificates"),
        DataType=NumericNodeId(344, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2076, 0),
        BrowseName=QualifiedName("ClientSoftwareCertificates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2075, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2075, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserIdentityToken"),
        DataType=NumericNodeId(316, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2077, 0),
        BrowseName=QualifiedName("UserIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2075, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2075, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecureChannelId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11485, 0),
        BrowseName=QualifiedName("SecureChannelId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2075, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2075, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCancelEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2078, 0),
        BrowseName=QualifiedName("AuditCancelEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2069, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2069, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RequestHandle"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2079, 0),
        BrowseName=QualifiedName("RequestHandle", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2078, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2078, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCertificateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2080, 0),
        BrowseName=QualifiedName("AuditCertificateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2058, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2081, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2058, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Certificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2081, 0),
        BrowseName=QualifiedName("Certificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2080, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCertificateDataMismatchEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2082, 0),
        BrowseName=QualifiedName("AuditCertificateDataMismatchEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2080, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2084, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InvalidHostname"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2083, 0),
        BrowseName=QualifiedName("InvalidHostname", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2082, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2082, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InvalidUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2084, 0),
        BrowseName=QualifiedName("InvalidUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2082, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2082, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCertificateExpiredEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2085, 0),
        BrowseName=QualifiedName("AuditCertificateExpiredEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2080, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCertificateInvalidEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2086, 0),
        BrowseName=QualifiedName("AuditCertificateInvalidEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2080, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCertificateUntrustedEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2087, 0),
        BrowseName=QualifiedName("AuditCertificateUntrustedEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2080, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCertificateRevokedEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2088, 0),
        BrowseName=QualifiedName("AuditCertificateRevokedEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2080, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditCertificateMismatchEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2089, 0),
        BrowseName=QualifiedName("AuditCertificateMismatchEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2080, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditNodeManagementEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2090, 0),
        BrowseName=QualifiedName("AuditNodeManagementEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditAddNodesEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2091, 0),
        BrowseName=QualifiedName("AuditAddNodesEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2090, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2090, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NodesToAdd"),
        DataType=NumericNodeId(376, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2092, 0),
        BrowseName=QualifiedName("NodesToAdd", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2091, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2091, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditDeleteNodesEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2093, 0),
        BrowseName=QualifiedName("AuditDeleteNodesEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2090, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2090, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NodesToDelete"),
        DataType=NumericNodeId(382, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2094, 0),
        BrowseName=QualifiedName("NodesToDelete", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2093, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2093, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditAddReferencesEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2095, 0),
        BrowseName=QualifiedName("AuditAddReferencesEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2090, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2096, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2090, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReferencesToAdd"),
        DataType=NumericNodeId(379, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2096, 0),
        BrowseName=QualifiedName("ReferencesToAdd", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2095, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditDeleteReferencesEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2097, 0),
        BrowseName=QualifiedName("AuditDeleteReferencesEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2090, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2090, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReferencesToDelete"),
        DataType=NumericNodeId(385, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2098, 0),
        BrowseName=QualifiedName("ReferencesToDelete", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2097, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2097, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditUpdateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2099, 0),
        BrowseName=QualifiedName("AuditUpdateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditWriteUpdateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2100, 0),
        BrowseName=QualifiedName("AuditWriteUpdateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2099, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2750, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2103, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2099, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AttributeId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2750, 0),
        BrowseName=QualifiedName("AttributeId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2100, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2100, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IndexRange"),
        DataType=NumericNodeId(291, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2101, 0),
        BrowseName=QualifiedName("IndexRange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2100, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2100, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldValue"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2102, 0),
        BrowseName=QualifiedName("OldValue", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2100, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2100, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NewValue"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2103, 0),
        BrowseName=QualifiedName("NewValue", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2100, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2100, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryUpdateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2104, 0),
        BrowseName=QualifiedName("AuditHistoryUpdateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2099, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2751, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2099, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ParameterDataTypeId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2751, 0),
        BrowseName=QualifiedName("ParameterDataTypeId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2104, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2104, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditUpdateMethodEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2127, 0),
        BrowseName=QualifiedName("AuditUpdateMethodEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2052, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2128, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2129, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MethodId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2128, 0),
        BrowseName=QualifiedName("MethodId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2129, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SystemEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2130, 0),
        BrowseName=QualifiedName("SystemEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DeviceFailureEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2131, 0),
        BrowseName=QualifiedName("DeviceFailureEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2130, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SystemStatusChangeEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11446, 0),
        BrowseName=QualifiedName("SystemStatusChangeEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2130, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11696, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SystemState"),
        DataType=NumericNodeId(852, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11696, 0),
        BrowseName=QualifiedName("SystemState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11446, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11696, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11696, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11696, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11446, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BaseModelChangeEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2132, 0),
        BrowseName=QualifiedName("BaseModelChangeEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("GeneralModelChangeEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2133, 0),
        BrowseName=QualifiedName("GeneralModelChangeEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2132, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2134, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2132, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Changes"),
        DataType=NumericNodeId(877, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2134, 0),
        BrowseName=QualifiedName("Changes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2133, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2133, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SemanticChangeEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2738, 0),
        BrowseName=QualifiedName("SemanticChangeEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2739, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Changes"),
        DataType=NumericNodeId(897, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2739, 0),
        BrowseName=QualifiedName("Changes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2738, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2738, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EventQueueOverflowEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3035, 0),
        BrowseName=QualifiedName("EventQueueOverflowEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ProgressEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11436, 0),
        BrowseName=QualifiedName("ProgressEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12503, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Context"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12502, 0),
        BrowseName=QualifiedName("Context", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11436, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11436, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Progress"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12503, 0),
        BrowseName=QualifiedName("Progress", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11436, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11436, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AggregateFunctionType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2340, 0),
        BrowseName=QualifiedName("AggregateFunctionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("ServerVendorCapabilityType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2137, 0),
        BrowseName=QualifiedName("ServerVendorCapabilityType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ServerStatusType"),
        DataType=NumericNodeId(862, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2138, 0),
        BrowseName=QualifiedName("ServerStatusType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2139, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2140, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2141, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2142, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2752, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2753, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StartTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2139, 0),
        BrowseName=QualifiedName("StartTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2138, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2140, 0),
        BrowseName=QualifiedName("CurrentTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2138, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(852, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2141, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2138, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BuildInfo"),
        DataType=NumericNodeId(338, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2142, 0),
        BrowseName=QualifiedName("BuildInfo", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2138, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(3051, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3699, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3700, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3701, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3702, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3703, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3698, 0),
        BrowseName=QualifiedName("ProductUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ManufacturerName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3699, 0),
        BrowseName=QualifiedName("ManufacturerName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3700, 0),
        BrowseName=QualifiedName("ProductName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("SoftwareVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3701, 0),
        BrowseName=QualifiedName("SoftwareVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildNumber"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3702, 0),
        BrowseName=QualifiedName("BuildNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildDate"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3703, 0),
        BrowseName=QualifiedName("BuildDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2142, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecondsTillShutdown"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2752, 0),
        BrowseName=QualifiedName("SecondsTillShutdown", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2138, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ShutdownReason"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2753, 0),
        BrowseName=QualifiedName("ShutdownReason", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2138, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("BuildInfoType"),
        DataType=NumericNodeId(338, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3051, 0),
        BrowseName=QualifiedName("BuildInfoType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3052, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3053, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3054, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3057, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3052, 0),
        BrowseName=QualifiedName("ProductUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3051, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ManufacturerName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3053, 0),
        BrowseName=QualifiedName("ManufacturerName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3051, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3054, 0),
        BrowseName=QualifiedName("ProductName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3051, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("SoftwareVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3055, 0),
        BrowseName=QualifiedName("SoftwareVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3051, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildNumber"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3056, 0),
        BrowseName=QualifiedName("BuildNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3051, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildDate"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3057, 0),
        BrowseName=QualifiedName("BuildDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3051, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsSummaryType"),
        DataType=NumericNodeId(859, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2150, 0),
        BrowseName=QualifiedName("ServerDiagnosticsSummaryType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2151, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2152, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2153, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2154, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2155, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2156, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2157, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2159, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2160, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2161, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2162, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2163, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerViewCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2151, 0),
        BrowseName=QualifiedName("ServerViewCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2152, 0),
        BrowseName=QualifiedName("CurrentSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2153, 0),
        BrowseName=QualifiedName("CumulatedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2154, 0),
        BrowseName=QualifiedName("SecurityRejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2155, 0),
        BrowseName=QualifiedName("RejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionTimeoutCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2156, 0),
        BrowseName=QualifiedName("SessionTimeoutCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionAbortCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2157, 0),
        BrowseName=QualifiedName("SessionAbortCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingIntervalCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2159, 0),
        BrowseName=QualifiedName("PublishingIntervalCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2160, 0),
        BrowseName=QualifiedName("CurrentSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2161, 0),
        BrowseName=QualifiedName("CumulatedSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2162, 0),
        BrowseName=QualifiedName("SecurityRejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2163, 0),
        BrowseName=QualifiedName("RejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2150, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnosticsArrayType"),
        DataType=NumericNodeId(856, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2164, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnosticsArrayType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12779, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnostics"),
        DataType=NumericNodeId(856, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12779, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2164, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2165, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12780, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12781, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12782, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12783, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2165, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(83, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2164, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12780, 0),
        BrowseName=QualifiedName("SamplingInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12779, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12779, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SampledMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12781, 0),
        BrowseName=QualifiedName("SampledMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12779, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12779, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxSampledMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12782, 0),
        BrowseName=QualifiedName("MaxSampledMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12779, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12779, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DisabledMonitoredItemsSamplingCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12783, 0),
        BrowseName=QualifiedName("DisabledMonitoredItemsSamplingCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12779, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12779, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnosticsType"),
        DataType=NumericNodeId(856, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2165, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnosticsType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2166, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11697, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11699, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2166, 0),
        BrowseName=QualifiedName("SamplingInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2165, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2165, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SampledMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11697, 0),
        BrowseName=QualifiedName("SampledMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2165, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2165, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxSampledMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11698, 0),
        BrowseName=QualifiedName("MaxSampledMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2165, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2165, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DisabledMonitoredItemsSamplingCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11699, 0),
        BrowseName=QualifiedName("DisabledMonitoredItemsSamplingCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2165, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2165, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsArrayType"),
        DataType=NumericNodeId(874, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2171, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsArrayType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnostics"),
        DataType=NumericNodeId(874, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12784, 0),
        BrowseName=QualifiedName("SubscriptionDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2171, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2172, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12785, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12786, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12787, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12788, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12789, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12790, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12791, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12792, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12793, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12794, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12795, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12796, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12797, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12798, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12799, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12800, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12801, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12802, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12804, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12805, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12806, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12807, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12808, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12809, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12810, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12811, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12812, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12813, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12814, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(83, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2171, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12785, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12786, 0),
        BrowseName=QualifiedName("SubscriptionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Priority"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12787, 0),
        BrowseName=QualifiedName("Priority", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12788, 0),
        BrowseName=QualifiedName("PublishingInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxKeepAliveCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12789, 0),
        BrowseName=QualifiedName("MaxKeepAliveCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxLifetimeCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12790, 0),
        BrowseName=QualifiedName("MaxLifetimeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNotificationsPerPublish"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12791, 0),
        BrowseName=QualifiedName("MaxNotificationsPerPublish", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12791, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12791, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12791, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingEnabled"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12792, 0),
        BrowseName=QualifiedName("PublishingEnabled", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifyCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12793, 0),
        BrowseName=QualifiedName("ModifyCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12793, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12793, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12793, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnableCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12794, 0),
        BrowseName=QualifiedName("EnableCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DisableCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12795, 0),
        BrowseName=QualifiedName("DisableCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12796, 0),
        BrowseName=QualifiedName("RepublishRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishMessageRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12797, 0),
        BrowseName=QualifiedName("RepublishMessageRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12797, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12797, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12797, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishMessageCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12798, 0),
        BrowseName=QualifiedName("RepublishMessageCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12799, 0),
        BrowseName=QualifiedName("TransferRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferredToAltClientCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12800, 0),
        BrowseName=QualifiedName("TransferredToAltClientCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferredToSameClientCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12801, 0),
        BrowseName=QualifiedName("TransferredToSameClientCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12802, 0),
        BrowseName=QualifiedName("PublishRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataChangeNotificationsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12803, 0),
        BrowseName=QualifiedName("DataChangeNotificationsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventNotificationsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12804, 0),
        BrowseName=QualifiedName("EventNotificationsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NotificationsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12805, 0),
        BrowseName=QualifiedName("NotificationsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LatePublishRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12806, 0),
        BrowseName=QualifiedName("LatePublishRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12806, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12806, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12806, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentKeepAliveCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12807, 0),
        BrowseName=QualifiedName("CurrentKeepAliveCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentLifetimeCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12808, 0),
        BrowseName=QualifiedName("CurrentLifetimeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnacknowledgedMessageCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12809, 0),
        BrowseName=QualifiedName("UnacknowledgedMessageCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12809, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12809, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12809, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiscardedMessageCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12810, 0),
        BrowseName=QualifiedName("DiscardedMessageCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12810, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12810, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12810, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MonitoredItemCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12811, 0),
        BrowseName=QualifiedName("MonitoredItemCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DisabledMonitoredItemCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12812, 0),
        BrowseName=QualifiedName("DisabledMonitoredItemCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MonitoringQueueOverflowCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12813, 0),
        BrowseName=QualifiedName("MonitoringQueueOverflowCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NextSequenceNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12814, 0),
        BrowseName=QualifiedName("NextSequenceNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventQueueOverflowCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12815, 0),
        BrowseName=QualifiedName("EventQueueOverflowCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12784, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsType"),
        DataType=NumericNodeId(874, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2172, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2173, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2174, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2176, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2177, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8888, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2179, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2180, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2181, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2182, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2183, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2184, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2185, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2186, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2187, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2188, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2190, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2191, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2998, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2193, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8889, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8890, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8892, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8893, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8895, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8896, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8902, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2173, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2174, 0),
        BrowseName=QualifiedName("SubscriptionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Priority"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2175, 0),
        BrowseName=QualifiedName("Priority", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2176, 0),
        BrowseName=QualifiedName("PublishingInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxKeepAliveCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2177, 0),
        BrowseName=QualifiedName("MaxKeepAliveCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxLifetimeCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8888, 0),
        BrowseName=QualifiedName("MaxLifetimeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNotificationsPerPublish"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2179, 0),
        BrowseName=QualifiedName("MaxNotificationsPerPublish", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingEnabled"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2180, 0),
        BrowseName=QualifiedName("PublishingEnabled", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifyCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2181, 0),
        BrowseName=QualifiedName("ModifyCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnableCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2182, 0),
        BrowseName=QualifiedName("EnableCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DisableCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2183, 0),
        BrowseName=QualifiedName("DisableCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2184, 0),
        BrowseName=QualifiedName("RepublishRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishMessageRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2185, 0),
        BrowseName=QualifiedName("RepublishMessageRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishMessageCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2186, 0),
        BrowseName=QualifiedName("RepublishMessageCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2187, 0),
        BrowseName=QualifiedName("TransferRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferredToAltClientCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2188, 0),
        BrowseName=QualifiedName("TransferredToAltClientCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferredToSameClientCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2189, 0),
        BrowseName=QualifiedName("TransferredToSameClientCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2190, 0),
        BrowseName=QualifiedName("PublishRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataChangeNotificationsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2191, 0),
        BrowseName=QualifiedName("DataChangeNotificationsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventNotificationsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2998, 0),
        BrowseName=QualifiedName("EventNotificationsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NotificationsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2193, 0),
        BrowseName=QualifiedName("NotificationsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LatePublishRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8889, 0),
        BrowseName=QualifiedName("LatePublishRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentKeepAliveCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8890, 0),
        BrowseName=QualifiedName("CurrentKeepAliveCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentLifetimeCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8891, 0),
        BrowseName=QualifiedName("CurrentLifetimeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnacknowledgedMessageCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8892, 0),
        BrowseName=QualifiedName("UnacknowledgedMessageCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiscardedMessageCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8893, 0),
        BrowseName=QualifiedName("DiscardedMessageCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MonitoredItemCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8894, 0),
        BrowseName=QualifiedName("MonitoredItemCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DisabledMonitoredItemCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8895, 0),
        BrowseName=QualifiedName("DisabledMonitoredItemCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MonitoringQueueOverflowCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8896, 0),
        BrowseName=QualifiedName("MonitoringQueueOverflowCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NextSequenceNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8897, 0),
        BrowseName=QualifiedName("NextSequenceNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventQueueOverflowCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8902, 0),
        BrowseName=QualifiedName("EventQueueOverflowCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2172, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2172, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsArrayType"),
        DataType=NumericNodeId(865, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2196, 0),
        BrowseName=QualifiedName("SessionDiagnosticsArrayType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnostics"),
        DataType=NumericNodeId(865, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12816, 0),
        BrowseName=QualifiedName("SessionDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2196, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2197, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12818, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12819, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12820, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12822, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12823, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12825, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12826, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12828, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12832, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12834, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12837, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12838, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12840, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12841, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12843, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12844, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12846, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12847, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12850, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12851, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12854, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12856, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12857, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12858, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12859, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(83, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2196, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12817, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12818, 0),
        BrowseName=QualifiedName("SessionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientDescription"),
        DataType=NumericNodeId(308, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12819, 0),
        BrowseName=QualifiedName("ClientDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12820, 0),
        BrowseName=QualifiedName("ServerUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrl"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12821, 0),
        BrowseName=QualifiedName("EndpointUrl", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIds"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12822, 0),
        BrowseName=QualifiedName("LocaleIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActualSessionTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12823, 0),
        BrowseName=QualifiedName("ActualSessionTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxResponseMessageSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12824, 0),
        BrowseName=QualifiedName("MaxResponseMessageSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientConnectionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12825, 0),
        BrowseName=QualifiedName("ClientConnectionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientLastContactTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12826, 0),
        BrowseName=QualifiedName("ClientLastContactTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12827, 0),
        BrowseName=QualifiedName("CurrentSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12828, 0),
        BrowseName=QualifiedName("CurrentMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentPublishRequestsInQueue"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12829, 0),
        BrowseName=QualifiedName("CurrentPublishRequestsInQueue", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalRequestCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12830, 0),
        BrowseName=QualifiedName("TotalRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnauthorizedRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12831, 0),
        BrowseName=QualifiedName("UnauthorizedRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12832, 0),
        BrowseName=QualifiedName("ReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12833, 0),
        BrowseName=QualifiedName("HistoryReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriteCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12834, 0),
        BrowseName=QualifiedName("WriteCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryUpdateCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12835, 0),
        BrowseName=QualifiedName("HistoryUpdateCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CallCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12836, 0),
        BrowseName=QualifiedName("CallCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12837, 0),
        BrowseName=QualifiedName("CreateMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifyMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12838, 0),
        BrowseName=QualifiedName("ModifyMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetMonitoringModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12839, 0),
        BrowseName=QualifiedName("SetMonitoringModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetTriggeringCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12840, 0),
        BrowseName=QualifiedName("SetTriggeringCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12841, 0),
        BrowseName=QualifiedName("DeleteMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateSubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12842, 0),
        BrowseName=QualifiedName("CreateSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifySubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12843, 0),
        BrowseName=QualifiedName("ModifySubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetPublishingModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12844, 0),
        BrowseName=QualifiedName("SetPublishingModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12845, 0),
        BrowseName=QualifiedName("PublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12846, 0),
        BrowseName=QualifiedName("RepublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12847, 0),
        BrowseName=QualifiedName("TransferSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12848, 0),
        BrowseName=QualifiedName("DeleteSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12849, 0),
        BrowseName=QualifiedName("AddNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12850, 0),
        BrowseName=QualifiedName("AddReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12851, 0),
        BrowseName=QualifiedName("DeleteNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12852, 0),
        BrowseName=QualifiedName("DeleteReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12853, 0),
        BrowseName=QualifiedName("BrowseCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12854, 0),
        BrowseName=QualifiedName("BrowseNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TranslateBrowsePathsToNodeIdsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12855, 0),
        BrowseName=QualifiedName("TranslateBrowsePathsToNodeIdsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryFirstCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12856, 0),
        BrowseName=QualifiedName("QueryFirstCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12857, 0),
        BrowseName=QualifiedName("QueryNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RegisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12858, 0),
        BrowseName=QualifiedName("RegisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnregisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12859, 0),
        BrowseName=QualifiedName("UnregisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12816, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12816, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsVariableType"),
        DataType=NumericNodeId(865, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2197, 0),
        BrowseName=QualifiedName("SessionDiagnosticsVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2198, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2199, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2200, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2201, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2202, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2203, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2204, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3050, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2205, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2206, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2207, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2208, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2209, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8900, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11892, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2218, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2219, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2220, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2221, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2222, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2223, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2224, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2225, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2226, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2227, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2228, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2229, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2230, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2231, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2232, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2233, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2234, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2235, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2236, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2237, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2238, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2239, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2240, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2241, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2242, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2730, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2198, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2199, 0),
        BrowseName=QualifiedName("SessionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientDescription"),
        DataType=NumericNodeId(308, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2200, 0),
        BrowseName=QualifiedName("ClientDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2201, 0),
        BrowseName=QualifiedName("ServerUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrl"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2202, 0),
        BrowseName=QualifiedName("EndpointUrl", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIds"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2203, 0),
        BrowseName=QualifiedName("LocaleIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActualSessionTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2204, 0),
        BrowseName=QualifiedName("ActualSessionTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxResponseMessageSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3050, 0),
        BrowseName=QualifiedName("MaxResponseMessageSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientConnectionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2205, 0),
        BrowseName=QualifiedName("ClientConnectionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientLastContactTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2206, 0),
        BrowseName=QualifiedName("ClientLastContactTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2207, 0),
        BrowseName=QualifiedName("CurrentSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentMonitoredItemsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2208, 0),
        BrowseName=QualifiedName("CurrentMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentPublishRequestsInQueue"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2209, 0),
        BrowseName=QualifiedName("CurrentPublishRequestsInQueue", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalRequestCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8900, 0),
        BrowseName=QualifiedName("TotalRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnauthorizedRequestCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11892, 0),
        BrowseName=QualifiedName("UnauthorizedRequestCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2217, 0),
        BrowseName=QualifiedName("ReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryReadCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2218, 0),
        BrowseName=QualifiedName("HistoryReadCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriteCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2219, 0),
        BrowseName=QualifiedName("WriteCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryUpdateCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2220, 0),
        BrowseName=QualifiedName("HistoryUpdateCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CallCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2221, 0),
        BrowseName=QualifiedName("CallCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2222, 0),
        BrowseName=QualifiedName("CreateMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifyMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2223, 0),
        BrowseName=QualifiedName("ModifyMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetMonitoringModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2224, 0),
        BrowseName=QualifiedName("SetMonitoringModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetTriggeringCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2225, 0),
        BrowseName=QualifiedName("SetTriggeringCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteMonitoredItemsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2226, 0),
        BrowseName=QualifiedName("DeleteMonitoredItemsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateSubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2227, 0),
        BrowseName=QualifiedName("CreateSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2227, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2227, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2227, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModifySubscriptionCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2228, 0),
        BrowseName=QualifiedName("ModifySubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2228, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2228, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2228, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetPublishingModeCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2229, 0),
        BrowseName=QualifiedName("SetPublishingModeCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2230, 0),
        BrowseName=QualifiedName("PublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2230, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2230, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2230, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RepublishCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2231, 0),
        BrowseName=QualifiedName("RepublishCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2231, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2231, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2231, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransferSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2232, 0),
        BrowseName=QualifiedName("TransferSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteSubscriptionsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2233, 0),
        BrowseName=QualifiedName("DeleteSubscriptionsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2233, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2233, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2233, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2234, 0),
        BrowseName=QualifiedName("AddNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2234, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2234, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2234, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2235, 0),
        BrowseName=QualifiedName("AddReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2235, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2235, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2235, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2236, 0),
        BrowseName=QualifiedName("DeleteNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2236, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2236, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2236, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteReferencesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2237, 0),
        BrowseName=QualifiedName("DeleteReferencesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2237, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2237, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2237, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2238, 0),
        BrowseName=QualifiedName("BrowseCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrowseNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2239, 0),
        BrowseName=QualifiedName("BrowseNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2239, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2239, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2239, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TranslateBrowsePathsToNodeIdsCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2240, 0),
        BrowseName=QualifiedName("TranslateBrowsePathsToNodeIdsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2240, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2240, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2240, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryFirstCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2241, 0),
        BrowseName=QualifiedName("QueryFirstCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueryNextCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2242, 0),
        BrowseName=QualifiedName("QueryNextCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RegisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2730, 0),
        BrowseName=QualifiedName("RegisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnregisterNodesCount"),
        DataType=NumericNodeId(871, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2731, 0),
        BrowseName=QualifiedName("UnregisterNodesCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2197, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsArrayType"),
        DataType=NumericNodeId(868, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2243, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsArrayType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnostics"),
        DataType=NumericNodeId(868, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12860, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnostics", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2243, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2244, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12861, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12862, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12863, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12864, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12865, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12866, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12867, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12868, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12869, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(83, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2243, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12861, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdOfSession"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12862, 0),
        BrowseName=QualifiedName("ClientUserIdOfSession", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdHistory"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12863, 0),
        BrowseName=QualifiedName("ClientUserIdHistory", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationMechanism"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12864, 0),
        BrowseName=QualifiedName("AuthenticationMechanism", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Encoding"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12865, 0),
        BrowseName=QualifiedName("Encoding", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransportProtocol"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12866, 0),
        BrowseName=QualifiedName("TransportProtocol", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12867, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityPolicyUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12868, 0),
        BrowseName=QualifiedName("SecurityPolicyUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12869, 0),
        BrowseName=QualifiedName("ClientCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12860, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsType"),
        DataType=NumericNodeId(868, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2244, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2245, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2246, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2247, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2248, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2249, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2250, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2251, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2252, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3058, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2245, 0),
        BrowseName=QualifiedName("SessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdOfSession"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2246, 0),
        BrowseName=QualifiedName("ClientUserIdOfSession", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserIdHistory"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2247, 0),
        BrowseName=QualifiedName("ClientUserIdHistory", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationMechanism"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2248, 0),
        BrowseName=QualifiedName("AuthenticationMechanism", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Encoding"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2249, 0),
        BrowseName=QualifiedName("Encoding", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransportProtocol"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2250, 0),
        BrowseName=QualifiedName("TransportProtocol", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2251, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityPolicyUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2252, 0),
        BrowseName=QualifiedName("SecurityPolicyUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3058, 0),
        BrowseName=QualifiedName("ClientCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2244, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2244, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("OptionSetType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11487, 0),
        BrowseName=QualifiedName("OptionSetType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11488, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11701, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11488, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11487, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11487, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BitMask"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11701, 0),
        BrowseName=QualifiedName("BitMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11487, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11487, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("SelectionListType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16309, 0),
        BrowseName=QualifiedName("SelectionListType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17632, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17633, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16312, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(16309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Selections"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17632, 0),
        BrowseName=QualifiedName("Selections", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16309, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SelectionDescriptions"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17633, 0),
        BrowseName=QualifiedName("SelectionDescriptions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16309, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RestrictToList"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16312, 0),
        BrowseName=QualifiedName("RestrictToList", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16309, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("AudioVariableType"),
        DataType=NumericNodeId(16307, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17986, 0),
        BrowseName=QualifiedName("AudioVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17986, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17988, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17986, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17989, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17986, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17990, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17986, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ListId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17988, 0),
        BrowseName=QualifiedName("ListId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17986, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17988, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17988, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17988, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17986, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AgencyId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17989, 0),
        BrowseName=QualifiedName("AgencyId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17986, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17989, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17989, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17989, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17986, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("VersionId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17990, 0),
        BrowseName=QualifiedName("VersionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17986, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17990, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17990, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17990, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17986, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("EventTypes"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3048, 0),
        BrowseName=QualifiedName("EventTypes", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(86, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(3048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(86, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(3048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Server"),
        EventNotifier=1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2253, 0),
        BrowseName=QualifiedName("Server", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(85, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(2004, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2254, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2255, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2256, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2267, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2994, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12885, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17634, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2274, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2295, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2296, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11715, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11492, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12873, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12749, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12886, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(85, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2004, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ServerArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2254, 0),
        BrowseName=QualifiedName("ServerArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("NamespaceArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2255, 0),
        BrowseName=QualifiedName("NamespaceArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ServerStatus"),
        DataType=NumericNodeId(862, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2256, 0),
        BrowseName=QualifiedName("ServerStatus", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2138, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2257, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2258, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2259, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2260, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2992, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2993, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2138, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StartTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2257, 0),
        BrowseName=QualifiedName("StartTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2256, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2258, 0),
        BrowseName=QualifiedName("CurrentTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2256, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(852, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2259, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2256, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BuildInfo"),
        DataType=NumericNodeId(338, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2260, 0),
        BrowseName=QualifiedName("BuildInfo", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2256, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(3051, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2262, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2263, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2261, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2264, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2265, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2266, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3051, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2262, 0),
        BrowseName=QualifiedName("ProductUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2260, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2260, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ManufacturerName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2263, 0),
        BrowseName=QualifiedName("ManufacturerName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2260, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2260, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ProductName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2261, 0),
        BrowseName=QualifiedName("ProductName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2260, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2260, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("SoftwareVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2264, 0),
        BrowseName=QualifiedName("SoftwareVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2260, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2260, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildNumber"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2265, 0),
        BrowseName=QualifiedName("BuildNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2260, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2260, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("BuildDate"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2266, 0),
        BrowseName=QualifiedName("BuildDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2260, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2260, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecondsTillShutdown"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2992, 0),
        BrowseName=QualifiedName("SecondsTillShutdown", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2256, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ShutdownReason"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2993, 0),
        BrowseName=QualifiedName("ShutdownReason", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2256, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("ServiceLevel"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2267, 0),
        BrowseName=QualifiedName("ServiceLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("Auditing"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2994, 0),
        BrowseName=QualifiedName("Auditing", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2994, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2994, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("EstimatedReturnTime"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12885, 0),
        BrowseName=QualifiedName("EstimatedReturnTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        MinimumSamplingInterval=1000,
        DisplayName=LocalizedText("LocalTime"),
        DataType=NumericNodeId(8912, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17634, 0),
        BrowseName=QualifiedName("LocalTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ServerCapabilities"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2268, 0),
        BrowseName=QualifiedName("ServerCapabilities", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2013, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2269, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2271, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2272, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2735, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2736, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3704, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11702, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11703, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12911, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2996, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2997, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2013, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerProfileArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2269, 0),
        BrowseName=QualifiedName("ServerProfileArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIdArray"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2271, 0),
        BrowseName=QualifiedName("LocaleIdArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MinSupportedSampleRate"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2272, 0),
        BrowseName=QualifiedName("MinSupportedSampleRate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxBrowseContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2735, 0),
        BrowseName=QualifiedName("MaxBrowseContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxQueryContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2736, 0),
        BrowseName=QualifiedName("MaxQueryContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxHistoryContinuationPoints"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2737, 0),
        BrowseName=QualifiedName("MaxHistoryContinuationPoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SoftwareCertificates"),
        DataType=NumericNodeId(344, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3704, 0),
        BrowseName=QualifiedName("SoftwareCertificates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxArrayLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11702, 0),
        BrowseName=QualifiedName("MaxArrayLength", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxStringLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11703, 0),
        BrowseName=QualifiedName("MaxStringLength", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxByteStringLength"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12911, 0),
        BrowseName=QualifiedName("MaxByteStringLength", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12911, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12911, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("OperationLimits"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11704, 0),
        BrowseName=QualifiedName("OperationLimits", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11564, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11705, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12165, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12166, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11707, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12167, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12168, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11709, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11711, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11712, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11713, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11714, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerRead"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11705, 0),
        BrowseName=QualifiedName("MaxNodesPerRead", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryReadData"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12165, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryReadData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryReadEvents"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12166, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryReadEvents", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerWrite"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11707, 0),
        BrowseName=QualifiedName("MaxNodesPerWrite", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryUpdateData"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12167, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryUpdateData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerHistoryUpdateEvents"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12168, 0),
        BrowseName=QualifiedName("MaxNodesPerHistoryUpdateEvents", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerMethodCall"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11709, 0),
        BrowseName=QualifiedName("MaxNodesPerMethodCall", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerBrowse"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11710, 0),
        BrowseName=QualifiedName("MaxNodesPerBrowse", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerRegisterNodes"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11711, 0),
        BrowseName=QualifiedName("MaxNodesPerRegisterNodes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerTranslateBrowsePathsToNodeIds"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11712, 0),
        BrowseName=QualifiedName("MaxNodesPerTranslateBrowsePathsToNodeIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNodesPerNodeManagement"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11713, 0),
        BrowseName=QualifiedName("MaxNodesPerNodeManagement", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxMonitoredItemsPerCall"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11714, 0),
        BrowseName=QualifiedName("MaxMonitoredItemsPerCall", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ModellingRules"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2996, 0),
        BrowseName=QualifiedName("ModellingRules", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateFunctions"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2997, 0),
        BrowseName=QualifiedName("AggregateFunctions", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("RoleSet"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15606, 0),
        BrowseName=QualifiedName("RoleSet", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15607, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16301, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16304, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddRole"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16301, 0),
        BrowseName=QualifiedName("AddRole", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16301, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16302, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16301, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16303, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16301, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='NamespaceUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16302, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16301, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16302, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16302, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16301, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16303, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16301, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16301, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveRole"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16304, 0),
        BrowseName=QualifiedName("RemoveRole", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16304, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16305, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16304, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16305, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16304, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16305, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16305, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16304, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ServerDiagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2274, 0),
        BrowseName=QualifiedName("ServerDiagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2020, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2289, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2290, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3706, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2294, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsSummary"),
        DataType=NumericNodeId(859, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2275, 0),
        BrowseName=QualifiedName("ServerDiagnosticsSummary", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2274, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2150, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2276, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2277, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2278, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2279, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3705, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2281, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2282, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2284, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2285, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2286, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2287, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2288, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2150, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2274, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerViewCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2276, 0),
        BrowseName=QualifiedName("ServerViewCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2277, 0),
        BrowseName=QualifiedName("CurrentSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2278, 0),
        BrowseName=QualifiedName("CumulatedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2279, 0),
        BrowseName=QualifiedName("SecurityRejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedSessionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3705, 0),
        BrowseName=QualifiedName("RejectedSessionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionTimeoutCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2281, 0),
        BrowseName=QualifiedName("SessionTimeoutCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionAbortCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2282, 0),
        BrowseName=QualifiedName("SessionAbortCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingIntervalCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2284, 0),
        BrowseName=QualifiedName("PublishingIntervalCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2285, 0),
        BrowseName=QualifiedName("CurrentSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CumulatedSubscriptionCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2286, 0),
        BrowseName=QualifiedName("CumulatedSubscriptionCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityRejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2287, 0),
        BrowseName=QualifiedName("SecurityRejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RejectedRequestsCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2288, 0),
        BrowseName=QualifiedName("RejectedRequestsCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2275, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2275, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnosticsArray"),
        DataType=NumericNodeId(856, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2289, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2274, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2164, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2164, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2274, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsArray"),
        DataType=NumericNodeId(874, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2290, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2274, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2171, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2290, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2290, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2274, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SessionsDiagnosticsSummary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3706, 0),
        BrowseName=QualifiedName("SessionsDiagnosticsSummary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2274, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2026, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3707, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3708, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2026, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2274, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsArray"),
        DataType=NumericNodeId(865, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3707, 0),
        BrowseName=QualifiedName("SessionDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3706, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2196, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2196, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3706, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsArray"),
        DataType=NumericNodeId(868, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3708, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsArray", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3706, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2243, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2243, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3706, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledFlag"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        AccessLevel=3,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2294, 0),
        BrowseName=QualifiedName("EnabledFlag", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2274, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2274, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("VendorServerInfo"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2295, 0),
        BrowseName=QualifiedName("VendorServerInfo", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2033, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2033, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ServerRedundancy"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2296, 0),
        BrowseName=QualifiedName("ServerRedundancy", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2034, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3709, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2034, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RedundancySupport"),
        DataType=NumericNodeId(851, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3709, 0),
        BrowseName=QualifiedName("RedundancySupport", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2296, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2296, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Namespaces"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11715, 0),
        BrowseName=QualifiedName("Namespaces", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11645, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11645, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetMonitoredItems"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11492, 0),
        BrowseName=QualifiedName("GetMonitoredItems", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11493, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11494, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11493, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11492, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11492, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ServerHandles',
                DataType=NumericNodeId(7, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='ClientHandles',
                DataType=NumericNodeId(7, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11494, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11492, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11492, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ResendData"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12873, 0),
        BrowseName=QualifiedName("ResendData", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12874, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12874, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12873, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12873, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetSubscriptionDurable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12749, 0),
        BrowseName=QualifiedName("SetSubscriptionDurable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12750, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12751, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='LifetimeInHours',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12750, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12749, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12749, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RevisedLifetimeInHours',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12751, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12749, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12749, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RequestServerStateChange"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12886, 0),
        BrowseName=QualifiedName("RequestServerStateChange", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12887, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='State',
                DataType=NumericNodeId(852, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='EstimatedReturnTime',
                DataType=NumericNodeId(13, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecondsTillShutdown',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Reason',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Restart',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12887, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12886, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12886, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Server_ServerRedundancy_CurrentServerId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11312, 0),
        BrowseName=QualifiedName("Server_ServerRedundancy_CurrentServerId", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Server_ServerRedundancy_RedundantServerArray"),
        DataType=NumericNodeId(853, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11313, 0),
        BrowseName=QualifiedName("Server_ServerRedundancy_RedundantServerArray", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11313, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Server_ServerRedundancy_ServerUriArray"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11314, 0),
        BrowseName=QualifiedName("Server_ServerRedundancy_ServerUriArray", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Server_ServerRedundancy_ServerNetworkGroups"),
        DataType=NumericNodeId(11944, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14415, 0),
        BrowseName=QualifiedName("Server_ServerRedundancy_ServerNetworkGroups", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("HistoryServerCapabilities"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11192, 0),
        BrowseName=QualifiedName("HistoryServerCapabilities", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2268, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2330, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11193, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11242, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11274, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11196, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11197, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11198, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11199, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11200, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11281, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11282, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11283, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11275, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11201, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19091, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2268, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AccessHistoryDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11193, 0),
        BrowseName=QualifiedName("AccessHistoryDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AccessHistoryEventsCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11242, 0),
        BrowseName=QualifiedName("AccessHistoryEventsCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxReturnDataValues"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11273, 0),
        BrowseName=QualifiedName("MaxReturnDataValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxReturnEventValues"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11274, 0),
        BrowseName=QualifiedName("MaxReturnEventValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InsertDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11196, 0),
        BrowseName=QualifiedName("InsertDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReplaceDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11197, 0),
        BrowseName=QualifiedName("ReplaceDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdateDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11198, 0),
        BrowseName=QualifiedName("UpdateDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteRawCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11199, 0),
        BrowseName=QualifiedName("DeleteRawCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteAtTimeCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11200, 0),
        BrowseName=QualifiedName("DeleteAtTimeCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InsertEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11281, 0),
        BrowseName=QualifiedName("InsertEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReplaceEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11282, 0),
        BrowseName=QualifiedName("ReplaceEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdateEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11283, 0),
        BrowseName=QualifiedName("UpdateEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11502, 0),
        BrowseName=QualifiedName("DeleteEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InsertAnnotationCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11275, 0),
        BrowseName=QualifiedName("InsertAnnotationCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateFunctions"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11201, 0),
        BrowseName=QualifiedName("AggregateFunctions", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerTimestampSupported"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19091, 0),
        BrowseName=QualifiedName("ServerTimestampSupported", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11192, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11192, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BitFieldMaskDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11737, 0),
        BrowseName=QualifiedName("BitFieldMaskDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(9, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("KeyValuePair"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14533, 0),
        BrowseName=QualifiedName("KeyValuePair", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14533, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AdditionalParametersType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16313, 0),
        BrowseName=QualifiedName("AdditionalParametersType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(16313, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EphemeralKeyType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17548, 0),
        BrowseName=QualifiedName("EphemeralKeyType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EndpointType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15528, 0),
        BrowseName=QualifiedName("EndpointType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15528, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("StateMachineType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2299, 0),
        BrowseName=QualifiedName("StateMachineType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2769, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2770, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2769, 0),
        BrowseName=QualifiedName("CurrentState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2299, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2755, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3720, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2299, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3720, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2769, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2769, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransition"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2770, 0),
        BrowseName=QualifiedName("LastTransition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2299, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2762, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3724, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2299, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3724, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2770, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2770, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("StateVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2755, 0),
        BrowseName=QualifiedName("StateVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2756, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2757, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2758, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2759, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2756, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Name"),
        DataType=ua.NodeId(ua.ObjectIds.QualifiedName),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2757, 0),
        BrowseName=QualifiedName("Name", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Number"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2758, 0),
        BrowseName=QualifiedName("Number", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EffectiveDisplayName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2759, 0),
        BrowseName=QualifiedName("EffectiveDisplayName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("TransitionVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2762, 0),
        BrowseName=QualifiedName("TransitionVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2763, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2764, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2765, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2766, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2763, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2762, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Name"),
        DataType=ua.NodeId(ua.ObjectIds.QualifiedName),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2764, 0),
        BrowseName=QualifiedName("Name", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2762, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Number"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2765, 0),
        BrowseName=QualifiedName("Number", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2762, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2766, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2762, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EffectiveTransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11456, 0),
        BrowseName=QualifiedName("EffectiveTransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2762, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("FiniteStateMachineType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2771, 0),
        BrowseName=QualifiedName("FiniteStateMachineType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2299, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2772, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2773, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17636, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2299, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2772, 0),
        BrowseName=QualifiedName("CurrentState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2760, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2760, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3728, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2772, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2772, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransition"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2773, 0),
        BrowseName=QualifiedName("LastTransition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2767, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3732, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3732, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2773, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2773, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AvailableStates"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17635, 0),
        BrowseName=QualifiedName("AvailableStates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AvailableTransitions"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17636, 0),
        BrowseName=QualifiedName("AvailableTransitions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("FiniteStateVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2760, 0),
        BrowseName=QualifiedName("FiniteStateVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2755, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2761, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2761, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2760, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2760, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("FiniteTransitionVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2767, 0),
        BrowseName=QualifiedName("FiniteTransitionVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2762, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2768, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2768, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2767, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2767, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("StateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2307, 0),
        BrowseName=QualifiedName("StateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2308, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2308, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2307, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("InitialStateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2309, 0),
        BrowseName=QualifiedName("InitialStateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2307, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TransitionType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2310, 0),
        BrowseName=QualifiedName("TransitionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2312, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2312, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ChoiceStateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15109, 0),
        BrowseName=QualifiedName("ChoiceStateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2307, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasGuard"),
        InverseName=LocalizedText("GuardOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15112, 0),
        BrowseName=QualifiedName("HasGuard", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("GuardVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15113, 0),
        BrowseName=QualifiedName("GuardVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ExpressionGuardVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15128, 0),
        BrowseName=QualifiedName("ExpressionGuardVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(15113, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15129, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15113, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Expression"),
        DataType=NumericNodeId(586, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15129, 0),
        BrowseName=QualifiedName("Expression", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15128, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15128, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ElseGuardVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15317, 0),
        BrowseName=QualifiedName("ElseGuardVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(15113, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15317, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15113, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("RationalNumberType"),
        DataType=NumericNodeId(18806, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17709, 0),
        BrowseName=QualifiedName("RationalNumberType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17712, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17713, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Numerator"),
        DataType=ua.NodeId(ua.ObjectIds.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17712, 0),
        BrowseName=QualifiedName("Numerator", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17709, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17709, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Denominator"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17713, 0),
        BrowseName=QualifiedName("Denominator", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17709, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17709, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("VectorType"),
        DataType=NumericNodeId(18807, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17714, 0),
        BrowseName=QualifiedName("VectorType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17715, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("VectorUnit"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17715, 0),
        BrowseName=QualifiedName("VectorUnit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17714, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17714, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("3DVectorType"),
        DataType=NumericNodeId(18808, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17716, 0),
        BrowseName=QualifiedName("3DVectorType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(17714, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18769, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18770, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18771, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17714, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("X"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18769, 0),
        BrowseName=QualifiedName("X", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Y"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18770, 0),
        BrowseName=QualifiedName("Y", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Z"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18771, 0),
        BrowseName=QualifiedName("Z", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("CartesianCoordinatesType"),
        DataType=NumericNodeId(18809, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18772, 0),
        BrowseName=QualifiedName("CartesianCoordinatesType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18773, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LengthUnit"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18773, 0),
        BrowseName=QualifiedName("LengthUnit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18772, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18772, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("3DCartesianCoordinatesType"),
        DataType=NumericNodeId(18810, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18774, 0),
        BrowseName=QualifiedName("3DCartesianCoordinatesType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(18772, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18776, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18777, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18778, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18772, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("X"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18776, 0),
        BrowseName=QualifiedName("X", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18774, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18774, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Y"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18777, 0),
        BrowseName=QualifiedName("Y", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18774, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18774, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Z"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18778, 0),
        BrowseName=QualifiedName("Z", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18774, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18774, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("OrientationType"),
        DataType=NumericNodeId(18811, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18779, 0),
        BrowseName=QualifiedName("OrientationType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18780, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AngleUnit"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18780, 0),
        BrowseName=QualifiedName("AngleUnit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18779, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18779, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("3DOrientationType"),
        DataType=NumericNodeId(18812, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18781, 0),
        BrowseName=QualifiedName("3DOrientationType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(18779, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18783, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18784, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18785, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18779, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("A"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18783, 0),
        BrowseName=QualifiedName("A", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18781, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18781, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("B"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18784, 0),
        BrowseName=QualifiedName("B", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18781, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18781, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("C"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18785, 0),
        BrowseName=QualifiedName("C", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18781, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18781, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("FrameType"),
        DataType=NumericNodeId(18813, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18786, 0),
        BrowseName=QualifiedName("FrameType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18801, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18787, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18788, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18789, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18790, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CartesianCoordinates"),
        DataType=NumericNodeId(18809, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18801, 0),
        BrowseName=QualifiedName("CartesianCoordinates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18786, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(18772, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18772, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18786, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Orientation"),
        DataType=NumericNodeId(18811, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18787, 0),
        BrowseName=QualifiedName("Orientation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18786, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(18779, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18779, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18786, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Constant"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18788, 0),
        BrowseName=QualifiedName("Constant", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18786, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18786, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BaseFrame"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18789, 0),
        BrowseName=QualifiedName("BaseFrame", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18786, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18786, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FixedBase"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18790, 0),
        BrowseName=QualifiedName("FixedBase", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18786, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18786, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("3DFrameType"),
        DataType=NumericNodeId(18814, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18791, 0),
        BrowseName=QualifiedName("3DFrameType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(18786, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18791, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18796, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18791, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18792, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18791, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18786, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CartesianCoordinates"),
        DataType=NumericNodeId(18810, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18796, 0),
        BrowseName=QualifiedName("CartesianCoordinates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18791, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(18774, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18798, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18799, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18800, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18774, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18791, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("X"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18798, 0),
        BrowseName=QualifiedName("X", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18796, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18796, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Y"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18799, 0),
        BrowseName=QualifiedName("Y", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18796, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18796, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Z"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18800, 0),
        BrowseName=QualifiedName("Z", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18796, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18796, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Orientation"),
        DataType=NumericNodeId(18812, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18792, 0),
        BrowseName=QualifiedName("Orientation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18791, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(18781, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19074, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19075, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18781, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18791, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("A"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19074, 0),
        BrowseName=QualifiedName("A", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18792, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18792, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("B"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19075, 0),
        BrowseName=QualifiedName("B", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18792, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18792, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("C"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19076, 0),
        BrowseName=QualifiedName("C", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18792, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18792, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RationalNumber"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18806, 0),
        BrowseName=QualifiedName("RationalNumber", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18806, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Vector"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18807, 0),
        BrowseName=QualifiedName("Vector", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("3DVector"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18808, 0),
        BrowseName=QualifiedName("3DVector", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(18807, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18807, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CartesianCoordinates"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18809, 0),
        BrowseName=QualifiedName("CartesianCoordinates", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18809, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("3DCartesianCoordinates"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18810, 0),
        BrowseName=QualifiedName("3DCartesianCoordinates", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(18809, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18810, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18809, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Orientation"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18811, 0),
        BrowseName=QualifiedName("Orientation", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("3DOrientation"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18812, 0),
        BrowseName=QualifiedName("3DOrientation", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(18811, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18811, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Frame"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18813, 0),
        BrowseName=QualifiedName("Frame", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("3DFrame"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18814, 0),
        BrowseName=QualifiedName("3DFrame", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(18813, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18813, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TransitionEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2311, 0),
        BrowseName=QualifiedName("TransitionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2774, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2775, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2776, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Transition"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2774, 0),
        BrowseName=QualifiedName("Transition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2311, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2762, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3754, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2762, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3754, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2774, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2774, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FromState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2775, 0),
        BrowseName=QualifiedName("FromState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2311, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2755, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3746, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2775, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2775, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ToState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2776, 0),
        BrowseName=QualifiedName("ToState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2311, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2755, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3750, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3750, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2776, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2776, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditUpdateStateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2315, 0),
        BrowseName=QualifiedName("AuditUpdateStateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2777, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2778, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldStateId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2777, 0),
        BrowseName=QualifiedName("OldStateId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2315, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2315, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NewStateId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2778, 0),
        BrowseName=QualifiedName("NewStateId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2315, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2315, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("OpenFileMode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11939, 0),
        BrowseName=QualifiedName("OpenFileMode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='Read'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='Write'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='EraseExisting'),
                ),
            ua.EnumValueType(
                Value=8,
                DisplayName=LocalizedText(Text='Append'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11940, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11939, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11939, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("FileDirectoryType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13353, 0),
        BrowseName=QualifiedName("FileDirectoryType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(13353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13354, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(13353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13387, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13390, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13393, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13395, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(13353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<FileDirectoryName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13354, 0),
        BrowseName=QualifiedName("<FileDirectoryName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13353, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(13353, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13355, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13358, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17718, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13363, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(13354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateDirectory"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13355, 0),
        BrowseName=QualifiedName("CreateDirectory", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13354, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13356, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13357, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13354, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DirectoryName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13356, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13355, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13355, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DirectoryNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13357, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13355, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13355, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateFile"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13358, 0),
        BrowseName=QualifiedName("CreateFile", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13354, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13359, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13360, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13354, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RequestFileOpen',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13359, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13358, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13358, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13360, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13358, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13358, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Delete"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17718, 0),
        BrowseName=QualifiedName("Delete", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13354, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17719, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13354, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ObjectToDelete',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17719, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17718, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17718, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("MoveOrCopy"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13363, 0),
        BrowseName=QualifiedName("MoveOrCopy", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13354, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13364, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13365, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13354, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ObjectToMoveOrCopy',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='TargetDirectory',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CreateCopy',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='NewName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13364, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13363, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13363, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='NewNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13365, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13363, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13363, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<FileName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13366, 0),
        BrowseName=QualifiedName("<FileName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13353, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(11575, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13367, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13368, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13369, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13370, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13372, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13375, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13377, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13380, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13382, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13385, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(13366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13367, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13368, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13369, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13370, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13372, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13373, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13374, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13373, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13372, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13372, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13374, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13372, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13372, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13375, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13376, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13376, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13375, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13375, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13377, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13379, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13378, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13377, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13377, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13379, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13377, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13377, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13380, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13381, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13381, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13382, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13383, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13384, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13383, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13382, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13382, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13384, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13382, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13382, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13385, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13366, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13386, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13386, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13385, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13385, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateDirectory"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13387, 0),
        BrowseName=QualifiedName("CreateDirectory", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13353, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13388, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DirectoryName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13388, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13387, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13387, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DirectoryNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13389, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13387, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13387, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateFile"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13390, 0),
        BrowseName=QualifiedName("CreateFile", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13353, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13391, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13392, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RequestFileOpen',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13391, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13390, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13390, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13392, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13390, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13390, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Delete"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13393, 0),
        BrowseName=QualifiedName("Delete", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13353, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13394, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ObjectToDelete',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13394, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13393, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13393, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("MoveOrCopy"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13395, 0),
        BrowseName=QualifiedName("MoveOrCopy", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13353, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13396, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13397, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ObjectToMoveOrCopy',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='TargetDirectory',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CreateCopy',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='NewName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13396, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13395, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13395, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='NewNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13397, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13395, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13395, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("FileSystem"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16314, 0),
        BrowseName=QualifiedName("FileSystem", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(13353, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16348, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16351, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16354, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16356, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13353, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateDirectory"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16348, 0),
        BrowseName=QualifiedName("CreateDirectory", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16314, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16349, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16350, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16314, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DirectoryName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16349, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16348, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16349, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16349, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16348, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DirectoryNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16350, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16348, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16350, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16350, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16348, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateFile"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16351, 0),
        BrowseName=QualifiedName("CreateFile", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16314, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16352, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16353, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16314, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RequestFileOpen',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16352, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16351, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16352, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16352, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16351, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16353, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16351, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16351, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Delete"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16354, 0),
        BrowseName=QualifiedName("Delete", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16314, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16355, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16314, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ObjectToDelete',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16355, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16354, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16354, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("MoveOrCopy"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16356, 0),
        BrowseName=QualifiedName("MoveOrCopy", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16314, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16357, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16358, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16314, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ObjectToMoveOrCopy',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='TargetDirectory',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CreateCopy',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='NewName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16357, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16356, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16356, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='NewNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16358, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16356, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16356, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TemporaryFileTransferType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15744, 0),
        BrowseName=QualifiedName("TemporaryFileTransferType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15745, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15749, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15751, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15754, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientProcessingTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15745, 0),
        BrowseName=QualifiedName("ClientProcessingTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15744, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15744, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GenerateFileForRead"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15746, 0),
        BrowseName=QualifiedName("GenerateFileForRead", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15744, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15747, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15748, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15744, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='GenerateOptions',
                DataType=NumericNodeId(24, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15747, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15746, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15746, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CompletionStateMachine',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15748, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15746, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15746, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GenerateFileForWrite"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15749, 0),
        BrowseName=QualifiedName("GenerateFileForWrite", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15744, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16359, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15750, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15744, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='GenerateOptions',
                DataType=NumericNodeId(24, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16359, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15749, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15749, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15750, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15749, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15749, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CloseAndCommit"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15751, 0),
        BrowseName=QualifiedName("CloseAndCommit", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15744, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15752, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15753, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15744, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15752, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15751, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15751, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CompletionStateMachine',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15753, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15751, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15751, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<TransferState>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15754, 0),
        BrowseName=QualifiedName("<TransferState>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15744, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15803, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15794, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15744, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15755, 0),
        BrowseName=QualifiedName("CurrentState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15754, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2760, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15756, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2760, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15754, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15756, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15794, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15754, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15754, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("FileTransferStateMachineType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15803, 0),
        BrowseName=QualifiedName("FileTransferStateMachineType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15819, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15823, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15825, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15837, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15841, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15843, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Idle"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15815, 0),
        BrowseName=QualifiedName("Idle", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2309, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15816, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15825, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15841, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(1, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15816, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15815, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadPrepare"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15817, 0),
        BrowseName=QualifiedName("ReadPrepare", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15818, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15825, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(2, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15818, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15817, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15817, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadTransfer"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15819, 0),
        BrowseName=QualifiedName("ReadTransfer", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15820, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15837, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(3, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15820, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15819, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15819, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ApplyWrite"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15821, 0),
        BrowseName=QualifiedName("ApplyWrite", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15822, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(4, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15822, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15821, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15821, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Error"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15823, 0),
        BrowseName=QualifiedName("Error", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15837, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15841, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(5, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15824, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15823, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15823, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("IdleToReadPrepare"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15825, 0),
        BrowseName=QualifiedName("IdleToReadPrepare", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15826, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(12, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15826, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15825, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15825, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadPrepareToReadTransfer"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15827, 0),
        BrowseName=QualifiedName("ReadPrepareToReadTransfer", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15828, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15819, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(23, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15828, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15827, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15827, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadTransferToIdle"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15829, 0),
        BrowseName=QualifiedName("ReadTransferToIdle", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15819, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(31, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15830, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15829, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15829, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("IdleToApplyWrite"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15831, 0),
        BrowseName=QualifiedName("IdleToApplyWrite", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15832, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(14, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15832, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15831, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15831, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ApplyWriteToIdle"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15833, 0),
        BrowseName=QualifiedName("ApplyWriteToIdle", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15834, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(41, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15834, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15833, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15833, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadPrepareToError"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15835, 0),
        BrowseName=QualifiedName("ReadPrepareToError", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15823, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(25, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15836, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15835, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15835, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadTransferToError"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15837, 0),
        BrowseName=QualifiedName("ReadTransferToError", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15838, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15819, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15823, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(35, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15838, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15837, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15837, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ApplyWriteToError"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15839, 0),
        BrowseName=QualifiedName("ApplyWriteToError", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15840, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15823, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(45, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15840, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15839, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15839, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ErrorToIdle"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15841, 0),
        BrowseName=QualifiedName("ErrorToIdle", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(15841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15823, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(15841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(15841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(51, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15842, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15841, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15841, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15843, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15803, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15803, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RoleSetType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15607, 0),
        BrowseName=QualifiedName("RoleSetType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15608, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15997, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16000, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<RoleName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15608, 0),
        BrowseName=QualifiedName("<RoleName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15607, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16162, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15607, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16162, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15608, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15608, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddRole"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15997, 0),
        BrowseName=QualifiedName("AddRole", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15607, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15998, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15999, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15607, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='NamespaceUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15998, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15997, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15997, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15999, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15997, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15997, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveRole"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16000, 0),
        BrowseName=QualifiedName("RemoveRole", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15607, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16001, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15607, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RoleNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16001, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16000, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16000, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RoleType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15620, 0),
        BrowseName=QualifiedName("RoleType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16173, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16174, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15410, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15411, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15624, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15626, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16176, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16180, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16182, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16173, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16174, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15410, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16175, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15411, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15624, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15625, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15625, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15626, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15626, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15627, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15626, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15626, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15627, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15626, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15626, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16176, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16177, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16177, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16176, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16176, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16178, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16179, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16179, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16178, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16180, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16181, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16181, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16180, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16180, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16182, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15620, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16183, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16183, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16182, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16182, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("IdentityCriteriaType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15632, 0),
        BrowseName=QualifiedName("IdentityCriteriaType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15633, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='UserName'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='Thumbprint'),
                ),
            ua.EnumValueType(
                Value=3,
                DisplayName=LocalizedText(Text='Role'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='GroupId'),
                ),
            ua.EnumValueType(
                Value=5,
                DisplayName=LocalizedText(Text='Anonymous'),
                ),
            ua.EnumValueType(
                Value=6,
                DisplayName=LocalizedText(Text='AuthenticatedUser'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15633, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15632, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15632, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("IdentityMappingRuleType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15634, 0),
        BrowseName=QualifiedName("IdentityMappingRuleType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RoleMappingRuleChangedAuditEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17641, 0),
        BrowseName=QualifiedName("RoleMappingRuleChangedAuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Anonymous"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15644, 0),
        BrowseName=QualifiedName("Anonymous", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16192, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16193, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15412, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16194, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15413, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15648, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16195, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16197, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16199, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16201, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16192, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16193, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15412, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16194, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16194, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16194, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15413, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15413, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15413, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15648, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15649, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15649, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15648, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15648, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15650, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15651, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15651, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15650, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15650, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16195, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16196, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16196, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16195, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16195, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16197, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16198, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16198, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16197, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16197, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16199, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16200, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16200, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16199, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16199, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16201, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15644, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16202, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16202, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16201, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16201, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AuthenticatedUser"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15656, 0),
        BrowseName=QualifiedName("AuthenticatedUser", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16203, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16204, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15414, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16205, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15415, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15660, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15662, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16206, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16208, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16210, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16212, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16203, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16204, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15414, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16205, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15415, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15660, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15660, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15661, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15660, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15661, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15660, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15661, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15661, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15660, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15662, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15662, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15663, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15662, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15663, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15662, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15663, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15663, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15662, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16206, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16207, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16207, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16206, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16206, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16208, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16209, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16209, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16208, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16208, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16210, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16210, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16211, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16210, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16211, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16210, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16210, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16212, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15656, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16212, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16213, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16212, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16213, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16212, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16212, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Observer"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15668, 0),
        BrowseName=QualifiedName("Observer", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16214, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16215, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15416, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16216, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15417, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15672, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15674, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16219, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16221, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16223, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16214, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16215, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15416, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16216, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15417, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15672, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15672, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15673, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15672, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15673, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15672, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15673, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15673, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15672, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15674, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15674, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15675, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15674, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15675, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15674, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15675, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15675, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15674, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16217, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16218, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16218, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16217, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16217, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16219, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16220, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16220, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16219, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16219, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16221, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16222, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16222, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16221, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16221, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16223, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15668, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16224, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16224, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16223, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16223, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Operator"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15680, 0),
        BrowseName=QualifiedName("Operator", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16225, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16226, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15418, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16227, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15684, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16228, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16230, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16232, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16234, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16225, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16226, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15418, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16227, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16227, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16227, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15423, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15684, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15685, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15685, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15684, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15685, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15685, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15684, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15686, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15687, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15687, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15686, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15686, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16228, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16228, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16229, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16228, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16229, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16228, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16228, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16230, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16230, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16231, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16230, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16231, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16230, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16231, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16231, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16230, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16232, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16233, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16233, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16232, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16233, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16233, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16234, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15680, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16234, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16235, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16234, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15680, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16235, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16234, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16235, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16235, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16234, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Engineer"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16036, 0),
        BrowseName=QualifiedName("Engineer", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16236, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16237, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15424, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16238, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15425, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16041, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16043, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16239, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16241, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16243, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16245, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16236, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16236, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16236, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16237, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16237, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16237, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15424, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16238, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15425, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16041, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16042, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16042, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16041, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16043, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16044, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16044, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16043, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16044, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16044, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16043, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16239, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16239, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16240, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16239, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16240, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16239, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16240, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16240, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16239, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16241, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16242, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16242, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16241, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16243, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16244, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16244, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16243, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16243, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16245, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16036, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16246, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16036, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16246, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16245, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16245, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Supervisor"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15692, 0),
        BrowseName=QualifiedName("Supervisor", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16247, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16248, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15426, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16249, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15427, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15696, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16250, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16252, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16254, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16256, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16247, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16248, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15426, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16249, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15427, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15696, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15696, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15697, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15696, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15697, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15696, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15696, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15698, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15699, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15699, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15698, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15698, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16250, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16251, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16251, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16250, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16250, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16252, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16253, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16253, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16252, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16254, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16255, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16255, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16254, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16254, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16256, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15692, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16257, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16257, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16256, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ConfigureAdmin"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15716, 0),
        BrowseName=QualifiedName("ConfigureAdmin", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16269, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16270, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15428, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16271, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15720, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15722, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16272, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16274, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16276, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16278, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16269, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16270, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15428, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16271, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15429, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15720, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15721, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15721, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15720, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15720, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15722, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15723, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15723, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15722, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15722, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16272, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16273, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16272, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16272, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16274, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16275, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16275, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16274, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16274, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16276, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16277, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16277, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16276, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16276, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16278, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15716, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16279, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16279, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16278, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16278, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SecurityAdmin"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15704, 0),
        BrowseName=QualifiedName("SecurityAdmin", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15606, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15620, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16258, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16259, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15430, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16260, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15527, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15708, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16261, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16263, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16265, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16267, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Identities"),
        DataType=NumericNodeId(15634, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16258, 0),
        BrowseName=QualifiedName("Identities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Applications"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16259, 0),
        BrowseName=QualifiedName("Applications", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15430, 0),
        BrowseName=QualifiedName("ApplicationsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15430, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15430, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Endpoints"),
        DataType=NumericNodeId(15528, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16260, 0),
        BrowseName=QualifiedName("Endpoints", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointsExclude"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15527, 0),
        BrowseName=QualifiedName("EndpointsExclude", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15708, 0),
        BrowseName=QualifiedName("AddIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15709, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15709, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15708, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15708, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveIdentity"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15710, 0),
        BrowseName=QualifiedName("RemoveIdentity", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15711, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Rule',
                DataType=NumericNodeId(15634, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15711, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15710, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15710, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16261, 0),
        BrowseName=QualifiedName("AddApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16262, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16262, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16261, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16261, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveApplication"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16263, 0),
        BrowseName=QualifiedName("RemoveApplication", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16264, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplicationUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16264, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16263, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16263, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16265, 0),
        BrowseName=QualifiedName("AddEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16266, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16266, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16265, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16265, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveEndpoint"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16267, 0),
        BrowseName=QualifiedName("RemoveEndpoint", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15704, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16268, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15704, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Endpoint',
                DataType=NumericNodeId(15528, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16268, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16267, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16267, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DictionaryEntryType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17589, 0),
        BrowseName=QualifiedName("DictionaryEntryType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17590, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<DictionaryEntryName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17590, 0),
        BrowseName=QualifiedName("<DictionaryEntryName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17589, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17589, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17589, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17589, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DictionaryFolderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17591, 0),
        BrowseName=QualifiedName("DictionaryFolderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17592, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17593, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<DictionaryFolderName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17592, 0),
        BrowseName=QualifiedName("<DictionaryFolderName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17591, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17591, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17591, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17591, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<DictionaryEntryName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17593, 0),
        BrowseName=QualifiedName("<DictionaryEntryName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17591, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17589, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17589, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17591, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Dictionaries"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17594, 0),
        BrowseName=QualifiedName("Dictionaries", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17591, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17591, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasDictionaryEntry"),
        InverseName=LocalizedText("DictionaryEntryOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17597, 0),
        BrowseName=QualifiedName("HasDictionaryEntry", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("IrdiDictionaryEntryType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17598, 0),
        BrowseName=QualifiedName("IrdiDictionaryEntryType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17589, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17589, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("UriDictionaryEntryType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17600, 0),
        BrowseName=QualifiedName("UriDictionaryEntryType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17589, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17589, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BaseInterfaceType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17602, 0),
        BrowseName=QualifiedName("BaseInterfaceType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("InterfaceTypes"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17708, 0),
        BrowseName=QualifiedName("InterfaceTypes", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(86, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(17708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17602, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(17708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(86, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasInterface"),
        InverseName=LocalizedText("InterfaceOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17603, 0),
        BrowseName=QualifiedName("HasInterface", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasAddIn"),
        InverseName=LocalizedText("AddInOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17604, 0),
        BrowseName=QualifiedName("HasAddIn", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17604, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CurrencyUnitType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23498, 0),
        BrowseName=QualifiedName("CurrencyUnitType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23498, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrencyUnit"),
        DataType=NumericNodeId(23498, 0),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23501, 0),
        BrowseName=QualifiedName("CurrencyUnit", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23501, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("IOrderedObjectType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23513, 0),
        BrowseName=QualifiedName("IOrderedObjectType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17602, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23517, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17602, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NumberInList"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23517, 0),
        BrowseName=QualifiedName("NumberInList", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23513, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23513, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("OrderedListType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23518, 0),
        BrowseName=QualifiedName("OrderedListType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(49, 0),
        SourceNodeId=NumericNodeId(23518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23519, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23525, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(41, 0),
        SourceNodeId=NumericNodeId(23518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2133, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<OrderedObject>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23519, 0),
        BrowseName=QualifiedName("<OrderedObject>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23518, 0),
        ReferenceTypeId=NumericNodeId(49, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23521, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(17603, 0),
        SourceNodeId=NumericNodeId(23519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23513, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(49, 0),
        SourceNodeId=NumericNodeId(23519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23518, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NumberInList"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23521, 0),
        BrowseName=QualifiedName("NumberInList", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23519, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23521, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23521, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23521, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23519, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NodeVersion"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23525, 0),
        BrowseName=QualifiedName("NodeVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23518, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23525, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23525, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23525, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23518, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("DataItemType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2365, 0),
        BrowseName=QualifiedName("DataItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2366, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2367, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Definition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2366, 0),
        BrowseName=QualifiedName("Definition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2365, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2365, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ValuePrecision"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2367, 0),
        BrowseName=QualifiedName("ValuePrecision", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2365, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2365, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("BaseAnalogType"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15318, 0),
        BrowseName=QualifiedName("BaseAnalogType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2365, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17568, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17569, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2365, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InstrumentRange"),
        DataType=NumericNodeId(884, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17567, 0),
        BrowseName=QualifiedName("InstrumentRange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EURange"),
        DataType=NumericNodeId(884, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17568, 0),
        BrowseName=QualifiedName("EURange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EngineeringUnits"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17569, 0),
        BrowseName=QualifiedName("EngineeringUnits", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("AnalogItemType"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2368, 0),
        BrowseName=QualifiedName("AnalogItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(15318, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2369, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EURange"),
        DataType=NumericNodeId(884, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2369, 0),
        BrowseName=QualifiedName("EURange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2368, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2368, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("AnalogUnitType"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17497, 0),
        BrowseName=QualifiedName("AnalogUnitType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(15318, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EngineeringUnits"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17502, 0),
        BrowseName=QualifiedName("EngineeringUnits", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17497, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17497, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("AnalogUnitRangeType"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17570, 0),
        BrowseName=QualifiedName("AnalogUnitRangeType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2368, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17575, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2368, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EngineeringUnits"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17575, 0),
        BrowseName=QualifiedName("EngineeringUnits", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17570, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17570, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("DiscreteItemType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2372, 0),
        BrowseName=QualifiedName("DiscreteItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2365, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2365, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("TwoStateDiscreteType"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2373, 0),
        BrowseName=QualifiedName("TwoStateDiscreteType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2372, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2374, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2375, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2372, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2374, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2373, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2373, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2375, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2373, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2373, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("MultiStateDiscreteType"),
        DataType=ua.NodeId(ua.ObjectIds.UInteger),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2376, 0),
        BrowseName=QualifiedName("MultiStateDiscreteType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2372, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2377, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2372, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2377, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2376, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2376, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("MultiStateValueDiscreteType"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11238, 0),
        BrowseName=QualifiedName("MultiStateValueDiscreteType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2372, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11241, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11461, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2372, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11241, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11238, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11238, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ValueAsText"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11461, 0),
        BrowseName=QualifiedName("ValueAsText", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11238, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11238, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        IsAbstract=True,
        DisplayName=LocalizedText("ArrayItemType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12021, 0),
        BrowseName=QualifiedName("ArrayItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2365, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12024, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12025, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12026, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12028, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2365, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InstrumentRange"),
        DataType=NumericNodeId(884, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12024, 0),
        BrowseName=QualifiedName("InstrumentRange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EURange"),
        DataType=NumericNodeId(884, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12025, 0),
        BrowseName=QualifiedName("EURange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EngineeringUnits"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12026, 0),
        BrowseName=QualifiedName("EngineeringUnits", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Title"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12027, 0),
        BrowseName=QualifiedName("Title", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AxisScaleType"),
        DataType=NumericNodeId(12077, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12028, 0),
        BrowseName=QualifiedName("AxisScaleType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("YArrayItemType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12029, 0),
        BrowseName=QualifiedName("YArrayItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12037, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("XAxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12037, 0),
        BrowseName=QualifiedName("XAxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12029, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("XYArrayItemType"),
        DataType=NumericNodeId(12080, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12038, 0),
        BrowseName=QualifiedName("XYArrayItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12046, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("XAxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12046, 0),
        BrowseName=QualifiedName("XAxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12038, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12038, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ImageItemType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=2,
        ArrayDimensions=[0, 0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12047, 0),
        BrowseName=QualifiedName("ImageItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("XAxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12055, 0),
        BrowseName=QualifiedName("XAxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12047, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12047, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("YAxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12056, 0),
        BrowseName=QualifiedName("YAxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12047, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12047, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("CubeItemType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=3,
        ArrayDimensions=[0, 0, 0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12057, 0),
        BrowseName=QualifiedName("CubeItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12065, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12066, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12067, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("XAxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12065, 0),
        BrowseName=QualifiedName("XAxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12057, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12057, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("YAxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12066, 0),
        BrowseName=QualifiedName("YAxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12057, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12057, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ZAxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12067, 0),
        BrowseName=QualifiedName("ZAxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12057, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12057, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("NDimensionArrayItemType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12068, 0),
        BrowseName=QualifiedName("NDimensionArrayItemType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(12021, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AxisDefinition"),
        DataType=NumericNodeId(12079, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12076, 0),
        BrowseName=QualifiedName("AxisDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12068, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12068, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("TwoStateVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8995, 0),
        BrowseName=QualifiedName("TwoStateVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(2755, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8996, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9000, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9001, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11111, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(8995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8996, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8995, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(8996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9000, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8995, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EffectiveTransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9001, 0),
        BrowseName=QualifiedName("EffectiveTransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8995, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11110, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8995, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11111, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8995, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ConditionVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9002, 0),
        BrowseName=QualifiedName("ConditionVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9002, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9003, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9002, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9003, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9002, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasTrueSubState"),
        InverseName=LocalizedText("IsTrueSubStateOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9004, 0),
        BrowseName=QualifiedName("HasTrueSubState", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasFalseSubState"),
        InverseName=LocalizedText("IsFalseSubStateOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9005, 0),
        BrowseName=QualifiedName("HasFalseSubState", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasAlarmSuppressionGroup"),
        InverseName=LocalizedText("IsAlarmSuppressionGroupOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16361, 0),
        BrowseName=QualifiedName("HasAlarmSuppressionGroup", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(16361, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("AlarmGroupMember"),
        InverseName=LocalizedText("MemberOfAlarmGroup"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16362, 0),
        BrowseName=QualifiedName("AlarmGroupMember", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(35, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(16362, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(35, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ConditionType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2782, 0),
        BrowseName=QualifiedName("ConditionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2041, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11112, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16363, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16364, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9009, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9010, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3874, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9011, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9022, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9024, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9026, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9028, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9029, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3875, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12912, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11112, 0),
        BrowseName=QualifiedName("ConditionClassId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11113, 0),
        BrowseName=QualifiedName("ConditionClassName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionSubClassId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16363, 0),
        BrowseName=QualifiedName("ConditionSubClassId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionSubClassName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16364, 0),
        BrowseName=QualifiedName("ConditionSubClassName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9009, 0),
        BrowseName=QualifiedName("ConditionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BranchId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9010, 0),
        BrowseName=QualifiedName("BranchId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Retain"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3874, 0),
        BrowseName=QualifiedName("Retain", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9011, 0),
        BrowseName=QualifiedName("EnabledState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9012, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9015, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9016, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9017, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9018, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9019, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9012, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9011, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EffectiveDisplayName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9015, 0),
        BrowseName=QualifiedName("EffectiveDisplayName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9011, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9016, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9011, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EffectiveTransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9017, 0),
        BrowseName=QualifiedName("EffectiveTransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9011, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Enabled"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9018, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9011, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Disabled"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9019, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9011, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Quality"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9020, 0),
        BrowseName=QualifiedName("Quality", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9021, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9021, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9020, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastSeverity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9022, 0),
        BrowseName=QualifiedName("LastSeverity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9023, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9023, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9022, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9022, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Comment"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9024, 0),
        BrowseName=QualifiedName("Comment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9025, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9025, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9024, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9024, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9026, 0),
        BrowseName=QualifiedName("ClientUserId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Disable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9028, 0),
        BrowseName=QualifiedName("Disable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Enable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9027, 0),
        BrowseName=QualifiedName("Enable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddComment"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9029, 0),
        BrowseName=QualifiedName("AddComment", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9030, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9029, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ConditionRefresh"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3875, 0),
        BrowseName=QualifiedName("ConditionRefresh", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3876, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(3875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2787, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(3875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2788, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the suscription to refresh.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3876, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3875, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3875, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ConditionRefresh2"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12912, 0),
        BrowseName=QualifiedName("ConditionRefresh2", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12913, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(12912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2787, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(12912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2788, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SubscriptionId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the suscription to refresh.'),
                ),
            ua.Argument(
                Name='MonitoredItemId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the monitored item to refresh.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12913, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12912, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12912, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DialogConditionType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2830, 0),
        BrowseName=QualifiedName("DialogConditionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9035, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9064, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9065, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9066, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9067, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9068, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9069, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9035, 0),
        BrowseName=QualifiedName("EnabledState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9036, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9036, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9035, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9035, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DialogState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9055, 0),
        BrowseName=QualifiedName("DialogState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9060, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9062, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9063, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9035, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9056, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9055, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9055, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9060, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9055, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9055, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Active"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9062, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9055, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9055, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Inactive"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9063, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9055, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9055, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Prompt"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2831, 0),
        BrowseName=QualifiedName("Prompt", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResponseOptionSet"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9064, 0),
        BrowseName=QualifiedName("ResponseOptionSet", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DefaultResponse"),
        DataType=ua.NodeId(ua.ObjectIds.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9065, 0),
        BrowseName=QualifiedName("DefaultResponse", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OkResponse"),
        DataType=ua.NodeId(ua.ObjectIds.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9066, 0),
        BrowseName=QualifiedName("OkResponse", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CancelResponse"),
        DataType=ua.NodeId(ua.ObjectIds.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9067, 0),
        BrowseName=QualifiedName("CancelResponse", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastResponse"),
        DataType=ua.NodeId(ua.ObjectIds.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9068, 0),
        BrowseName=QualifiedName("LastResponse", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Respond"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9069, 0),
        BrowseName=QualifiedName("Respond", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2830, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9070, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8927, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SelectedResponse',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The response to the dialog condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9070, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9069, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9069, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AcknowledgeableConditionType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2881, 0),
        BrowseName=QualifiedName("AcknowledgeableConditionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2782, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9111, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9073, 0),
        BrowseName=QualifiedName("EnabledState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2881, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9074, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9074, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9073, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9073, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AckedState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9093, 0),
        BrowseName=QualifiedName("AckedState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2881, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9094, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9093, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9093, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9098, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9093, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9093, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Acknowledged"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9100, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9093, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9093, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Unacknowledged"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9101, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9093, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9093, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfirmedState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9102, 0),
        BrowseName=QualifiedName("ConfirmedState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2881, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9103, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9109, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9103, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9102, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9102, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9107, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9102, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9102, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Confirmed"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9109, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9102, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9102, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Unconfirmed"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9110, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9102, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9102, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Acknowledge"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9111, 0),
        BrowseName=QualifiedName("Acknowledge", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2881, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9112, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9112, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Confirm"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9113, 0),
        BrowseName=QualifiedName("Confirm", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2881, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9114, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8961, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9114, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9113, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9113, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AlarmConditionType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2915, 0),
        BrowseName=QualifiedName("AlarmConditionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2881, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11120, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16371, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9215, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9216, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16390, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16380, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16395, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16396, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16397, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16398, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18190, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(16361, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16399, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16400, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16401, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16403, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17868, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17869, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17870, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18199, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9118, 0),
        BrowseName=QualifiedName("EnabledState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9119, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9118, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9118, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActiveState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9160, 0),
        BrowseName=QualifiedName("ActiveState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9161, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9164, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9165, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9166, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9167, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9168, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9161, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9160, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EffectiveDisplayName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9164, 0),
        BrowseName=QualifiedName("EffectiveDisplayName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9160, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9165, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9160, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EffectiveTransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9166, 0),
        BrowseName=QualifiedName("EffectiveTransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9160, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Active"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9167, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9160, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Inactive"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9168, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9160, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9160, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11120, 0),
        BrowseName=QualifiedName("InputNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SuppressedState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9169, 0),
        BrowseName=QualifiedName("SuppressedState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9170, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9174, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9176, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9177, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9170, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9169, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9169, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9174, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9169, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9169, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Suppressed"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9176, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9169, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9169, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Unsuppressed"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9177, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9169, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9169, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutOfServiceState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16371, 0),
        BrowseName=QualifiedName("OutOfServiceState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16372, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16376, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16379, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16372, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16371, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16371, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16376, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16371, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16371, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Out of Service"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16378, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16371, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16371, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("In Service"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16379, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16371, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16371, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ShelvingState"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9178, 0),
        BrowseName=QualifiedName("ShelvingState", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2929, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9179, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9184, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9213, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9211, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9212, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9179, 0),
        BrowseName=QualifiedName("CurrentState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9178, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2760, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9180, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2760, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9180, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9179, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9179, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransition"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9184, 0),
        BrowseName=QualifiedName("LastTransition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9178, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2767, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9185, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9188, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9185, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9184, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9184, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9188, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9184, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9184, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnshelveTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9189, 0),
        BrowseName=QualifiedName("UnshelveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9178, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("TimedShelve"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9213, 0),
        BrowseName=QualifiedName("TimedShelve", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(9178, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9214, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ShelvingTime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='If not 0, this parameter specifies a fixed time for which the Alarm is to be shelved.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9214, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9213, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9213, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Unshelve"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9211, 0),
        BrowseName=QualifiedName("Unshelve", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(9178, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OneShotShelve"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9212, 0),
        BrowseName=QualifiedName("OneShotShelve", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(9178, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(9212, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9212, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9212, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SuppressedOrShelved"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9215, 0),
        BrowseName=QualifiedName("SuppressedOrShelved", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxTimeShelved"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9216, 0),
        BrowseName=QualifiedName("MaxTimeShelved", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AudibleEnabled"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16389, 0),
        BrowseName=QualifiedName("AudibleEnabled", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AudibleSound"),
        DataType=NumericNodeId(16307, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16390, 0),
        BrowseName=QualifiedName("AudibleSound", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(17986, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17986, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SilenceState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16380, 0),
        BrowseName=QualifiedName("SilenceState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16381, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16385, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16387, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16388, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16381, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16385, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Silenced"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16387, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Not Silenced"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16388, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OnDelay"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16395, 0),
        BrowseName=QualifiedName("OnDelay", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OffDelay"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16396, 0),
        BrowseName=QualifiedName("OffDelay", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FirstInGroupFlag"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16397, 0),
        BrowseName=QualifiedName("FirstInGroupFlag", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("FirstInGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16398, 0),
        BrowseName=QualifiedName("FirstInGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(16405, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16405, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LatchedState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18190, 0),
        BrowseName=QualifiedName("LatchedState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18191, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18195, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18197, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18198, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18191, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18190, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18190, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18195, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18190, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18190, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Latched"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18197, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18190, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18190, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Unlatched"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18198, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18190, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18190, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<AlarmGroup>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16399, 0),
        BrowseName=QualifiedName("<AlarmGroup>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(16361, 0),
        TypeDefinition=NumericNodeId(16405, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16405, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(16361, 0),
        SourceNodeId=NumericNodeId(16399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReAlarmTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16400, 0),
        BrowseName=QualifiedName("ReAlarmTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReAlarmRepeatCount"),
        DataType=ua.NodeId(ua.ObjectIds.Int16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16401, 0),
        BrowseName=QualifiedName("ReAlarmRepeatCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Silence"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16402, 0),
        BrowseName=QualifiedName("Silence", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(16402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17242, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Suppress"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16403, 0),
        BrowseName=QualifiedName("Suppress", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(16403, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17225, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16403, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16403, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Unsuppress"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17868, 0),
        BrowseName=QualifiedName("Unsuppress", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(17868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17225, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveFromService"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17869, 0),
        BrowseName=QualifiedName("RemoveFromService", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(17869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17259, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("PlaceInService"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17870, 0),
        BrowseName=QualifiedName("PlaceInService", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(17870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17259, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18199, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(18199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15013, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AlarmGroupType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16405, 0),
        BrowseName=QualifiedName("AlarmGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(16362, 0),
        SourceNodeId=NumericNodeId(16405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(16405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<AlarmConditionInstance>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16406, 0),
        BrowseName=QualifiedName("<AlarmConditionInstance>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(16405, 0),
        ReferenceTypeId=NumericNodeId(16362, 0),
        TypeDefinition=NumericNodeId(2915, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16407, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16408, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16409, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16410, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16411, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16412, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16414, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16415, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16416, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16417, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16421, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16422, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16432, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16434, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16436, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16438, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16439, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16440, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16441, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16443, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16461, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16474, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16519, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(16362, 0),
        SourceNodeId=NumericNodeId(16406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16405, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventId"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16407, 0),
        BrowseName=QualifiedName("EventId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventType"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16408, 0),
        BrowseName=QualifiedName("EventType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16409, 0),
        BrowseName=QualifiedName("SourceNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16410, 0),
        BrowseName=QualifiedName("SourceName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Time"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16411, 0),
        BrowseName=QualifiedName("Time", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceiveTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16412, 0),
        BrowseName=QualifiedName("ReceiveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Message"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16414, 0),
        BrowseName=QualifiedName("Message", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Severity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16415, 0),
        BrowseName=QualifiedName("Severity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16416, 0),
        BrowseName=QualifiedName("ConditionClassId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16417, 0),
        BrowseName=QualifiedName("ConditionClassName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16420, 0),
        BrowseName=QualifiedName("ConditionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BranchId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16421, 0),
        BrowseName=QualifiedName("BranchId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Retain"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16422, 0),
        BrowseName=QualifiedName("Retain", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16423, 0),
        BrowseName=QualifiedName("EnabledState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16424, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16424, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16423, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Quality"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16432, 0),
        BrowseName=QualifiedName("Quality", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16433, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16433, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16432, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16432, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastSeverity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16434, 0),
        BrowseName=QualifiedName("LastSeverity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16435, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16435, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16434, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16435, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16435, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16435, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16434, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Comment"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16436, 0),
        BrowseName=QualifiedName("Comment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16437, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16437, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16436, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16437, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16437, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16437, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16436, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16438, 0),
        BrowseName=QualifiedName("ClientUserId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16438, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16438, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16438, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Disable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16439, 0),
        BrowseName=QualifiedName("Disable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(16439, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16439, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16439, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Enable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16440, 0),
        BrowseName=QualifiedName("Enable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(16440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddComment"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16441, 0),
        BrowseName=QualifiedName("AddComment", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16442, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(16441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16442, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16441, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16441, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AckedState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16443, 0),
        BrowseName=QualifiedName("AckedState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16444, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16444, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16443, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Acknowledge"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16461, 0),
        BrowseName=QualifiedName("Acknowledge", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(16461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16462, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16461, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16461, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActiveState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16465, 0),
        BrowseName=QualifiedName("ActiveState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16466, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16466, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16465, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16465, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16474, 0),
        BrowseName=QualifiedName("InputNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16474, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16474, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16474, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SuppressedOrShelved"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16519, 0),
        BrowseName=QualifiedName("SuppressedOrShelved", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ShelvedStateMachineType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2929, 0),
        BrowseName=QualifiedName("ShelvedStateMachineType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9115, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2930, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2933, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2936, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2942, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2945, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2949, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2947, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UnshelveTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9115, 0),
        BrowseName=QualifiedName("UnshelveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Unshelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2930, 0),
        BrowseName=QualifiedName("Unshelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(6098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2936, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(1, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(6098, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2930, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(6098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(6098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(6098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2930, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TimedShelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2932, 0),
        BrowseName=QualifiedName("TimedShelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(6100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2942, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2945, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(2, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(6100, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2932, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(6100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(6100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(6100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2932, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("OneShotShelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2933, 0),
        BrowseName=QualifiedName("OneShotShelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(6101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2936, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2942, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2945, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(3, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(6101, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2933, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(6101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(6101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(6101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2933, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("UnshelvedToTimedShelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2935, 0),
        BrowseName=QualifiedName("UnshelvedToTimedShelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11322, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2930, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2949, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(12, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11322, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2935, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2935, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("UnshelvedToOneShotShelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2936, 0),
        BrowseName=QualifiedName("UnshelvedToOneShotShelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11323, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2930, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2933, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(13, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11323, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2936, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2936, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TimedShelvedToUnshelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2940, 0),
        BrowseName=QualifiedName("TimedShelvedToUnshelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11324, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2930, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2947, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(21, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11324, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2940, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2940, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TimedShelvedToOneShotShelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2942, 0),
        BrowseName=QualifiedName("TimedShelvedToOneShotShelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11325, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2933, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(23, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11325, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2942, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2942, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("OneShotShelvedToUnshelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2943, 0),
        BrowseName=QualifiedName("OneShotShelvedToUnshelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11326, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2933, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2930, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2947, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(31, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11326, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2943, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2943, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("OneShotShelvedToTimedShelved"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2945, 0),
        BrowseName=QualifiedName("OneShotShelvedToTimedShelved", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11327, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2933, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2949, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(32, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11327, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2945, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2945, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("TimedShelve"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2949, 0),
        BrowseName=QualifiedName("TimedShelve", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2991, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2945, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(2949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ShelvingTime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='If not 0, this parameter specifies a fixed time for which the Alarm is to be shelved.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2991, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2949, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2949, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Unshelve"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2947, 0),
        BrowseName=QualifiedName("Unshelve", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(2947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OneShotShelve"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2948, 0),
        BrowseName=QualifiedName("OneShotShelve", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2929, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2936, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2942, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(2948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2929, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("LimitAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2955, 0),
        BrowseName=QualifiedName("LimitAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11124, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11125, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11126, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16572, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16574, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16575, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HighHighLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11124, 0),
        BrowseName=QualifiedName("HighHighLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HighLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11125, 0),
        BrowseName=QualifiedName("HighLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LowLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11126, 0),
        BrowseName=QualifiedName("LowLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LowLowLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11127, 0),
        BrowseName=QualifiedName("LowLowLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BaseHighHighLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16572, 0),
        BrowseName=QualifiedName("BaseHighHighLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BaseHighLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16573, 0),
        BrowseName=QualifiedName("BaseHighLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BaseLowLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16574, 0),
        BrowseName=QualifiedName("BaseLowLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BaseLowLowLimit"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16575, 0),
        BrowseName=QualifiedName("BaseLowLowLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ExclusiveLimitStateMachineType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9318, 0),
        BrowseName=QualifiedName("ExclusiveLimitStateMachineType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9329, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9331, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9333, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9335, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9337, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9338, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9339, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9340, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("HighHigh"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9329, 0),
        BrowseName=QualifiedName("HighHigh", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9330, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9339, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9340, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(1, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9330, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9329, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9329, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("High"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9331, 0),
        BrowseName=QualifiedName("High", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9332, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9339, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9340, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(2, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9332, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9331, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9331, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Low"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9333, 0),
        BrowseName=QualifiedName("Low", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9334, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9337, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9338, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(3, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9334, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9333, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9333, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LowLow"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9335, 0),
        BrowseName=QualifiedName("LowLow", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9336, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9337, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9338, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(4, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9336, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9335, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9335, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LowLowToLow"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9337, 0),
        BrowseName=QualifiedName("LowLowToLow", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11340, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9335, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9333, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(43, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11340, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9337, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9337, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LowToLowLow"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9338, 0),
        BrowseName=QualifiedName("LowToLowLow", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11341, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9333, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9335, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(34, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11341, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9338, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9338, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("HighHighToHigh"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9339, 0),
        BrowseName=QualifiedName("HighHighToHigh", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11342, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9329, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9331, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(12, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11342, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9339, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11342, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11342, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11342, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9339, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("HighToHighHigh"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9340, 0),
        BrowseName=QualifiedName("HighToHighHigh", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11343, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(9340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9331, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(9340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9329, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(21, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11343, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9340, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11343, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11343, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11343, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ExclusiveLimitAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9341, 0),
        BrowseName=QualifiedName("ExclusiveLimitAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9398, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9455, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActiveState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9398, 0),
        BrowseName=QualifiedName("ActiveState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9341, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9399, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9455, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9341, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9399, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9398, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9398, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LimitState"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9455, 0),
        BrowseName=QualifiedName("LimitState", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(9341, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9318, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9461, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9398, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9318, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9341, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9456, 0),
        BrowseName=QualifiedName("CurrentState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9455, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2760, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9457, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2760, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9455, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9457, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9456, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9456, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransition"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9461, 0),
        BrowseName=QualifiedName("LastTransition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9455, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2767, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9455, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9462, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9461, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9461, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9465, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9461, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9461, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NonExclusiveLimitAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9906, 0),
        BrowseName=QualifiedName("NonExclusiveLimitAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2955, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9963, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10029, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10038, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10047, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2955, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActiveState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9963, 0),
        BrowseName=QualifiedName("ActiveState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9964, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10029, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10038, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10047, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(9963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9964, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9963, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9963, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HighHighState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10020, 0),
        BrowseName=QualifiedName("HighHighState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10021, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10025, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10028, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9963, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(10020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10021, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10020, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10025, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10020, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("HighHigh active"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10027, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10020, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("HighHigh inactive"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10028, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10020, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10020, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HighState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10029, 0),
        BrowseName=QualifiedName("HighState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10034, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10036, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10037, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9963, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(10029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10030, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10029, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10034, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10029, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("High active"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10036, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10029, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("High inactive"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10037, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10029, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LowState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10038, 0),
        BrowseName=QualifiedName("LowState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10039, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10043, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10045, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10046, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9963, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(10038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10039, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10038, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10039, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10039, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10039, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10038, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10043, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10038, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10038, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Low active"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10045, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10038, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10038, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("Low inactive"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10046, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10038, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10038, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LowLowState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10047, 0),
        BrowseName=QualifiedName("LowLowState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10048, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10052, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10054, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(9004, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9963, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(10047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10048, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10047, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10047, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10052, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10047, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10047, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrueState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("LowLow active"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10054, 0),
        BrowseName=QualifiedName("TrueState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10047, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10054, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10047, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FalseState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=ua.Variant(LocalizedText("LowLow inactive"), ua.VariantType.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10055, 0),
        BrowseName=QualifiedName("FalseState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10047, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10047, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NonExclusiveLevelAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10060, 0),
        BrowseName=QualifiedName("NonExclusiveLevelAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(10060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ExclusiveLevelAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9482, 0),
        BrowseName=QualifiedName("ExclusiveLevelAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(9341, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9341, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NonExclusiveDeviationAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10368, 0),
        BrowseName=QualifiedName("NonExclusiveDeviationAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16776, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(10368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetpointNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10522, 0),
        BrowseName=QualifiedName("SetpointNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10368, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(10522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(10522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10368, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BaseSetpointNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16776, 0),
        BrowseName=QualifiedName("BaseSetpointNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10368, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10368, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NonExclusiveRateOfChangeAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10214, 0),
        BrowseName=QualifiedName("NonExclusiveRateOfChangeAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(9906, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16858, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(10214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EngineeringUnits"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16858, 0),
        BrowseName=QualifiedName("EngineeringUnits", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10214, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10214, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ExclusiveDeviationAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9764, 0),
        BrowseName=QualifiedName("ExclusiveDeviationAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(9341, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9905, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9341, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SetpointNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9905, 0),
        BrowseName=QualifiedName("SetpointNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9764, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(9905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(9905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9764, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BaseSetpointNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16817, 0),
        BrowseName=QualifiedName("BaseSetpointNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9764, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9764, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ExclusiveRateOfChangeAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9623, 0),
        BrowseName=QualifiedName("ExclusiveRateOfChangeAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(9341, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(9623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16899, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9341, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EngineeringUnits"),
        DataType=NumericNodeId(887, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16899, 0),
        BrowseName=QualifiedName("EngineeringUnits", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(9623, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DiscreteAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10523, 0),
        BrowseName=QualifiedName("DiscreteAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(10523, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("OffNormalAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10637, 0),
        BrowseName=QualifiedName("OffNormalAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(10523, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(10637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11158, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(10637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10523, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NormalState"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11158, 0),
        BrowseName=QualifiedName("NormalState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(10637, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SystemOffNormalAlarmType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11753, 0),
        BrowseName=QualifiedName("SystemOffNormalAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(10637, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TripAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(10751, 0),
        BrowseName=QualifiedName("TripAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(10637, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(10751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("InstrumentDiagnosticAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18347, 0),
        BrowseName=QualifiedName("InstrumentDiagnosticAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(10637, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18347, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SystemDiagnosticAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18496, 0),
        BrowseName=QualifiedName("SystemDiagnosticAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(10637, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(10637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("CertificateExpirationAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13225, 0),
        BrowseName=QualifiedName("CertificateExpirationAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11753, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13325, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14900, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13326, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13327, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(13225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11753, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ExpirationDate"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13325, 0),
        BrowseName=QualifiedName("ExpirationDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13225, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13225, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ExpirationLimit"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14900, 0),
        BrowseName=QualifiedName("ExpirationLimit", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13225, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13225, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateType"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13326, 0),
        BrowseName=QualifiedName("CertificateType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13225, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13225, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Certificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13327, 0),
        BrowseName=QualifiedName("Certificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13225, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13225, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DiscrepancyAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17080, 0),
        BrowseName=QualifiedName("DiscrepancyAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2915, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17215, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17216, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2915, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TargetValueNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17215, 0),
        BrowseName=QualifiedName("TargetValueNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17080, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ExpectedTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17216, 0),
        BrowseName=QualifiedName("ExpectedTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17080, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Tolerance"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17217, 0),
        BrowseName=QualifiedName("Tolerance", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17080, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17080, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BaseConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11163, 0),
        BrowseName=QualifiedName("BaseConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ProcessConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11164, 0),
        BrowseName=QualifiedName("ProcessConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("MaintenanceConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11165, 0),
        BrowseName=QualifiedName("MaintenanceConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SystemConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11166, 0),
        BrowseName=QualifiedName("SystemConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SafetyConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17218, 0),
        BrowseName=QualifiedName("SafetyConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("HighlyManagedAlarmConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17219, 0),
        BrowseName=QualifiedName("HighlyManagedAlarmConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TrainingConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17220, 0),
        BrowseName=QualifiedName("TrainingConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("StatisticalConditionClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18665, 0),
        BrowseName=QualifiedName("StatisticalConditionClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18665, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TestingConditionSubClassType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17221, 0),
        BrowseName=QualifiedName("TestingConditionSubClassType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11163, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11163, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2790, 0),
        BrowseName=QualifiedName("AuditConditionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionEnableEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2803, 0),
        BrowseName=QualifiedName("AuditConditionEnableEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionCommentEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2829, 0),
        BrowseName=QualifiedName("AuditConditionCommentEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17222, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11851, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionEventId"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17222, 0),
        BrowseName=QualifiedName("ConditionEventId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2829, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2829, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Comment"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11851, 0),
        BrowseName=QualifiedName("Comment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2829, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2829, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionRespondEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8927, 0),
        BrowseName=QualifiedName("AuditConditionRespondEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(8927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SelectedResponse"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11852, 0),
        BrowseName=QualifiedName("SelectedResponse", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8927, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8927, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionAcknowledgeEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8944, 0),
        BrowseName=QualifiedName("AuditConditionAcknowledgeEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8944, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17223, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8944, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(8944, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionEventId"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17223, 0),
        BrowseName=QualifiedName("ConditionEventId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8944, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8944, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Comment"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11853, 0),
        BrowseName=QualifiedName("Comment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8944, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8944, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionConfirmEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8961, 0),
        BrowseName=QualifiedName("AuditConditionConfirmEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17224, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11854, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(8961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionEventId"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17224, 0),
        BrowseName=QualifiedName("ConditionEventId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8961, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8961, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Comment"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11854, 0),
        BrowseName=QualifiedName("Comment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8961, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8961, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionShelvingEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11093, 0),
        BrowseName=QualifiedName("AuditConditionShelvingEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ShelvingTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11855, 0),
        BrowseName=QualifiedName("ShelvingTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11093, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11093, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionSuppressionEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17225, 0),
        BrowseName=QualifiedName("AuditConditionSuppressionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionSilenceEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17242, 0),
        BrowseName=QualifiedName("AuditConditionSilenceEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionResetEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15013, 0),
        BrowseName=QualifiedName("AuditConditionResetEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditConditionOutOfServiceEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17259, 0),
        BrowseName=QualifiedName("AuditConditionOutOfServiceEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2790, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2790, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RefreshStartEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2787, 0),
        BrowseName=QualifiedName("RefreshStartEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2130, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RefreshEndEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2788, 0),
        BrowseName=QualifiedName("RefreshEndEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2130, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RefreshRequiredEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2789, 0),
        BrowseName=QualifiedName("RefreshRequiredEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2130, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasCondition"),
        InverseName=LocalizedText("IsConditionOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(9006, 0),
        BrowseName=QualifiedName("HasCondition", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(9006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasEffectDisable"),
        InverseName=LocalizedText("MayBeDisabledBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17276, 0),
        BrowseName=QualifiedName("HasEffectDisable", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(54, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(54, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasEffectEnable"),
        InverseName=LocalizedText("MayBeEnabledBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17983, 0),
        BrowseName=QualifiedName("HasEffectEnable", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(54, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17983, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(54, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasEffectSuppressed"),
        InverseName=LocalizedText("MayBeSuppressedBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17984, 0),
        BrowseName=QualifiedName("HasEffectSuppressed", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(54, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17984, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(54, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasEffectUnsuppressed"),
        InverseName=LocalizedText("MayBeUnsuppressedBy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17985, 0),
        BrowseName=QualifiedName("HasEffectUnsuppressed", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(54, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17985, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(54, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AlarmMetricsType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17279, 0),
        BrowseName=QualifiedName("AlarmMetricsType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17280, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17991, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17281, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17282, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17284, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17286, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17283, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17288, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18666, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AlarmCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17280, 0),
        BrowseName=QualifiedName("AlarmCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StartTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17991, 0),
        BrowseName=QualifiedName("StartTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaximumActiveState"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17281, 0),
        BrowseName=QualifiedName("MaximumActiveState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaximumUnAck"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17282, 0),
        BrowseName=QualifiedName("MaximumUnAck", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentAlarmRate"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17284, 0),
        BrowseName=QualifiedName("CurrentAlarmRate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17277, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17285, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17277, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Rate"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17285, 0),
        BrowseName=QualifiedName("Rate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17284, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17284, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaximumAlarmRate"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17286, 0),
        BrowseName=QualifiedName("MaximumAlarmRate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17277, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17287, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17277, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Rate"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17287, 0),
        BrowseName=QualifiedName("Rate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17286, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17286, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaximumReAlarmCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17283, 0),
        BrowseName=QualifiedName("MaximumReAlarmCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AverageAlarmRate"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17288, 0),
        BrowseName=QualifiedName("AverageAlarmRate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17277, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17289, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17277, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Rate"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17289, 0),
        BrowseName=QualifiedName("Rate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17288, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17288, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18666, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17279, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(18666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17279, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("AlarmRateVariableType"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17277, 0),
        BrowseName=QualifiedName("AlarmRateVariableType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17278, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Rate"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17278, 0),
        BrowseName=QualifiedName("Rate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17277, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17277, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ProgramStateMachineType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2391, 0),
        BrowseName=QualifiedName("ProgramStateMachineType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2771, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2392, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2393, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2394, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2395, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2396, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2397, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2398, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3850, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2406, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2400, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2404, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2408, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2410, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2412, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2414, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2416, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2418, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2422, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2424, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2426, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2427, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2428, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2430, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2771, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrentState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3830, 0),
        BrowseName=QualifiedName("CurrentState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2760, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2760, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3831, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Number"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3833, 0),
        BrowseName=QualifiedName("Number", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3830, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3830, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransition"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3835, 0),
        BrowseName=QualifiedName("LastTransition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2767, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3838, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3836, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3835, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3835, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Number"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3838, 0),
        BrowseName=QualifiedName("Number", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3835, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3835, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3839, 0),
        BrowseName=QualifiedName("TransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3835, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3835, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Creatable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2392, 0),
        BrowseName=QualifiedName("Creatable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Deletable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2393, 0),
        BrowseName=QualifiedName("Deletable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AutoDelete"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2394, 0),
        BrowseName=QualifiedName("AutoDelete", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RecycleCount"),
        DataType=ua.NodeId(ua.ObjectIds.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2395, 0),
        BrowseName=QualifiedName("RecycleCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InstanceCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2396, 0),
        BrowseName=QualifiedName("InstanceCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxInstanceCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2397, 0),
        BrowseName=QualifiedName("MaxInstanceCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxRecycleCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2398, 0),
        BrowseName=QualifiedName("MaxRecycleCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProgramDiagnostic"),
        DataType=NumericNodeId(15396, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2399, 0),
        BrowseName=QualifiedName("ProgramDiagnostic", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15383, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3840, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3841, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3843, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3844, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3846, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3847, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15038, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15040, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2399, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateSessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3840, 0),
        BrowseName=QualifiedName("CreateSessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateClientName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3841, 0),
        BrowseName=QualifiedName("CreateClientName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InvocationCreationTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3842, 0),
        BrowseName=QualifiedName("InvocationCreationTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3843, 0),
        BrowseName=QualifiedName("LastTransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodCall"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3844, 0),
        BrowseName=QualifiedName("LastMethodCall", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodSessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3845, 0),
        BrowseName=QualifiedName("LastMethodSessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodInputArguments"),
        DataType=NumericNodeId(296, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3846, 0),
        BrowseName=QualifiedName("LastMethodInputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodOutputArguments"),
        DataType=NumericNodeId(296, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3847, 0),
        BrowseName=QualifiedName("LastMethodOutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodInputValues"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15038, 0),
        BrowseName=QualifiedName("LastMethodInputValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodOutputValues"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15040, 0),
        BrowseName=QualifiedName("LastMethodOutputValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodCallTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3848, 0),
        BrowseName=QualifiedName("LastMethodCallTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodReturnStatus"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3849, 0),
        BrowseName=QualifiedName("LastMethodReturnStatus", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2399, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2399, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("FinalResultData"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3850, 0),
        BrowseName=QualifiedName("FinalResultData", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Halted"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2406, 0),
        BrowseName=QualifiedName("Halted", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2407, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2408, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2412, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2424, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(11, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2407, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Ready"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2400, 0),
        BrowseName=QualifiedName("Ready", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2401, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2408, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2410, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2414, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2422, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2424, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2400, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(12, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2401, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2400, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2400, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Running"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2402, 0),
        BrowseName=QualifiedName("Running", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2403, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2410, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2412, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2414, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2416, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2418, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(13, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2403, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2402, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2403, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2403, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2403, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2402, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Suspended"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2404, 0),
        BrowseName=QualifiedName("Suspended", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2307, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2405, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2416, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2418, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2422, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(14, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2405, 0),
        BrowseName=QualifiedName("StateNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2404, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2404, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("HaltedToReady"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2408, 0),
        BrowseName=QualifiedName("HaltedToReady", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2409, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2406, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2400, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2430, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(1, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2409, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2408, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2408, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadyToRunning"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2410, 0),
        BrowseName=QualifiedName("ReadyToRunning", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2411, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2400, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2426, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(2, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2411, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2410, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2410, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("RunningToHalted"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2412, 0),
        BrowseName=QualifiedName("RunningToHalted", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2413, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2406, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(3, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2413, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2412, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2413, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2413, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2413, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2412, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("RunningToReady"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2414, 0),
        BrowseName=QualifiedName("RunningToReady", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2415, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2400, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(4, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2415, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2414, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2415, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2414, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("RunningToSuspended"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2416, 0),
        BrowseName=QualifiedName("RunningToSuspended", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2417, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2404, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2427, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(5, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2417, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2416, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SuspendedToRunning"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2418, 0),
        BrowseName=QualifiedName("SuspendedToRunning", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2419, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2404, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2428, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(6, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2419, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2418, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2418, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SuspendedToHalted"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2420, 0),
        BrowseName=QualifiedName("SuspendedToHalted", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2421, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2404, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2406, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(7, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2421, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2420, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2420, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SuspendedToReady"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2422, 0),
        BrowseName=QualifiedName("SuspendedToReady", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2404, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2400, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(8, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2423, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2422, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2422, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ReadyToHalted"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2424, 0),
        BrowseName=QualifiedName("ReadyToHalted", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2310, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2425, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(51, 0),
        SourceNodeId=NumericNodeId(2424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2400, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(52, 0),
        SourceNodeId=NumericNodeId(2424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2406, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(54, 0),
        SourceNodeId=NumericNodeId(2424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        Value=ua.Variant(9, ua.VariantType.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2425, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2424, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2424, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Start"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2426, 0),
        BrowseName=QualifiedName("Start", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2410, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Suspend"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2427, 0),
        BrowseName=QualifiedName("Suspend", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2416, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Resume"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2428, 0),
        BrowseName=QualifiedName("Resume", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2418, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Halt"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2429, 0),
        BrowseName=QualifiedName("Halt", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2412, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2424, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2430, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(2391, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(53, 0),
        SourceNodeId=NumericNodeId(2430, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2408, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2430, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2430, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2391, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ProgramTransitionEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2378, 0),
        BrowseName=QualifiedName("ProgramTransitionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2311, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2379, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2311, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IntermediateResult"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2379, 0),
        BrowseName=QualifiedName("IntermediateResult", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2378, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2378, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditProgramTransitionEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11856, 0),
        BrowseName=QualifiedName("AuditProgramTransitionEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2315, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11875, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2315, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransitionNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11875, 0),
        BrowseName=QualifiedName("TransitionNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11856, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11856, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ProgramTransitionAuditEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3806, 0),
        BrowseName=QualifiedName("ProgramTransitionAuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2315, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3806, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3825, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3806, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2315, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Transition"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3825, 0),
        BrowseName=QualifiedName("Transition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3806, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(2767, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3826, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3806, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3826, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3825, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3825, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ProgramDiagnosticType"),
        DataType=NumericNodeId(894, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2380, 0),
        BrowseName=QualifiedName("ProgramDiagnosticType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2381, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2382, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2383, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2384, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2385, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2386, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2387, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2388, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2390, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateSessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2381, 0),
        BrowseName=QualifiedName("CreateSessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateClientName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2382, 0),
        BrowseName=QualifiedName("CreateClientName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InvocationCreationTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2383, 0),
        BrowseName=QualifiedName("InvocationCreationTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2384, 0),
        BrowseName=QualifiedName("LastTransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodCall"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2385, 0),
        BrowseName=QualifiedName("LastMethodCall", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodSessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2386, 0),
        BrowseName=QualifiedName("LastMethodSessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodInputArguments"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2387, 0),
        BrowseName=QualifiedName("LastMethodInputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodOutputArguments"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2388, 0),
        BrowseName=QualifiedName("LastMethodOutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodCallTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2389, 0),
        BrowseName=QualifiedName("LastMethodCallTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodReturnStatus"),
        DataType=NumericNodeId(299, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2390, 0),
        BrowseName=QualifiedName("LastMethodReturnStatus", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2380, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2380, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("ProgramDiagnostic2Type"),
        DataType=NumericNodeId(15396, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15383, 0),
        BrowseName=QualifiedName("ProgramDiagnostic2Type", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15384, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15385, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15386, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15387, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15388, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15390, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15391, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15392, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15393, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15394, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15395, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateSessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15384, 0),
        BrowseName=QualifiedName("CreateSessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CreateClientName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15385, 0),
        BrowseName=QualifiedName("CreateClientName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InvocationCreationTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15386, 0),
        BrowseName=QualifiedName("InvocationCreationTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastTransitionTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15387, 0),
        BrowseName=QualifiedName("LastTransitionTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodCall"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15388, 0),
        BrowseName=QualifiedName("LastMethodCall", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodSessionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15389, 0),
        BrowseName=QualifiedName("LastMethodSessionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodInputArguments"),
        DataType=NumericNodeId(296, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15390, 0),
        BrowseName=QualifiedName("LastMethodInputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodOutputArguments"),
        DataType=NumericNodeId(296, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15391, 0),
        BrowseName=QualifiedName("LastMethodOutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodInputValues"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15392, 0),
        BrowseName=QualifiedName("LastMethodInputValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodOutputValues"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15393, 0),
        BrowseName=QualifiedName("LastMethodOutputValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodCallTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15394, 0),
        BrowseName=QualifiedName("LastMethodCallTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastMethodReturnStatus"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15395, 0),
        BrowseName=QualifiedName("LastMethodReturnStatus", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15383, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15383, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Annotations"),
        DataType=NumericNodeId(891, 0),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11214, 0),
        BrowseName=QualifiedName("Annotations", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("HistoricalDataConfigurationType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2318, 0),
        BrowseName=QualifiedName("HistoricalDataConfigurationType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3059, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11876, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2323, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2324, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2325, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2326, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2327, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2328, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11499, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11500, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateConfiguration"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3059, 0),
        BrowseName=QualifiedName("AggregateConfiguration", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11187, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11168, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11170, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11187, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(3059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TreatUncertainAsBad"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11168, 0),
        BrowseName=QualifiedName("TreatUncertainAsBad", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3059, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3059, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PercentDataBad"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11169, 0),
        BrowseName=QualifiedName("PercentDataBad", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3059, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3059, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PercentDataGood"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11170, 0),
        BrowseName=QualifiedName("PercentDataGood", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3059, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3059, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UseSlopedExtrapolation"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11171, 0),
        BrowseName=QualifiedName("UseSlopedExtrapolation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3059, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3059, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateFunctions"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11876, 0),
        BrowseName=QualifiedName("AggregateFunctions", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Stepped"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2323, 0),
        BrowseName=QualifiedName("Stepped", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Definition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2324, 0),
        BrowseName=QualifiedName("Definition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxTimeInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2325, 0),
        BrowseName=QualifiedName("MaxTimeInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MinTimeInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2326, 0),
        BrowseName=QualifiedName("MinTimeInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ExceptionDeviation"),
        DataType=ua.NodeId(ua.ObjectIds.Double),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2327, 0),
        BrowseName=QualifiedName("ExceptionDeviation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ExceptionDeviationFormat"),
        DataType=NumericNodeId(890, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2328, 0),
        BrowseName=QualifiedName("ExceptionDeviationFormat", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2328, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2328, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2328, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StartOfArchive"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11499, 0),
        BrowseName=QualifiedName("StartOfArchive", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StartOfOnlineArchive"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11500, 0),
        BrowseName=QualifiedName("StartOfOnlineArchive", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerTimestampSupported"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19092, 0),
        BrowseName=QualifiedName("ServerTimestampSupported", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2318, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("HA Configuration"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11202, 0),
        BrowseName=QualifiedName("HA Configuration", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2318, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11203, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11208, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2318, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateConfiguration"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11203, 0),
        BrowseName=QualifiedName("AggregateConfiguration", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11202, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(11187, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11204, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11205, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11206, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11207, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11187, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11202, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TreatUncertainAsBad"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11204, 0),
        BrowseName=QualifiedName("TreatUncertainAsBad", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11203, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11203, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PercentDataBad"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11205, 0),
        BrowseName=QualifiedName("PercentDataBad", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11203, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11203, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PercentDataGood"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11206, 0),
        BrowseName=QualifiedName("PercentDataGood", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11203, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11203, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UseSlopedExtrapolation"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11207, 0),
        BrowseName=QualifiedName("UseSlopedExtrapolation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11203, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11203, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Stepped"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11208, 0),
        BrowseName=QualifiedName("Stepped", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11202, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11202, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoricalEventFilter"),
        DataType=NumericNodeId(725, 0),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11215, 0),
        BrowseName=QualifiedName("HistoricalEventFilter", 0),
        NodeClass_=NodeClass.Variable,
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("HistoryServerCapabilitiesType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2330, 0),
        BrowseName=QualifiedName("HistoryServerCapabilitiesType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2331, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2332, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11268, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11269, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2334, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2335, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2336, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2337, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2338, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11278, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11279, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11280, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11501, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11270, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AccessHistoryDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2331, 0),
        BrowseName=QualifiedName("AccessHistoryDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AccessHistoryEventsCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2332, 0),
        BrowseName=QualifiedName("AccessHistoryEventsCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxReturnDataValues"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11268, 0),
        BrowseName=QualifiedName("MaxReturnDataValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxReturnEventValues"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11269, 0),
        BrowseName=QualifiedName("MaxReturnEventValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InsertDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2334, 0),
        BrowseName=QualifiedName("InsertDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReplaceDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2335, 0),
        BrowseName=QualifiedName("ReplaceDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdateDataCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2336, 0),
        BrowseName=QualifiedName("UpdateDataCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteRawCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2337, 0),
        BrowseName=QualifiedName("DeleteRawCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteAtTimeCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2338, 0),
        BrowseName=QualifiedName("DeleteAtTimeCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(2338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InsertEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11278, 0),
        BrowseName=QualifiedName("InsertEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReplaceEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11279, 0),
        BrowseName=QualifiedName("ReplaceEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdateEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11280, 0),
        BrowseName=QualifiedName("UpdateEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteEventCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11501, 0),
        BrowseName=QualifiedName("DeleteEventCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11501, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11501, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11501, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InsertAnnotationCapability"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11270, 0),
        BrowseName=QualifiedName("InsertAnnotationCapability", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AggregateFunctions"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11172, 0),
        BrowseName=QualifiedName("AggregateFunctions", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerTimestampSupported"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19094, 0),
        BrowseName=QualifiedName("ServerTimestampSupported", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2330, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2330, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryEventUpdateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2999, 0),
        BrowseName=QualifiedName("AuditHistoryEventUpdateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2104, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3025, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3028, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3003, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3029, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(2999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(2999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2104, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdatedNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3025, 0),
        BrowseName=QualifiedName("UpdatedNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2999, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PerformInsertReplace"),
        DataType=NumericNodeId(11293, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3028, 0),
        BrowseName=QualifiedName("PerformInsertReplace", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2999, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Filter"),
        DataType=NumericNodeId(725, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3003, 0),
        BrowseName=QualifiedName("Filter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2999, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3003, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NewValues"),
        DataType=NumericNodeId(920, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3029, 0),
        BrowseName=QualifiedName("NewValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2999, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldValues"),
        DataType=NumericNodeId(920, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3030, 0),
        BrowseName=QualifiedName("OldValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(2999, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryValueUpdateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3006, 0),
        BrowseName=QualifiedName("AuditHistoryValueUpdateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2104, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3026, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3031, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3032, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3033, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2104, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdatedNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3026, 0),
        BrowseName=QualifiedName("UpdatedNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3006, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3006, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PerformInsertReplace"),
        DataType=NumericNodeId(11293, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3031, 0),
        BrowseName=QualifiedName("PerformInsertReplace", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3006, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3006, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NewValues"),
        DataType=NumericNodeId(23, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3032, 0),
        BrowseName=QualifiedName("NewValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3006, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3006, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldValues"),
        DataType=NumericNodeId(23, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3033, 0),
        BrowseName=QualifiedName("OldValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3006, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3006, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryAnnotationUpdateEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19095, 0),
        BrowseName=QualifiedName("AuditHistoryAnnotationUpdateEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2104, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19293, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19294, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19295, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2104, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PerformInsertReplace"),
        DataType=NumericNodeId(11293, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19293, 0),
        BrowseName=QualifiedName("PerformInsertReplace", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19095, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19293, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19293, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19293, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NewValues"),
        DataType=NumericNodeId(23, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19294, 0),
        BrowseName=QualifiedName("NewValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19095, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldValues"),
        DataType=NumericNodeId(23, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19295, 0),
        BrowseName=QualifiedName("OldValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19095, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryDeleteEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3012, 0),
        BrowseName=QualifiedName("AuditHistoryDeleteEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2104, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2104, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdatedNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3027, 0),
        BrowseName=QualifiedName("UpdatedNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3012, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3012, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryRawModifyDeleteEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3014, 0),
        BrowseName=QualifiedName("AuditHistoryRawModifyDeleteEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(3012, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3015, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3016, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3017, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3034, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3012, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IsDeleteModified"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3015, 0),
        BrowseName=QualifiedName("IsDeleteModified", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3014, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3014, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StartTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3016, 0),
        BrowseName=QualifiedName("StartTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3014, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3014, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3017, 0),
        BrowseName=QualifiedName("EndTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3014, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3014, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldValues"),
        DataType=NumericNodeId(23, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3034, 0),
        BrowseName=QualifiedName("OldValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3014, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3014, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryAtTimeDeleteEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3019, 0),
        BrowseName=QualifiedName("AuditHistoryAtTimeDeleteEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(3012, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3021, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3012, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReqTimes"),
        DataType=NumericNodeId(294, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3020, 0),
        BrowseName=QualifiedName("ReqTimes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3019, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3019, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldValues"),
        DataType=NumericNodeId(23, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3021, 0),
        BrowseName=QualifiedName("OldValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3019, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3019, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuditHistoryEventDeleteEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3022, 0),
        BrowseName=QualifiedName("AuditHistoryEventDeleteEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(3012, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3023, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3024, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(3022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3012, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventIds"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3023, 0),
        BrowseName=QualifiedName("EventIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3022, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3022, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OldValues"),
        DataType=NumericNodeId(920, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(3024, 0),
        BrowseName=QualifiedName("OldValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(3022, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(3024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(3024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(3024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3022, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TrustListType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12522, 0),
        BrowseName=QualifiedName("TrustListType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11575, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12542, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19296, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12543, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12546, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12548, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12550, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12542, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12522, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdateFrequency"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19296, 0),
        BrowseName=QualifiedName("UpdateFrequency", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12522, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12543, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12522, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12543, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12544, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12543, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12545, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12543, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12543, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12544, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12543, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12544, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12544, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12544, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12543, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12545, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12543, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12543, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CloseAndUpdate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12546, 0),
        BrowseName=QualifiedName("CloseAndUpdate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12522, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12705, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12547, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12705, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12546, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12546, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplyChangesRequired',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12547, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12546, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12546, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12548, 0),
        BrowseName=QualifiedName("AddCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12522, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12549, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Certificate',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12549, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12548, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12548, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12550, 0),
        BrowseName=QualifiedName("RemoveCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12522, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12551, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Thumbprint',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12551, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12550, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TrustListMasks"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12552, 0),
        BrowseName=QualifiedName("TrustListMasks", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12553, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=0,
                DisplayName=LocalizedText(Text='None'),
                ),
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='TrustedCertificates'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='TrustedCrls'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='IssuerCertificates'),
                ),
            ua.EnumValueType(
                Value=8,
                DisplayName=LocalizedText(Text='IssuerCrls'),
                ),
            ua.EnumValueType(
                Value=15,
                DisplayName=LocalizedText(Text='All'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12553, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12552, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12552, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TrustListDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12554, 0),
        BrowseName=QualifiedName("TrustListDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TrustListOutOfDateAlarmType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19297, 0),
        BrowseName=QualifiedName("TrustListOutOfDateAlarmType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(11753, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19446, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19447, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19448, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11753, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrustListId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19446, 0),
        BrowseName=QualifiedName("TrustListId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19297, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19297, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19447, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19297, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19297, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdateFrequency"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19448, 0),
        BrowseName=QualifiedName("UpdateFrequency", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19297, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19297, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("CertificateGroupType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12555, 0),
        BrowseName=QualifiedName("CertificateGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13631, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23526, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13599, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12555, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13600, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13601, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13602, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13603, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13605, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13608, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13610, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13613, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13615, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13618, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13620, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13621, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13600, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13601, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13601, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13601, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13601, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13602, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13603, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13605, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13606, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13605, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13605, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13607, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13605, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13605, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13608, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13609, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13609, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13608, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13608, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13610, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13611, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13612, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13611, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13610, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13610, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13612, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13610, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13610, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13613, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13613, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13614, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13613, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13613, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13614, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13613, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13613, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13615, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13617, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13616, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13615, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13615, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13617, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13615, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13615, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13618, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13619, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13619, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13618, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13618, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13620, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13621, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13599, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13623, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13599, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13622, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13621, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13621, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13623, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13621, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13621, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13631, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12555, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("CertificateExpired"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19450, 0),
        BrowseName=QualifiedName("CertificateExpired", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12555, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(13225, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19451, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19452, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19453, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19454, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19455, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19458, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19459, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19460, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19461, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19464, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19466, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19467, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19476, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19478, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19482, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19483, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19484, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19485, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19487, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19505, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19509, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19518, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20138, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20139, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20141, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20142, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13225, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventId"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19451, 0),
        BrowseName=QualifiedName("EventId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventType"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19452, 0),
        BrowseName=QualifiedName("EventType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19453, 0),
        BrowseName=QualifiedName("SourceNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19454, 0),
        BrowseName=QualifiedName("SourceName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Time"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19455, 0),
        BrowseName=QualifiedName("Time", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceiveTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19456, 0),
        BrowseName=QualifiedName("ReceiveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Message"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19458, 0),
        BrowseName=QualifiedName("Message", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Severity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19459, 0),
        BrowseName=QualifiedName("Severity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19460, 0),
        BrowseName=QualifiedName("ConditionClassId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19461, 0),
        BrowseName=QualifiedName("ConditionClassName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19464, 0),
        BrowseName=QualifiedName("ConditionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BranchId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19465, 0),
        BrowseName=QualifiedName("BranchId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Retain"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19466, 0),
        BrowseName=QualifiedName("Retain", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19467, 0),
        BrowseName=QualifiedName("EnabledState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19467, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19468, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19467, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19467, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19467, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19468, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19467, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19467, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Quality"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19476, 0),
        BrowseName=QualifiedName("Quality", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19477, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19476, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19476, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastSeverity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19478, 0),
        BrowseName=QualifiedName("LastSeverity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19479, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19479, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19478, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Comment"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19480, 0),
        BrowseName=QualifiedName("Comment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19481, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19481, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19480, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19480, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19482, 0),
        BrowseName=QualifiedName("ClientUserId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Disable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19483, 0),
        BrowseName=QualifiedName("Disable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(19483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Enable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19484, 0),
        BrowseName=QualifiedName("Enable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(19484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddComment"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19485, 0),
        BrowseName=QualifiedName("AddComment", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(19485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19486, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19485, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19485, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AckedState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19487, 0),
        BrowseName=QualifiedName("AckedState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19488, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19488, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19487, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19487, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Acknowledge"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19505, 0),
        BrowseName=QualifiedName("Acknowledge", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19506, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(19505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19506, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19505, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19505, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActiveState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19509, 0),
        BrowseName=QualifiedName("ActiveState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19510, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19510, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19509, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19518, 0),
        BrowseName=QualifiedName("InputNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SuppressedOrShelved"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20101, 0),
        BrowseName=QualifiedName("SuppressedOrShelved", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NormalState"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20138, 0),
        BrowseName=QualifiedName("NormalState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ExpirationDate"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20139, 0),
        BrowseName=QualifiedName("ExpirationDate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateType"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20141, 0),
        BrowseName=QualifiedName("CertificateType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Certificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20142, 0),
        BrowseName=QualifiedName("Certificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19450, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19450, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustListOutOfDate"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20143, 0),
        BrowseName=QualifiedName("TrustListOutOfDate", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12555, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19297, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20144, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20145, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20146, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20147, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20148, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20149, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20151, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20152, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20153, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20154, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20157, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20158, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20159, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20160, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20173, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20176, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20177, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20180, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20198, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20202, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20211, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20249, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20286, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20287, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20288, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20289, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19297, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventId"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20144, 0),
        BrowseName=QualifiedName("EventId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventType"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20145, 0),
        BrowseName=QualifiedName("EventType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20146, 0),
        BrowseName=QualifiedName("SourceNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20147, 0),
        BrowseName=QualifiedName("SourceName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Time"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20148, 0),
        BrowseName=QualifiedName("Time", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceiveTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20149, 0),
        BrowseName=QualifiedName("ReceiveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Message"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20151, 0),
        BrowseName=QualifiedName("Message", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Severity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20152, 0),
        BrowseName=QualifiedName("Severity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20153, 0),
        BrowseName=QualifiedName("ConditionClassId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionClassName"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20154, 0),
        BrowseName=QualifiedName("ConditionClassName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConditionName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20157, 0),
        BrowseName=QualifiedName("ConditionName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BranchId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20158, 0),
        BrowseName=QualifiedName("BranchId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Retain"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20159, 0),
        BrowseName=QualifiedName("Retain", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnabledState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20160, 0),
        BrowseName=QualifiedName("EnabledState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20161, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20161, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20160, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20160, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Quality"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20169, 0),
        BrowseName=QualifiedName("Quality", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20170, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20170, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20169, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20169, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastSeverity"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20171, 0),
        BrowseName=QualifiedName("LastSeverity", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20172, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20171, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20171, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Comment"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20173, 0),
        BrowseName=QualifiedName("Comment", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(9002, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20174, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(9002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SourceTimestamp"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20174, 0),
        BrowseName=QualifiedName("SourceTimestamp", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20173, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20173, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ClientUserId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20175, 0),
        BrowseName=QualifiedName("ClientUserId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Disable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20176, 0),
        BrowseName=QualifiedName("Disable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(20176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Enable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20177, 0),
        BrowseName=QualifiedName("Enable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(20177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddComment"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20178, 0),
        BrowseName=QualifiedName("AddComment", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20179, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(20178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20179, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20178, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20178, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AckedState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20180, 0),
        BrowseName=QualifiedName("AckedState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20181, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20181, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20180, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20180, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Acknowledge"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20198, 0),
        BrowseName=QualifiedName("Acknowledge", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20199, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(3065, 0),
        SourceNodeId=NumericNodeId(20198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='EventId',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The identifier for the event to comment.'),
                ),
            ua.Argument(
                Name='Comment',
                DataType=NumericNodeId(21, 0),
                ValueRank=-1,
                Description=LocalizedText(Text='The comment to add to the condition.'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20199, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20198, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20198, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ActiveState"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20202, 0),
        BrowseName=QualifiedName("ActiveState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(8995, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20203, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Id"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20203, 0),
        BrowseName=QualifiedName("Id", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20202, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20202, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputNode"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20211, 0),
        BrowseName=QualifiedName("InputNode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SuppressedOrShelved"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20249, 0),
        BrowseName=QualifiedName("SuppressedOrShelved", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NormalState"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20286, 0),
        BrowseName=QualifiedName("NormalState", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrustListId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20287, 0),
        BrowseName=QualifiedName("TrustListId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20288, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UpdateFrequency"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20289, 0),
        BrowseName=QualifiedName("UpdateFrequency", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20143, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20143, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetRejectedList"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23526, 0),
        BrowseName=QualifiedName("GetRejectedList", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12555, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23526, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23527, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23526, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23526, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Certificates',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23527, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23526, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23526, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("CertificateGroupFolderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13813, 0),
        BrowseName=QualifiedName("CertificateGroupFolderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13814, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13882, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(13813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13916, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(13813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DefaultApplicationGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13814, 0),
        BrowseName=QualifiedName("DefaultApplicationGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13813, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13847, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13813, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13815, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13814, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13816, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13818, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13819, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13826, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13834, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13837, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13814, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13816, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13817, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13818, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13819, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13821, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13822, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13823, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13822, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13821, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13821, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13823, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13821, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13821, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13824, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13825, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13825, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13824, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13824, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13826, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13828, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13827, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13826, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13828, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13826, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13829, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13830, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13829, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13829, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13831, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13832, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13832, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13831, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13831, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13833, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13831, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13831, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13834, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13835, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13834, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13834, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13836, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13837, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13815, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13838, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13815, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13838, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13837, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13837, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13839, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13837, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13837, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13847, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13814, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13814, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DefaultHttpsGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13848, 0),
        BrowseName=QualifiedName("DefaultHttpsGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13813, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13881, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13813, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13849, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13848, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13850, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13851, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13858, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13860, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13863, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13865, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13868, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13870, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13871, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13848, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13850, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13851, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13852, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13853, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13855, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13856, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13857, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13856, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13855, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13855, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13857, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13855, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13855, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13858, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13859, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13859, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13858, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13858, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13860, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13861, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13862, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13861, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13860, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13862, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13860, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13860, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13863, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13864, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13864, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13863, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13863, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13865, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13866, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13867, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13866, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13865, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13865, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13867, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13865, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13865, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13868, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13869, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13869, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13868, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13868, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13870, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13871, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13849, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13872, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13873, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13849, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13872, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13871, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13871, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13873, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13871, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13871, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13881, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13848, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13848, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DefaultUserTokenGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13882, 0),
        BrowseName=QualifiedName("DefaultUserTokenGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13813, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13813, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13883, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13882, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13884, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13885, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13886, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13887, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13889, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13892, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13899, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13902, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13904, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13905, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13882, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13884, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13885, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13886, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13887, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13889, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13890, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13890, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13889, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13889, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13891, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13889, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13889, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13892, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13893, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13893, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13892, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13892, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13894, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13895, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13896, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13895, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13894, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13894, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13896, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13894, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13894, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13897, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13898, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13898, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13897, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13897, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13899, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13900, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13901, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13900, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13899, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13899, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13901, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13899, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13899, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13902, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13903, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13903, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13902, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13902, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13904, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13904, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13904, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13904, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13905, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13883, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13906, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13907, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13883, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13906, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13905, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13905, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13907, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13905, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13907, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13907, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13907, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13905, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13915, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13882, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13882, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<AdditionalGroup>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13916, 0),
        BrowseName=QualifiedName("<AdditionalGroup>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13813, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13949, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(13916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13813, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13917, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13916, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13918, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13919, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13920, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13921, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13923, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13926, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13928, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13931, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13933, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13936, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13938, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13939, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13916, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13918, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13919, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13919, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13919, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13919, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13920, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13920, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13920, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13920, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13921, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13921, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13921, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13921, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13923, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13923, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13924, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13923, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13925, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13923, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13923, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13924, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13923, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13924, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13924, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13924, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13923, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13925, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13923, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13925, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13925, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13925, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13923, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13926, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13926, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13927, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13926, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13926, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13927, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13926, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13926, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13928, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13928, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13929, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13928, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13930, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13928, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13928, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13929, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13928, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13928, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13930, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13928, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13930, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13928, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13931, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13931, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13931, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13931, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13932, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13931, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13931, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13933, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13934, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13934, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13933, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13933, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13935, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13933, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13933, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13936, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13937, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13937, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13936, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13937, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13937, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13937, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13936, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13938, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13938, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13938, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13938, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13939, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13941, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13940, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13939, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13939, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13941, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13939, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13941, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13941, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13941, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13939, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13949, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13916, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13916, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("CertificateType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12556, 0),
        BrowseName=QualifiedName("CertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ApplicationCertificateType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12557, 0),
        BrowseName=QualifiedName("ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(12556, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12556, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("HttpsCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12558, 0),
        BrowseName=QualifiedName("HttpsCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(12556, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12556, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("UserCredentialCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15181, 0),
        BrowseName=QualifiedName("UserCredentialCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(12556, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12556, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RsaMinApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12559, 0),
        BrowseName=QualifiedName("RsaMinApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(12557, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("RsaSha256ApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12560, 0),
        BrowseName=QualifiedName("RsaSha256ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(12557, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EccApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23537, 0),
        BrowseName=QualifiedName("EccApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(12557, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23537, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EccNistP256ApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23538, 0),
        BrowseName=QualifiedName("EccNistP256ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(23537, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23538, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23537, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EccNistP384ApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23539, 0),
        BrowseName=QualifiedName("EccNistP384ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(23537, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23539, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23537, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EccBrainpoolP256r1ApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23540, 0),
        BrowseName=QualifiedName("EccBrainpoolP256r1ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(23537, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23540, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23537, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EccBrainpoolP384r1ApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23541, 0),
        BrowseName=QualifiedName("EccBrainpoolP384r1ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(23537, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23541, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23537, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EccCurve25519ApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23542, 0),
        BrowseName=QualifiedName("EccCurve25519ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(23537, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23537, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("EccCurve448ApplicationCertificateType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23543, 0),
        BrowseName=QualifiedName("EccCurve448ApplicationCertificateType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(23537, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23543, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23537, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TrustListUpdatedAuditEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12561, 0),
        BrowseName=QualifiedName("TrustListUpdatedAuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12561, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ServerConfigurationType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12581, 0),
        BrowseName=QualifiedName("ServerConfigurationType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13950, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12708, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12583, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12584, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12585, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12775, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("CertificateGroups"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13950, 0),
        BrowseName=QualifiedName("CertificateGroups", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(13813, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13951, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13813, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DefaultApplicationGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13951, 0),
        BrowseName=QualifiedName("DefaultApplicationGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13950, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13984, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13950, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13952, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(13951, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13953, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13954, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13955, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13956, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13958, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13961, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13963, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13966, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13968, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13971, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13973, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13974, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13951, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13953, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13953, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13953, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13953, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13954, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13955, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13956, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13958, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13959, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13960, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13959, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13958, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13958, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13960, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13958, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13958, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13961, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13962, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13962, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13961, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13961, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13963, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13964, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13965, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13963, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13964, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13963, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13963, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13965, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13963, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13965, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13965, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13965, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13963, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13966, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13966, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13967, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13966, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13966, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13967, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13966, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13967, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13967, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13967, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13966, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13968, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13968, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13969, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13968, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13970, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13968, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13968, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13969, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13968, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13969, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13969, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13969, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13968, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13970, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13968, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13970, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13970, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13970, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13968, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13971, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13971, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13972, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13971, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13971, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13972, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13971, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13972, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13972, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13972, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13971, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13973, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13973, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13973, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13973, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13974, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(13952, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13974, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13975, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13974, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13976, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13974, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13974, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13952, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13975, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13974, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13975, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13975, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13975, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13974, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13976, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13974, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13976, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13976, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13976, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13974, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13984, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13951, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13984, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13984, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13984, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13951, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerCapabilities"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12708, 0),
        BrowseName=QualifiedName("ServerCapabilities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SupportedPrivateKeyFormats"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12583, 0),
        BrowseName=QualifiedName("SupportedPrivateKeyFormats", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxTrustListSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12584, 0),
        BrowseName=QualifiedName("MaxTrustListSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MulticastDnsEnabled"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12585, 0),
        BrowseName=QualifiedName("MulticastDnsEnabled", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("UpdateCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12616, 0),
        BrowseName=QualifiedName("UpdateCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12617, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12618, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CertificateGroupId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CertificateTypeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Certificate',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IssuerCertificates',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='PrivateKeyFormat',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='PrivateKey',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12617, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplyChangesRequired',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12618, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ApplyChanges"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12734, 0),
        BrowseName=QualifiedName("ApplyChanges", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateSigningRequest"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12731, 0),
        BrowseName=QualifiedName("CreateSigningRequest", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12732, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12733, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CertificateGroupId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CertificateTypeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SubjectName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RegeneratePrivateKey',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Nonce',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12732, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12731, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12731, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CertificateRequest',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12733, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12731, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12731, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetRejectedList"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12775, 0),
        BrowseName=QualifiedName("GetRejectedList", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12581, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12776, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Certificates',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12776, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12775, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12775, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("CertificateUpdatedAuditEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12620, 0),
        BrowseName=QualifiedName("CertificateUpdatedAuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13735, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13736, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateGroup"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13735, 0),
        BrowseName=QualifiedName("CertificateGroup", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12620, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateType"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13736, 0),
        BrowseName=QualifiedName("CertificateType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12620, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(13736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12620, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ServerConfiguration"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12637, 0),
        BrowseName=QualifiedName("ServerConfiguration", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12581, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14053, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12639, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12640, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12641, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12740, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12777, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12581, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("CertificateGroups"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14053, 0),
        BrowseName=QualifiedName("CertificateGroups", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(13813, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14156, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14088, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14122, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13813, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DefaultApplicationGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14156, 0),
        BrowseName=QualifiedName("DefaultApplicationGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14053, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14161, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14053, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12642, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14156, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14157, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14158, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12646, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12647, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12652, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12655, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12657, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12660, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12662, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12663, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12666, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12668, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12670, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14156, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12643, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14157, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14158, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12646, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12647, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12648, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12649, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12648, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12647, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12647, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12649, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12647, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12647, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12650, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12651, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12651, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12650, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12650, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12652, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12654, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12653, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12652, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12653, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12653, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12652, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12654, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12652, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12652, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12655, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12656, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12656, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12655, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12655, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12657, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12658, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12659, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12658, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12657, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12657, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12659, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12657, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12657, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12660, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12660, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12661, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12660, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12661, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12660, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12661, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12661, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12660, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12662, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12662, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12662, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12663, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12663, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12664, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12663, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12663, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12664, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12663, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12664, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12664, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12663, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12665, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12663, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12665, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12665, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12663, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CloseAndUpdate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12666, 0),
        BrowseName=QualifiedName("CloseAndUpdate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14160, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12667, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14160, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12666, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12666, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplyChangesRequired',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12667, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12666, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12667, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12667, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12666, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12668, 0),
        BrowseName=QualifiedName("AddCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12669, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Certificate',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12669, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12668, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12669, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12669, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12668, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12670, 0),
        BrowseName=QualifiedName("RemoveCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12642, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12670, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12671, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12670, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Thumbprint',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12671, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12670, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12670, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14161, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14156, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14156, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DefaultHttpsGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14088, 0),
        BrowseName=QualifiedName("DefaultHttpsGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14053, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14121, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14053, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14089, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14088, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14090, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14091, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14095, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14103, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14105, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14108, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14111, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14114, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14117, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14088, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14090, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14091, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14092, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14093, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14095, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14096, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14097, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14096, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14095, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14097, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14095, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14095, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14098, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14099, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14099, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14098, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14098, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14100, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14101, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14100, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14100, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14102, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14100, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14100, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14103, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14104, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14104, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14103, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14103, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14105, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14106, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14106, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14105, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14105, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14107, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14105, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14105, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14108, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14109, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14109, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14108, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14108, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14110, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14111, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14112, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14112, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14113, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CloseAndUpdate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14114, 0),
        BrowseName=QualifiedName("CloseAndUpdate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14115, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14116, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14115, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14114, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14114, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplyChangesRequired',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14116, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14114, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14114, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14117, 0),
        BrowseName=QualifiedName("AddCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Certificate',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14118, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14117, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14117, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14119, 0),
        BrowseName=QualifiedName("RemoveCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14089, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14120, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Thumbprint',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14120, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14119, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14119, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14121, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14088, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14088, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DefaultUserTokenGroup"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14122, 0),
        BrowseName=QualifiedName("DefaultUserTokenGroup", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14053, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12555, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14155, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14053, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TrustList"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14123, 0),
        BrowseName=QualifiedName("TrustList", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14122, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(12522, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14124, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14125, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14126, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14129, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14132, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14134, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14137, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14139, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14142, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14144, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14145, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14148, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14151, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14153, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14122, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Size"),
        DataType=ua.NodeId(ua.ObjectIds.UInt64),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14124, 0),
        BrowseName=QualifiedName("Size", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Writable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14125, 0),
        BrowseName=QualifiedName("Writable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserWritable"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14126, 0),
        BrowseName=QualifiedName("UserWritable", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OpenCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14127, 0),
        BrowseName=QualifiedName("OpenCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Open"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14129, 0),
        BrowseName=QualifiedName("Open", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14130, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14131, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Mode',
                DataType=NumericNodeId(3, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14130, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14129, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14129, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14131, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14129, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14129, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Close"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14132, 0),
        BrowseName=QualifiedName("Close", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14133, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14133, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14132, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14132, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Read"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14134, 0),
        BrowseName=QualifiedName("Read", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14135, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14136, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Length',
                DataType=NumericNodeId(6, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14135, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14134, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14134, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14136, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14134, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14134, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Write"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14137, 0),
        BrowseName=QualifiedName("Write", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14138, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Data',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14138, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14137, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14137, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14139, 0),
        BrowseName=QualifiedName("GetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14140, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14141, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14140, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14139, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14139, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14141, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14139, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14139, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetPosition"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14142, 0),
        BrowseName=QualifiedName("SetPosition", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14143, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Position',
                DataType=NumericNodeId(9, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14143, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14142, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LastUpdateTime"),
        DataType=NumericNodeId(294, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14144, 0),
        BrowseName=QualifiedName("LastUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("OpenWithMasks"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14145, 0),
        BrowseName=QualifiedName("OpenWithMasks", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14146, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14147, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Masks',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14146, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14145, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14145, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14147, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14145, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14145, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CloseAndUpdate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14148, 0),
        BrowseName=QualifiedName("CloseAndUpdate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14149, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14150, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FileHandle',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14149, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14148, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14148, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplyChangesRequired',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14150, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14148, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14148, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14151, 0),
        BrowseName=QualifiedName("AddCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14152, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Certificate',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14152, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14151, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14151, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14153, 0),
        BrowseName=QualifiedName("RemoveCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14123, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14154, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14123, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Thumbprint',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IsTrustedCertificate',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14154, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14153, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14153, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CertificateTypes"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14155, 0),
        BrowseName=QualifiedName("CertificateTypes", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14122, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14122, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerCapabilities"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12710, 0),
        BrowseName=QualifiedName("ServerCapabilities", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SupportedPrivateKeyFormats"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12639, 0),
        BrowseName=QualifiedName("SupportedPrivateKeyFormats", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxTrustListSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12640, 0),
        BrowseName=QualifiedName("MaxTrustListSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MulticastDnsEnabled"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12641, 0),
        BrowseName=QualifiedName("MulticastDnsEnabled", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("UpdateCertificate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13737, 0),
        BrowseName=QualifiedName("UpdateCertificate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13738, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13739, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(13737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CertificateGroupId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CertificateTypeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Certificate',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='IssuerCertificates',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='PrivateKeyFormat',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='PrivateKey',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13738, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13737, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13737, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ApplyChangesRequired',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(13739, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(13737, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(13739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(13739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13737, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ApplyChanges"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12740, 0),
        BrowseName=QualifiedName("ApplyChanges", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateSigningRequest"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12737, 0),
        BrowseName=QualifiedName("CreateSigningRequest", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12738, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12739, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CertificateGroupId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CertificateTypeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SubjectName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RegeneratePrivateKey',
                DataType=NumericNodeId(1, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Nonce',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12738, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12737, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12737, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CertificateRequest',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12739, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12737, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12737, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetRejectedList"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12777, 0),
        BrowseName=QualifiedName("GetRejectedList", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12778, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Certificates',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12778, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12777, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12777, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("KeyCredentialConfigurationFolderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17496, 0),
        BrowseName=QualifiedName("KeyCredentialConfigurationFolderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17511, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<ServiceName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17511, 0),
        BrowseName=QualifiedName("<ServiceName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17496, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(18001, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17512, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17513, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17496, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResourceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17512, 0),
        BrowseName=QualifiedName("ResourceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17511, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17511, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17513, 0),
        BrowseName=QualifiedName("ProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17511, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17511, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateCredential"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17522, 0),
        BrowseName=QualifiedName("CreateCredential", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17496, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17523, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17524, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17496, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ResourceUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ProfileUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='EndpointUrls',
                DataType=NumericNodeId(12, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17523, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17522, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17523, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17523, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17523, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CredentialNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17524, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17522, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17522, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("KeyCredentialConfiguration"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18155, 0),
        BrowseName=QualifiedName("KeyCredentialConfiguration", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17496, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17496, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("KeyCredentialConfigurationType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18001, 0),
        BrowseName=QualifiedName("KeyCredentialConfigurationType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18069, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18165, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18004, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18005, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17534, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18006, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18008, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResourceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18069, 0),
        BrowseName=QualifiedName("ResourceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18001, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18165, 0),
        BrowseName=QualifiedName("ProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18001, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrls"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18004, 0),
        BrowseName=QualifiedName("EndpointUrls", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18001, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServiceStatus"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18005, 0),
        BrowseName=QualifiedName("ServiceStatus", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18001, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetEncryptingKey"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17534, 0),
        BrowseName=QualifiedName("GetEncryptingKey", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(18001, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17535, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17536, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CredentialId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RequestedSecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17535, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17534, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17534, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='PublicKey',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RevisedSecurityPolicyUri',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17536, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17534, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17536, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17536, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17536, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17534, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("UpdateCredential"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18006, 0),
        BrowseName=QualifiedName("UpdateCredential", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(18001, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18007, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='CredentialId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CredentialSecret',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CertificateThumbprint',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18007, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18006, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18006, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("DeleteCredential"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18008, 0),
        BrowseName=QualifiedName("DeleteCredential", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(18001, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18001, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("KeyCredentialAuditEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18011, 0),
        BrowseName=QualifiedName("KeyCredentialAuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2127, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18028, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2127, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResourceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18028, 0),
        BrowseName=QualifiedName("ResourceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18011, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("KeyCredentialUpdatedAuditEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18029, 0),
        BrowseName=QualifiedName("KeyCredentialUpdatedAuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(18011, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("KeyCredentialDeletedAuditEventType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18047, 0),
        BrowseName=QualifiedName("KeyCredentialDeletedAuditEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(18011, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18011, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuthorizationServicesConfigurationFolderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23556, 0),
        BrowseName=QualifiedName("AuthorizationServicesConfigurationFolderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23557, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<ServiceName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23557, 0),
        BrowseName=QualifiedName("<ServiceName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23556, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(17852, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23558, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23559, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23560, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23556, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServiceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23558, 0),
        BrowseName=QualifiedName("ServiceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23557, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServiceCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23559, 0),
        BrowseName=QualifiedName("ServiceCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23557, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IssuerEndpointUrl"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23560, 0),
        BrowseName=QualifiedName("IssuerEndpointUrl", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23557, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AuthorizationServices"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17732, 0),
        BrowseName=QualifiedName("AuthorizationServices", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12637, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(61, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12637, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AuthorizationServiceConfigurationType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17852, 0),
        BrowseName=QualifiedName("AuthorizationServiceConfigurationType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18072, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17860, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServiceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18072, 0),
        BrowseName=QualifiedName("ServiceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17852, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17852, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServiceCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17860, 0),
        BrowseName=QualifiedName("ServiceCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17852, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17852, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IssuerEndpointUrl"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18073, 0),
        BrowseName=QualifiedName("IssuerEndpointUrl", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17852, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17852, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AggregateConfigurationType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11187, 0),
        BrowseName=QualifiedName("AggregateConfigurationType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11188, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11190, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11191, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TreatUncertainAsBad"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11188, 0),
        BrowseName=QualifiedName("TreatUncertainAsBad", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11187, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11187, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PercentDataBad"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11189, 0),
        BrowseName=QualifiedName("PercentDataBad", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11187, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11187, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PercentDataGood"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11190, 0),
        BrowseName=QualifiedName("PercentDataGood", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11187, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11187, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UseSlopedExtrapolation"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11191, 0),
        BrowseName=QualifiedName("UseSlopedExtrapolation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11187, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11187, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Interpolative"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2341, 0),
        BrowseName=QualifiedName("Interpolative", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Average"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2342, 0),
        BrowseName=QualifiedName("Average", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2342, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TimeAverage"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2343, 0),
        BrowseName=QualifiedName("TimeAverage", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2343, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TimeAverage2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11285, 0),
        BrowseName=QualifiedName("TimeAverage2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Total"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2344, 0),
        BrowseName=QualifiedName("Total", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2344, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Total2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11304, 0),
        BrowseName=QualifiedName("Total2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11304, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Minimum"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2346, 0),
        BrowseName=QualifiedName("Minimum", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2346, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Maximum"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2347, 0),
        BrowseName=QualifiedName("Maximum", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2347, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MinimumActualTime"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2348, 0),
        BrowseName=QualifiedName("MinimumActualTime", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MaximumActualTime"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2349, 0),
        BrowseName=QualifiedName("MaximumActualTime", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2349, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Range"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2350, 0),
        BrowseName=QualifiedName("Range", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2350, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Minimum2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11286, 0),
        BrowseName=QualifiedName("Minimum2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Maximum2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11287, 0),
        BrowseName=QualifiedName("Maximum2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MinimumActualTime2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11305, 0),
        BrowseName=QualifiedName("MinimumActualTime2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11305, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MaximumActualTime2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11306, 0),
        BrowseName=QualifiedName("MaximumActualTime2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Range2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11288, 0),
        BrowseName=QualifiedName("Range2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("AnnotationCount"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2351, 0),
        BrowseName=QualifiedName("AnnotationCount", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Count"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2352, 0),
        BrowseName=QualifiedName("Count", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2352, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DurationInStateZero"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11307, 0),
        BrowseName=QualifiedName("DurationInStateZero", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DurationInStateNonZero"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11308, 0),
        BrowseName=QualifiedName("DurationInStateNonZero", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("NumberOfTransitions"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2355, 0),
        BrowseName=QualifiedName("NumberOfTransitions", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Start"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2357, 0),
        BrowseName=QualifiedName("Start", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("End"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2358, 0),
        BrowseName=QualifiedName("End", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Delta"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2359, 0),
        BrowseName=QualifiedName("Delta", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("StartBound"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11505, 0),
        BrowseName=QualifiedName("StartBound", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("EndBound"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11506, 0),
        BrowseName=QualifiedName("EndBound", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DeltaBounds"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11507, 0),
        BrowseName=QualifiedName("DeltaBounds", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DurationGood"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2360, 0),
        BrowseName=QualifiedName("DurationGood", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DurationBad"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2361, 0),
        BrowseName=QualifiedName("DurationBad", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2361, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("PercentGood"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2362, 0),
        BrowseName=QualifiedName("PercentGood", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2362, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("PercentBad"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2363, 0),
        BrowseName=QualifiedName("PercentBad", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("WorstQuality"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(2364, 0),
        BrowseName=QualifiedName("WorstQuality", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(2364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("WorstQuality2"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11292, 0),
        BrowseName=QualifiedName("WorstQuality2", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11292, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("StandardDeviationSample"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11426, 0),
        BrowseName=QualifiedName("StandardDeviationSample", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("StandardDeviationPopulation"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11427, 0),
        BrowseName=QualifiedName("StandardDeviationPopulation", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("VarianceSample"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11428, 0),
        BrowseName=QualifiedName("VarianceSample", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("VariancePopulation"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11429, 0),
        BrowseName=QualifiedName("VariancePopulation", 0),
        NodeClass_=NodeClass.Object,
        TypeDefinition=NumericNodeId(2340, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2340, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DecimalDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17861, 0),
        BrowseName=QualifiedName("DecimalDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataTypeSchemaHeader"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15534, 0),
        BrowseName=QualifiedName("DataTypeSchemaHeader", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataTypeDescription"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14525, 0),
        BrowseName=QualifiedName("DataTypeDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14525, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("StructureDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15487, 0),
        BrowseName=QualifiedName("StructureDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(14525, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14525, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EnumDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15488, 0),
        BrowseName=QualifiedName("EnumDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(14525, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14525, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SimpleTypeDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15005, 0),
        BrowseName=QualifiedName("SimpleTypeDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(14525, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15005, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14525, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UABinaryFileDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15006, 0),
        BrowseName=QualifiedName("UABinaryFileDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15534, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15534, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PubSubState"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14647, 0),
        BrowseName=QualifiedName("PubSubState", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14648, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Disabled'), LocalizedText('Paused'), LocalizedText('Operational'), LocalizedText('Error')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14648, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14647, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14647, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetMetaDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14523, 0),
        BrowseName=QualifiedName("DataSetMetaDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15534, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14523, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15534, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FieldMetaData"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14524, 0),
        BrowseName=QualifiedName("FieldMetaData", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetFieldFlags"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15904, 0),
        BrowseName=QualifiedName("DataSetFieldFlags", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(5, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15904, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15577, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15904, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(5, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('PromotedField')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15577, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15904, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15577, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15577, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15577, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15904, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ConfigurationVersionDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14593, 0),
        BrowseName=QualifiedName("ConfigurationVersionDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PublishedDataSetDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15578, 0),
        BrowseName=QualifiedName("PublishedDataSetDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15578, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PublishedDataSetSourceDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15580, 0),
        BrowseName=QualifiedName("PublishedDataSetSourceDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PublishedVariableDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14273, 0),
        BrowseName=QualifiedName("PublishedVariableDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PublishedDataItemsDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15581, 0),
        BrowseName=QualifiedName("PublishedDataItemsDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15580, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15580, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PublishedEventsDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15582, 0),
        BrowseName=QualifiedName("PublishedEventsDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15580, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15580, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetFieldContentMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15583, 0),
        BrowseName=QualifiedName("DataSetFieldContentMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15584, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('StatusCode'), LocalizedText('SourceTimestamp'), LocalizedText('ServerTimestamp'), LocalizedText('SourcePicoSeconds'), LocalizedText('ServerPicoSeconds'), LocalizedText('RawData')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15584, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15583, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15583, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetWriterDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15597, 0),
        BrowseName=QualifiedName("DataSetWriterDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetWriterTransportDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15598, 0),
        BrowseName=QualifiedName("DataSetWriterTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetWriterMessageDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15605, 0),
        BrowseName=QualifiedName("DataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PubSubGroupDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15609, 0),
        BrowseName=QualifiedName("PubSubGroupDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("WriterGroupDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15480, 0),
        BrowseName=QualifiedName("WriterGroupDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15609, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("WriterGroupTransportDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15611, 0),
        BrowseName=QualifiedName("WriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("WriterGroupMessageDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15616, 0),
        BrowseName=QualifiedName("WriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PubSubConnectionDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15617, 0),
        BrowseName=QualifiedName("PubSubConnectionDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ConnectionTransportDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15618, 0),
        BrowseName=QualifiedName("ConnectionTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NetworkAddressDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15502, 0),
        BrowseName=QualifiedName("NetworkAddressDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NetworkAddressUrlDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15510, 0),
        BrowseName=QualifiedName("NetworkAddressUrlDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15502, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15502, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReaderGroupDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15520, 0),
        BrowseName=QualifiedName("ReaderGroupDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15609, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15520, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReaderGroupTransportDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15621, 0),
        BrowseName=QualifiedName("ReaderGroupTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReaderGroupMessageDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15622, 0),
        BrowseName=QualifiedName("ReaderGroupMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetReaderDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15623, 0),
        BrowseName=QualifiedName("DataSetReaderDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetReaderTransportDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15628, 0),
        BrowseName=QualifiedName("DataSetReaderTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15628, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetReaderMessageDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15629, 0),
        BrowseName=QualifiedName("DataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SubscribedDataSetDataType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15630, 0),
        BrowseName=QualifiedName("SubscribedDataSetDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TargetVariablesDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15631, 0),
        BrowseName=QualifiedName("TargetVariablesDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15630, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15630, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FieldTargetDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14744, 0),
        BrowseName=QualifiedName("FieldTargetDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("OverrideValueHandling"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15874, 0),
        BrowseName=QualifiedName("OverrideValueHandling", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15875, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Disabled'), LocalizedText('LastUsableValue'), LocalizedText('OverrideValue')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15875, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15874, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15874, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SubscribedDataSetMirrorDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15635, 0),
        BrowseName=QualifiedName("SubscribedDataSetMirrorDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15630, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15630, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PubSubConfigurationDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15530, 0),
        BrowseName=QualifiedName("PubSubConfigurationDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15530, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataSetOrderingType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20408, 0),
        BrowseName=QualifiedName("DataSetOrderingType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15641, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(20408, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Undefined'), LocalizedText('AscendingWriterId'), LocalizedText('AscendingWriterIdSingle')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15641, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20408, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20408, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UadpNetworkMessageContentMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15642, 0),
        BrowseName=QualifiedName("UadpNetworkMessageContentMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('PublisherId'), LocalizedText('GroupHeader'), LocalizedText('WriterGroupId'), LocalizedText('GroupVersion'), LocalizedText('NetworkMessageNumber'), LocalizedText('SequenceNumber'), LocalizedText('PayloadHeader'), LocalizedText('Timestamp'), LocalizedText('PicoSeconds'), LocalizedText('DataSetClassId'), LocalizedText('PromotedFields')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15643, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15642, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15642, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UadpWriterGroupMessageDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15645, 0),
        BrowseName=QualifiedName("UadpWriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15616, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UadpDataSetMessageContentMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15646, 0),
        BrowseName=QualifiedName("UadpDataSetMessageContentMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15647, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Timestamp'), LocalizedText('PicoSeconds'), LocalizedText('Status'), LocalizedText('MajorVersion'), LocalizedText('MinorVersion'), LocalizedText('SequenceNumber')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15647, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15646, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15646, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UadpDataSetWriterMessageDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15652, 0),
        BrowseName=QualifiedName("UadpDataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15605, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15605, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UadpDataSetReaderMessageDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15653, 0),
        BrowseName=QualifiedName("UadpDataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15629, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15653, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15629, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("JsonNetworkMessageContentMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15654, 0),
        BrowseName=QualifiedName("JsonNetworkMessageContentMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15655, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('NetworkMessageHeader'), LocalizedText('DataSetMessageHeader'), LocalizedText('SingleDataSetMessage'), LocalizedText('PublisherId'), LocalizedText('DataSetClassId'), LocalizedText('ReplyTo')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15655, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15654, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15654, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("JsonWriterGroupMessageDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15657, 0),
        BrowseName=QualifiedName("JsonWriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15616, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("JsonDataSetMessageContentMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15658, 0),
        BrowseName=QualifiedName("JsonDataSetMessageContentMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15659, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('DataSetWriterId'), LocalizedText('MetaDataVersion'), LocalizedText('SequenceNumber'), LocalizedText('Timestamp'), LocalizedText('Status')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15659, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15658, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15658, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("JsonDataSetWriterMessageDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15664, 0),
        BrowseName=QualifiedName("JsonDataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15605, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15664, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15605, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("JsonDataSetReaderMessageDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15665, 0),
        BrowseName=QualifiedName("JsonDataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15629, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15665, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15629, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DatagramConnectionTransportDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17467, 0),
        BrowseName=QualifiedName("DatagramConnectionTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15618, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17467, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15618, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DatagramWriterGroupTransportDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15532, 0),
        BrowseName=QualifiedName("DatagramWriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15611, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15532, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15611, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrokerConnectionTransportDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15007, 0),
        BrowseName=QualifiedName("BrokerConnectionTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15618, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15618, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrokerTransportQualityOfService"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15008, 0),
        BrowseName=QualifiedName("BrokerTransportQualityOfService", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15009, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('NotSpecified'), LocalizedText('BestEffort'), LocalizedText('AtLeastOnce'), LocalizedText('AtMostOnce'), LocalizedText('ExactlyOnce')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15009, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15008, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15008, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrokerWriterGroupTransportDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15667, 0),
        BrowseName=QualifiedName("BrokerWriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15611, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15667, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15611, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrokerDataSetWriterTransportDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15669, 0),
        BrowseName=QualifiedName("BrokerDataSetWriterTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15598, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15669, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15598, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrokerDataSetReaderTransportDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15670, 0),
        BrowseName=QualifiedName("BrokerDataSetReaderTransportDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15628, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15670, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15628, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubKeyServiceType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15906, 0),
        BrowseName=QualifiedName("PubSubKeyServiceType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15907, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15910, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15913, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15906, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetSecurityKeys"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15907, 0),
        BrowseName=QualifiedName("GetSecurityKeys", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15907, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15908, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15907, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15909, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15907, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15907, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='StartingTokenId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RequestedKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15908, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15907, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15908, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15908, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15908, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15907, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FirstTokenId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Keys',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='TimeToNextKey',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='KeyLifetime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15909, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15907, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15909, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15909, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15909, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15907, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15910, 0),
        BrowseName=QualifiedName("GetSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15910, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15911, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15910, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15912, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15910, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15910, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15911, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15910, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15911, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15911, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15911, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15910, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15912, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15910, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15910, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SecurityGroups"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15913, 0),
        BrowseName=QualifiedName("SecurityGroups", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15906, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15452, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15914, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15917, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15452, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15914, 0),
        BrowseName=QualifiedName("AddSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15913, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15914, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15915, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15914, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15916, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15914, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15914, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15913, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='KeyLifetime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxFutureKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxPastKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15915, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15914, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15914, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15916, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15914, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15914, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15917, 0),
        BrowseName=QualifiedName("RemoveSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15913, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15918, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15913, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15918, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15917, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SecurityGroupFolderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15452, 0),
        BrowseName=QualifiedName("SecurityGroupFolderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(15452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15453, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15459, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15461, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15464, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<SecurityGroupFolderName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15453, 0),
        BrowseName=QualifiedName("<SecurityGroupFolderName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15452, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(15452, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15454, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15457, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15452, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(15453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15452, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15454, 0),
        BrowseName=QualifiedName("AddSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15453, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15455, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15453, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='KeyLifetime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxFutureKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxPastKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15455, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15454, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15454, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15456, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15454, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15454, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15457, 0),
        BrowseName=QualifiedName("RemoveSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15453, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15458, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15453, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15458, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15457, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15457, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<SecurityGroupName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15459, 0),
        BrowseName=QualifiedName("<SecurityGroupName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15452, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15471, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15460, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15010, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15011, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15012, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15043, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15471, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15452, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15460, 0),
        BrowseName=QualifiedName("SecurityGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15459, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15459, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeyLifetime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15010, 0),
        BrowseName=QualifiedName("KeyLifetime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15459, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15459, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityPolicyUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15011, 0),
        BrowseName=QualifiedName("SecurityPolicyUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15459, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15459, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxFutureKeyCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15012, 0),
        BrowseName=QualifiedName("MaxFutureKeyCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15459, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15459, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxPastKeyCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15043, 0),
        BrowseName=QualifiedName("MaxPastKeyCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15459, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15459, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15461, 0),
        BrowseName=QualifiedName("AddSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15452, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15463, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15452, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='KeyLifetime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxFutureKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxPastKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15462, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15461, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15461, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15463, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15461, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15461, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15464, 0),
        BrowseName=QualifiedName("RemoveSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15452, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15452, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15465, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15464, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15464, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SecurityGroupType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15471, 0),
        BrowseName=QualifiedName("SecurityGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15471, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15472, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15471, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15046, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15471, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15047, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15471, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15048, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15471, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15471, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15472, 0),
        BrowseName=QualifiedName("SecurityGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15471, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15471, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeyLifetime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15046, 0),
        BrowseName=QualifiedName("KeyLifetime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15471, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15471, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityPolicyUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15047, 0),
        BrowseName=QualifiedName("SecurityPolicyUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15471, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15471, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxFutureKeyCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15048, 0),
        BrowseName=QualifiedName("MaxFutureKeyCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15471, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15471, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxPastKeyCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15056, 0),
        BrowseName=QualifiedName("MaxPastKeyCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15471, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15471, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PublishSubscribeType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14416, 0),
        BrowseName=QualifiedName("PublishSubscribeType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(15906, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(14476, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14417, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17296, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14432, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14434, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15844, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17479, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15906, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<ConnectionName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14417, 0),
        BrowseName=QualifiedName("<ConnectionName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(14476, 0),
        TypeDefinition=NumericNodeId(14209, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14418, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17292, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17478, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14419, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(14476, 0),
        SourceNodeId=NumericNodeId(14417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublisherId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14418, 0),
        BrowseName=QualifiedName("PublisherId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14417, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14417, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransportProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17292, 0),
        BrowseName=QualifiedName("TransportProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14417, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(16309, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17292, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17706, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17292, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17292, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17292, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14417, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Selections"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17706, 0),
        BrowseName=QualifiedName("Selections", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17292, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17292, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConnectionProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17478, 0),
        BrowseName=QualifiedName("ConnectionProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14417, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14417, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Address"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14423, 0),
        BrowseName=QualifiedName("Address", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14417, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(21145, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15533, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21145, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14417, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkInterface"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15533, 0),
        BrowseName=QualifiedName("NetworkInterface", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14423, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(16309, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15533, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17503, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15533, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15533, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15533, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Selections"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17503, 0),
        BrowseName=QualifiedName("Selections", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15533, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15533, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14419, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14417, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14417, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14420, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14419, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14419, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("SetSecurityKeys"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17296, 0),
        BrowseName=QualifiedName("SetSecurityKeys", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17297, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CurrentTokenId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='CurrentKey',
                DataType=NumericNodeId(15, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FutureKeys',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='TimeToNextKey',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='KeyLifetime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17297, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17296, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17296, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddConnection"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16598, 0),
        BrowseName=QualifiedName("AddConnection", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16599, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16600, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Configuration',
                DataType=NumericNodeId(15617, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16599, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16598, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16598, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConnectionId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16600, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16598, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16598, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveConnection"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14432, 0),
        BrowseName=QualifiedName("RemoveConnection", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14433, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConnectionId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14433, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14432, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14432, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("PublishedDataSets"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14434, 0),
        BrowseName=QualifiedName("PublishedDataSets", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14477, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15844, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15845, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15844, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15844, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Diagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18715, 0),
        BrowseName=QualifiedName("Diagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19732, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18716, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18717, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18722, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18727, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18729, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18760, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19732, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18716, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18717, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18718, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18719, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18720, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18718, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18717, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18717, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18719, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18717, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18717, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18720, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18717, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18717, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18722, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18723, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18724, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18723, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18722, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18722, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18724, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18722, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18722, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18725, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18722, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18722, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18727, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(18715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18728, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18729, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18730, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18735, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18740, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18745, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18750, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18730, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18729, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18732, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18733, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18729, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18731, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18730, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18730, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18732, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18730, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18730, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18733, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18730, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18730, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18735, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18729, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18736, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18738, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18729, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18736, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18735, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18735, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18737, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18735, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18735, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18738, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18735, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18735, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18740, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18729, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18741, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18742, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18743, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18729, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18741, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18740, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18740, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18742, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18740, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18740, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18743, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18740, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18740, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18745, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18729, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18747, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18748, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18729, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18746, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18745, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18745, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18747, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18745, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18745, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18748, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18745, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18745, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18750, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18729, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18751, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18752, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18753, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18729, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18751, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18750, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18750, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18752, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18750, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18750, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18753, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18750, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18750, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18755, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18729, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18756, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18757, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18758, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18729, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18756, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18757, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18758, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18755, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18755, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18760, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18715, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18761, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18763, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18765, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18715, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18761, 0),
        BrowseName=QualifiedName("ConfiguredDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18760, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18762, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18760, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18762, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18761, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18761, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18763, 0),
        BrowseName=QualifiedName("ConfiguredDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18760, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18764, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18760, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18764, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18763, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18763, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18765, 0),
        BrowseName=QualifiedName("OperationalDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18760, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18766, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18760, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18766, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18765, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18765, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18767, 0),
        BrowseName=QualifiedName("OperationalDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18760, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18768, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18760, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18768, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18767, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18767, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SupportedTransportProfiles"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17479, 0),
        BrowseName=QualifiedName("SupportedTransportProfiles", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14416, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("PublishSubscribe"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14443, 0),
        BrowseName=QualifiedName("PublishSubscribe", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(2253, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14416, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15215, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15440, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15443, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17366, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17369, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17371, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17405, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17481, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2253, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetSecurityKeys"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15215, 0),
        BrowseName=QualifiedName("GetSecurityKeys", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15216, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15215, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='StartingTokenId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RequestedKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15216, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15215, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15215, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FirstTokenId',
                DataType=NumericNodeId(288, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='Keys',
                DataType=NumericNodeId(15, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='TimeToNextKey',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='KeyLifetime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15217, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15215, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15215, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("GetSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15440, 0),
        BrowseName=QualifiedName("GetSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15441, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15442, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15441, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15440, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15440, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15442, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15440, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15440, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SecurityGroups"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15443, 0),
        BrowseName=QualifiedName("SecurityGroups", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15452, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15444, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15447, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15452, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15444, 0),
        BrowseName=QualifiedName("AddSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15445, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15446, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='KeyLifetime',
                DataType=NumericNodeId(290, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityPolicyUri',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxFutureKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='MaxPastKeyCount',
                DataType=NumericNodeId(7, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15445, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15444, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15445, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15445, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15444, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupId',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15446, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15444, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15444, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveSecurityGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15447, 0),
        BrowseName=QualifiedName("RemoveSecurityGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15448, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='SecurityGroupNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15448, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15447, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15447, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddConnection"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17366, 0),
        BrowseName=QualifiedName("AddConnection", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17367, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17368, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Configuration',
                DataType=NumericNodeId(15617, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17367, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17366, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConnectionId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17368, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17366, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17366, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveConnection"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17369, 0),
        BrowseName=QualifiedName("RemoveConnection", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17370, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConnectionId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17370, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17369, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17369, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("PublishedDataSets"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17371, 0),
        BrowseName=QualifiedName("PublishedDataSets", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14477, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17405, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17406, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17406, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17405, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17405, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Diagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17409, 0),
        BrowseName=QualifiedName("Diagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19732, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17410, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17411, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17416, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17421, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17422, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17457, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19732, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17409, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17410, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17409, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17410, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17411, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17409, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17412, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17413, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17414, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17411, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17412, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17411, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17412, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17411, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17413, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17411, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17413, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17413, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17411, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17414, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17411, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17414, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17411, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17416, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17409, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17417, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17418, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17419, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17416, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17417, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17416, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17418, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17416, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17418, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17419, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17416, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17416, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17421, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17409, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17422, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17409, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17423, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17409, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17424, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17431, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17436, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17441, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17446, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17451, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17424, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17423, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17425, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17426, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17425, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17424, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17424, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17426, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17424, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17424, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17429, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17424, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17424, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17431, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17423, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17431, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17432, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17431, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17433, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17431, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17434, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17431, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17431, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17432, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17431, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17431, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17433, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17431, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17431, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17434, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17431, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17431, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17436, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17423, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17437, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17438, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17439, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17437, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17436, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17437, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17437, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17436, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17438, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17436, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17438, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17438, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17436, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17439, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17436, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17439, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17439, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17436, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17441, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17423, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17442, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17443, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17444, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17442, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17441, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17441, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17443, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17441, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17441, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17444, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17441, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17441, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17446, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17423, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17447, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17448, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17449, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17447, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17446, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17446, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17448, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17446, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17446, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17449, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17446, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17449, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17449, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17446, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17451, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17423, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17452, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17453, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17454, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17423, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17452, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17451, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17451, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17453, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17451, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17451, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17454, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17451, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17451, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17457, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17409, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17458, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17460, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17464, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17409, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17458, 0),
        BrowseName=QualifiedName("ConfiguredDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17457, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17459, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17457, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17459, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17458, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17458, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17460, 0),
        BrowseName=QualifiedName("ConfiguredDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17457, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17461, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17457, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17461, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17460, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17460, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17462, 0),
        BrowseName=QualifiedName("OperationalDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17457, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17463, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17457, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17463, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17462, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17462, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17464, 0),
        BrowseName=QualifiedName("OperationalDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17457, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17466, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17457, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17466, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17464, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17464, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SupportedTransportProfiles"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17481, 0),
        BrowseName=QualifiedName("SupportedTransportProfiles", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14443, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14443, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasPubSubConnection"),
        InverseName=LocalizedText("PubSubConnectionOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14476, 0),
        BrowseName=QualifiedName("HasPubSubConnection", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PublishedDataSetType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14509, 0),
        BrowseName=QualifiedName("PublishedDataSetType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(14936, 0),
        SourceNodeId=NumericNodeId(14509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15222, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14519, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15229, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16759, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15481, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<DataSetWriterName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15222, 0),
        BrowseName=QualifiedName("<DataSetWriterName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14509, 0),
        ReferenceTypeId=NumericNodeId(14936, 0),
        TypeDefinition=NumericNodeId(15298, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16720, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16721, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17482, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15223, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(14936, 0),
        SourceNodeId=NumericNodeId(15222, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16720, 0),
        BrowseName=QualifiedName("DataSetWriterId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15222, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15222, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetFieldContentMask"),
        DataType=NumericNodeId(15583, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16721, 0),
        BrowseName=QualifiedName("DataSetFieldContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15222, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15222, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17482, 0),
        BrowseName=QualifiedName("DataSetWriterProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15222, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15222, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15223, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15222, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15224, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15222, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15224, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15223, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15224, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15223, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfigurationVersion"),
        DataType=NumericNodeId(14593, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14519, 0),
        BrowseName=QualifiedName("ConfigurationVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14509, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMetaData"),
        DataType=NumericNodeId(14523, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15229, 0),
        BrowseName=QualifiedName("DataSetMetaData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14509, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetClassId"),
        DataType=ua.NodeId(ua.ObjectIds.Guid),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16759, 0),
        BrowseName=QualifiedName("DataSetClassId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14509, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("ExtensionFields"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15481, 0),
        BrowseName=QualifiedName("ExtensionFields", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14509, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15489, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15482, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15485, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15489, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddExtensionField"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15482, 0),
        BrowseName=QualifiedName("AddExtensionField", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15481, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15483, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15484, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15481, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FieldName',
                DataType=NumericNodeId(20, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldValue',
                DataType=NumericNodeId(24, 0),
                ValueRank=-2,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15483, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15482, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15482, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FieldId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15484, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15482, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15482, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveExtensionField"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15485, 0),
        BrowseName=QualifiedName("RemoveExtensionField", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15481, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15481, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FieldId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15486, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15485, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15485, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ExtensionFieldsType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15489, 0),
        BrowseName=QualifiedName("ExtensionFieldsType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15490, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15491, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15494, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("<ExtensionFieldName>"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15490, 0),
        BrowseName=QualifiedName("<ExtensionFieldName>", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15489, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15489, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddExtensionField"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15491, 0),
        BrowseName=QualifiedName("AddExtensionField", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15489, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15492, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15493, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15489, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FieldName',
                DataType=NumericNodeId(20, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldValue',
                DataType=NumericNodeId(24, 0),
                ValueRank=-2,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15492, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15491, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15491, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FieldId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15493, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15491, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15491, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveExtensionField"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15494, 0),
        BrowseName=QualifiedName("RemoveExtensionField", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15489, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15495, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15489, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='FieldId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15495, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15494, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15494, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("DataSetToWriter"),
        InverseName=LocalizedText("WriterToDataSet"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14936, 0),
        BrowseName=QualifiedName("DataSetToWriter", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(33, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14936, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(33, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PublishedDataItemsType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14534, 0),
        BrowseName=QualifiedName("PublishedDataItemsType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(14509, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14548, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14558, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedData"),
        DataType=NumericNodeId(14273, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14548, 0),
        BrowseName=QualifiedName("PublishedData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14534, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14534, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddVariables"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14555, 0),
        BrowseName=QualifiedName("AddVariables", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14534, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14556, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14557, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14534, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldNameAliases',
                DataType=NumericNodeId(12, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='PromotedFields',
                DataType=NumericNodeId(1, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='VariablesToAdd',
                DataType=NumericNodeId(14273, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14556, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14555, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14555, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='NewConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='AddResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14557, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14555, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14555, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveVariables"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14558, 0),
        BrowseName=QualifiedName("RemoveVariables", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14534, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14559, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14560, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14534, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='VariablesToRemove',
                DataType=NumericNodeId(7, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14559, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14558, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14558, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='NewConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RemoveResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14560, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14558, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14558, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PublishedEventsType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14572, 0),
        BrowseName=QualifiedName("PublishedEventsType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(14509, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14586, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14587, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14588, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15052, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventNotifier"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14586, 0),
        BrowseName=QualifiedName("EventNotifier", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14572, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14572, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SelectedFields"),
        DataType=NumericNodeId(601, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14587, 0),
        BrowseName=QualifiedName("SelectedFields", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14572, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14572, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Filter"),
        DataType=NumericNodeId(586, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14588, 0),
        BrowseName=QualifiedName("Filter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14572, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14572, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("ModifyFieldSelection"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15052, 0),
        BrowseName=QualifiedName("ModifyFieldSelection", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14572, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15053, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15517, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14572, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldNameAliases',
                DataType=NumericNodeId(12, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='PromotedFields',
                DataType=NumericNodeId(1, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='SelectedFields',
                DataType=NumericNodeId(601, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15053, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15052, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='NewConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15517, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15052, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15052, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataSetFolderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14477, 0),
        BrowseName=QualifiedName("DataSetFolderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14487, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14493, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14496, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16960, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14499, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16994, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16997, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<DataSetFolderName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14478, 0),
        BrowseName=QualifiedName("<DataSetFolderName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(14477, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14479, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14482, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16881, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14485, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16884, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16923, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(14478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedDataItems"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14479, 0),
        BrowseName=QualifiedName("AddPublishedDataItems", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14478, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14481, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldNameAliases',
                DataType=NumericNodeId(12, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='FieldFlags',
                DataType=NumericNodeId(15904, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='VariablesToAdd',
                DataType=NumericNodeId(14273, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14480, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14479, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14479, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='AddResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14481, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14479, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14479, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedEvents"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14482, 0),
        BrowseName=QualifiedName("AddPublishedEvents", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14478, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14483, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14484, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='EventNotifier',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldNameAliases',
                DataType=NumericNodeId(12, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='FieldFlags',
                DataType=NumericNodeId(15904, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='SelectedFields',
                DataType=NumericNodeId(601, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='Filter',
                DataType=NumericNodeId(586, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14483, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14482, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14482, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14484, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14482, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14482, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedDataItemsTemplate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16842, 0),
        BrowseName=QualifiedName("AddPublishedDataItemsTemplate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14478, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16843, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='DataSetMetaData',
                DataType=NumericNodeId(14523, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='VariablesToAdd',
                DataType=NumericNodeId(14273, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16843, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16842, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16842, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='AddResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16853, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16842, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16842, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedEventsTemplate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16881, 0),
        BrowseName=QualifiedName("AddPublishedEventsTemplate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14478, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16882, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16883, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='DataSetMetaData',
                DataType=NumericNodeId(14523, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='EventNotifier',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SelectedFields',
                DataType=NumericNodeId(601, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='Filter',
                DataType=NumericNodeId(586, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16882, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16881, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16883, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16881, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16881, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemovePublishedDataSet"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14485, 0),
        BrowseName=QualifiedName("RemovePublishedDataSet", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14478, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14486, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14485, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14485, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddDataSetFolder"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16884, 0),
        BrowseName=QualifiedName("AddDataSetFolder", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14478, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16922, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16894, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16884, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16884, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetFolderNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16922, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16884, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16884, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveDataSetFolder"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16923, 0),
        BrowseName=QualifiedName("RemoveDataSetFolder", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14478, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16923, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16924, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16923, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16923, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14478, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetFolderNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16924, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16923, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16924, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16924, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16924, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16923, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<PublishedDataSetName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14487, 0),
        BrowseName=QualifiedName("<PublishedDataSetName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14509, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14489, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15221, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14509, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfigurationVersion"),
        DataType=NumericNodeId(14593, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14489, 0),
        BrowseName=QualifiedName("ConfigurationVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14487, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14487, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMetaData"),
        DataType=NumericNodeId(14523, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15221, 0),
        BrowseName=QualifiedName("DataSetMetaData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14487, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14487, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedDataItems"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14493, 0),
        BrowseName=QualifiedName("AddPublishedDataItems", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14494, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14495, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldNameAliases',
                DataType=NumericNodeId(12, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='FieldFlags',
                DataType=NumericNodeId(15904, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='VariablesToAdd',
                DataType=NumericNodeId(14273, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14494, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14493, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14493, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='AddResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14495, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14493, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14493, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedEvents"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14496, 0),
        BrowseName=QualifiedName("AddPublishedEvents", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14497, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14498, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='EventNotifier',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='FieldNameAliases',
                DataType=NumericNodeId(12, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='FieldFlags',
                DataType=NumericNodeId(15904, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='SelectedFields',
                DataType=NumericNodeId(601, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='Filter',
                DataType=NumericNodeId(586, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14497, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14496, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14496, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14498, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14496, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14498, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14498, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14498, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14496, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedDataItemsTemplate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16935, 0),
        BrowseName=QualifiedName("AddPublishedDataItemsTemplate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16958, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16959, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='DataSetMetaData',
                DataType=NumericNodeId(14523, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='VariablesToAdd',
                DataType=NumericNodeId(14273, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16958, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16935, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16935, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='AddResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16959, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16935, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16935, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddPublishedEventsTemplate"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16960, 0),
        BrowseName=QualifiedName("AddPublishedEventsTemplate", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16961, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16971, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='DataSetMetaData',
                DataType=NumericNodeId(14523, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='EventNotifier',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='SelectedFields',
                DataType=NumericNodeId(601, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ua.Argument(
                Name='Filter',
                DataType=NumericNodeId(586, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16961, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16960, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16960, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16971, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16960, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16971, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16971, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16971, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16960, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemovePublishedDataSet"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14499, 0),
        BrowseName=QualifiedName("RemovePublishedDataSet", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14500, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14500, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14499, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14499, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddDataSetFolder"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16994, 0),
        BrowseName=QualifiedName("AddDataSetFolder", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16994, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16995, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16994, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16996, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16994, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16994, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Name',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16995, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16994, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16994, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetFolderNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16996, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16994, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16994, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveDataSetFolder"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16997, 0),
        BrowseName=QualifiedName("RemoveDataSetFolder", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14477, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(16997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17007, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(16997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14477, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetFolderNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17007, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(16997, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16997, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubConnectionType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14209, 0),
        BrowseName=QualifiedName("PubSubConnectionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14595, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17306, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17485, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14221, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17203, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(18804, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(18805, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17325, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14600, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17427, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14225, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublisherId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14595, 0),
        BrowseName=QualifiedName("PublisherId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TransportProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17306, 0),
        BrowseName=QualifiedName("TransportProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(16309, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Selections"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17710, 0),
        BrowseName=QualifiedName("Selections", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConnectionProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17485, 0),
        BrowseName=QualifiedName("ConnectionProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Address"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14221, 0),
        BrowseName=QualifiedName("Address", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(21145, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17202, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21145, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14221, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkInterface"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17202, 0),
        BrowseName=QualifiedName("NetworkInterface", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14221, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(16309, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17576, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14221, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Selections"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17576, 0),
        BrowseName=QualifiedName("Selections", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17202, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17202, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TransportSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17203, 0),
        BrowseName=QualifiedName("TransportSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17721, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17721, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<WriterGroupName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17310, 0),
        BrowseName=QualifiedName("<WriterGroupName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(18804, 0),
        TypeDefinition=NumericNodeId(17725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17204, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17314, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17214, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17318, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17319, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17321, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17322, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17558, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(18804, 0),
        SourceNodeId=NumericNodeId(17310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17311, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNetworkMessageSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17204, 0),
        BrowseName=QualifiedName("MaxNetworkMessageSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("GroupProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17486, 0),
        BrowseName=QualifiedName("GroupProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17314, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17315, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17315, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17314, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17314, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17214, 0),
        BrowseName=QualifiedName("WriterGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17318, 0),
        BrowseName=QualifiedName("PublishingInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeepAliveTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17319, 0),
        BrowseName=QualifiedName("KeepAliveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Priority"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17321, 0),
        BrowseName=QualifiedName("Priority", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIds"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17322, 0),
        BrowseName=QualifiedName("LocaleIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HeaderLayoutUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17558, 0),
        BrowseName=QualifiedName("HeaderLayoutUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17310, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17310, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<ReaderGroupName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17325, 0),
        BrowseName=QualifiedName("<ReaderGroupName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(18805, 0),
        TypeDefinition=NumericNodeId(17999, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17326, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17302, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17487, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17329, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17999, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(18805, 0),
        SourceNodeId=NumericNodeId(17325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17326, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17325, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17325, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNetworkMessageSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17302, 0),
        BrowseName=QualifiedName("MaxNetworkMessageSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17325, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17302, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17302, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17302, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17325, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("GroupProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17487, 0),
        BrowseName=QualifiedName("GroupProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17325, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17325, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17329, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17325, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17330, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17325, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17330, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17329, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17329, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14600, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14601, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14601, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14600, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14601, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14601, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14601, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14600, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Diagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19241, 0),
        BrowseName=QualifiedName("Diagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19786, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19242, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19243, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19248, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19253, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19254, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19255, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19286, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19786, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19242, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19241, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19242, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19243, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19241, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19244, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19245, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19246, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19243, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19244, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19243, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19243, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19245, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19243, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19245, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19243, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19246, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19243, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19243, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19248, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19241, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19249, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19250, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19251, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19248, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19249, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19248, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19248, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19250, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19248, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19248, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19251, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19248, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19248, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19253, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19241, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19253, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19254, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19241, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19255, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19241, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19256, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19261, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19266, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19271, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19276, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19281, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19256, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19255, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19257, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19258, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19259, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19255, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19257, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19256, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19258, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19256, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19259, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19256, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19261, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19255, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19262, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19263, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19264, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19255, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19262, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19261, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19261, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19263, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19261, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19261, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19264, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19261, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19261, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19266, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19255, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19267, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19268, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19269, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19255, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19267, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19266, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19266, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19268, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19266, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19266, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19269, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19266, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19266, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19271, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19255, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19272, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19274, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19255, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19272, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19271, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19271, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19273, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19271, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19271, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19274, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19271, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19271, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19276, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19255, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19277, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19278, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19279, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19255, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19277, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19276, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19276, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19278, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19276, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19276, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19279, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19276, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19276, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19281, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19255, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19282, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19283, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19284, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19255, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19282, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19281, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19281, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19283, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19281, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19281, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19284, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19281, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19281, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19286, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19241, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19287, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19241, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResolvedAddress"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19287, 0),
        BrowseName=QualifiedName("ResolvedAddress", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19286, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19288, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19286, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19288, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19287, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19287, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddWriterGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17427, 0),
        BrowseName=QualifiedName("AddWriterGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17428, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Configuration',
                DataType=NumericNodeId(15480, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17428, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17427, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17427, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='GroupId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17456, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17427, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17427, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddReaderGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17465, 0),
        BrowseName=QualifiedName("AddReaderGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17507, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Configuration',
                DataType=NumericNodeId(15520, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17507, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17465, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17465, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='GroupId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17508, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17465, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17465, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveGroup"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14225, 0),
        BrowseName=QualifiedName("RemoveGroup", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14209, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14226, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14225, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14209, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='GroupId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14226, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14225, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14225, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ConnectionTransportType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17721, 0),
        BrowseName=QualifiedName("ConnectionTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubGroupType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14232, 0),
        BrowseName=QualifiedName("PubSubGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15926, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15927, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15928, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17724, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17488, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15265, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15926, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15926, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15926, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15926, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15927, 0),
        BrowseName=QualifiedName("SecurityGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityKeyServices"),
        DataType=NumericNodeId(312, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15928, 0),
        BrowseName=QualifiedName("SecurityKeyServices", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15928, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15928, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15928, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MaxNetworkMessageSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17724, 0),
        BrowseName=QualifiedName("MaxNetworkMessageSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("GroupProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17488, 0),
        BrowseName=QualifiedName("GroupProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15265, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15266, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15266, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15265, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15265, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("WriterGroupType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17725, 0),
        BrowseName=QualifiedName("WriterGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17736, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17738, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17739, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17740, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17559, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17741, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17742, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(15296, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17743, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17969, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17992, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17736, 0),
        BrowseName=QualifiedName("WriterGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17737, 0),
        BrowseName=QualifiedName("PublishingInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeepAliveTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17738, 0),
        BrowseName=QualifiedName("KeepAliveTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Priority"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17739, 0),
        BrowseName=QualifiedName("Priority", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LocaleIds"),
        DataType=NumericNodeId(295, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17740, 0),
        BrowseName=QualifiedName("LocaleIds", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HeaderLayoutUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17559, 0),
        BrowseName=QualifiedName("HeaderLayoutUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TransportSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17741, 0),
        BrowseName=QualifiedName("TransportSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17997, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17997, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MessageSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17742, 0),
        BrowseName=QualifiedName("MessageSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(17998, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17998, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<DataSetWriterName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17743, 0),
        BrowseName=QualifiedName("<DataSetWriterName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(15296, 0),
        TypeDefinition=NumericNodeId(15298, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17744, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17745, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17490, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17749, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(15296, 0),
        SourceNodeId=NumericNodeId(17743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17744, 0),
        BrowseName=QualifiedName("DataSetWriterId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17743, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17743, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetFieldContentMask"),
        DataType=NumericNodeId(15583, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17745, 0),
        BrowseName=QualifiedName("DataSetFieldContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17743, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17743, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17490, 0),
        BrowseName=QualifiedName("DataSetWriterProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17743, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17743, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17749, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17743, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17750, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17743, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17750, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17749, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17749, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Diagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17812, 0),
        BrowseName=QualifiedName("Diagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19834, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17813, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17814, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17819, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17825, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17858, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19834, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17812, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17813, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17812, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17813, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17814, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17812, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17815, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17816, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17817, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17814, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17815, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17814, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17814, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17816, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17814, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17814, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17817, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17814, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17814, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17819, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17812, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17820, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17822, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17820, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17819, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17819, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17821, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17819, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17819, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17822, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17819, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17819, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17824, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17812, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17825, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17812, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17826, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17812, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17832, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17837, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17847, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17859, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17874, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17900, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17827, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17828, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17828, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17827, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17827, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17829, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17827, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17827, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17830, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17827, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17827, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17832, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17834, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17833, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17832, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17832, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17834, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17832, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17832, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17835, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17832, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17832, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17837, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17838, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17840, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17838, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17837, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17837, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17839, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17837, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17837, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17840, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17837, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17837, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17842, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17843, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17844, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17843, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17842, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17842, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17844, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17842, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17842, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17845, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17842, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17842, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17847, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17850, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17848, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17847, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17847, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17849, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17847, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17847, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17850, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17847, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17847, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17853, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17854, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17856, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17854, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17853, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17853, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17855, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17853, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17853, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17856, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17853, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17853, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SentNetworkMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17859, 0),
        BrowseName=QualifiedName("SentNetworkMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17864, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17871, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17872, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17864, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17859, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17859, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17871, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17859, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17859, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17872, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17859, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17859, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FailedTransmissions"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17874, 0),
        BrowseName=QualifiedName("FailedTransmissions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17878, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17885, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17892, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17878, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17874, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17874, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17885, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17874, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17874, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17892, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17874, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17874, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EncryptionErrors"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17900, 0),
        BrowseName=QualifiedName("EncryptionErrors", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17826, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17901, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17902, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17903, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17826, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17901, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17900, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17900, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17902, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17900, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17900, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17903, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17900, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17900, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17858, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17812, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17913, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17927, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17812, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17913, 0),
        BrowseName=QualifiedName("ConfiguredDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17858, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17920, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17858, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17920, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17913, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17920, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17920, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17920, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17913, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17927, 0),
        BrowseName=QualifiedName("OperationalDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17858, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17934, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17927, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17858, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17934, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17927, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17927, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddDataSetWriter"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17969, 0),
        BrowseName=QualifiedName("AddDataSetWriter", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17969, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17976, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17969, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17987, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17969, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17969, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Configuration',
                DataType=NumericNodeId(15597, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17976, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17969, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17976, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17976, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17976, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17969, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetWriterNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17987, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17969, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17987, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17987, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17987, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17969, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveDataSetWriter"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17992, 0),
        BrowseName=QualifiedName("RemoveDataSetWriter", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17725, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17993, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetWriterNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17993, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17992, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17992, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasDataSetWriter"),
        InverseName=LocalizedText("IsWriterInGroup"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15296, 0),
        BrowseName=QualifiedName("HasDataSetWriter", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasWriterGroup"),
        InverseName=LocalizedText("IsWriterGroupOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18804, 0),
        BrowseName=QualifiedName("HasWriterGroup", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("WriterGroupTransportType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17997, 0),
        BrowseName=QualifiedName("WriterGroupTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17997, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("WriterGroupMessageType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17998, 0),
        BrowseName=QualifiedName("WriterGroupMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ReaderGroupType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17999, 0),
        BrowseName=QualifiedName("ReaderGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(14232, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(15297, 0),
        SourceNodeId=NumericNodeId(17999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21081, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21082, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21085, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14232, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<DataSetReaderName>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18076, 0),
        BrowseName=QualifiedName("<DataSetReaderName>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17999, 0),
        ReferenceTypeId=NumericNodeId(15297, 0),
        TypeDefinition=NumericNodeId(15306, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18077, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18078, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18081, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18082, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17560, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17562, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17492, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18088, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21006, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(15297, 0),
        SourceNodeId=NumericNodeId(18076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublisherId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18077, 0),
        BrowseName=QualifiedName("PublisherId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18078, 0),
        BrowseName=QualifiedName("WriterGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18079, 0),
        BrowseName=QualifiedName("DataSetWriterId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMetaData"),
        DataType=NumericNodeId(14523, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18080, 0),
        BrowseName=QualifiedName("DataSetMetaData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetFieldContentMask"),
        DataType=NumericNodeId(15583, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18081, 0),
        BrowseName=QualifiedName("DataSetFieldContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MessageReceiveTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18082, 0),
        BrowseName=QualifiedName("MessageReceiveTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(18082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeyFrameCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17560, 0),
        BrowseName=QualifiedName("KeyFrameCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HeaderLayoutUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17562, 0),
        BrowseName=QualifiedName("HeaderLayoutUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17492, 0),
        BrowseName=QualifiedName("DataSetReaderProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18088, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18089, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18089, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(18088, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(18089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18088, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SubscribedDataSet"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21006, 0),
        BrowseName=QualifiedName("SubscribedDataSet", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18076, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15108, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15108, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21006, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Diagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21015, 0),
        BrowseName=QualifiedName("Diagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17999, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19903, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21016, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21017, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21022, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21028, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21060, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19903, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21016, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21015, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21017, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21015, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21018, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21019, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21018, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21017, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21017, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21019, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21017, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21017, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21020, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21017, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21017, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21022, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21015, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21023, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21024, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21025, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21023, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21022, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21022, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21024, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21022, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21022, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21025, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21022, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21022, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21027, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(21015, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21028, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21015, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21028, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21029, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(21015, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21035, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21040, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21045, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21050, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21061, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21029, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21030, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21031, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21032, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21033, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21031, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21030, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21032, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21030, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21033, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21030, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21030, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21035, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21036, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21037, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21038, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21036, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21035, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21035, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21037, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21035, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21035, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21038, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21035, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21038, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21035, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21040, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21041, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21042, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21043, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21041, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21040, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21040, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21042, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21040, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21040, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21043, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21040, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21043, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21040, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21045, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21046, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21047, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21048, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21046, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21045, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21046, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21045, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21047, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21045, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21045, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21048, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21045, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21048, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21045, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21050, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21051, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21052, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21053, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21051, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21050, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21050, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21052, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21050, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21050, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21053, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21050, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21050, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21055, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21057, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21058, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21056, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21055, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21055, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21057, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21055, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21055, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21058, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21055, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21055, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceivedNetworkMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21061, 0),
        BrowseName=QualifiedName("ReceivedNetworkMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21029, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21062, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21063, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21064, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21029, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21062, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21061, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21061, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21063, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21061, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21061, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21064, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21061, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21061, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21060, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(21015, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21078, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21015, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21076, 0),
        BrowseName=QualifiedName("ConfiguredDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21060, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21077, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21077, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21076, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21076, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21078, 0),
        BrowseName=QualifiedName("OperationalDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21060, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21060, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21079, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21078, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21078, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TransportSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21080, 0),
        BrowseName=QualifiedName("TransportSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17999, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(21090, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21090, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MessageSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21081, 0),
        BrowseName=QualifiedName("MessageSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17999, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(21091, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21091, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddDataSetReader"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21082, 0),
        BrowseName=QualifiedName("AddDataSetReader", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17999, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21084, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='Configuration',
                DataType=NumericNodeId(15623, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21083, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21082, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21082, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetReaderNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21084, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21082, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21082, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveDataSetReader"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21085, 0),
        BrowseName=QualifiedName("RemoveDataSetReader", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(17999, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21086, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17999, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='DataSetReaderNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21086, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21085, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21085, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasDataSetReader"),
        InverseName=LocalizedText("IsReaderInGroup"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15297, 0),
        BrowseName=QualifiedName("HasDataSetReader", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("HasReaderGroup"),
        InverseName=LocalizedText("IsReaderGroupOf"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18805, 0),
        BrowseName=QualifiedName("HasReaderGroup", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(47, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(18805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(47, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ReaderGroupTransportType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21090, 0),
        BrowseName=QualifiedName("ReaderGroupTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("ReaderGroupMessageType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21091, 0),
        BrowseName=QualifiedName("ReaderGroupMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataSetWriterType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15298, 0),
        BrowseName=QualifiedName("DataSetWriterType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17493, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15303, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21095, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15299, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21092, 0),
        BrowseName=QualifiedName("DataSetWriterId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetFieldContentMask"),
        DataType=NumericNodeId(15583, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21093, 0),
        BrowseName=QualifiedName("DataSetFieldContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeyFrameCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21094, 0),
        BrowseName=QualifiedName("KeyFrameCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17493, 0),
        BrowseName=QualifiedName("DataSetWriterProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TransportSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15303, 0),
        BrowseName=QualifiedName("TransportSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15305, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15305, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MessageSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21095, 0),
        BrowseName=QualifiedName("MessageSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(21096, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21096, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15299, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15300, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15300, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15299, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15299, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Diagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19550, 0),
        BrowseName=QualifiedName("Diagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15298, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19968, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19551, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19552, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19557, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19562, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19563, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19595, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19968, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15298, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19551, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19550, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19552, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19550, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19553, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19554, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19553, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19552, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19552, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19554, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19552, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19552, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19555, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19552, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19552, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19557, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19550, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19558, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19559, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19560, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19558, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19557, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19559, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19557, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19560, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19557, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19557, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19562, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19550, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19563, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19550, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19564, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19550, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19565, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19570, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19575, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19585, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19590, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19596, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19565, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19564, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19566, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19568, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19566, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19565, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19566, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19566, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19566, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19565, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19567, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19565, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19565, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19568, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19565, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19565, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19570, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19564, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19571, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19572, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19571, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19570, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19570, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19572, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19570, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19570, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19573, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19570, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19570, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19575, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19564, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19576, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19577, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19578, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19576, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19577, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19577, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19577, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19577, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19578, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19575, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19578, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19578, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19578, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19575, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19580, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19564, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19581, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19583, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19581, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19580, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19580, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19582, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19580, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19580, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19583, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19580, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19580, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19585, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19564, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19586, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19587, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19588, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19586, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19585, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19585, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19587, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19585, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19585, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19588, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19585, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19585, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19590, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19564, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19591, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19592, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19593, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19591, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19590, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19590, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19592, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19590, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19590, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19593, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19590, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19590, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FailedDataSetMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19596, 0),
        BrowseName=QualifiedName("FailedDataSetMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19564, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19597, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19599, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19564, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19597, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19596, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19596, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19598, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19596, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19596, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19599, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19596, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19596, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19595, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19550, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19550, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataSetWriterTransportType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15305, 0),
        BrowseName=QualifiedName("DataSetWriterTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15305, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataSetWriterMessageType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21096, 0),
        BrowseName=QualifiedName("DataSetWriterMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataSetReaderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15306, 0),
        BrowseName=QualifiedName("DataSetReaderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21097, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21099, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17563, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15933, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15934, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17494, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15311, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21103, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15307, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15316, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17386, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublisherId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21097, 0),
        BrowseName=QualifiedName("PublisherId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21098, 0),
        BrowseName=QualifiedName("WriterGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterId"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21099, 0),
        BrowseName=QualifiedName("DataSetWriterId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMetaData"),
        DataType=NumericNodeId(14523, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21100, 0),
        BrowseName=QualifiedName("DataSetMetaData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetFieldContentMask"),
        DataType=NumericNodeId(15583, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21101, 0),
        BrowseName=QualifiedName("DataSetFieldContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MessageReceiveTimeout"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21102, 0),
        BrowseName=QualifiedName("MessageReceiveTimeout", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeyFrameCount"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17563, 0),
        BrowseName=QualifiedName("KeyFrameCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HeaderLayoutUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17564, 0),
        BrowseName=QualifiedName("HeaderLayoutUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityMode"),
        DataType=NumericNodeId(302, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15932, 0),
        BrowseName=QualifiedName("SecurityMode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityGroupId"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15933, 0),
        BrowseName=QualifiedName("SecurityGroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15933, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityKeyServices"),
        DataType=NumericNodeId(312, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15934, 0),
        BrowseName=QualifiedName("SecurityKeyServices", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15934, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderProperties"),
        DataType=NumericNodeId(14533, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17494, 0),
        BrowseName=QualifiedName("DataSetReaderProperties", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TransportSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15311, 0),
        BrowseName=QualifiedName("TransportSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15319, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15319, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("MessageSettings"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21103, 0),
        BrowseName=QualifiedName("MessageSettings", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(21104, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21104, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Status"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15307, 0),
        BrowseName=QualifiedName("Status", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(14643, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15308, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15308, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15307, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15307, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Diagnostics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19609, 0),
        BrowseName=QualifiedName("Diagnostics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(20027, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19610, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19611, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19621, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19654, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19610, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19609, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19611, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19609, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19612, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19613, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19614, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19612, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19611, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19611, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19613, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19611, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19613, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19613, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19613, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19611, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19614, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19611, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19611, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19616, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19609, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19617, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19618, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19619, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19617, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19618, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19619, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19616, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19616, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19621, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19609, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19622, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19609, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19623, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19609, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19624, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19634, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19639, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19644, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19649, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19655, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19624, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19623, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19625, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19626, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19627, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19625, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19626, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19626, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19626, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19626, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19627, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19624, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19624, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19629, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19623, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19630, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19631, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19632, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19630, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19629, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19629, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19631, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19629, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19629, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19632, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19629, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19629, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19634, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19623, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19636, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19637, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19635, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19634, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19634, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19636, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19634, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19634, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19637, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19634, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19634, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19639, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19623, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19640, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19641, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19642, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19640, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19639, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19639, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19641, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19639, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19639, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19642, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19639, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19639, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19644, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19623, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19645, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19646, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19647, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19645, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19646, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19647, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19644, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19644, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19649, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19623, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19651, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19652, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19650, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19649, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19649, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19651, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19649, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19649, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19652, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19649, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19649, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FailedDataSetMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19655, 0),
        BrowseName=QualifiedName("FailedDataSetMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19623, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19656, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19657, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19658, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19623, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19656, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19655, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19655, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19657, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19655, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19655, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19658, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19655, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19655, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19654, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19609, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19609, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("SubscribedDataSet"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15316, 0),
        BrowseName=QualifiedName("SubscribedDataSet", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(15108, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15316, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15108, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15316, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15316, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateTargetVariables"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17386, 0),
        BrowseName=QualifiedName("CreateTargetVariables", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17387, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17388, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='TargetVariablesToAdd',
                DataType=NumericNodeId(14744, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17387, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17386, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17386, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AddResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17388, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17386, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17386, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("CreateDataSetMirror"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17389, 0),
        BrowseName=QualifiedName("CreateDataSetMirror", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15306, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17390, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17391, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15306, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ParentNodeName',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='RolePermissions',
                DataType=NumericNodeId(96, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17390, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17389, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17389, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ParentNodeId',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17391, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(17389, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17389, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataSetReaderTransportType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15319, 0),
        BrowseName=QualifiedName("DataSetReaderTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DataSetReaderMessageType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21104, 0),
        BrowseName=QualifiedName("DataSetReaderMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SubscribedDataSetType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15108, 0),
        BrowseName=QualifiedName("SubscribedDataSetType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("TargetVariablesType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15111, 0),
        BrowseName=QualifiedName("TargetVariablesType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(15108, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15114, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15115, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15118, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15108, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TargetVariables"),
        DataType=NumericNodeId(14744, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15114, 0),
        BrowseName=QualifiedName("TargetVariables", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("AddTargetVariables"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15115, 0),
        BrowseName=QualifiedName("AddTargetVariables", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15111, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15116, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15117, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='TargetVariablesToAdd',
                DataType=NumericNodeId(14744, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15116, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15115, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15115, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AddResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15117, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15115, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15115, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("RemoveTargetVariables"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15118, 0),
        BrowseName=QualifiedName("RemoveTargetVariables", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(15111, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15120, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15118, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='ConfigurationVersion',
                DataType=NumericNodeId(14593, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='TargetsToRemove',
                DataType=NumericNodeId(7, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15119, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15118, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15118, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='RemoveResults',
                DataType=NumericNodeId(19, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15120, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15118, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15118, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("SubscribedDataSetMirrorType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15127, 0),
        BrowseName=QualifiedName("SubscribedDataSetMirrorType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(15108, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15108, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubStatusType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14643, 0),
        BrowseName=QualifiedName("PubSubStatusType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14644, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14645, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14646, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(14643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14644, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(14643, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Enable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14645, 0),
        BrowseName=QualifiedName("Enable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14643, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Disable"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14646, 0),
        BrowseName=QualifiedName("Disable", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(14643, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14643, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19677, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19678, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19679, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19684, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19689, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19690, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19691, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19722, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19678, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalInformation"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19679, 0),
        BrowseName=QualifiedName("TotalInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19680, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19681, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19682, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19680, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19679, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19679, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19681, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19679, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19679, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19682, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19679, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19679, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TotalError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19684, 0),
        BrowseName=QualifiedName("TotalError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19685, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19687, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19685, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19684, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19685, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19685, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19685, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19684, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19686, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19684, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19684, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19687, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19684, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19684, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("Reset"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19689, 0),
        BrowseName=QualifiedName("Reset", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19689, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19689, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubError"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19690, 0),
        BrowseName=QualifiedName("SubError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19691, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19692, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19697, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19702, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19707, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19712, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19717, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19692, 0),
        BrowseName=QualifiedName("StateError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19691, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19693, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19694, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19695, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19691, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19693, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19694, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19695, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19692, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19695, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19695, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19695, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19692, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19697, 0),
        BrowseName=QualifiedName("StateOperationalByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19691, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19699, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19700, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19691, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19698, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19697, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19697, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19699, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19697, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19697, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19700, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19697, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19697, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19702, 0),
        BrowseName=QualifiedName("StateOperationalByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19691, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19703, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19704, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19705, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19691, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19703, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19702, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19702, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19704, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19702, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19702, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19705, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19702, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19702, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateOperationalFromError"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19707, 0),
        BrowseName=QualifiedName("StateOperationalFromError", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19691, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19708, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19709, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19691, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19708, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19707, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19707, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19709, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19707, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19707, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19710, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19707, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19707, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatePausedByParent"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19712, 0),
        BrowseName=QualifiedName("StatePausedByParent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19691, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19713, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19714, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19715, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19691, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19713, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19712, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19712, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19714, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19712, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19712, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19715, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19712, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19712, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StateDisabledByMethod"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19717, 0),
        BrowseName=QualifiedName("StateDisabledByMethod", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19691, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19718, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19719, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19720, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19691, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19718, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19717, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19717, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19719, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19717, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19717, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19720, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19717, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19717, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19722, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19723, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19724, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Basic'), LocalizedText('Advanced'), LocalizedText('Info'), LocalizedText('Log'), LocalizedText('Debug')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19724, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19723, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19723, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsCounterType"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19725, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsCounterType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(63, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19726, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19727, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19729, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19726, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19727, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19728, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TimeFirstChange"),
        DataType=ua.NodeId(ua.ObjectIds.DateTime),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19729, 0),
        BrowseName=QualifiedName("TimeFirstChange", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19725, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsCounterClassification"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19730, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsCounterClassification", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Information'), LocalizedText('Error')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19731, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19730, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19730, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsRootType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19732, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsRootType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19777, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19777, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19732, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19778, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19780, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19782, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19784, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19732, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19778, 0),
        BrowseName=QualifiedName("ConfiguredDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19777, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19779, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19777, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19779, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19778, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19778, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19780, 0),
        BrowseName=QualifiedName("ConfiguredDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19777, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19781, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19777, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19781, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19780, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19780, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19782, 0),
        BrowseName=QualifiedName("OperationalDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19777, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19783, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19777, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19783, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19782, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19782, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19784, 0),
        BrowseName=QualifiedName("OperationalDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19777, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19785, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19777, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19785, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19784, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19784, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsConnectionType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19786, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsConnectionType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19831, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19831, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19786, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19832, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19786, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResolvedAddress"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19832, 0),
        BrowseName=QualifiedName("ResolvedAddress", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19831, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19831, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19833, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19832, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19832, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsWriterGroupType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19834, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsWriterGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19879, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19848, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19834, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19880, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19885, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19890, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19834, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SentNetworkMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19880, 0),
        BrowseName=QualifiedName("SentNetworkMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19848, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19881, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19882, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19883, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19848, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19881, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19880, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19880, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19882, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19880, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19880, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19883, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19880, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19880, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FailedTransmissions"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19885, 0),
        BrowseName=QualifiedName("FailedTransmissions", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19848, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19886, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19887, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19888, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19848, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19886, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19885, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19885, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19887, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19885, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19885, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19888, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19885, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19885, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EncryptionErrors"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19890, 0),
        BrowseName=QualifiedName("EncryptionErrors", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19848, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19892, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19893, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19848, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19891, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19890, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19890, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19892, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19890, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19890, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19893, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19890, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19890, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19879, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19834, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19895, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19899, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19901, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19834, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19895, 0),
        BrowseName=QualifiedName("ConfiguredDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19879, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19896, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19896, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19895, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19895, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetWriters"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19897, 0),
        BrowseName=QualifiedName("OperationalDataSetWriters", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19879, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19898, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19898, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19897, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19897, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityTokenID"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19899, 0),
        BrowseName=QualifiedName("SecurityTokenID", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19879, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19900, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19900, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19899, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19899, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TimeToNextTokenID"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19901, 0),
        BrowseName=QualifiedName("TimeToNextTokenID", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19879, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19902, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19902, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19901, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19901, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsReaderGroupType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19903, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsReaderGroupType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19917, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19917, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19903, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19949, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19954, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19959, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19903, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceivedNetworkMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19949, 0),
        BrowseName=QualifiedName("ReceivedNetworkMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19950, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19951, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19952, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19950, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19949, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19949, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19951, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19949, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19949, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19952, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19949, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19949, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceivedInvalidNetworkMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19954, 0),
        BrowseName=QualifiedName("ReceivedInvalidNetworkMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19955, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19956, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19957, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19955, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19954, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19954, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19956, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19954, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19954, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19957, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19954, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19954, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DecryptionErrors"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19959, 0),
        BrowseName=QualifiedName("DecryptionErrors", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19917, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19960, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19961, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19962, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19917, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19960, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19959, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19960, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19959, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19961, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19959, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19961, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19959, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19962, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19959, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19959, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19948, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19903, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19964, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19966, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19903, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19964, 0),
        BrowseName=QualifiedName("ConfiguredDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19948, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19965, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19964, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19948, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19965, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19964, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19965, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19965, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19965, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19964, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OperationalDataSetReaders"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19966, 0),
        BrowseName=QualifiedName("OperationalDataSetReaders", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19948, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19966, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19967, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19966, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19966, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19966, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19948, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19967, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19966, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19967, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19967, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19967, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19966, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsDataSetWriterType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19968, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsDataSetWriterType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19968, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19982, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19968, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20013, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19968, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19982, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19968, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19982, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20014, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19982, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19982, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19982, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19968, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FailedDataSetMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20014, 0),
        BrowseName=QualifiedName("FailedDataSetMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19982, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20015, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20016, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20017, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19982, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20015, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20014, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20014, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20016, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20014, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20014, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20017, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20014, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20014, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20013, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(19968, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20019, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20021, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20023, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20025, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20013, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19968, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MessageSequenceNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20019, 0),
        BrowseName=QualifiedName("MessageSequenceNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20020, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20020, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20019, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20020, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20019, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatusCode"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20021, 0),
        BrowseName=QualifiedName("StatusCode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20022, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20022, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20021, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20021, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MajorVersion"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20023, 0),
        BrowseName=QualifiedName("MajorVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20024, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20024, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20023, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20023, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MinorVersion"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20025, 0),
        BrowseName=QualifiedName("MinorVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20013, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20026, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20025, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20013, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20026, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20025, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20026, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20025, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubDiagnosticsDataSetReaderType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20027, 0),
        BrowseName=QualifiedName("PubSubDiagnosticsDataSetReaderType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(19677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20041, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20072, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(20027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Counters"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20041, 0),
        BrowseName=QualifiedName("Counters", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(20027, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20078, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20027, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FailedDataSetMessages"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20073, 0),
        BrowseName=QualifiedName("FailedDataSetMessages", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20041, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20074, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20075, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20074, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20073, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20073, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20075, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20073, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20073, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(0, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20076, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20073, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20073, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DecryptionErrors"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20078, 0),
        BrowseName=QualifiedName("DecryptionErrors", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20041, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(19725, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20081, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20041, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Active"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20079, 0),
        BrowseName=QualifiedName("Active", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20078, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20078, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Classification"),
        DataType=NumericNodeId(19730, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20080, 0),
        BrowseName=QualifiedName("Classification", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20078, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20078, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(1, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20081, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20078, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20078, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("LiveValues"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20072, 0),
        BrowseName=QualifiedName("LiveValues", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(20027, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(58, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20085, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20087, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20089, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20091, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20093, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20027, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MessageSequenceNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20083, 0),
        BrowseName=QualifiedName("MessageSequenceNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20072, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20084, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20072, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20084, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20083, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20083, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatusCode"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20085, 0),
        BrowseName=QualifiedName("StatusCode", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20072, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20086, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20072, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20086, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20085, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20085, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MajorVersion"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20087, 0),
        BrowseName=QualifiedName("MajorVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20072, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20088, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20072, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20088, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20087, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20087, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MinorVersion"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20089, 0),
        BrowseName=QualifiedName("MinorVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20072, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20090, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20072, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20090, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20089, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20089, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SecurityTokenID"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20091, 0),
        BrowseName=QualifiedName("SecurityTokenID", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20072, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20072, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20092, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20091, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20091, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TimeToNextTokenID"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20093, 0),
        BrowseName=QualifiedName("TimeToNextTokenID", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20072, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(20093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20072, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiagnosticsLevel"),
        DataType=NumericNodeId(19723, 0),
        Value=ua.Variant(2, ua.VariantType.Int32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20094, 0),
        BrowseName=QualifiedName("DiagnosticsLevel", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(20093, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(20094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(20094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(20094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20093, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubStatusEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15535, 0),
        BrowseName=QualifiedName("PubSubStatusEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(2130, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15545, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15546, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15547, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(2130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConnectionId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15545, 0),
        BrowseName=QualifiedName("ConnectionId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15535, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15535, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("GroupId"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15546, 0),
        BrowseName=QualifiedName("GroupId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15535, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15535, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("State"),
        DataType=NumericNodeId(14647, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15547, 0),
        BrowseName=QualifiedName("State", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15535, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15535, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubTransportLimitsExceedEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15548, 0),
        BrowseName=QualifiedName("PubSubTransportLimitsExceedEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(15535, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15561, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15562, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15535, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Actual"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15561, 0),
        BrowseName=QualifiedName("Actual", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15548, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15561, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15561, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15561, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15548, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Maximum"),
        DataType=ua.NodeId(ua.ObjectIds.UInt32),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15562, 0),
        BrowseName=QualifiedName("Maximum", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15548, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15548, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("PubSubCommunicationFailureEventType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15563, 0),
        BrowseName=QualifiedName("PubSubCommunicationFailureEventType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(15535, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15576, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15535, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Error"),
        DataType=ua.NodeId(ua.ObjectIds.StatusCode),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15576, 0),
        BrowseName=QualifiedName("Error", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15563, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15563, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("UadpWriterGroupMessageType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21105, 0),
        BrowseName=QualifiedName("UadpWriterGroupMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17998, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21106, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21108, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21109, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17998, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("GroupVersion"),
        DataType=NumericNodeId(20998, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21106, 0),
        BrowseName=QualifiedName("GroupVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21105, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21105, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetOrdering"),
        DataType=NumericNodeId(20408, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21107, 0),
        BrowseName=QualifiedName("DataSetOrdering", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21105, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21105, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkMessageContentMask"),
        DataType=NumericNodeId(15642, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21108, 0),
        BrowseName=QualifiedName("NetworkMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21105, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21108, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21105, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingOffset"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21109, 0),
        BrowseName=QualifiedName("SamplingOffset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21105, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21109, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21105, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingOffset"),
        DataType=NumericNodeId(290, 0),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21110, 0),
        BrowseName=QualifiedName("PublishingOffset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21105, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21105, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("UadpDataSetWriterMessageType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21111, 0),
        BrowseName=QualifiedName("UadpDataSetWriterMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(21096, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21112, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21114, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21115, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21111, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21096, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMessageContentMask"),
        DataType=NumericNodeId(15646, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21112, 0),
        BrowseName=QualifiedName("DataSetMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21112, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfiguredSize"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21113, 0),
        BrowseName=QualifiedName("ConfiguredSize", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkMessageNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21114, 0),
        BrowseName=QualifiedName("NetworkMessageNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21114, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetOffset"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21115, 0),
        BrowseName=QualifiedName("DataSetOffset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21111, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21115, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21111, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("UadpDataSetReaderMessageType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21116, 0),
        BrowseName=QualifiedName("UadpDataSetReaderMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(21104, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21117, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21120, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21121, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21122, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21123, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21124, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21125, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21104, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("GroupVersion"),
        DataType=NumericNodeId(20998, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21117, 0),
        BrowseName=QualifiedName("GroupVersion", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21117, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkMessageNumber"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21119, 0),
        BrowseName=QualifiedName("NetworkMessageNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetOffset"),
        DataType=ua.NodeId(ua.ObjectIds.UInt16),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17477, 0),
        BrowseName=QualifiedName("DataSetOffset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetClassId"),
        DataType=ua.NodeId(ua.ObjectIds.Guid),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21120, 0),
        BrowseName=QualifiedName("DataSetClassId", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21120, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkMessageContentMask"),
        DataType=NumericNodeId(15642, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21121, 0),
        BrowseName=QualifiedName("NetworkMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMessageContentMask"),
        DataType=NumericNodeId(15646, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21122, 0),
        BrowseName=QualifiedName("DataSetMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishingInterval"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21123, 0),
        BrowseName=QualifiedName("PublishingInterval", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProcessingOffset"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21124, 0),
        BrowseName=QualifiedName("ProcessingOffset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReceiveOffset"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21125, 0),
        BrowseName=QualifiedName("ReceiveOffset", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21116, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21116, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("JsonWriterGroupMessageType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21126, 0),
        BrowseName=QualifiedName("JsonWriterGroupMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17998, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17998, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkMessageContentMask"),
        DataType=NumericNodeId(15654, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21127, 0),
        BrowseName=QualifiedName("NetworkMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21126, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21126, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("JsonDataSetWriterMessageType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21128, 0),
        BrowseName=QualifiedName("JsonDataSetWriterMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(21096, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21129, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21096, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMessageContentMask"),
        DataType=NumericNodeId(15658, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21129, 0),
        BrowseName=QualifiedName("DataSetMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21128, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21129, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21128, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("JsonDataSetReaderMessageType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21130, 0),
        BrowseName=QualifiedName("JsonDataSetReaderMessageType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(21104, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21131, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21132, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21104, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkMessageContentMask"),
        DataType=NumericNodeId(15654, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21131, 0),
        BrowseName=QualifiedName("NetworkMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21130, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMessageContentMask"),
        DataType=NumericNodeId(15658, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21132, 0),
        BrowseName=QualifiedName("DataSetMessageContentMask", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21130, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21130, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DatagramConnectionTransportType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15064, 0),
        BrowseName=QualifiedName("DatagramConnectionTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17721, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15072, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17721, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("DiscoveryAddress"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15072, 0),
        BrowseName=QualifiedName("DiscoveryAddress", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15064, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(21145, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15154, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21145, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15064, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkInterface"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15154, 0),
        BrowseName=QualifiedName("NetworkInterface", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15072, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(16309, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17579, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15072, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Selections"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17579, 0),
        BrowseName=QualifiedName("Selections", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15154, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15154, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("DatagramWriterGroupTransportType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21133, 0),
        BrowseName=QualifiedName("DatagramWriterGroupTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17997, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21134, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21135, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17997, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MessageRepeatCount"),
        DataType=ua.NodeId(ua.ObjectIds.Byte),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21134, 0),
        BrowseName=QualifiedName("MessageRepeatCount", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21133, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21133, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MessageRepeatDelay"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21135, 0),
        BrowseName=QualifiedName("MessageRepeatDelay", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21133, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21133, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BrokerConnectionTransportType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15155, 0),
        BrowseName=QualifiedName("BrokerConnectionTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17721, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15156, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17721, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResourceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15156, 0),
        BrowseName=QualifiedName("ResourceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15155, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15155, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15178, 0),
        BrowseName=QualifiedName("AuthenticationProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15155, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15155, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BrokerWriterGroupTransportType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21136, 0),
        BrowseName=QualifiedName("BrokerWriterGroupTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(17997, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21137, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15246, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15247, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15249, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17997, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueueName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21137, 0),
        BrowseName=QualifiedName("QueueName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21136, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21136, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResourceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15246, 0),
        BrowseName=QualifiedName("ResourceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21136, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15246, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21136, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15247, 0),
        BrowseName=QualifiedName("AuthenticationProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21136, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21136, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RequestedDeliveryGuarantee"),
        DataType=NumericNodeId(15008, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15249, 0),
        BrowseName=QualifiedName("RequestedDeliveryGuarantee", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21136, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15249, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21136, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BrokerDataSetWriterTransportType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21138, 0),
        BrowseName=QualifiedName("BrokerDataSetWriterTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(15305, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21139, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21140, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15250, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15251, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15330, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21141, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15305, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueueName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21139, 0),
        BrowseName=QualifiedName("QueueName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21138, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MetaDataQueueName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21140, 0),
        BrowseName=QualifiedName("MetaDataQueueName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21138, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResourceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15250, 0),
        BrowseName=QualifiedName("ResourceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21138, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15250, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15251, 0),
        BrowseName=QualifiedName("AuthenticationProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21138, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RequestedDeliveryGuarantee"),
        DataType=NumericNodeId(15008, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15330, 0),
        BrowseName=QualifiedName("RequestedDeliveryGuarantee", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21138, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15330, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MetaDataUpdateTime"),
        DataType=NumericNodeId(290, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21141, 0),
        BrowseName=QualifiedName("MetaDataUpdateTime", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21138, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21138, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("BrokerDataSetReaderTransportType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21142, 0),
        BrowseName=QualifiedName("BrokerDataSetReaderTransportType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(15319, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21143, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15334, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15419, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21144, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15319, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("QueueName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21143, 0),
        BrowseName=QualifiedName("QueueName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21142, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ResourceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15334, 0),
        BrowseName=QualifiedName("ResourceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21142, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15334, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AuthenticationProfileUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15419, 0),
        BrowseName=QualifiedName("AuthenticationProfileUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21142, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15419, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RequestedDeliveryGuarantee"),
        DataType=NumericNodeId(15008, 0),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15420, 0),
        BrowseName=QualifiedName("RequestedDeliveryGuarantee", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21142, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MetaDataQueueName"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21144, 0),
        BrowseName=QualifiedName("MetaDataQueueName", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21142, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21142, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NetworkAddressType"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21145, 0),
        BrowseName=QualifiedName("NetworkAddressType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21146, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkInterface"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21146, 0),
        BrowseName=QualifiedName("NetworkInterface", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21145, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(16309, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(21146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21145, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Selections"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17582, 0),
        BrowseName=QualifiedName("Selections", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21146, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(17582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(17582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21146, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("NetworkAddressUrlType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21147, 0),
        BrowseName=QualifiedName("NetworkAddressUrlType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(21145, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21149, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(21147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21145, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Url"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21149, 0),
        BrowseName=QualifiedName("Url", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(21147, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(63, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(63, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(21149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21147, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AliasNameType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23455, 0),
        BrowseName=QualifiedName("AliasNameType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(58, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(58, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectTypeAttributes(
        DisplayName=LocalizedText("AliasNameCategoryType"),
        IsAbstract=False,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23456, 0),
        BrowseName=QualifiedName("AliasNameCategoryType", 0),
        NodeClass_=NodeClass.ObjectType,
        ParentNodeId=NumericNodeId(61, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23457, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23458, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(61, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<Alias>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23457, 0),
        BrowseName=QualifiedName("<Alias>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23456, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(23455, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23455, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23456, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("<SubAliasNameCategories>"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23458, 0),
        BrowseName=QualifiedName("<SubAliasNameCategories>", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23456, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(23456, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23459, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23456, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("FindAlias"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23459, 0),
        BrowseName=QualifiedName("FindAlias", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(23458, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23460, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23461, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23458, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNameSearchPattern',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ReferenceTypeFilter',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23460, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23459, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23459, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNodeList',
                DataType=NumericNodeId(23468, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23461, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23459, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23459, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("FindAlias"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23462, 0),
        BrowseName=QualifiedName("FindAlias", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(23456, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23463, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23464, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23456, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNameSearchPattern',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ReferenceTypeFilter',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23463, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23462, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23462, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNodeList',
                DataType=NumericNodeId(23468, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23464, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23462, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(23464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23462, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AliasNameDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23468, 0),
        BrowseName=QualifiedName("AliasNameDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ReferenceTypeAttributes(
        DisplayName=LocalizedText("AliasFor"),
        InverseName=LocalizedText("HasAlias"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23469, 0),
        BrowseName=QualifiedName("AliasFor", 0),
        NodeClass_=NodeClass.ReferenceType,
        ParentNodeId=NumericNodeId(32, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23469, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(32, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Aliases"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23470, 0),
        BrowseName=QualifiedName("Aliases", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(85, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(23456, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23470, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23476, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23470, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(85, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23470, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23456, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("FindAlias"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23476, 0),
        BrowseName=QualifiedName("FindAlias", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(23470, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23478, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23470, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNameSearchPattern',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ReferenceTypeFilter',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23477, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23476, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23476, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNodeList',
                DataType=NumericNodeId(23468, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23478, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23476, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23476, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("TagVariables"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23479, 0),
        BrowseName=QualifiedName("TagVariables", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23470, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(23456, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23485, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23470, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23456, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("FindAlias"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23485, 0),
        BrowseName=QualifiedName("FindAlias", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(23479, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23487, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23479, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNameSearchPattern',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ReferenceTypeFilter',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23486, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23485, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23485, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNodeList',
                DataType=NumericNodeId(23468, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23487, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23485, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23485, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Topics"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23488, 0),
        BrowseName=QualifiedName("Topics", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23470, 0),
        ReferenceTypeId=NumericNodeId(35, 0),
        TypeDefinition=NumericNodeId(23456, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23494, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(35, 0),
        SourceNodeId=NumericNodeId(23488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23470, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23456, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.MethodAttributes(
        DisplayName=LocalizedText("FindAlias"),
    )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23494, 0),
        BrowseName=QualifiedName("FindAlias", 0),
        NodeClass_=NodeClass.Method,
        ParentNodeId=NumericNodeId(23488, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23495, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23496, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23488, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("InputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNameSearchPattern',
                DataType=NumericNodeId(12, 0),
                ValueRank=-1,
                ),
            ua.Argument(
                Name='ReferenceTypeFilter',
                DataType=NumericNodeId(17, 0),
                ValueRank=-1,
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23495, 0),
        BrowseName=QualifiedName("InputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23494, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23494, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OutputArguments"),
        DataType=NumericNodeId(296, 0),
        Value=ua.Variant([
            ua.Argument(
                Name='AliasNodeList',
                DataType=NumericNodeId(23468, 0),
                ValueRank=1,
                ArrayDimensions=[0],
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23496, 0),
        BrowseName=QualifiedName("OutputArguments", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(23494, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(23496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23494, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("MultiStateDictionaryEntryDiscreteBaseType"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-2,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19077, 0),
        BrowseName=QualifiedName("MultiStateDictionaryEntryDiscreteBaseType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(11238, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19082, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11238, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumDictionaryEntries"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=2,
        ArrayDimensions=[0, 0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19082, 0),
        BrowseName=QualifiedName("EnumDictionaryEntries", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19077, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ValueAsDictionaryEntries"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19083, 0),
        BrowseName=QualifiedName("ValueAsDictionaryEntries", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19077, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(80, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableTypeAttributes(
        DisplayName=LocalizedText("MultiStateDictionaryEntryDiscreteType"),
        DataType=ua.NodeId(ua.ObjectIds.Number),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19084, 0),
        BrowseName=QualifiedName("MultiStateDictionaryEntryDiscreteType", 0),
        NodeClass_=NodeClass.VariableType,
        ParentNodeId=NumericNodeId(19077, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19090, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(19084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ValueAsDictionaryEntries"),
        DataType=ua.NodeId(ua.ObjectIds.NodeId),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19090, 0),
        BrowseName=QualifiedName("ValueAsDictionaryEntries", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(19084, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(19090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(19090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19084, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("IdType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(256, 0),
        BrowseName=QualifiedName("IdType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7591, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Numeric'), LocalizedText('String'), LocalizedText('Guid'), LocalizedText('Opaque')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7591, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(256, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(256, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NodeClass"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(257, 0),
        BrowseName=QualifiedName("NodeClass", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11878, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=0,
                DisplayName=LocalizedText(Text='Unspecified'),
                ),
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='Object'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='Variable'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='Method'),
                ),
            ua.EnumValueType(
                Value=8,
                DisplayName=LocalizedText(Text='ObjectType'),
                ),
            ua.EnumValueType(
                Value=16,
                DisplayName=LocalizedText(Text='VariableType'),
                ),
            ua.EnumValueType(
                Value=32,
                DisplayName=LocalizedText(Text='ReferenceType'),
                ),
            ua.EnumValueType(
                Value=64,
                DisplayName=LocalizedText(Text='DataType'),
                ),
            ua.EnumValueType(
                Value=128,
                DisplayName=LocalizedText(Text='View'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11878, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(257, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(257, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PermissionType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(94, 0),
        BrowseName=QualifiedName("PermissionType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(94, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(94, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Browse'), LocalizedText('ReadRolePermissions'), LocalizedText('WriteAttribute'), LocalizedText('WriteRolePermissions'), LocalizedText('WriteHistorizing'), LocalizedText('Read'), LocalizedText('Write'), LocalizedText('ReadHistory'), LocalizedText('InsertHistory'), LocalizedText('ModifyHistory'), LocalizedText('DeleteHistory'), LocalizedText('ReceiveEvents'), LocalizedText('Call'), LocalizedText('AddReference'), LocalizedText('RemoveReference'), LocalizedText('DeleteNode'), LocalizedText('AddNode')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15030, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(94, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(94, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AccessLevelType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15031, 0),
        BrowseName=QualifiedName("AccessLevelType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(3, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15032, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15031, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('CurrentRead'), LocalizedText('CurrentWrite'), LocalizedText('HistoryRead'), LocalizedText('HistoryWrite'), LocalizedText('SemanticChange'), LocalizedText('StatusWrite'), LocalizedText('TimestampWrite')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15032, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15031, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15032, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15031, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AccessLevelExType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15406, 0),
        BrowseName=QualifiedName("AccessLevelExType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15407, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15406, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('CurrentRead'), LocalizedText('CurrentWrite'), LocalizedText('HistoryRead'), LocalizedText('HistoryWrite'), LocalizedText('SemanticChange'), LocalizedText('StatusWrite'), LocalizedText('TimestampWrite'), LocalizedText('Reserved'), LocalizedText('NonatomicRead'), LocalizedText('NonatomicWrite'), LocalizedText('WriteFullArrayOnly'), LocalizedText('NoSubDataTypes')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15407, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15406, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15407, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15406, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EventNotifierType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15033, 0),
        BrowseName=QualifiedName("EventNotifierType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(3, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15034, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(3, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('SubscribeToEvents'), LocalizedText('Reserved'), LocalizedText('HistoryRead'), LocalizedText('HistoryWrite')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15034, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(15033, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15034, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15033, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AccessRestrictionType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(95, 0),
        BrowseName=QualifiedName("AccessRestrictionType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(5, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(95, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15035, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(95, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(5, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('SigningRequired'), LocalizedText('EncryptionRequired'), LocalizedText('SessionRequired'), LocalizedText('ApplyRestrictionsToBrowse')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15035, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(95, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15035, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(95, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RolePermissionType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(96, 0),
        BrowseName=QualifiedName("RolePermissionType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(96, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataTypeDefinition"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(97, 0),
        BrowseName=QualifiedName("DataTypeDefinition", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(97, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("StructureType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(98, 0),
        BrowseName=QualifiedName("StructureType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(98, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14528, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(98, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Structure'), LocalizedText('StructureWithOptionalFields'), LocalizedText('Union')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14528, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(98, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14528, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(14528, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(14528, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(98, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("StructureField"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(101, 0),
        BrowseName=QualifiedName("StructureField", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("StructureDefinition"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(99, 0),
        BrowseName=QualifiedName("StructureDefinition", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(97, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(99, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(97, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EnumDefinition"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(100, 0),
        BrowseName=QualifiedName("EnumDefinition", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(97, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(97, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Node"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(258, 0),
        BrowseName=QualifiedName("Node", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("InstanceNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11879, 0),
        BrowseName=QualifiedName("InstanceNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(258, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(258, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TypeNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11880, 0),
        BrowseName=QualifiedName("TypeNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(258, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(258, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ObjectNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(261, 0),
        BrowseName=QualifiedName("ObjectNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11879, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ObjectTypeNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(264, 0),
        BrowseName=QualifiedName("ObjectTypeNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11880, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11880, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("VariableNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(267, 0),
        BrowseName=QualifiedName("VariableNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11879, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(267, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("VariableTypeNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(270, 0),
        BrowseName=QualifiedName("VariableTypeNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11880, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11880, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReferenceTypeNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(273, 0),
        BrowseName=QualifiedName("ReferenceTypeNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11880, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11880, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MethodNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(276, 0),
        BrowseName=QualifiedName("MethodNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11879, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ViewNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(279, 0),
        BrowseName=QualifiedName("ViewNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11879, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11879, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataTypeNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(282, 0),
        BrowseName=QualifiedName("DataTypeNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11880, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11880, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReferenceNode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(285, 0),
        BrowseName=QualifiedName("ReferenceNode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Argument"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(296, 0),
        BrowseName=QualifiedName("Argument", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EnumValueType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7594, 0),
        BrowseName=QualifiedName("EnumValueType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(7594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EnumField"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(102, 0),
        BrowseName=QualifiedName("EnumField", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7594, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7594, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("OptionSet"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12755, 0),
        BrowseName=QualifiedName("OptionSet", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Union"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12756, 0),
        BrowseName=QualifiedName("Union", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NormalizedString"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12877, 0),
        BrowseName=QualifiedName("NormalizedString", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12877, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DecimalString"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12878, 0),
        BrowseName=QualifiedName("DecimalString", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DurationString"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12879, 0),
        BrowseName=QualifiedName("DurationString", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TimeString"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12880, 0),
        BrowseName=QualifiedName("TimeString", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DateString"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12881, 0),
        BrowseName=QualifiedName("DateString", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Duration"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(290, 0),
        BrowseName=QualifiedName("Duration", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(11, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(290, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UtcTime"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(294, 0),
        BrowseName=QualifiedName("UtcTime", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(13, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(13, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("LocaleId"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(295, 0),
        BrowseName=QualifiedName("LocaleId", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TimeZoneDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8912, 0),
        BrowseName=QualifiedName("TimeZoneDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(8912, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Index"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17588, 0),
        BrowseName=QualifiedName("Index", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("IntegerId"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(288, 0),
        BrowseName=QualifiedName("IntegerId", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ApplicationType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(307, 0),
        BrowseName=QualifiedName("ApplicationType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7597, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(307, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Server'), LocalizedText('Client'), LocalizedText('ClientAndServer'), LocalizedText('DiscoveryServer')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7597, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(307, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(307, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ApplicationDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(308, 0),
        BrowseName=QualifiedName("ApplicationDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RequestHeader"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(389, 0),
        BrowseName=QualifiedName("RequestHeader", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(389, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ResponseHeader"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(392, 0),
        BrowseName=QualifiedName("ResponseHeader", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("VersionTime"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20998, 0),
        BrowseName=QualifiedName("VersionTime", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(20998, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ServiceFault"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(395, 0),
        BrowseName=QualifiedName("ServiceFault", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(395, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SessionlessInvokeRequestType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15901, 0),
        BrowseName=QualifiedName("SessionlessInvokeRequestType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SessionlessInvokeResponseType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(20999, 0),
        BrowseName=QualifiedName("SessionlessInvokeResponseType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(20999, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FindServersRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(420, 0),
        BrowseName=QualifiedName("FindServersRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(420, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FindServersResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(423, 0),
        BrowseName=QualifiedName("FindServersResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(423, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ServerOnNetwork"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12189, 0),
        BrowseName=QualifiedName("ServerOnNetwork", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FindServersOnNetworkRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12190, 0),
        BrowseName=QualifiedName("FindServersOnNetworkRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FindServersOnNetworkResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12191, 0),
        BrowseName=QualifiedName("FindServersOnNetworkResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ApplicationInstanceCertificate"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(311, 0),
        BrowseName=QualifiedName("ApplicationInstanceCertificate", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MessageSecurityMode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(302, 0),
        BrowseName=QualifiedName("MessageSecurityMode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(302, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7595, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(302, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Invalid'), LocalizedText('None'), LocalizedText('Sign'), LocalizedText('SignAndEncrypt')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7595, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(302, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(302, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UserTokenType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(303, 0),
        BrowseName=QualifiedName("UserTokenType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7596, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Anonymous'), LocalizedText('UserName'), LocalizedText('Certificate'), LocalizedText('IssuedToken')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7596, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(303, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(303, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UserTokenPolicy"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(304, 0),
        BrowseName=QualifiedName("UserTokenPolicy", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(304, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EndpointDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(312, 0),
        BrowseName=QualifiedName("EndpointDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("GetEndpointsRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(426, 0),
        BrowseName=QualifiedName("GetEndpointsRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(426, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("GetEndpointsResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(429, 0),
        BrowseName=QualifiedName("GetEndpointsResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(429, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RegisteredServer"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(432, 0),
        BrowseName=QualifiedName("RegisteredServer", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(432, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RegisterServerRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(435, 0),
        BrowseName=QualifiedName("RegisterServerRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(435, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RegisterServerResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(438, 0),
        BrowseName=QualifiedName("RegisterServerResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(438, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DiscoveryConfiguration"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12890, 0),
        BrowseName=QualifiedName("DiscoveryConfiguration", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MdnsDiscoveryConfiguration"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12891, 0),
        BrowseName=QualifiedName("MdnsDiscoveryConfiguration", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12890, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12890, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RegisterServer2Request"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12193, 0),
        BrowseName=QualifiedName("RegisterServer2Request", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RegisterServer2Response"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12194, 0),
        BrowseName=QualifiedName("RegisterServer2Response", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12194, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SecurityTokenRequestType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(315, 0),
        BrowseName=QualifiedName("SecurityTokenRequestType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Issue'), LocalizedText('Renew')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7598, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(315, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(315, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ChannelSecurityToken"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(441, 0),
        BrowseName=QualifiedName("ChannelSecurityToken", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(441, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("OpenSecureChannelRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(444, 0),
        BrowseName=QualifiedName("OpenSecureChannelRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(444, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("OpenSecureChannelResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(447, 0),
        BrowseName=QualifiedName("OpenSecureChannelResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(447, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CloseSecureChannelRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(450, 0),
        BrowseName=QualifiedName("CloseSecureChannelRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(450, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CloseSecureChannelResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(453, 0),
        BrowseName=QualifiedName("CloseSecureChannelResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(453, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SignedSoftwareCertificate"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(344, 0),
        BrowseName=QualifiedName("SignedSoftwareCertificate", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(344, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SessionAuthenticationToken"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(388, 0),
        BrowseName=QualifiedName("SessionAuthenticationToken", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(17, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(388, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SignatureData"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(456, 0),
        BrowseName=QualifiedName("SignatureData", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(456, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CreateSessionRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(459, 0),
        BrowseName=QualifiedName("CreateSessionRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(459, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CreateSessionResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(462, 0),
        BrowseName=QualifiedName("CreateSessionResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(462, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UserIdentityToken"),
        IsAbstract=True,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(316, 0),
        BrowseName=QualifiedName("UserIdentityToken", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(316, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AnonymousIdentityToken"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(319, 0),
        BrowseName=QualifiedName("AnonymousIdentityToken", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(316, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(316, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UserNameIdentityToken"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(322, 0),
        BrowseName=QualifiedName("UserNameIdentityToken", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(316, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(316, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("X509IdentityToken"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(325, 0),
        BrowseName=QualifiedName("X509IdentityToken", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(316, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(316, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("IssuedIdentityToken"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(938, 0),
        BrowseName=QualifiedName("IssuedIdentityToken", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(316, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(938, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(316, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RsaEncryptedSecret"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17545, 0),
        BrowseName=QualifiedName("RsaEncryptedSecret", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EccEncryptedSecret"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17546, 0),
        BrowseName=QualifiedName("EccEncryptedSecret", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(24, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(24, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ActivateSessionRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(465, 0),
        BrowseName=QualifiedName("ActivateSessionRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(465, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ActivateSessionResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(468, 0),
        BrowseName=QualifiedName("ActivateSessionResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CloseSessionRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(471, 0),
        BrowseName=QualifiedName("CloseSessionRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(471, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CloseSessionResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(474, 0),
        BrowseName=QualifiedName("CloseSessionResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(474, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CancelRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(477, 0),
        BrowseName=QualifiedName("CancelRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(477, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CancelResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(480, 0),
        BrowseName=QualifiedName("CancelResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(480, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NodeAttributesMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(348, 0),
        BrowseName=QualifiedName("NodeAttributesMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11881, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=0,
                DisplayName=LocalizedText(Text='None'),
                ),
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='AccessLevel'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='ArrayDimensions'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='BrowseName'),
                ),
            ua.EnumValueType(
                Value=8,
                DisplayName=LocalizedText(Text='ContainsNoLoops'),
                ),
            ua.EnumValueType(
                Value=16,
                DisplayName=LocalizedText(Text='DataType'),
                ),
            ua.EnumValueType(
                Value=32,
                DisplayName=LocalizedText(Text='Description'),
                ),
            ua.EnumValueType(
                Value=64,
                DisplayName=LocalizedText(Text='DisplayName'),
                ),
            ua.EnumValueType(
                Value=128,
                DisplayName=LocalizedText(Text='EventNotifier'),
                ),
            ua.EnumValueType(
                Value=256,
                DisplayName=LocalizedText(Text='Executable'),
                ),
            ua.EnumValueType(
                Value=512,
                DisplayName=LocalizedText(Text='Historizing'),
                ),
            ua.EnumValueType(
                Value=1024,
                DisplayName=LocalizedText(Text='InverseName'),
                ),
            ua.EnumValueType(
                Value=2048,
                DisplayName=LocalizedText(Text='IsAbstract'),
                ),
            ua.EnumValueType(
                Value=4096,
                DisplayName=LocalizedText(Text='MinimumSamplingInterval'),
                ),
            ua.EnumValueType(
                Value=8192,
                DisplayName=LocalizedText(Text='NodeClass'),
                ),
            ua.EnumValueType(
                Value=16384,
                DisplayName=LocalizedText(Text='NodeId'),
                ),
            ua.EnumValueType(
                Value=32768,
                DisplayName=LocalizedText(Text='Symmetric'),
                ),
            ua.EnumValueType(
                Value=65536,
                DisplayName=LocalizedText(Text='UserAccessLevel'),
                ),
            ua.EnumValueType(
                Value=131072,
                DisplayName=LocalizedText(Text='UserExecutable'),
                ),
            ua.EnumValueType(
                Value=262144,
                DisplayName=LocalizedText(Text='UserWriteMask'),
                ),
            ua.EnumValueType(
                Value=524288,
                DisplayName=LocalizedText(Text='ValueRank'),
                ),
            ua.EnumValueType(
                Value=1048576,
                DisplayName=LocalizedText(Text='WriteMask'),
                ),
            ua.EnumValueType(
                Value=2097152,
                DisplayName=LocalizedText(Text='Value'),
                ),
            ua.EnumValueType(
                Value=4194304,
                DisplayName=LocalizedText(Text='DataTypeDefinition'),
                ),
            ua.EnumValueType(
                Value=8388608,
                DisplayName=LocalizedText(Text='RolePermissions'),
                ),
            ua.EnumValueType(
                Value=16777216,
                DisplayName=LocalizedText(Text='AccessRestrictions'),
                ),
            ua.EnumValueType(
                Value=33554431,
                DisplayName=LocalizedText(Text='All'),
                ),
            ua.EnumValueType(
                Value=26501220,
                DisplayName=LocalizedText(Text='BaseNode'),
                ),
            ua.EnumValueType(
                Value=26501348,
                DisplayName=LocalizedText(Text='Object'),
                ),
            ua.EnumValueType(
                Value=26503268,
                DisplayName=LocalizedText(Text='ObjectType'),
                ),
            ua.EnumValueType(
                Value=26571383,
                DisplayName=LocalizedText(Text='Variable'),
                ),
            ua.EnumValueType(
                Value=28600438,
                DisplayName=LocalizedText(Text='VariableType'),
                ),
            ua.EnumValueType(
                Value=26632548,
                DisplayName=LocalizedText(Text='Method'),
                ),
            ua.EnumValueType(
                Value=26537060,
                DisplayName=LocalizedText(Text='ReferenceType'),
                ),
            ua.EnumValueType(
                Value=26501356,
                DisplayName=LocalizedText(Text='View'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11881, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(348, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11881, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(348, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NodeAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(349, 0),
        BrowseName=QualifiedName("NodeAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(349, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ObjectAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(352, 0),
        BrowseName=QualifiedName("ObjectAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(352, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("VariableAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(355, 0),
        BrowseName=QualifiedName("VariableAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MethodAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(358, 0),
        BrowseName=QualifiedName("MethodAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ObjectTypeAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(361, 0),
        BrowseName=QualifiedName("ObjectTypeAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(361, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("VariableTypeAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(364, 0),
        BrowseName=QualifiedName("VariableTypeAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReferenceTypeAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(367, 0),
        BrowseName=QualifiedName("ReferenceTypeAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataTypeAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(370, 0),
        BrowseName=QualifiedName("DataTypeAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ViewAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(373, 0),
        BrowseName=QualifiedName("ViewAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("GenericAttributeValue"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17606, 0),
        BrowseName=QualifiedName("GenericAttributeValue", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("GenericAttributes"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17607, 0),
        BrowseName=QualifiedName("GenericAttributes", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(17607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AddNodesItem"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(376, 0),
        BrowseName=QualifiedName("AddNodesItem", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AddNodesResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(483, 0),
        BrowseName=QualifiedName("AddNodesResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(483, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AddNodesRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(486, 0),
        BrowseName=QualifiedName("AddNodesRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(486, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AddNodesResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(489, 0),
        BrowseName=QualifiedName("AddNodesResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(489, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AddReferencesItem"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(379, 0),
        BrowseName=QualifiedName("AddReferencesItem", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AddReferencesRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(492, 0),
        BrowseName=QualifiedName("AddReferencesRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(492, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AddReferencesResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(495, 0),
        BrowseName=QualifiedName("AddReferencesResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(495, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteNodesItem"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(382, 0),
        BrowseName=QualifiedName("DeleteNodesItem", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteNodesRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(498, 0),
        BrowseName=QualifiedName("DeleteNodesRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(498, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteNodesResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(501, 0),
        BrowseName=QualifiedName("DeleteNodesResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(501, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteReferencesItem"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(385, 0),
        BrowseName=QualifiedName("DeleteReferencesItem", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(385, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteReferencesRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(504, 0),
        BrowseName=QualifiedName("DeleteReferencesRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(504, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteReferencesResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(507, 0),
        BrowseName=QualifiedName("DeleteReferencesResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AttributeWriteMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(347, 0),
        BrowseName=QualifiedName("AttributeWriteMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(347, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15036, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(347, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSetValues"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('AccessLevel'), LocalizedText('ArrayDimensions'), LocalizedText('BrowseName'), LocalizedText('ContainsNoLoops'), LocalizedText('DataType'), LocalizedText('Description'), LocalizedText('DisplayName'), LocalizedText('EventNotifier'), LocalizedText('Executable'), LocalizedText('Historizing'), LocalizedText('InverseName'), LocalizedText('IsAbstract'), LocalizedText('MinimumSamplingInterval'), LocalizedText('NodeClass'), LocalizedText('NodeId'), LocalizedText('Symmetric'), LocalizedText('UserAccessLevel'), LocalizedText('UserExecutable'), LocalizedText('UserWriteMask'), LocalizedText('ValueRank'), LocalizedText('WriteMask'), LocalizedText('ValueForVariableType'), LocalizedText('DataTypeDefinition'), LocalizedText('RolePermissions'), LocalizedText('AccessRestrictions'), LocalizedText('AccessLevelEx')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15036, 0),
        BrowseName=QualifiedName("OptionSetValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(347, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(15036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15036, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(347, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseDirection"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(510, 0),
        BrowseName=QualifiedName("BrowseDirection", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7603, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(510, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Forward'), LocalizedText('Inverse'), LocalizedText('Both'), LocalizedText('Invalid')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7603, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(510, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(510, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ViewDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(511, 0),
        BrowseName=QualifiedName("ViewDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(514, 0),
        BrowseName=QualifiedName("BrowseDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(514, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseResultMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(517, 0),
        BrowseName=QualifiedName("BrowseResultMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11883, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(517, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=0,
                DisplayName=LocalizedText(Text='None'),
                ),
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='ReferenceTypeId'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='IsForward'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='NodeClass'),
                ),
            ua.EnumValueType(
                Value=8,
                DisplayName=LocalizedText(Text='BrowseName'),
                ),
            ua.EnumValueType(
                Value=16,
                DisplayName=LocalizedText(Text='DisplayName'),
                ),
            ua.EnumValueType(
                Value=32,
                DisplayName=LocalizedText(Text='TypeDefinition'),
                ),
            ua.EnumValueType(
                Value=63,
                DisplayName=LocalizedText(Text='All'),
                ),
            ua.EnumValueType(
                Value=3,
                DisplayName=LocalizedText(Text='ReferenceTypeInfo'),
                ),
            ua.EnumValueType(
                Value=60,
                DisplayName=LocalizedText(Text='TargetInfo'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11883, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(517, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(517, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReferenceDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(518, 0),
        BrowseName=QualifiedName("ReferenceDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(518, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ContinuationPoint"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(521, 0),
        BrowseName=QualifiedName("ContinuationPoint", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(15, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(521, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(522, 0),
        BrowseName=QualifiedName("BrowseResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(525, 0),
        BrowseName=QualifiedName("BrowseRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(525, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(528, 0),
        BrowseName=QualifiedName("BrowseResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(528, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseNextRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(531, 0),
        BrowseName=QualifiedName("BrowseNextRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(531, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowseNextResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(534, 0),
        BrowseName=QualifiedName("BrowseNextResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(534, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RelativePathElement"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(537, 0),
        BrowseName=QualifiedName("RelativePathElement", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(537, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RelativePath"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(540, 0),
        BrowseName=QualifiedName("RelativePath", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(540, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowsePath"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(543, 0),
        BrowseName=QualifiedName("BrowsePath", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(543, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowsePathTarget"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(546, 0),
        BrowseName=QualifiedName("BrowsePathTarget", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(546, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BrowsePathResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(549, 0),
        BrowseName=QualifiedName("BrowsePathResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TranslateBrowsePathsToNodeIdsRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(552, 0),
        BrowseName=QualifiedName("TranslateBrowsePathsToNodeIdsRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(552, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TranslateBrowsePathsToNodeIdsResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(555, 0),
        BrowseName=QualifiedName("TranslateBrowsePathsToNodeIdsResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(555, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RegisterNodesRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(558, 0),
        BrowseName=QualifiedName("RegisterNodesRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(558, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RegisterNodesResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(561, 0),
        BrowseName=QualifiedName("RegisterNodesResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(561, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UnregisterNodesRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(564, 0),
        BrowseName=QualifiedName("UnregisterNodesRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UnregisterNodesResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(567, 0),
        BrowseName=QualifiedName("UnregisterNodesResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Counter"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(289, 0),
        BrowseName=QualifiedName("Counter", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(7, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NumericRange"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(291, 0),
        BrowseName=QualifiedName("NumericRange", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(12, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(291, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EndpointConfiguration"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(331, 0),
        BrowseName=QualifiedName("EndpointConfiguration", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("QueryDataDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(570, 0),
        BrowseName=QualifiedName("QueryDataDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NodeTypeDescription"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(573, 0),
        BrowseName=QualifiedName("NodeTypeDescription", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FilterOperator"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(576, 0),
        BrowseName=QualifiedName("FilterOperator", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7605, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Equals'), LocalizedText('IsNull'), LocalizedText('GreaterThan'), LocalizedText('LessThan'), LocalizedText('GreaterThanOrEqual'), LocalizedText('LessThanOrEqual'), LocalizedText('Like'), LocalizedText('Not'), LocalizedText('Between'), LocalizedText('InList'), LocalizedText('And'), LocalizedText('Or'), LocalizedText('Cast'), LocalizedText('InView'), LocalizedText('OfType'), LocalizedText('RelatedTo'), LocalizedText('BitwiseAnd'), LocalizedText('BitwiseOr')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7605, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(576, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(576, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("QueryDataSet"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(577, 0),
        BrowseName=QualifiedName("QueryDataSet", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(577, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NodeReference"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(580, 0),
        BrowseName=QualifiedName("NodeReference", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(580, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ContentFilterElement"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(583, 0),
        BrowseName=QualifiedName("ContentFilterElement", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(583, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ContentFilter"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(586, 0),
        BrowseName=QualifiedName("ContentFilter", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(586, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("FilterOperand"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(589, 0),
        BrowseName=QualifiedName("FilterOperand", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ElementOperand"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(592, 0),
        BrowseName=QualifiedName("ElementOperand", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(589, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(592, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(589, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("LiteralOperand"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(595, 0),
        BrowseName=QualifiedName("LiteralOperand", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(589, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(595, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(589, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AttributeOperand"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(598, 0),
        BrowseName=QualifiedName("AttributeOperand", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(589, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(598, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(589, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SimpleAttributeOperand"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(601, 0),
        BrowseName=QualifiedName("SimpleAttributeOperand", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(589, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(601, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(589, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ContentFilterElementResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(604, 0),
        BrowseName=QualifiedName("ContentFilterElementResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(604, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ContentFilterResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(607, 0),
        BrowseName=QualifiedName("ContentFilterResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(607, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ParsingResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(610, 0),
        BrowseName=QualifiedName("ParsingResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("QueryFirstRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(613, 0),
        BrowseName=QualifiedName("QueryFirstRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(613, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("QueryFirstResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(616, 0),
        BrowseName=QualifiedName("QueryFirstResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("QueryNextRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(619, 0),
        BrowseName=QualifiedName("QueryNextRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("QueryNextResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(622, 0),
        BrowseName=QualifiedName("QueryNextResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(622, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TimestampsToReturn"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(625, 0),
        BrowseName=QualifiedName("TimestampsToReturn", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(625, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Source'), LocalizedText('Server'), LocalizedText('Both'), LocalizedText('Neither'), LocalizedText('Invalid')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7606, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(625, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(625, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadValueId"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(626, 0),
        BrowseName=QualifiedName("ReadValueId", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(626, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(629, 0),
        BrowseName=QualifiedName("ReadRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(629, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(632, 0),
        BrowseName=QualifiedName("ReadResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(632, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryReadValueId"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(635, 0),
        BrowseName=QualifiedName("HistoryReadValueId", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(635, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryReadResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(638, 0),
        BrowseName=QualifiedName("HistoryReadResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(638, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryReadDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(641, 0),
        BrowseName=QualifiedName("HistoryReadDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(641, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadEventDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(644, 0),
        BrowseName=QualifiedName("ReadEventDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(644, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadRawModifiedDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(647, 0),
        BrowseName=QualifiedName("ReadRawModifiedDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(647, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadProcessedDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(650, 0),
        BrowseName=QualifiedName("ReadProcessedDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadAtTimeDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(653, 0),
        BrowseName=QualifiedName("ReadAtTimeDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(653, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ReadAnnotationDataDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23497, 0),
        BrowseName=QualifiedName("ReadAnnotationDataDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(23497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryData"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(656, 0),
        BrowseName=QualifiedName("HistoryData", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ModificationInfo"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11216, 0),
        BrowseName=QualifiedName("ModificationInfo", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11216, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryModifiedData"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11217, 0),
        BrowseName=QualifiedName("HistoryModifiedData", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(656, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(656, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryEvent"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(659, 0),
        BrowseName=QualifiedName("HistoryEvent", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryReadRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(662, 0),
        BrowseName=QualifiedName("HistoryReadRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(662, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryReadResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(665, 0),
        BrowseName=QualifiedName("HistoryReadResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(665, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("WriteValue"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(668, 0),
        BrowseName=QualifiedName("WriteValue", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("WriteRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(671, 0),
        BrowseName=QualifiedName("WriteRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("WriteResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(674, 0),
        BrowseName=QualifiedName("WriteResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(674, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryUpdateDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(677, 0),
        BrowseName=QualifiedName("HistoryUpdateDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryUpdateType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11234, 0),
        BrowseName=QualifiedName("HistoryUpdateType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11234, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11884, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11234, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='Insert'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='Replace'),
                ),
            ua.EnumValueType(
                Value=3,
                DisplayName=LocalizedText(Text='Update'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='Delete'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11884, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11234, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11234, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PerformUpdateType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11293, 0),
        BrowseName=QualifiedName("PerformUpdateType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11293, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11885, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11293, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='Insert'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='Replace'),
                ),
            ua.EnumValueType(
                Value=3,
                DisplayName=LocalizedText(Text='Update'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='Remove'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11885, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11293, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11293, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UpdateDataDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(680, 0),
        BrowseName=QualifiedName("UpdateDataDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UpdateStructureDataDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11295, 0),
        BrowseName=QualifiedName("UpdateStructureDataDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("UpdateEventDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(683, 0),
        BrowseName=QualifiedName("UpdateEventDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(683, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteRawModifiedDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(686, 0),
        BrowseName=QualifiedName("DeleteRawModifiedDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteAtTimeDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(689, 0),
        BrowseName=QualifiedName("DeleteAtTimeDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(689, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteEventDetails"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(692, 0),
        BrowseName=QualifiedName("DeleteEventDetails", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryUpdateResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(695, 0),
        BrowseName=QualifiedName("HistoryUpdateResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(695, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryUpdateRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(698, 0),
        BrowseName=QualifiedName("HistoryUpdateRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryUpdateResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(701, 0),
        BrowseName=QualifiedName("HistoryUpdateResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CallMethodRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(704, 0),
        BrowseName=QualifiedName("CallMethodRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(704, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CallMethodResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(707, 0),
        BrowseName=QualifiedName("CallMethodResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CallRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(710, 0),
        BrowseName=QualifiedName("CallRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(710, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CallResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(713, 0),
        BrowseName=QualifiedName("CallResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoringMode"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(716, 0),
        BrowseName=QualifiedName("MonitoringMode", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7608, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(716, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Disabled'), LocalizedText('Sampling'), LocalizedText('Reporting')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7608, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(716, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(716, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataChangeTrigger"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(717, 0),
        BrowseName=QualifiedName("DataChangeTrigger", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7609, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Status'), LocalizedText('StatusValue'), LocalizedText('StatusValueTimestamp')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7609, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(717, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(717, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeadbandType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(718, 0),
        BrowseName=QualifiedName("DeadbandType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7610, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('None'), LocalizedText('Absolute'), LocalizedText('Percent')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7610, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(718, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(718, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoringFilter"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(719, 0),
        BrowseName=QualifiedName("MonitoringFilter", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataChangeFilter"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(722, 0),
        BrowseName=QualifiedName("DataChangeFilter", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(719, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(722, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(719, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EventFilter"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(725, 0),
        BrowseName=QualifiedName("EventFilter", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(719, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(719, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AggregateConfiguration"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(948, 0),
        BrowseName=QualifiedName("AggregateConfiguration", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(948, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AggregateFilter"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(728, 0),
        BrowseName=QualifiedName("AggregateFilter", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(719, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(719, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoringFilterResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(731, 0),
        BrowseName=QualifiedName("MonitoringFilterResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EventFilterResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(734, 0),
        BrowseName=QualifiedName("EventFilterResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(731, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(731, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AggregateFilterResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(737, 0),
        BrowseName=QualifiedName("AggregateFilterResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(731, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(731, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoringParameters"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(740, 0),
        BrowseName=QualifiedName("MonitoringParameters", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(740, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoredItemCreateRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(743, 0),
        BrowseName=QualifiedName("MonitoredItemCreateRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(743, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoredItemCreateResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(746, 0),
        BrowseName=QualifiedName("MonitoredItemCreateResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(746, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CreateMonitoredItemsRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(749, 0),
        BrowseName=QualifiedName("CreateMonitoredItemsRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(749, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CreateMonitoredItemsResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(752, 0),
        BrowseName=QualifiedName("CreateMonitoredItemsResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(752, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoredItemModifyRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(755, 0),
        BrowseName=QualifiedName("MonitoredItemModifyRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(755, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoredItemModifyResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(758, 0),
        BrowseName=QualifiedName("MonitoredItemModifyResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ModifyMonitoredItemsRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(761, 0),
        BrowseName=QualifiedName("ModifyMonitoredItemsRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(761, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ModifyMonitoredItemsResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(764, 0),
        BrowseName=QualifiedName("ModifyMonitoredItemsResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(764, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SetMonitoringModeRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(767, 0),
        BrowseName=QualifiedName("SetMonitoringModeRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SetMonitoringModeResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(770, 0),
        BrowseName=QualifiedName("SetMonitoringModeResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SetTriggeringRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(773, 0),
        BrowseName=QualifiedName("SetTriggeringRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(773, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SetTriggeringResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(776, 0),
        BrowseName=QualifiedName("SetTriggeringResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(776, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteMonitoredItemsRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(779, 0),
        BrowseName=QualifiedName("DeleteMonitoredItemsRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(779, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteMonitoredItemsResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(782, 0),
        BrowseName=QualifiedName("DeleteMonitoredItemsResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CreateSubscriptionRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(785, 0),
        BrowseName=QualifiedName("CreateSubscriptionRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(785, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("CreateSubscriptionResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(788, 0),
        BrowseName=QualifiedName("CreateSubscriptionResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(788, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ModifySubscriptionRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(791, 0),
        BrowseName=QualifiedName("ModifySubscriptionRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(791, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ModifySubscriptionResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(794, 0),
        BrowseName=QualifiedName("ModifySubscriptionResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SetPublishingModeRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(797, 0),
        BrowseName=QualifiedName("SetPublishingModeRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(797, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SetPublishingModeResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(800, 0),
        BrowseName=QualifiedName("SetPublishingModeResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NotificationMessage"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(803, 0),
        BrowseName=QualifiedName("NotificationMessage", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NotificationData"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(945, 0),
        BrowseName=QualifiedName("NotificationData", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(945, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DataChangeNotification"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(809, 0),
        BrowseName=QualifiedName("DataChangeNotification", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(945, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(809, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(945, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("MonitoredItemNotification"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(806, 0),
        BrowseName=QualifiedName("MonitoredItemNotification", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(806, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EventNotificationList"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(914, 0),
        BrowseName=QualifiedName("EventNotificationList", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(945, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(914, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(945, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EventFieldList"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(917, 0),
        BrowseName=QualifiedName("EventFieldList", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("HistoryEventFieldList"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(920, 0),
        BrowseName=QualifiedName("HistoryEventFieldList", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(920, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("StatusChangeNotification"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(818, 0),
        BrowseName=QualifiedName("StatusChangeNotification", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(945, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(945, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SubscriptionAcknowledgement"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(821, 0),
        BrowseName=QualifiedName("SubscriptionAcknowledgement", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PublishRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(824, 0),
        BrowseName=QualifiedName("PublishRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("PublishResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(827, 0),
        BrowseName=QualifiedName("PublishResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RepublishRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(830, 0),
        BrowseName=QualifiedName("RepublishRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RepublishResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(833, 0),
        BrowseName=QualifiedName("RepublishResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TransferResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(836, 0),
        BrowseName=QualifiedName("TransferResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TransferSubscriptionsRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(839, 0),
        BrowseName=QualifiedName("TransferSubscriptionsRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("TransferSubscriptionsResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(842, 0),
        BrowseName=QualifiedName("TransferSubscriptionsResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteSubscriptionsRequest"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(845, 0),
        BrowseName=QualifiedName("DeleteSubscriptionsRequest", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DeleteSubscriptionsResponse"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(848, 0),
        BrowseName=QualifiedName("DeleteSubscriptionsResponse", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("BuildInfo"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(338, 0),
        BrowseName=QualifiedName("BuildInfo", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RedundancySupport"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(851, 0),
        BrowseName=QualifiedName("RedundancySupport", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7611, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('None'), LocalizedText('Cold'), LocalizedText('Warm'), LocalizedText('Hot'), LocalizedText('Transparent'), LocalizedText('HotAndMirrored')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7611, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(851, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(851, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ServerState"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(852, 0),
        BrowseName=QualifiedName("ServerState", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7612, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Running'), LocalizedText('Failed'), LocalizedText('NoConfiguration'), LocalizedText('Suspended'), LocalizedText('Shutdown'), LocalizedText('Test'), LocalizedText('CommunicationFault'), LocalizedText('Unknown')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7612, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(852, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(852, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("RedundantServerDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(853, 0),
        BrowseName=QualifiedName("RedundantServerDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EndpointUrlListDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11943, 0),
        BrowseName=QualifiedName("EndpointUrlListDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("NetworkGroupDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11944, 0),
        BrowseName=QualifiedName("NetworkGroupDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11944, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnosticsDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(856, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnosticsDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsSummaryDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(859, 0),
        BrowseName=QualifiedName("ServerDiagnosticsSummaryDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ServerStatusDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(862, 0),
        BrowseName=QualifiedName("ServerStatusDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(865, 0),
        BrowseName=QualifiedName("SessionDiagnosticsDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(865, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(868, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(868, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ServiceCounterDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(871, 0),
        BrowseName=QualifiedName("ServiceCounterDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(871, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("StatusResult"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(299, 0),
        BrowseName=QualifiedName("StatusResult", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(299, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(874, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(874, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ModelChangeStructureVerbMask"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11941, 0),
        BrowseName=QualifiedName("ModelChangeStructureVerbMask", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11941, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11942, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(11941, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValues"),
        DataType=NumericNodeId(7594, 0),
        Value=ua.Variant([
            ua.EnumValueType(
                Value=1,
                DisplayName=LocalizedText(Text='NodeAdded'),
                ),
            ua.EnumValueType(
                Value=2,
                DisplayName=LocalizedText(Text='NodeDeleted'),
                ),
            ua.EnumValueType(
                Value=4,
                DisplayName=LocalizedText(Text='ReferenceAdded'),
                ),
            ua.EnumValueType(
                Value=8,
                DisplayName=LocalizedText(Text='ReferenceDeleted'),
                ),
            ua.EnumValueType(
                Value=16,
                DisplayName=LocalizedText(Text='DataTypeChanged'),
                ),
            ],
            ua.VariantType.ExtensionObject),
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11942, 0),
        BrowseName=QualifiedName("EnumValues", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(11941, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(11942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(11942, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11941, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ModelChangeStructureDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(877, 0),
        BrowseName=QualifiedName("ModelChangeStructureDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(877, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("SemanticChangeStructureDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(897, 0),
        BrowseName=QualifiedName("SemanticChangeStructureDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Range"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(884, 0),
        BrowseName=QualifiedName("Range", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(884, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("EUInformation"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(887, 0),
        BrowseName=QualifiedName("EUInformation", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AxisScaleEnumeration"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12077, 0),
        BrowseName=QualifiedName("AxisScaleEnumeration", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12078, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('Linear'), LocalizedText('Log'), LocalizedText('Ln')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12078, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(12077, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(12078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(12078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12077, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ComplexNumberType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12171, 0),
        BrowseName=QualifiedName("ComplexNumberType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("DoubleComplexNumberType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12172, 0),
        BrowseName=QualifiedName("DoubleComplexNumberType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("AxisInformation"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12079, 0),
        BrowseName=QualifiedName("AxisInformation", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("XVType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12080, 0),
        BrowseName=QualifiedName("XVType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(12080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ProgramDiagnosticDataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(894, 0),
        BrowseName=QualifiedName("ProgramDiagnosticDataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ProgramDiagnostic2DataType"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15396, 0),
        BrowseName=QualifiedName("ProgramDiagnostic2DataType", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(15396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("Annotation"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(891, 0),
        BrowseName=QualifiedName("Annotation", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(22, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(891, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(22, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.DataTypeAttributes(
        DisplayName=LocalizedText("ExceptionDeviationFormat"),
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(890, 0),
        BrowseName=QualifiedName("ExceptionDeviationFormat", 0),
        NodeClass_=NodeClass.DataType,
        ParentNodeId=NumericNodeId(29, 0),
        ReferenceTypeId=NumericNodeId(45, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7614, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(45, 0),
        SourceNodeId=NumericNodeId(890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(29, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumStrings"),
        DataType=ua.NodeId(ua.ObjectIds.LocalizedText),
        Value=[LocalizedText('AbsoluteValue'), LocalizedText('PercentOfValue'), LocalizedText('PercentOfRange'), LocalizedText('PercentOfEURange'), LocalizedText('Unknown')],
        ValueRank=1,
        ArrayDimensions=[0],
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7614, 0),
        BrowseName=QualifiedName("EnumStrings", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(890, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(37, 0),
        SourceNodeId=NumericNodeId(7614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(78, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(890, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14846, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14533, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14533, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14873, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17537, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(16313, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17537, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16313, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(17537, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17538, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17537, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17549, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17548, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17548, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(17549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17550, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17549, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15671, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15528, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15528, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18815, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18806, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18806, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18815, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18816, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18807, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18807, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18816, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18817, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18808, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18808, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18817, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18818, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18809, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18809, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18818, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18819, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18810, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18810, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18820, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18811, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18811, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18821, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18812, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18812, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18821, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18822, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18813, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18813, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18823, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18814, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18814, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15736, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15634, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15634, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15738, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23507, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23498, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23498, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(23507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23514, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23507, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12680, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12554, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12554, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12681, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17863, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17861, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17861, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15676, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15534, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15534, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15741, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(125, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14525, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14525, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(126, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15487, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15487, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15599, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(127, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15488, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15488, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15602, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15421, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15005, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15005, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15501, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15422, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15006, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15006, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15521, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(124, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14523, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14523, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(124, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14839, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14524, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14524, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14847, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14593, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14593, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14876, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15677, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15578, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15578, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15766, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15678, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15580, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15769, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14323, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14273, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14324, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15679, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15581, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15581, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15772, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15681, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15582, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15775, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15682, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15597, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15597, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15778, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15683, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15598, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15683, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15683, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15781, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15683, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15688, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15605, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15688, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15605, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15688, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15784, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15688, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15689, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15609, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15689, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15609, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15689, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15787, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15689, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21150, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15480, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21156, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15691, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15611, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15611, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15793, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15693, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15616, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15854, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15694, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15617, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15617, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15857, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15695, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15618, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15695, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15618, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15695, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15860, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15695, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21151, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15502, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21159, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21152, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15510, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15510, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21162, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21153, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15520, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15520, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21165, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15701, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15621, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15621, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15866, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15701, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15702, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15622, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15869, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15703, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15623, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15623, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15872, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15705, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15628, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15628, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15877, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15706, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15629, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15880, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15707, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15630, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15630, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15883, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15707, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15712, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15631, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15631, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15886, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14848, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14744, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14744, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15713, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15635, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15889, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15713, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21154, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15530, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15530, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21168, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15715, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15645, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15645, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15895, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15717, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15652, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15652, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15898, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15717, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15718, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15653, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15919, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15719, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15657, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15657, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15922, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15719, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15724, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15664, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15664, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15925, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15725, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15665, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15931, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15725, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17468, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17467, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17467, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(17468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17469, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17468, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21155, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15532, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15532, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15479, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15007, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15007, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15524, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15727, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15667, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15667, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15729, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15669, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15669, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15733, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15670, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15670, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15946, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23499, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23468, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23468, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(23499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(128, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(96, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(96, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16131, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(128, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(121, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(97, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(97, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(121, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14844, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(101, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18181, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(122, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(99, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(99, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18184, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(122, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(123, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(100, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18187, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(123, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(260, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(258, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(258, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11889, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11879, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11879, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11890, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11880, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11880, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11890, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(263, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(261, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(261, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(266, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(264, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(264, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(266, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(269, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(267, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(267, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(272, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(270, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(270, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(275, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(273, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(278, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(276, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(276, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(281, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(279, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(279, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(284, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(282, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(282, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(287, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(285, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(285, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(298, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(296, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(296, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(298, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8251, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(7594, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(8251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7594, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(8251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7656, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8251, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14845, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(102, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14870, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12765, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12755, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12766, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12756, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12756, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12770, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8917, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(8912, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(8917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8912, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(8917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8914, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8917, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(310, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(308, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(308, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(391, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(389, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(394, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(392, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(392, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(397, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(395, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(395, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15903, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15901, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15901, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15903, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21001, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(20999, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20999, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21001, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(422, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(420, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(422, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(425, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(423, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(425, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12207, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12189, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12213, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12208, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12190, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12190, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12209, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12191, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12191, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(306, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(304, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(304, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7662, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(314, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(312, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(312, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7668, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(428, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(426, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(426, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(428, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(431, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(429, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(431, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(431, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(434, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(432, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(432, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7782, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(434, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(437, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(435, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(437, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(435, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(437, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(440, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(438, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(438, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(440, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12900, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12890, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12890, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12902, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12900, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12901, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12891, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12905, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12901, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12211, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12193, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12193, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12212, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12194, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12212, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12194, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12212, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(443, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(441, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(441, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(443, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(446, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(444, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(444, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(446, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(449, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(447, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(449, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(447, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(449, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(452, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(450, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(450, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(452, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(455, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(453, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(453, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(455, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(346, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(344, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(346, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(344, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(346, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(346, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(458, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(456, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(458, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(461, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(459, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(459, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(461, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(464, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(462, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(464, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(318, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(316, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(316, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7671, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(321, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(319, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(319, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7674, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(324, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(322, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(322, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7677, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(327, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(325, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(325, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7680, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(940, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(938, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(938, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7683, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(467, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(465, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(467, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(467, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(470, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(468, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(470, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(468, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(470, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(473, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(471, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(473, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(471, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(473, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(476, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(474, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(474, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(479, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(477, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(479, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(482, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(480, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(482, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(351, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(354, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(352, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(352, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(357, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(355, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(355, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(360, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(358, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(358, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(363, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(361, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(361, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(366, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(364, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(364, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(369, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(367, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(367, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(372, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(370, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(370, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(375, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(373, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(373, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17610, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17606, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17610, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17611, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17607, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(378, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(376, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(376, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(485, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(483, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(483, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(485, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(488, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(486, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(488, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(491, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(489, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(489, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(491, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(381, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(379, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(379, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(494, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(492, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(492, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(494, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(497, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(495, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(495, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(497, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(384, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(382, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(382, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(384, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(500, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(498, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(498, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(503, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(501, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(501, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(503, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(387, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(385, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(385, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(387, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(506, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(504, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(504, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(509, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(507, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(507, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(509, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(513, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(511, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(511, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(513, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(516, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(514, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(516, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(514, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(516, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(520, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(518, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(520, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(518, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(520, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(524, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(522, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(527, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(525, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(525, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(527, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(530, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(528, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(530, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(528, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(530, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(533, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(531, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(533, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(531, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(533, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(536, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(534, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(536, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(534, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(536, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(539, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(537, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(539, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(537, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(539, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12718, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(539, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(542, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(540, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(540, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12721, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(545, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(543, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(543, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(545, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(548, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(546, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(546, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(548, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(551, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(549, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(549, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(551, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(554, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(552, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(552, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(557, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(555, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(560, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(558, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(558, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(560, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(563, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(561, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(561, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(563, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(566, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(564, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(566, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(566, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(569, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(567, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(569, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(333, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(331, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(331, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(572, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(570, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(570, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(572, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(575, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(573, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(575, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(579, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(577, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(577, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(582, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(580, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(585, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(583, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(583, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7929, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(588, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(586, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(586, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(591, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(589, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(589, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(594, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(592, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(592, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7938, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(597, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(595, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(595, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7941, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(597, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(600, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(598, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(600, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(603, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(601, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(601, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7947, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(603, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(606, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(604, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(604, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(606, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(609, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(607, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(612, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(610, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(610, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(612, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(615, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(613, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(613, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(615, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(618, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(616, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(618, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(621, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(619, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(619, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(621, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(624, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(622, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(624, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(628, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(626, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(628, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(626, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(628, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(631, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(629, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(631, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(634, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(632, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(632, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(634, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(637, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(635, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(637, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(640, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(638, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(638, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(643, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(643, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(646, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(644, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(644, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(646, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(649, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(647, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(647, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(649, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(652, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(650, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(652, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(655, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(653, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(655, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23500, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23497, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23497, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23500, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(658, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(656, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(656, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(658, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11226, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11216, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11216, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11227, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11217, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11227, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11227, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(661, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(659, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(661, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(659, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(661, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8004, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(661, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(664, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(662, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(664, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(662, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(664, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(667, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(665, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(667, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(667, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(670, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(668, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(670, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(668, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(670, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(673, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(671, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(673, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(671, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(673, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(676, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(674, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(674, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(679, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(679, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(682, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(680, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(680, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(682, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11300, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11295, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11295, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(685, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(683, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(685, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(683, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(685, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(688, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(686, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(688, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(688, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(691, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(689, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(689, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(691, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(694, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(692, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(692, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(694, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(697, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(695, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(695, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(697, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(700, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(698, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(703, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(701, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(701, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(703, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(706, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(704, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(704, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(706, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(709, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(707, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(707, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(709, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(712, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(710, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(715, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(713, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(713, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(721, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(719, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(719, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8067, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(724, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(722, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(722, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(724, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(727, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(725, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(727, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(950, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(948, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(730, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(728, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(733, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(731, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(733, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(736, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(734, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(736, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(739, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(737, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(739, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(742, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(740, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(740, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(742, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(745, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(743, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(743, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(745, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(748, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(746, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(748, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(751, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(749, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(749, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(751, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(754, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(752, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(752, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(754, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(757, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(755, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(760, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(758, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(758, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(760, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(763, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(761, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(761, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(763, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(766, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(764, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(764, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(769, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(767, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(772, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(770, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(770, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(775, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(773, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(773, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(778, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(776, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(776, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(781, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(779, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(779, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(784, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(782, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(782, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(787, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(785, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(785, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(790, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(788, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(788, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(790, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(793, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(791, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(793, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(791, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(793, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(796, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(794, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(794, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(799, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(797, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(797, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(802, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(800, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(800, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(805, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(803, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(947, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(945, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(945, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(811, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(809, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(809, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(808, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(806, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(806, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(916, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(914, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(914, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(916, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(919, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(917, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(919, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(917, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(919, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(922, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(920, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(920, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(820, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(818, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(818, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(820, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(823, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(821, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(823, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(826, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(824, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(829, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(827, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(832, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(830, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(835, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(833, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(838, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(836, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(838, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(841, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(839, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(841, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(844, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(842, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(844, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(847, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(845, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(847, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(850, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(848, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(850, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(340, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(338, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(338, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7692, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(855, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(853, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8208, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11957, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11943, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(11957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11959, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11957, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11958, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11944, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(11958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11962, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11958, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(858, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(856, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(856, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8211, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(861, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(859, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(859, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8214, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(864, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(862, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(862, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(867, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(865, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(865, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8220, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(870, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(868, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(868, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8223, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(873, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(871, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(871, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8226, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(301, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(299, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(301, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(299, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(301, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7659, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(301, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(876, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(874, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(874, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8229, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(879, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(877, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(877, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8232, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(899, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(897, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8235, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(899, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(886, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(884, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(884, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8238, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(889, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(887, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(887, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8241, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12181, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12171, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12183, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12182, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12172, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12186, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12089, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12079, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12091, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12090, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12080, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(896, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(894, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8247, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(896, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15397, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15396, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15396, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15398, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15397, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default Binary"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(893, 0),
        BrowseName=QualifiedName("Default Binary", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(891, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8244, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Opc.Ua"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        Value=ua.Variant(b'<opc:TypeDictionary\r\n  xmlns:opc="http://opcfoundation.org/BinarySchema/"\r\n  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"\r\n  xmlns:ua="http://opcfoundation.org/UA/"\r\n  xmlns:tns="http://opcfoundation.org/UA/"\r\n  DefaultByteOrder="LittleEndian"\r\n  TargetNamespace="http://opcfoundation.org/UA/"\r\n>\r\n\r\n  <opc:Import Namespace="http://opcfoundation.org/BinarySchema/" />\r\n\r\n  <opc:StructuredType Name="XmlElement">\r\n    <opc:Documentation>An XML element encoded as a UTF-8 string.</opc:Documentation>\r\n    <opc:Field Name="Length" TypeName="opc:Int32" />\r\n    <opc:Field Name="Value" TypeName="opc:Char" LengthField="Length" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="NodeIdType" LengthInBits="6">\r\n    <opc:Documentation>The possible encodings for a NodeId value.</opc:Documentation>\r\n    <opc:EnumeratedValue Name="TwoByte" Value="0" />\r\n    <opc:EnumeratedValue Name="FourByte" Value="1" />\r\n    <opc:EnumeratedValue Name="Numeric" Value="2" />\r\n    <opc:EnumeratedValue Name="String" Value="3" />\r\n    <opc:EnumeratedValue Name="Guid" Value="4" />\r\n    <opc:EnumeratedValue Name="ByteString" Value="5" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="TwoByteNodeId">\r\n    <opc:Field Name="Identifier" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FourByteNodeId">\r\n    <opc:Field Name="NamespaceIndex" TypeName="opc:Byte" />\r\n    <opc:Field Name="Identifier" TypeName="opc:UInt16" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NumericNodeId">\r\n    <opc:Field Name="NamespaceIndex" TypeName="opc:UInt16" />\r\n    <opc:Field Name="Identifier" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="StringNodeId">\r\n    <opc:Field Name="NamespaceIndex" TypeName="opc:UInt16" />\r\n    <opc:Field Name="Identifier" TypeName="opc:CharArray" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="GuidNodeId">\r\n    <opc:Field Name="NamespaceIndex" TypeName="opc:UInt16" />\r\n    <opc:Field Name="Identifier" TypeName="opc:Guid" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ByteStringNodeId">\r\n    <opc:Field Name="NamespaceIndex" TypeName="opc:UInt16" />\r\n    <opc:Field Name="Identifier" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NodeId">\r\n    <opc:Documentation>An identifier for a node in a UA server address space.</opc:Documentation>\r\n    <opc:Field Name="NodeIdType" TypeName="ua:NodeIdType" />\r\n    <opc:Field Name="Reserved1" TypeName="opc:Bit" Length="2" />\r\n    <opc:Field Name="TwoByte" TypeName="ua:TwoByteNodeId" SwitchField="NodeIdType" SwitchValue="0" />\r\n    <opc:Field Name="FourByte" TypeName="ua:FourByteNodeId" SwitchField="NodeIdType" SwitchValue="1" />\r\n    <opc:Field Name="Numeric" TypeName="ua:NumericNodeId" SwitchField="NodeIdType" SwitchValue="2" />\r\n    <opc:Field Name="String" TypeName="ua:StringNodeId" SwitchField="NodeIdType" SwitchValue="3" />\r\n    <opc:Field Name="Guid" TypeName="ua:GuidNodeId" SwitchField="NodeIdType" SwitchValue="4" />\r\n    <opc:Field Name="ByteString" TypeName="ua:ByteStringNodeId" SwitchField="NodeIdType" SwitchValue="5" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ExpandedNodeId">\r\n    <opc:Documentation>An identifier for a node in a UA server address space qualified with a complete namespace string.</opc:Documentation>\r\n    <opc:Field Name="NodeIdType" TypeName="ua:NodeIdType" />\r\n    <opc:Field Name="ServerIndexSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="NamespaceURISpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="TwoByte" TypeName="ua:TwoByteNodeId" SwitchField="NodeIdType" SwitchValue="0" />\r\n    <opc:Field Name="FourByte" TypeName="ua:FourByteNodeId" SwitchField="NodeIdType" SwitchValue="1" />\r\n    <opc:Field Name="Numeric" TypeName="ua:NumericNodeId" SwitchField="NodeIdType" SwitchValue="2" />\r\n    <opc:Field Name="String" TypeName="ua:StringNodeId" SwitchField="NodeIdType" SwitchValue="3" />\r\n    <opc:Field Name="Guid" TypeName="ua:GuidNodeId" SwitchField="NodeIdType" SwitchValue="4" />\r\n    <opc:Field Name="ByteString" TypeName="ua:ByteStringNodeId" SwitchField="NodeIdType" SwitchValue="5" />\r\n    <opc:Field Name="NamespaceURI" TypeName="opc:CharArray" SwitchField="NamespaceURISpecified"/>\r\n    <opc:Field Name="ServerIndex" TypeName="opc:UInt32" SwitchField="ServerIndexSpecified"/>\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="StatusCode" LengthInBits="32" ByteOrderSignificant="true">\r\n    <opc:Documentation>A 32-bit status code value.</opc:Documentation>\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="DiagnosticInfo">\r\n    <opc:Documentation>A recursive structure containing diagnostic information associated with a status code.</opc:Documentation>\r\n    <opc:Field Name="SymbolicIdSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="NamespaceURISpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="LocalizedTextSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="LocaleSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="AdditionalInfoSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="InnerStatusCodeSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="InnerDiagnosticInfoSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="Reserved1" TypeName="opc:Bit" Length="1" />\r\n    <opc:Field Name="SymbolicId" TypeName="opc:Int32" SwitchField="SymbolicIdSpecified" />\r\n    <opc:Field Name="NamespaceURI" TypeName="opc:Int32" SwitchField="NamespaceURISpecified" />\r\n    <opc:Field Name="Locale" TypeName="opc:Int32" SwitchField="LocaleSpecified" />\r\n    <opc:Field Name="LocalizedText" TypeName="opc:Int32" SwitchField="LocalizedTextSpecified" />\r\n    <opc:Field Name="AdditionalInfo" TypeName="opc:CharArray" SwitchField="AdditionalInfoSpecified" />\r\n    <opc:Field Name="InnerStatusCode" TypeName="ua:StatusCode" SwitchField="InnerStatusCodeSpecified" />\r\n    <opc:Field Name="InnerDiagnosticInfo" TypeName="ua:DiagnosticInfo" SwitchField="InnerDiagnosticInfoSpecified" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="QualifiedName">\r\n    <opc:Documentation>A string qualified with a namespace index.</opc:Documentation>\r\n    <opc:Field Name="NamespaceIndex" TypeName="opc:UInt16" />\r\n    <opc:Field Name="Name" TypeName="opc:CharArray" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="LocalizedText">\r\n    <opc:Documentation>A string qualified with a namespace index.</opc:Documentation>\r\n    <opc:Field Name="LocaleSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="TextSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="Reserved1" TypeName="opc:Bit" Length="6" />\r\n    <opc:Field Name="Locale" TypeName="opc:CharArray" SwitchField="LocaleSpecified" />\r\n    <opc:Field Name="Text" TypeName="opc:CharArray" SwitchField="TextSpecified" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataValue">\r\n    <opc:Documentation>A value with an associated timestamp, and quality.</opc:Documentation>\r\n    <opc:Field Name="ValueSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="StatusCodeSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="SourceTimestampSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="ServerTimestampSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="SourcePicosecondsSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="ServerPicosecondsSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="Reserved1" TypeName="opc:Bit" Length="2" />\r\n    <opc:Field Name="Value" TypeName="ua:Variant" SwitchField="ValueSpecified" />\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" SwitchField="StatusCodeSpecified" />\r\n    <opc:Field Name="SourceTimestamp" TypeName="opc:DateTime" SwitchField="SourceTimestampSpecified" />\r\n    <opc:Field Name="SourcePicoseconds" TypeName="opc:UInt16" SwitchField="SourcePicosecondsSpecified" />\r\n    <opc:Field Name="ServerTimestamp" TypeName="opc:DateTime" SwitchField="ServerTimestampSpecified" />\r\n    <opc:Field Name="ServerPicoseconds" TypeName="opc:UInt16" SwitchField="ServerPicosecondsSpecified" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ExtensionObject">\r\n    <opc:Documentation>A serialized object prefixed with its data type identifier.</opc:Documentation>\r\n    <opc:Field Name="TypeIdSpecified" TypeName="opc:Bit" />\r\n    <opc:Field Name="BinaryBody" TypeName="opc:Bit" />\r\n    <opc:Field Name="XmlBody" TypeName="opc:Bit" />\r\n    <opc:Field Name="Reserved1" TypeName="opc:Bit" Length="5" />\r\n    <opc:Field Name="TypeId" TypeName="ua:ExpandedNodeId" SwitchField="TypeIdSpecified" />\r\n    <opc:Field Name="BodyLength" TypeName="opc:Int32" />\r\n    <opc:Field Name="Body" TypeName="opc:Byte" LengthField="BodyLength" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Variant">\r\n    <opc:Documentation>A union of several types.</opc:Documentation>\r\n    <opc:Field Name="VariantType" TypeName="opc:Bit" Length="6" />\r\n    <opc:Field Name="ArrayDimensionsSpecified" TypeName="opc:Bit" Length="1"/>\r\n    <opc:Field Name="ArrayLengthSpecified" TypeName="opc:Bit" Length="1"/>\r\n    <opc:Field Name="ArrayLength" TypeName="opc:Int32" SwitchField="ArrayLengthSpecified" />\r\n    <opc:Field Name="Boolean" TypeName="opc:Boolean" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="1" />\r\n    <opc:Field Name="SByte" TypeName="opc:SByte" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="2" />\r\n    <opc:Field Name="Byte" TypeName="opc:Byte" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="3" />\r\n    <opc:Field Name="Int16" TypeName="opc:Int16" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="4" />\r\n    <opc:Field Name="UInt16" TypeName="opc:UInt16" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="5" />\r\n    <opc:Field Name="Int32" TypeName="opc:Int32" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="6" />\r\n    <opc:Field Name="UInt32" TypeName="opc:UInt32" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="7" />\r\n    <opc:Field Name="Int64" TypeName="opc:Int64" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="8" />\r\n    <opc:Field Name="UInt64" TypeName="opc:UInt64" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="9" />\r\n    <opc:Field Name="Float" TypeName="opc:Float" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="10" />\r\n    <opc:Field Name="Double" TypeName="opc:Double" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="11" />\r\n    <opc:Field Name="String" TypeName="opc:CharArray" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="12" />\r\n    <opc:Field Name="DateTime" TypeName="opc:DateTime" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="13" />\r\n    <opc:Field Name="Guid" TypeName="opc:Guid" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="14" />\r\n    <opc:Field Name="ByteString" TypeName="opc:ByteString" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="15" />\r\n    <opc:Field Name="XmlElement" TypeName="ua:XmlElement" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="16" />\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="17" />\r\n    <opc:Field Name="ExpandedNodeId" TypeName="ua:ExpandedNodeId" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="18" />\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="19" />\r\n    <opc:Field Name="QualifiedName" TypeName="ua:QualifiedName" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="20" />\r\n    <opc:Field Name="LocalizedText" TypeName="ua:LocalizedText" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="21" />\r\n    <opc:Field Name="ExtensionObject" TypeName="ua:ExtensionObject" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="22" />\r\n    <opc:Field Name="DataValue" TypeName="ua:DataValue" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="23" />\r\n    <opc:Field Name="Variant" TypeName="ua:Variant" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="24" />\r\n    <opc:Field Name="DiagnosticInfo" TypeName="ua:DiagnosticInfo" LengthField="ArrayLength" SwitchField="VariantType" SwitchValue="25" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" SwitchField="ArrayDimensionsSpecified" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:Int32" LengthField="NoOfArrayDimensions" SwitchField="ArrayDimensionsSpecified" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="NamingRuleType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Mandatory" Value="1" />\r\n    <opc:EnumeratedValue Name="Optional" Value="2" />\r\n    <opc:EnumeratedValue Name="Constraint" Value="3" />\r\n  </opc:EnumeratedType>\r\n    \r\n\r\n  <opc:OpaqueType Name="ImageBMP">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="ImageGIF">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="ImageJPG">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="ImagePNG">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="AudioDataType">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="BitFieldMaskDataType">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="KeyValuePair" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Key" TypeName="ua:QualifiedName" />\r\n    <opc:Field Name="Value" TypeName="ua:Variant" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AdditionalParametersType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfParameters" TypeName="opc:Int32" />\r\n    <opc:Field Name="Parameters" TypeName="tns:KeyValuePair" LengthField="NoOfParameters" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EphemeralKeyType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="PublicKey" TypeName="opc:ByteString" />\r\n    <opc:Field Name="Signature" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EndpointType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="EndpointUrl" TypeName="opc:String" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" />\r\n    <opc:Field Name="SecurityPolicyUri" TypeName="opc:String" />\r\n    <opc:Field Name="TransportProfileUri" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RationalNumber" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Numerator" TypeName="opc:Int32" />\r\n    <opc:Field Name="Denominator" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Vector" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ThreeDVector" BaseType="tns:Vector">\r\n    <opc:Field Name="X" TypeName="opc:Double" />\r\n    <opc:Field Name="Y" TypeName="opc:Double" />\r\n    <opc:Field Name="Z" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CartesianCoordinates" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ThreeDCartesianCoordinates" BaseType="tns:CartesianCoordinates">\r\n    <opc:Field Name="X" TypeName="opc:Double" />\r\n    <opc:Field Name="Y" TypeName="opc:Double" />\r\n    <opc:Field Name="Z" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Orientation" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ThreeDOrientation" BaseType="tns:Orientation">\r\n    <opc:Field Name="A" TypeName="opc:Double" />\r\n    <opc:Field Name="B" TypeName="opc:Double" />\r\n    <opc:Field Name="C" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Frame" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ThreeDFrame" BaseType="tns:Frame">\r\n    <opc:Field Name="CartesianCoordinates" TypeName="tns:ThreeDCartesianCoordinates" />\r\n    <opc:Field Name="Orientation" TypeName="tns:ThreeDOrientation" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="OpenFileMode" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Read" Value="1" />\r\n    <opc:EnumeratedValue Name="Write" Value="2" />\r\n    <opc:EnumeratedValue Name="EraseExisting" Value="4" />\r\n    <opc:EnumeratedValue Name="Append" Value="8" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="IdentityCriteriaType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="UserName" Value="1" />\r\n    <opc:EnumeratedValue Name="Thumbprint" Value="2" />\r\n    <opc:EnumeratedValue Name="Role" Value="3" />\r\n    <opc:EnumeratedValue Name="GroupId" Value="4" />\r\n    <opc:EnumeratedValue Name="Anonymous" Value="5" />\r\n    <opc:EnumeratedValue Name="AuthenticatedUser" Value="6" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="IdentityMappingRuleType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="CriteriaType" TypeName="tns:IdentityCriteriaType" />\r\n    <opc:Field Name="Criteria" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CurrencyUnitType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NumericCode" TypeName="opc:Int16" />\r\n    <opc:Field Name="Exponent" TypeName="opc:SByte" />\r\n    <opc:Field Name="AlphabeticCode" TypeName="opc:String" />\r\n    <opc:Field Name="Currency" TypeName="ua:LocalizedText" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="TrustListMasks" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="TrustedCertificates" Value="1" />\r\n    <opc:EnumeratedValue Name="TrustedCrls" Value="2" />\r\n    <opc:EnumeratedValue Name="IssuerCertificates" Value="4" />\r\n    <opc:EnumeratedValue Name="IssuerCrls" Value="8" />\r\n    <opc:EnumeratedValue Name="All" Value="15" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="TrustListDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SpecifiedLists" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfTrustedCertificates" TypeName="opc:Int32" />\r\n    <opc:Field Name="TrustedCertificates" TypeName="opc:ByteString" LengthField="NoOfTrustedCertificates" />\r\n    <opc:Field Name="NoOfTrustedCrls" TypeName="opc:Int32" />\r\n    <opc:Field Name="TrustedCrls" TypeName="opc:ByteString" LengthField="NoOfTrustedCrls" />\r\n    <opc:Field Name="NoOfIssuerCertificates" TypeName="opc:Int32" />\r\n    <opc:Field Name="IssuerCertificates" TypeName="opc:ByteString" LengthField="NoOfIssuerCertificates" />\r\n    <opc:Field Name="NoOfIssuerCrls" TypeName="opc:Int32" />\r\n    <opc:Field Name="IssuerCrls" TypeName="opc:ByteString" LengthField="NoOfIssuerCrls" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DecimalDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Scale" TypeName="opc:Int16" />\r\n    <opc:Field Name="Value" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataTypeSchemaHeader" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfNamespaces" TypeName="opc:Int32" />\r\n    <opc:Field Name="Namespaces" TypeName="opc:String" LengthField="NoOfNamespaces" />\r\n    <opc:Field Name="NoOfStructureDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="StructureDataTypes" TypeName="tns:StructureDescription" LengthField="NoOfStructureDataTypes" />\r\n    <opc:Field Name="NoOfEnumDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="EnumDataTypes" TypeName="tns:EnumDescription" LengthField="NoOfEnumDataTypes" />\r\n    <opc:Field Name="NoOfSimpleDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="SimpleDataTypes" TypeName="tns:SimpleTypeDescription" LengthField="NoOfSimpleDataTypes" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataTypeDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="DataTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="Name" TypeName="ua:QualifiedName" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="StructureDescription" BaseType="tns:DataTypeDescription">\r\n    <opc:Field Name="DataTypeId" TypeName="ua:NodeId" SourceType="tns:DataTypeDescription" />\r\n    <opc:Field Name="Name" TypeName="ua:QualifiedName" SourceType="tns:DataTypeDescription" />\r\n    <opc:Field Name="StructureDefinition" TypeName="tns:StructureDefinition" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EnumDescription" BaseType="tns:DataTypeDescription">\r\n    <opc:Field Name="DataTypeId" TypeName="ua:NodeId" SourceType="tns:DataTypeDescription" />\r\n    <opc:Field Name="Name" TypeName="ua:QualifiedName" SourceType="tns:DataTypeDescription" />\r\n    <opc:Field Name="EnumDefinition" TypeName="tns:EnumDefinition" />\r\n    <opc:Field Name="BuiltInType" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SimpleTypeDescription" BaseType="tns:DataTypeDescription">\r\n    <opc:Field Name="DataTypeId" TypeName="ua:NodeId" SourceType="tns:DataTypeDescription" />\r\n    <opc:Field Name="Name" TypeName="ua:QualifiedName" SourceType="tns:DataTypeDescription" />\r\n    <opc:Field Name="BaseDataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="BuiltInType" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UABinaryFileDataType" BaseType="tns:DataTypeSchemaHeader">\r\n    <opc:Field Name="NoOfNamespaces" TypeName="opc:Int32" />\r\n    <opc:Field Name="Namespaces" TypeName="opc:String" LengthField="NoOfNamespaces" />\r\n    <opc:Field Name="NoOfStructureDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="StructureDataTypes" TypeName="tns:StructureDescription" LengthField="NoOfStructureDataTypes" />\r\n    <opc:Field Name="NoOfEnumDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="EnumDataTypes" TypeName="tns:EnumDescription" LengthField="NoOfEnumDataTypes" />\r\n    <opc:Field Name="NoOfSimpleDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="SimpleDataTypes" TypeName="tns:SimpleTypeDescription" LengthField="NoOfSimpleDataTypes" />\r\n    <opc:Field Name="SchemaLocation" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfFileHeader" TypeName="opc:Int32" />\r\n    <opc:Field Name="FileHeader" TypeName="tns:KeyValuePair" LengthField="NoOfFileHeader" />\r\n    <opc:Field Name="Body" TypeName="ua:Variant" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="PubSubState" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Disabled" Value="0" />\r\n    <opc:EnumeratedValue Name="Paused" Value="1" />\r\n    <opc:EnumeratedValue Name="Operational" Value="2" />\r\n    <opc:EnumeratedValue Name="Error" Value="3" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="DataSetMetaDataType" BaseType="tns:DataTypeSchemaHeader">\r\n    <opc:Field Name="NoOfNamespaces" TypeName="opc:Int32" />\r\n    <opc:Field Name="Namespaces" TypeName="opc:String" LengthField="NoOfNamespaces" />\r\n    <opc:Field Name="NoOfStructureDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="StructureDataTypes" TypeName="tns:StructureDescription" LengthField="NoOfStructureDataTypes" />\r\n    <opc:Field Name="NoOfEnumDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="EnumDataTypes" TypeName="tns:EnumDescription" LengthField="NoOfEnumDataTypes" />\r\n    <opc:Field Name="NoOfSimpleDataTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="SimpleDataTypes" TypeName="tns:SimpleTypeDescription" LengthField="NoOfSimpleDataTypes" />\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="NoOfFields" TypeName="opc:Int32" />\r\n    <opc:Field Name="Fields" TypeName="tns:FieldMetaData" LengthField="NoOfFields" />\r\n    <opc:Field Name="DataSetClassId" TypeName="opc:Guid" />\r\n    <opc:Field Name="ConfigurationVersion" TypeName="tns:ConfigurationVersionDataType" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FieldMetaData" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="FieldFlags" TypeName="tns:DataSetFieldFlags" />\r\n    <opc:Field Name="BuiltInType" TypeName="opc:Byte" />\r\n    <opc:Field Name="DataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ValueRank" TypeName="opc:Int32" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:UInt32" LengthField="NoOfArrayDimensions" />\r\n    <opc:Field Name="MaxStringLength" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DataSetFieldId" TypeName="opc:Guid" />\r\n    <opc:Field Name="NoOfProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="Properties" TypeName="tns:KeyValuePair" LengthField="NoOfProperties" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="DataSetFieldFlags" LengthInBits="16" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="PromotedField" Value="1" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ConfigurationVersionDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="MajorVersion" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MinorVersion" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PublishedDataSetDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfDataSetFolder" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataSetFolder" TypeName="opc:String" LengthField="NoOfDataSetFolder" />\r\n    <opc:Field Name="DataSetMetaData" TypeName="tns:DataSetMetaDataType" />\r\n    <opc:Field Name="NoOfExtensionFields" TypeName="opc:Int32" />\r\n    <opc:Field Name="ExtensionFields" TypeName="tns:KeyValuePair" LengthField="NoOfExtensionFields" />\r\n    <opc:Field Name="DataSetSource" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PublishedDataSetSourceDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PublishedVariableDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="PublishedVariable" TypeName="ua:NodeId" />\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SamplingIntervalHint" TypeName="opc:Double" />\r\n    <opc:Field Name="DeadbandType" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DeadbandValue" TypeName="opc:Double" />\r\n    <opc:Field Name="IndexRange" TypeName="opc:String" />\r\n    <opc:Field Name="SubstituteValue" TypeName="ua:Variant" />\r\n    <opc:Field Name="NoOfMetaDataProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="MetaDataProperties" TypeName="ua:QualifiedName" LengthField="NoOfMetaDataProperties" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PublishedDataItemsDataType" BaseType="tns:PublishedDataSetSourceDataType">\r\n    <opc:Field Name="NoOfPublishedData" TypeName="opc:Int32" />\r\n    <opc:Field Name="PublishedData" TypeName="tns:PublishedVariableDataType" LengthField="NoOfPublishedData" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PublishedEventsDataType" BaseType="tns:PublishedDataSetSourceDataType">\r\n    <opc:Field Name="EventNotifier" TypeName="ua:NodeId" />\r\n    <opc:Field Name="NoOfSelectedFields" TypeName="opc:Int32" />\r\n    <opc:Field Name="SelectedFields" TypeName="tns:SimpleAttributeOperand" LengthField="NoOfSelectedFields" />\r\n    <opc:Field Name="Filter" TypeName="tns:ContentFilter" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="DataSetFieldContentMask" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="StatusCode" Value="1" />\r\n    <opc:EnumeratedValue Name="SourceTimestamp" Value="2" />\r\n    <opc:EnumeratedValue Name="ServerTimestamp" Value="4" />\r\n    <opc:EnumeratedValue Name="SourcePicoSeconds" Value="8" />\r\n    <opc:EnumeratedValue Name="ServerPicoSeconds" Value="16" />\r\n    <opc:EnumeratedValue Name="RawData" Value="32" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="DataSetWriterDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="Enabled" TypeName="opc:Boolean" />\r\n    <opc:Field Name="DataSetWriterId" TypeName="opc:UInt16" />\r\n    <opc:Field Name="DataSetFieldContentMask" TypeName="tns:DataSetFieldContentMask" />\r\n    <opc:Field Name="KeyFrameCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DataSetName" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfDataSetWriterProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataSetWriterProperties" TypeName="tns:KeyValuePair" LengthField="NoOfDataSetWriterProperties" />\r\n    <opc:Field Name="TransportSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="MessageSettings" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataSetWriterTransportDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataSetWriterMessageDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PubSubGroupDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="Enabled" TypeName="opc:Boolean" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" />\r\n    <opc:Field Name="SecurityGroupId" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfSecurityKeyServices" TypeName="opc:Int32" />\r\n    <opc:Field Name="SecurityKeyServices" TypeName="tns:EndpointDescription" LengthField="NoOfSecurityKeyServices" />\r\n    <opc:Field Name="MaxNetworkMessageSize" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfGroupProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="GroupProperties" TypeName="tns:KeyValuePair" LengthField="NoOfGroupProperties" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="WriterGroupDataType" BaseType="tns:PubSubGroupDataType">\r\n    <opc:Field Name="Name" TypeName="opc:String" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="Enabled" TypeName="opc:Boolean" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="SecurityGroupId" TypeName="opc:String" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="NoOfSecurityKeyServices" TypeName="opc:Int32" />\r\n    <opc:Field Name="SecurityKeyServices" TypeName="tns:EndpointDescription" LengthField="NoOfSecurityKeyServices" />\r\n    <opc:Field Name="MaxNetworkMessageSize" TypeName="opc:UInt32" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="NoOfGroupProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="GroupProperties" TypeName="tns:KeyValuePair" LengthField="NoOfGroupProperties" />\r\n    <opc:Field Name="WriterGroupId" TypeName="opc:UInt16" />\r\n    <opc:Field Name="PublishingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="KeepAliveTime" TypeName="opc:Double" />\r\n    <opc:Field Name="Priority" TypeName="opc:Byte" />\r\n    <opc:Field Name="NoOfLocaleIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="LocaleIds" TypeName="opc:String" LengthField="NoOfLocaleIds" />\r\n    <opc:Field Name="HeaderLayoutUri" TypeName="opc:String" />\r\n    <opc:Field Name="TransportSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="MessageSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="NoOfDataSetWriters" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataSetWriters" TypeName="tns:DataSetWriterDataType" LengthField="NoOfDataSetWriters" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="WriterGroupTransportDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="WriterGroupMessageDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PubSubConnectionDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="Enabled" TypeName="opc:Boolean" />\r\n    <opc:Field Name="PublisherId" TypeName="ua:Variant" />\r\n    <opc:Field Name="TransportProfileUri" TypeName="opc:String" />\r\n    <opc:Field Name="Address" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="NoOfConnectionProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="ConnectionProperties" TypeName="tns:KeyValuePair" LengthField="NoOfConnectionProperties" />\r\n    <opc:Field Name="TransportSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="NoOfWriterGroups" TypeName="opc:Int32" />\r\n    <opc:Field Name="WriterGroups" TypeName="tns:WriterGroupDataType" LengthField="NoOfWriterGroups" />\r\n    <opc:Field Name="NoOfReaderGroups" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReaderGroups" TypeName="tns:ReaderGroupDataType" LengthField="NoOfReaderGroups" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ConnectionTransportDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NetworkAddressDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NetworkInterface" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NetworkAddressUrlDataType" BaseType="tns:NetworkAddressDataType">\r\n    <opc:Field Name="NetworkInterface" TypeName="opc:String" SourceType="tns:NetworkAddressDataType" />\r\n    <opc:Field Name="Url" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReaderGroupDataType" BaseType="tns:PubSubGroupDataType">\r\n    <opc:Field Name="Name" TypeName="opc:String" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="Enabled" TypeName="opc:Boolean" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="SecurityGroupId" TypeName="opc:String" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="NoOfSecurityKeyServices" TypeName="opc:Int32" />\r\n    <opc:Field Name="SecurityKeyServices" TypeName="tns:EndpointDescription" LengthField="NoOfSecurityKeyServices" />\r\n    <opc:Field Name="MaxNetworkMessageSize" TypeName="opc:UInt32" SourceType="tns:PubSubGroupDataType" />\r\n    <opc:Field Name="NoOfGroupProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="GroupProperties" TypeName="tns:KeyValuePair" LengthField="NoOfGroupProperties" />\r\n    <opc:Field Name="TransportSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="MessageSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="NoOfDataSetReaders" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataSetReaders" TypeName="tns:DataSetReaderDataType" LengthField="NoOfDataSetReaders" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReaderGroupTransportDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReaderGroupMessageDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataSetReaderDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="Enabled" TypeName="opc:Boolean" />\r\n    <opc:Field Name="PublisherId" TypeName="ua:Variant" />\r\n    <opc:Field Name="WriterGroupId" TypeName="opc:UInt16" />\r\n    <opc:Field Name="DataSetWriterId" TypeName="opc:UInt16" />\r\n    <opc:Field Name="DataSetMetaData" TypeName="tns:DataSetMetaDataType" />\r\n    <opc:Field Name="DataSetFieldContentMask" TypeName="tns:DataSetFieldContentMask" />\r\n    <opc:Field Name="MessageReceiveTimeout" TypeName="opc:Double" />\r\n    <opc:Field Name="KeyFrameCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="HeaderLayoutUri" TypeName="opc:String" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" />\r\n    <opc:Field Name="SecurityGroupId" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfSecurityKeyServices" TypeName="opc:Int32" />\r\n    <opc:Field Name="SecurityKeyServices" TypeName="tns:EndpointDescription" LengthField="NoOfSecurityKeyServices" />\r\n    <opc:Field Name="NoOfDataSetReaderProperties" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataSetReaderProperties" TypeName="tns:KeyValuePair" LengthField="NoOfDataSetReaderProperties" />\r\n    <opc:Field Name="TransportSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="MessageSettings" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="SubscribedDataSet" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataSetReaderTransportDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataSetReaderMessageDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SubscribedDataSetDataType" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="TargetVariablesDataType" BaseType="tns:SubscribedDataSetDataType">\r\n    <opc:Field Name="NoOfTargetVariables" TypeName="opc:Int32" />\r\n    <opc:Field Name="TargetVariables" TypeName="tns:FieldTargetDataType" LengthField="NoOfTargetVariables" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FieldTargetDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="DataSetFieldId" TypeName="opc:Guid" />\r\n    <opc:Field Name="ReceiverIndexRange" TypeName="opc:String" />\r\n    <opc:Field Name="TargetNodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="WriteIndexRange" TypeName="opc:String" />\r\n    <opc:Field Name="OverrideValueHandling" TypeName="tns:OverrideValueHandling" />\r\n    <opc:Field Name="OverrideValue" TypeName="ua:Variant" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="OverrideValueHandling" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Disabled" Value="0" />\r\n    <opc:EnumeratedValue Name="LastUsableValue" Value="1" />\r\n    <opc:EnumeratedValue Name="OverrideValue" Value="2" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="SubscribedDataSetMirrorDataType" BaseType="tns:SubscribedDataSetDataType">\r\n    <opc:Field Name="ParentNodeName" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PubSubConfigurationDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfPublishedDataSets" TypeName="opc:Int32" />\r\n    <opc:Field Name="PublishedDataSets" TypeName="tns:PublishedDataSetDataType" LengthField="NoOfPublishedDataSets" />\r\n    <opc:Field Name="NoOfConnections" TypeName="opc:Int32" />\r\n    <opc:Field Name="Connections" TypeName="tns:PubSubConnectionDataType" LengthField="NoOfConnections" />\r\n    <opc:Field Name="Enabled" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="DataSetOrderingType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Undefined" Value="0" />\r\n    <opc:EnumeratedValue Name="AscendingWriterId" Value="1" />\r\n    <opc:EnumeratedValue Name="AscendingWriterIdSingle" Value="2" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="UadpNetworkMessageContentMask" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="PublisherId" Value="1" />\r\n    <opc:EnumeratedValue Name="GroupHeader" Value="2" />\r\n    <opc:EnumeratedValue Name="WriterGroupId" Value="4" />\r\n    <opc:EnumeratedValue Name="GroupVersion" Value="8" />\r\n    <opc:EnumeratedValue Name="NetworkMessageNumber" Value="16" />\r\n    <opc:EnumeratedValue Name="SequenceNumber" Value="32" />\r\n    <opc:EnumeratedValue Name="PayloadHeader" Value="64" />\r\n    <opc:EnumeratedValue Name="Timestamp" Value="128" />\r\n    <opc:EnumeratedValue Name="PicoSeconds" Value="256" />\r\n    <opc:EnumeratedValue Name="DataSetClassId" Value="512" />\r\n    <opc:EnumeratedValue Name="PromotedFields" Value="1024" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="UadpWriterGroupMessageDataType" BaseType="tns:WriterGroupMessageDataType">\r\n    <opc:Field Name="GroupVersion" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DataSetOrdering" TypeName="tns:DataSetOrderingType" />\r\n    <opc:Field Name="NetworkMessageContentMask" TypeName="tns:UadpNetworkMessageContentMask" />\r\n    <opc:Field Name="SamplingOffset" TypeName="opc:Double" />\r\n    <opc:Field Name="NoOfPublishingOffset" TypeName="opc:Int32" />\r\n    <opc:Field Name="PublishingOffset" TypeName="opc:Double" LengthField="NoOfPublishingOffset" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="UadpDataSetMessageContentMask" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="Timestamp" Value="1" />\r\n    <opc:EnumeratedValue Name="PicoSeconds" Value="2" />\r\n    <opc:EnumeratedValue Name="Status" Value="4" />\r\n    <opc:EnumeratedValue Name="MajorVersion" Value="8" />\r\n    <opc:EnumeratedValue Name="MinorVersion" Value="16" />\r\n    <opc:EnumeratedValue Name="SequenceNumber" Value="32" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="UadpDataSetWriterMessageDataType" BaseType="tns:DataSetWriterMessageDataType">\r\n    <opc:Field Name="DataSetMessageContentMask" TypeName="tns:UadpDataSetMessageContentMask" />\r\n    <opc:Field Name="ConfiguredSize" TypeName="opc:UInt16" />\r\n    <opc:Field Name="NetworkMessageNumber" TypeName="opc:UInt16" />\r\n    <opc:Field Name="DataSetOffset" TypeName="opc:UInt16" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UadpDataSetReaderMessageDataType" BaseType="tns:DataSetReaderMessageDataType">\r\n    <opc:Field Name="GroupVersion" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NetworkMessageNumber" TypeName="opc:UInt16" />\r\n    <opc:Field Name="DataSetOffset" TypeName="opc:UInt16" />\r\n    <opc:Field Name="DataSetClassId" TypeName="opc:Guid" />\r\n    <opc:Field Name="NetworkMessageContentMask" TypeName="tns:UadpNetworkMessageContentMask" />\r\n    <opc:Field Name="DataSetMessageContentMask" TypeName="tns:UadpDataSetMessageContentMask" />\r\n    <opc:Field Name="PublishingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="ReceiveOffset" TypeName="opc:Double" />\r\n    <opc:Field Name="ProcessingOffset" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="JsonNetworkMessageContentMask" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="NetworkMessageHeader" Value="1" />\r\n    <opc:EnumeratedValue Name="DataSetMessageHeader" Value="2" />\r\n    <opc:EnumeratedValue Name="SingleDataSetMessage" Value="4" />\r\n    <opc:EnumeratedValue Name="PublisherId" Value="8" />\r\n    <opc:EnumeratedValue Name="DataSetClassId" Value="16" />\r\n    <opc:EnumeratedValue Name="ReplyTo" Value="32" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="JsonWriterGroupMessageDataType" BaseType="tns:WriterGroupMessageDataType">\r\n    <opc:Field Name="NetworkMessageContentMask" TypeName="tns:JsonNetworkMessageContentMask" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="JsonDataSetMessageContentMask" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="DataSetWriterId" Value="1" />\r\n    <opc:EnumeratedValue Name="MetaDataVersion" Value="2" />\r\n    <opc:EnumeratedValue Name="SequenceNumber" Value="4" />\r\n    <opc:EnumeratedValue Name="Timestamp" Value="8" />\r\n    <opc:EnumeratedValue Name="Status" Value="16" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="JsonDataSetWriterMessageDataType" BaseType="tns:DataSetWriterMessageDataType">\r\n    <opc:Field Name="DataSetMessageContentMask" TypeName="tns:JsonDataSetMessageContentMask" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="JsonDataSetReaderMessageDataType" BaseType="tns:DataSetReaderMessageDataType">\r\n    <opc:Field Name="NetworkMessageContentMask" TypeName="tns:JsonNetworkMessageContentMask" />\r\n    <opc:Field Name="DataSetMessageContentMask" TypeName="tns:JsonDataSetMessageContentMask" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DatagramConnectionTransportDataType" BaseType="tns:ConnectionTransportDataType">\r\n    <opc:Field Name="DiscoveryAddress" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DatagramWriterGroupTransportDataType" BaseType="tns:WriterGroupTransportDataType">\r\n    <opc:Field Name="MessageRepeatCount" TypeName="opc:Byte" />\r\n    <opc:Field Name="MessageRepeatDelay" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrokerConnectionTransportDataType" BaseType="tns:ConnectionTransportDataType">\r\n    <opc:Field Name="ResourceUri" TypeName="opc:String" />\r\n    <opc:Field Name="AuthenticationProfileUri" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="BrokerTransportQualityOfService" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="NotSpecified" Value="0" />\r\n    <opc:EnumeratedValue Name="BestEffort" Value="1" />\r\n    <opc:EnumeratedValue Name="AtLeastOnce" Value="2" />\r\n    <opc:EnumeratedValue Name="AtMostOnce" Value="3" />\r\n    <opc:EnumeratedValue Name="ExactlyOnce" Value="4" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="BrokerWriterGroupTransportDataType" BaseType="tns:WriterGroupTransportDataType">\r\n    <opc:Field Name="QueueName" TypeName="opc:String" />\r\n    <opc:Field Name="ResourceUri" TypeName="opc:String" />\r\n    <opc:Field Name="AuthenticationProfileUri" TypeName="opc:String" />\r\n    <opc:Field Name="RequestedDeliveryGuarantee" TypeName="tns:BrokerTransportQualityOfService" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrokerDataSetWriterTransportDataType" BaseType="tns:DataSetWriterTransportDataType">\r\n    <opc:Field Name="QueueName" TypeName="opc:String" />\r\n    <opc:Field Name="ResourceUri" TypeName="opc:String" />\r\n    <opc:Field Name="AuthenticationProfileUri" TypeName="opc:String" />\r\n    <opc:Field Name="RequestedDeliveryGuarantee" TypeName="tns:BrokerTransportQualityOfService" />\r\n    <opc:Field Name="MetaDataQueueName" TypeName="opc:String" />\r\n    <opc:Field Name="MetaDataUpdateTime" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrokerDataSetReaderTransportDataType" BaseType="tns:DataSetReaderTransportDataType">\r\n    <opc:Field Name="QueueName" TypeName="opc:String" />\r\n    <opc:Field Name="ResourceUri" TypeName="opc:String" />\r\n    <opc:Field Name="AuthenticationProfileUri" TypeName="opc:String" />\r\n    <opc:Field Name="RequestedDeliveryGuarantee" TypeName="tns:BrokerTransportQualityOfService" />\r\n    <opc:Field Name="MetaDataQueueName" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="DiagnosticsLevel" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Basic" Value="0" />\r\n    <opc:EnumeratedValue Name="Advanced" Value="1" />\r\n    <opc:EnumeratedValue Name="Info" Value="2" />\r\n    <opc:EnumeratedValue Name="Log" Value="3" />\r\n    <opc:EnumeratedValue Name="Debug" Value="4" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="PubSubDiagnosticsCounterClassification" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Information" Value="0" />\r\n    <opc:EnumeratedValue Name="Error" Value="1" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="AliasNameDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="AliasName" TypeName="ua:QualifiedName" />\r\n    <opc:Field Name="NoOfReferencedNodes" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReferencedNodes" TypeName="ua:ExpandedNodeId" LengthField="NoOfReferencedNodes" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="IdType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Numeric" Value="0" />\r\n    <opc:EnumeratedValue Name="String" Value="1" />\r\n    <opc:EnumeratedValue Name="Guid" Value="2" />\r\n    <opc:EnumeratedValue Name="Opaque" Value="3" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="NodeClass" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Unspecified" Value="0" />\r\n    <opc:EnumeratedValue Name="Object" Value="1" />\r\n    <opc:EnumeratedValue Name="Variable" Value="2" />\r\n    <opc:EnumeratedValue Name="Method" Value="4" />\r\n    <opc:EnumeratedValue Name="ObjectType" Value="8" />\r\n    <opc:EnumeratedValue Name="VariableType" Value="16" />\r\n    <opc:EnumeratedValue Name="ReferenceType" Value="32" />\r\n    <opc:EnumeratedValue Name="DataType" Value="64" />\r\n    <opc:EnumeratedValue Name="View" Value="128" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="PermissionType" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="Browse" Value="1" />\r\n    <opc:EnumeratedValue Name="ReadRolePermissions" Value="2" />\r\n    <opc:EnumeratedValue Name="WriteAttribute" Value="4" />\r\n    <opc:EnumeratedValue Name="WriteRolePermissions" Value="8" />\r\n    <opc:EnumeratedValue Name="WriteHistorizing" Value="16" />\r\n    <opc:EnumeratedValue Name="Read" Value="32" />\r\n    <opc:EnumeratedValue Name="Write" Value="64" />\r\n    <opc:EnumeratedValue Name="ReadHistory" Value="128" />\r\n    <opc:EnumeratedValue Name="InsertHistory" Value="256" />\r\n    <opc:EnumeratedValue Name="ModifyHistory" Value="512" />\r\n    <opc:EnumeratedValue Name="DeleteHistory" Value="1024" />\r\n    <opc:EnumeratedValue Name="ReceiveEvents" Value="2048" />\r\n    <opc:EnumeratedValue Name="Call" Value="4096" />\r\n    <opc:EnumeratedValue Name="AddReference" Value="8192" />\r\n    <opc:EnumeratedValue Name="RemoveReference" Value="16384" />\r\n    <opc:EnumeratedValue Name="DeleteNode" Value="32768" />\r\n    <opc:EnumeratedValue Name="AddNode" Value="65536" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="AccessLevelType" LengthInBits="8" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="CurrentRead" Value="1" />\r\n    <opc:EnumeratedValue Name="CurrentWrite" Value="2" />\r\n    <opc:EnumeratedValue Name="HistoryRead" Value="4" />\r\n    <opc:EnumeratedValue Name="HistoryWrite" Value="8" />\r\n    <opc:EnumeratedValue Name="SemanticChange" Value="16" />\r\n    <opc:EnumeratedValue Name="StatusWrite" Value="32" />\r\n    <opc:EnumeratedValue Name="TimestampWrite" Value="64" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="AccessLevelExType" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="CurrentRead" Value="1" />\r\n    <opc:EnumeratedValue Name="CurrentWrite" Value="2" />\r\n    <opc:EnumeratedValue Name="HistoryRead" Value="4" />\r\n    <opc:EnumeratedValue Name="HistoryWrite" Value="8" />\r\n    <opc:EnumeratedValue Name="SemanticChange" Value="16" />\r\n    <opc:EnumeratedValue Name="StatusWrite" Value="32" />\r\n    <opc:EnumeratedValue Name="TimestampWrite" Value="64" />\r\n    <opc:EnumeratedValue Name="NonatomicRead" Value="256" />\r\n    <opc:EnumeratedValue Name="NonatomicWrite" Value="512" />\r\n    <opc:EnumeratedValue Name="WriteFullArrayOnly" Value="1024" />\r\n    <opc:EnumeratedValue Name="NoSubDataTypes" Value="2048" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="EventNotifierType" LengthInBits="8" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="SubscribeToEvents" Value="1" />\r\n    <opc:EnumeratedValue Name="HistoryRead" Value="4" />\r\n    <opc:EnumeratedValue Name="HistoryWrite" Value="8" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="AccessRestrictionType" LengthInBits="16" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="SigningRequired" Value="1" />\r\n    <opc:EnumeratedValue Name="EncryptionRequired" Value="2" />\r\n    <opc:EnumeratedValue Name="SessionRequired" Value="4" />\r\n    <opc:EnumeratedValue Name="ApplyRestrictionsToBrowse" Value="8" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="RolePermissionType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RoleId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="Permissions" TypeName="tns:PermissionType" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="StructureType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Structure" Value="0" />\r\n    <opc:EnumeratedValue Name="StructureWithOptionalFields" Value="1" />\r\n    <opc:EnumeratedValue Name="Union" Value="2" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="StructureField" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="DataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ValueRank" TypeName="opc:Int32" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:UInt32" LengthField="NoOfArrayDimensions" />\r\n    <opc:Field Name="MaxStringLength" TypeName="opc:UInt32" />\r\n    <opc:Field Name="IsOptional" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="StructureDefinition" BaseType="tns:DataTypeDefinition">\r\n    <opc:Field Name="DefaultEncodingId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="BaseDataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="StructureType" TypeName="tns:StructureType" />\r\n    <opc:Field Name="NoOfFields" TypeName="opc:Int32" />\r\n    <opc:Field Name="Fields" TypeName="tns:StructureField" LengthField="NoOfFields" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EnumDefinition" BaseType="tns:DataTypeDefinition">\r\n    <opc:Field Name="NoOfFields" TypeName="opc:Int32" />\r\n    <opc:Field Name="Fields" TypeName="tns:EnumField" LengthField="NoOfFields" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Node" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="InstanceNode" BaseType="tns:Node">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="TypeNode" BaseType="tns:Node">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ObjectNode" BaseType="tns:InstanceNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="EventNotifier" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ObjectTypeNode" BaseType="tns:TypeNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="VariableNode" BaseType="tns:InstanceNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="Value" TypeName="ua:Variant" />\r\n    <opc:Field Name="DataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ValueRank" TypeName="opc:Int32" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:UInt32" LengthField="NoOfArrayDimensions" />\r\n    <opc:Field Name="AccessLevel" TypeName="opc:Byte" />\r\n    <opc:Field Name="UserAccessLevel" TypeName="opc:Byte" />\r\n    <opc:Field Name="MinimumSamplingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="Historizing" TypeName="opc:Boolean" />\r\n    <opc:Field Name="AccessLevelEx" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="VariableTypeNode" BaseType="tns:TypeNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="Value" TypeName="ua:Variant" />\r\n    <opc:Field Name="DataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ValueRank" TypeName="opc:Int32" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:UInt32" LengthField="NoOfArrayDimensions" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReferenceTypeNode" BaseType="tns:TypeNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n    <opc:Field Name="Symmetric" TypeName="opc:Boolean" />\r\n    <opc:Field Name="InverseName" TypeName="ua:LocalizedText" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MethodNode" BaseType="tns:InstanceNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="Executable" TypeName="opc:Boolean" />\r\n    <opc:Field Name="UserExecutable" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ViewNode" BaseType="tns:InstanceNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="ContainsNoLoops" TypeName="opc:Boolean" />\r\n    <opc:Field Name="EventNotifier" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataTypeNode" BaseType="tns:TypeNode">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:Node" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" SourceType="tns:Node" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" SourceType="tns:Node" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:Node" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="RolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfRolePermissions" />\r\n    <opc:Field Name="NoOfUserRolePermissions" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserRolePermissions" TypeName="tns:RolePermissionType" LengthField="NoOfUserRolePermissions" />\r\n    <opc:Field Name="AccessRestrictions" TypeName="opc:UInt16" SourceType="tns:Node" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceNode" LengthField="NoOfReferences" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n    <opc:Field Name="DataTypeDefinition" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReferenceNode" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IsInverse" TypeName="opc:Boolean" />\r\n    <opc:Field Name="TargetId" TypeName="ua:ExpandedNodeId" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Argument" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n    <opc:Field Name="DataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ValueRank" TypeName="opc:Int32" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:UInt32" LengthField="NoOfArrayDimensions" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EnumValueType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Value" TypeName="opc:Int64" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EnumField" BaseType="tns:EnumValueType">\r\n    <opc:Field Name="Value" TypeName="opc:Int64" SourceType="tns:EnumValueType" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:EnumValueType" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:EnumValueType" />\r\n    <opc:Field Name="Name" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="OptionSet" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Value" TypeName="opc:ByteString" />\r\n    <opc:Field Name="ValidBits" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Union" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="NormalizedString">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="DecimalString">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="DurationString">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="TimeString">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="DateString">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="Duration">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="UtcTime">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="LocaleId">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="TimeZoneDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Offset" TypeName="opc:Int16" />\r\n    <opc:Field Name="DaylightSavingInOffset" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="Index">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="IntegerId">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:EnumeratedType Name="ApplicationType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Server" Value="0" />\r\n    <opc:EnumeratedValue Name="Client" Value="1" />\r\n    <opc:EnumeratedValue Name="ClientAndServer" Value="2" />\r\n    <opc:EnumeratedValue Name="DiscoveryServer" Value="3" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ApplicationDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ApplicationUri" TypeName="opc:String" />\r\n    <opc:Field Name="ProductUri" TypeName="opc:String" />\r\n    <opc:Field Name="ApplicationName" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="ApplicationType" TypeName="tns:ApplicationType" />\r\n    <opc:Field Name="GatewayServerUri" TypeName="opc:String" />\r\n    <opc:Field Name="DiscoveryProfileUri" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfDiscoveryUrls" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiscoveryUrls" TypeName="opc:String" LengthField="NoOfDiscoveryUrls" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RequestHeader" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="AuthenticationToken" TypeName="ua:NodeId" />\r\n    <opc:Field Name="Timestamp" TypeName="opc:DateTime" />\r\n    <opc:Field Name="RequestHandle" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ReturnDiagnostics" TypeName="opc:UInt32" />\r\n    <opc:Field Name="AuditEntryId" TypeName="opc:String" />\r\n    <opc:Field Name="TimeoutHint" TypeName="opc:UInt32" />\r\n    <opc:Field Name="AdditionalHeader" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ResponseHeader" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Timestamp" TypeName="opc:DateTime" />\r\n    <opc:Field Name="RequestHandle" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ServiceResult" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="ServiceDiagnostics" TypeName="ua:DiagnosticInfo" />\r\n    <opc:Field Name="NoOfStringTable" TypeName="opc:Int32" />\r\n    <opc:Field Name="StringTable" TypeName="opc:String" LengthField="NoOfStringTable" />\r\n    <opc:Field Name="AdditionalHeader" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="VersionTime">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="ServiceFault" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SessionlessInvokeRequestType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="UrisVersion" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfNamespaceUris" TypeName="opc:Int32" />\r\n    <opc:Field Name="NamespaceUris" TypeName="opc:String" LengthField="NoOfNamespaceUris" />\r\n    <opc:Field Name="NoOfServerUris" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerUris" TypeName="opc:String" LengthField="NoOfServerUris" />\r\n    <opc:Field Name="NoOfLocaleIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="LocaleIds" TypeName="opc:String" LengthField="NoOfLocaleIds" />\r\n    <opc:Field Name="ServiceId" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SessionlessInvokeResponseType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfNamespaceUris" TypeName="opc:Int32" />\r\n    <opc:Field Name="NamespaceUris" TypeName="opc:String" LengthField="NoOfNamespaceUris" />\r\n    <opc:Field Name="NoOfServerUris" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerUris" TypeName="opc:String" LengthField="NoOfServerUris" />\r\n    <opc:Field Name="ServiceId" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FindServersRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="EndpointUrl" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfLocaleIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="LocaleIds" TypeName="opc:String" LengthField="NoOfLocaleIds" />\r\n    <opc:Field Name="NoOfServerUris" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerUris" TypeName="opc:String" LengthField="NoOfServerUris" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FindServersResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfServers" TypeName="opc:Int32" />\r\n    <opc:Field Name="Servers" TypeName="tns:ApplicationDescription" LengthField="NoOfServers" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ServerOnNetwork" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RecordId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ServerName" TypeName="opc:String" />\r\n    <opc:Field Name="DiscoveryUrl" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfServerCapabilities" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerCapabilities" TypeName="opc:String" LengthField="NoOfServerCapabilities" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FindServersOnNetworkRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="StartingRecordId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MaxRecordsToReturn" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfServerCapabilityFilter" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerCapabilityFilter" TypeName="opc:String" LengthField="NoOfServerCapabilityFilter" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FindServersOnNetworkResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="LastCounterResetTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="NoOfServers" TypeName="opc:Int32" />\r\n    <opc:Field Name="Servers" TypeName="tns:ServerOnNetwork" LengthField="NoOfServers" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="ApplicationInstanceCertificate">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:EnumeratedType Name="MessageSecurityMode" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Invalid" Value="0" />\r\n    <opc:EnumeratedValue Name="None" Value="1" />\r\n    <opc:EnumeratedValue Name="Sign" Value="2" />\r\n    <opc:EnumeratedValue Name="SignAndEncrypt" Value="3" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="UserTokenType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Anonymous" Value="0" />\r\n    <opc:EnumeratedValue Name="UserName" Value="1" />\r\n    <opc:EnumeratedValue Name="Certificate" Value="2" />\r\n    <opc:EnumeratedValue Name="IssuedToken" Value="3" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="UserTokenPolicy" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="PolicyId" TypeName="opc:String" />\r\n    <opc:Field Name="TokenType" TypeName="tns:UserTokenType" />\r\n    <opc:Field Name="IssuedTokenType" TypeName="opc:String" />\r\n    <opc:Field Name="IssuerEndpointUrl" TypeName="opc:String" />\r\n    <opc:Field Name="SecurityPolicyUri" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EndpointDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="EndpointUrl" TypeName="opc:String" />\r\n    <opc:Field Name="Server" TypeName="tns:ApplicationDescription" />\r\n    <opc:Field Name="ServerCertificate" TypeName="opc:ByteString" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" />\r\n    <opc:Field Name="SecurityPolicyUri" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfUserIdentityTokens" TypeName="opc:Int32" />\r\n    <opc:Field Name="UserIdentityTokens" TypeName="tns:UserTokenPolicy" LengthField="NoOfUserIdentityTokens" />\r\n    <opc:Field Name="TransportProfileUri" TypeName="opc:String" />\r\n    <opc:Field Name="SecurityLevel" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="GetEndpointsRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="EndpointUrl" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfLocaleIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="LocaleIds" TypeName="opc:String" LengthField="NoOfLocaleIds" />\r\n    <opc:Field Name="NoOfProfileUris" TypeName="opc:Int32" />\r\n    <opc:Field Name="ProfileUris" TypeName="opc:String" LengthField="NoOfProfileUris" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="GetEndpointsResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfEndpoints" TypeName="opc:Int32" />\r\n    <opc:Field Name="Endpoints" TypeName="tns:EndpointDescription" LengthField="NoOfEndpoints" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RegisteredServer" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ServerUri" TypeName="opc:String" />\r\n    <opc:Field Name="ProductUri" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfServerNames" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerNames" TypeName="ua:LocalizedText" LengthField="NoOfServerNames" />\r\n    <opc:Field Name="ServerType" TypeName="tns:ApplicationType" />\r\n    <opc:Field Name="GatewayServerUri" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfDiscoveryUrls" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiscoveryUrls" TypeName="opc:String" LengthField="NoOfDiscoveryUrls" />\r\n    <opc:Field Name="SemaphoreFilePath" TypeName="opc:String" />\r\n    <opc:Field Name="IsOnline" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RegisterServerRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="Server" TypeName="tns:RegisteredServer" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RegisterServerResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DiscoveryConfiguration" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MdnsDiscoveryConfiguration" BaseType="tns:DiscoveryConfiguration">\r\n    <opc:Field Name="MdnsServerName" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfServerCapabilities" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerCapabilities" TypeName="opc:String" LengthField="NoOfServerCapabilities" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RegisterServer2Request" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="Server" TypeName="tns:RegisteredServer" />\r\n    <opc:Field Name="NoOfDiscoveryConfiguration" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiscoveryConfiguration" TypeName="ua:ExtensionObject" LengthField="NoOfDiscoveryConfiguration" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RegisterServer2Response" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfConfigurationResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="ConfigurationResults" TypeName="ua:StatusCode" LengthField="NoOfConfigurationResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="SecurityTokenRequestType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Issue" Value="0" />\r\n    <opc:EnumeratedValue Name="Renew" Value="1" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ChannelSecurityToken" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ChannelId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TokenId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CreatedAt" TypeName="opc:DateTime" />\r\n    <opc:Field Name="RevisedLifetime" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="OpenSecureChannelRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="ClientProtocolVersion" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RequestType" TypeName="tns:SecurityTokenRequestType" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" />\r\n    <opc:Field Name="ClientNonce" TypeName="opc:ByteString" />\r\n    <opc:Field Name="RequestedLifetime" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="OpenSecureChannelResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="ServerProtocolVersion" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SecurityToken" TypeName="tns:ChannelSecurityToken" />\r\n    <opc:Field Name="ServerNonce" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CloseSecureChannelRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CloseSecureChannelResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SignedSoftwareCertificate" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="CertificateData" TypeName="opc:ByteString" />\r\n    <opc:Field Name="Signature" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="SessionAuthenticationToken">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="SignatureData" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Algorithm" TypeName="opc:String" />\r\n    <opc:Field Name="Signature" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CreateSessionRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="ClientDescription" TypeName="tns:ApplicationDescription" />\r\n    <opc:Field Name="ServerUri" TypeName="opc:String" />\r\n    <opc:Field Name="EndpointUrl" TypeName="opc:String" />\r\n    <opc:Field Name="SessionName" TypeName="opc:String" />\r\n    <opc:Field Name="ClientNonce" TypeName="opc:ByteString" />\r\n    <opc:Field Name="ClientCertificate" TypeName="opc:ByteString" />\r\n    <opc:Field Name="RequestedSessionTimeout" TypeName="opc:Double" />\r\n    <opc:Field Name="MaxResponseMessageSize" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CreateSessionResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="SessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="AuthenticationToken" TypeName="ua:NodeId" />\r\n    <opc:Field Name="RevisedSessionTimeout" TypeName="opc:Double" />\r\n    <opc:Field Name="ServerNonce" TypeName="opc:ByteString" />\r\n    <opc:Field Name="ServerCertificate" TypeName="opc:ByteString" />\r\n    <opc:Field Name="NoOfServerEndpoints" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerEndpoints" TypeName="tns:EndpointDescription" LengthField="NoOfServerEndpoints" />\r\n    <opc:Field Name="NoOfServerSoftwareCertificates" TypeName="opc:Int32" />\r\n    <opc:Field Name="ServerSoftwareCertificates" TypeName="tns:SignedSoftwareCertificate" LengthField="NoOfServerSoftwareCertificates" />\r\n    <opc:Field Name="ServerSignature" TypeName="tns:SignatureData" />\r\n    <opc:Field Name="MaxRequestMessageSize" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UserIdentityToken" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="PolicyId" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AnonymousIdentityToken" BaseType="tns:UserIdentityToken">\r\n    <opc:Field Name="PolicyId" TypeName="opc:String" SourceType="tns:UserIdentityToken" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UserNameIdentityToken" BaseType="tns:UserIdentityToken">\r\n    <opc:Field Name="PolicyId" TypeName="opc:String" SourceType="tns:UserIdentityToken" />\r\n    <opc:Field Name="UserName" TypeName="opc:String" />\r\n    <opc:Field Name="Password" TypeName="opc:ByteString" />\r\n    <opc:Field Name="EncryptionAlgorithm" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="X509IdentityToken" BaseType="tns:UserIdentityToken">\r\n    <opc:Field Name="PolicyId" TypeName="opc:String" SourceType="tns:UserIdentityToken" />\r\n    <opc:Field Name="CertificateData" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="IssuedIdentityToken" BaseType="tns:UserIdentityToken">\r\n    <opc:Field Name="PolicyId" TypeName="opc:String" SourceType="tns:UserIdentityToken" />\r\n    <opc:Field Name="TokenData" TypeName="opc:ByteString" />\r\n    <opc:Field Name="EncryptionAlgorithm" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="RsaEncryptedSecret">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="EccEncryptedSecret">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="ActivateSessionRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="ClientSignature" TypeName="tns:SignatureData" />\r\n    <opc:Field Name="NoOfClientSoftwareCertificates" TypeName="opc:Int32" />\r\n    <opc:Field Name="ClientSoftwareCertificates" TypeName="tns:SignedSoftwareCertificate" LengthField="NoOfClientSoftwareCertificates" />\r\n    <opc:Field Name="NoOfLocaleIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="LocaleIds" TypeName="opc:String" LengthField="NoOfLocaleIds" />\r\n    <opc:Field Name="UserIdentityToken" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="UserTokenSignature" TypeName="tns:SignatureData" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ActivateSessionResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="ServerNonce" TypeName="opc:ByteString" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CloseSessionRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="DeleteSubscriptions" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CloseSessionResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CancelRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="RequestHandle" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CancelResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="CancelCount" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="NodeAttributesMask" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="AccessLevel" Value="1" />\r\n    <opc:EnumeratedValue Name="ArrayDimensions" Value="2" />\r\n    <opc:EnumeratedValue Name="BrowseName" Value="4" />\r\n    <opc:EnumeratedValue Name="ContainsNoLoops" Value="8" />\r\n    <opc:EnumeratedValue Name="DataType" Value="16" />\r\n    <opc:EnumeratedValue Name="Description" Value="32" />\r\n    <opc:EnumeratedValue Name="DisplayName" Value="64" />\r\n    <opc:EnumeratedValue Name="EventNotifier" Value="128" />\r\n    <opc:EnumeratedValue Name="Executable" Value="256" />\r\n    <opc:EnumeratedValue Name="Historizing" Value="512" />\r\n    <opc:EnumeratedValue Name="InverseName" Value="1024" />\r\n    <opc:EnumeratedValue Name="IsAbstract" Value="2048" />\r\n    <opc:EnumeratedValue Name="MinimumSamplingInterval" Value="4096" />\r\n    <opc:EnumeratedValue Name="NodeClass" Value="8192" />\r\n    <opc:EnumeratedValue Name="NodeId" Value="16384" />\r\n    <opc:EnumeratedValue Name="Symmetric" Value="32768" />\r\n    <opc:EnumeratedValue Name="UserAccessLevel" Value="65536" />\r\n    <opc:EnumeratedValue Name="UserExecutable" Value="131072" />\r\n    <opc:EnumeratedValue Name="UserWriteMask" Value="262144" />\r\n    <opc:EnumeratedValue Name="ValueRank" Value="524288" />\r\n    <opc:EnumeratedValue Name="WriteMask" Value="1048576" />\r\n    <opc:EnumeratedValue Name="Value" Value="2097152" />\r\n    <opc:EnumeratedValue Name="DataTypeDefinition" Value="4194304" />\r\n    <opc:EnumeratedValue Name="RolePermissions" Value="8388608" />\r\n    <opc:EnumeratedValue Name="AccessRestrictions" Value="16777216" />\r\n    <opc:EnumeratedValue Name="All" Value="33554431" />\r\n    <opc:EnumeratedValue Name="BaseNode" Value="26501220" />\r\n    <opc:EnumeratedValue Name="Object" Value="26501348" />\r\n    <opc:EnumeratedValue Name="ObjectType" Value="26503268" />\r\n    <opc:EnumeratedValue Name="Variable" Value="26571383" />\r\n    <opc:EnumeratedValue Name="VariableType" Value="28600438" />\r\n    <opc:EnumeratedValue Name="Method" Value="26632548" />\r\n    <opc:EnumeratedValue Name="ReferenceType" Value="26537060" />\r\n    <opc:EnumeratedValue Name="View" Value="26501356" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="NodeAttributes" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ObjectAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="EventNotifier" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="VariableAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Value" TypeName="ua:Variant" />\r\n    <opc:Field Name="DataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ValueRank" TypeName="opc:Int32" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:UInt32" LengthField="NoOfArrayDimensions" />\r\n    <opc:Field Name="AccessLevel" TypeName="opc:Byte" />\r\n    <opc:Field Name="UserAccessLevel" TypeName="opc:Byte" />\r\n    <opc:Field Name="MinimumSamplingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="Historizing" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MethodAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Executable" TypeName="opc:Boolean" />\r\n    <opc:Field Name="UserExecutable" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ObjectTypeAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="VariableTypeAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Value" TypeName="ua:Variant" />\r\n    <opc:Field Name="DataType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ValueRank" TypeName="opc:Int32" />\r\n    <opc:Field Name="NoOfArrayDimensions" TypeName="opc:Int32" />\r\n    <opc:Field Name="ArrayDimensions" TypeName="opc:UInt32" LengthField="NoOfArrayDimensions" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReferenceTypeAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n    <opc:Field Name="Symmetric" TypeName="opc:Boolean" />\r\n    <opc:Field Name="InverseName" TypeName="ua:LocalizedText" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataTypeAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="IsAbstract" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ViewAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="ContainsNoLoops" TypeName="opc:Boolean" />\r\n    <opc:Field Name="EventNotifier" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="GenericAttributeValue" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="Value" TypeName="ua:Variant" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="GenericAttributes" BaseType="tns:NodeAttributes">\r\n    <opc:Field Name="SpecifiedAttributes" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="WriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="UserWriteMask" TypeName="opc:UInt32" SourceType="tns:NodeAttributes" />\r\n    <opc:Field Name="NoOfAttributeValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="AttributeValues" TypeName="tns:GenericAttributeValue" LengthField="NoOfAttributeValues" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AddNodesItem" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ParentNodeId" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="RequestedNewNodeId" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" />\r\n    <opc:Field Name="NodeAttributes" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="TypeDefinition" TypeName="ua:ExpandedNodeId" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AddNodesResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="AddedNodeId" TypeName="ua:NodeId" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AddNodesRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfNodesToAdd" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToAdd" TypeName="tns:AddNodesItem" LengthField="NoOfNodesToAdd" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AddNodesResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:AddNodesResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AddReferencesItem" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SourceNodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IsForward" TypeName="opc:Boolean" />\r\n    <opc:Field Name="TargetServerUri" TypeName="opc:String" />\r\n    <opc:Field Name="TargetNodeId" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="TargetNodeClass" TypeName="tns:NodeClass" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AddReferencesRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfReferencesToAdd" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReferencesToAdd" TypeName="tns:AddReferencesItem" LengthField="NoOfReferencesToAdd" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AddReferencesResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteNodesItem" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="DeleteTargetReferences" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteNodesRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfNodesToDelete" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToDelete" TypeName="tns:DeleteNodesItem" LengthField="NoOfNodesToDelete" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteNodesResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteReferencesItem" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SourceNodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IsForward" TypeName="opc:Boolean" />\r\n    <opc:Field Name="TargetNodeId" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="DeleteBidirectional" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteReferencesRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfReferencesToDelete" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReferencesToDelete" TypeName="tns:DeleteReferencesItem" LengthField="NoOfReferencesToDelete" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteReferencesResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="AttributeWriteMask" LengthInBits="32" IsOptionSet="true">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="AccessLevel" Value="1" />\r\n    <opc:EnumeratedValue Name="ArrayDimensions" Value="2" />\r\n    <opc:EnumeratedValue Name="BrowseName" Value="4" />\r\n    <opc:EnumeratedValue Name="ContainsNoLoops" Value="8" />\r\n    <opc:EnumeratedValue Name="DataType" Value="16" />\r\n    <opc:EnumeratedValue Name="Description" Value="32" />\r\n    <opc:EnumeratedValue Name="DisplayName" Value="64" />\r\n    <opc:EnumeratedValue Name="EventNotifier" Value="128" />\r\n    <opc:EnumeratedValue Name="Executable" Value="256" />\r\n    <opc:EnumeratedValue Name="Historizing" Value="512" />\r\n    <opc:EnumeratedValue Name="InverseName" Value="1024" />\r\n    <opc:EnumeratedValue Name="IsAbstract" Value="2048" />\r\n    <opc:EnumeratedValue Name="MinimumSamplingInterval" Value="4096" />\r\n    <opc:EnumeratedValue Name="NodeClass" Value="8192" />\r\n    <opc:EnumeratedValue Name="NodeId" Value="16384" />\r\n    <opc:EnumeratedValue Name="Symmetric" Value="32768" />\r\n    <opc:EnumeratedValue Name="UserAccessLevel" Value="65536" />\r\n    <opc:EnumeratedValue Name="UserExecutable" Value="131072" />\r\n    <opc:EnumeratedValue Name="UserWriteMask" Value="262144" />\r\n    <opc:EnumeratedValue Name="ValueRank" Value="524288" />\r\n    <opc:EnumeratedValue Name="WriteMask" Value="1048576" />\r\n    <opc:EnumeratedValue Name="ValueForVariableType" Value="2097152" />\r\n    <opc:EnumeratedValue Name="DataTypeDefinition" Value="4194304" />\r\n    <opc:EnumeratedValue Name="RolePermissions" Value="8388608" />\r\n    <opc:EnumeratedValue Name="AccessRestrictions" Value="16777216" />\r\n    <opc:EnumeratedValue Name="AccessLevelEx" Value="33554432" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="BrowseDirection" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Forward" Value="0" />\r\n    <opc:EnumeratedValue Name="Inverse" Value="1" />\r\n    <opc:EnumeratedValue Name="Both" Value="2" />\r\n    <opc:EnumeratedValue Name="Invalid" Value="3" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ViewDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ViewId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="Timestamp" TypeName="opc:DateTime" />\r\n    <opc:Field Name="ViewVersion" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowseDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="BrowseDirection" TypeName="tns:BrowseDirection" />\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IncludeSubtypes" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NodeClassMask" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ResultMask" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="BrowseResultMask" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="ReferenceTypeId" Value="1" />\r\n    <opc:EnumeratedValue Name="IsForward" Value="2" />\r\n    <opc:EnumeratedValue Name="NodeClass" Value="4" />\r\n    <opc:EnumeratedValue Name="BrowseName" Value="8" />\r\n    <opc:EnumeratedValue Name="DisplayName" Value="16" />\r\n    <opc:EnumeratedValue Name="TypeDefinition" Value="32" />\r\n    <opc:EnumeratedValue Name="All" Value="63" />\r\n    <opc:EnumeratedValue Name="ReferenceTypeInfo" Value="3" />\r\n    <opc:EnumeratedValue Name="TargetInfo" Value="60" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ReferenceDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IsForward" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NodeId" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="BrowseName" TypeName="ua:QualifiedName" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="NodeClass" TypeName="tns:NodeClass" />\r\n    <opc:Field Name="TypeDefinition" TypeName="ua:ExpandedNodeId" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="ContinuationPoint">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="BrowseResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="ContinuationPoint" TypeName="opc:ByteString" />\r\n    <opc:Field Name="NoOfReferences" TypeName="opc:Int32" />\r\n    <opc:Field Name="References" TypeName="tns:ReferenceDescription" LengthField="NoOfReferences" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowseRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="View" TypeName="tns:ViewDescription" />\r\n    <opc:Field Name="RequestedMaxReferencesPerNode" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfNodesToBrowse" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToBrowse" TypeName="tns:BrowseDescription" LengthField="NoOfNodesToBrowse" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowseResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:BrowseResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowseNextRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="ReleaseContinuationPoints" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NoOfContinuationPoints" TypeName="opc:Int32" />\r\n    <opc:Field Name="ContinuationPoints" TypeName="opc:ByteString" LengthField="NoOfContinuationPoints" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowseNextResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:BrowseResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RelativePathElement" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IsInverse" TypeName="opc:Boolean" />\r\n    <opc:Field Name="IncludeSubtypes" TypeName="opc:Boolean" />\r\n    <opc:Field Name="TargetName" TypeName="ua:QualifiedName" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RelativePath" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfElements" TypeName="opc:Int32" />\r\n    <opc:Field Name="Elements" TypeName="tns:RelativePathElement" LengthField="NoOfElements" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowsePath" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StartingNode" TypeName="ua:NodeId" />\r\n    <opc:Field Name="RelativePath" TypeName="tns:RelativePath" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowsePathTarget" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="TargetId" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="RemainingPathIndex" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BrowsePathResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="NoOfTargets" TypeName="opc:Int32" />\r\n    <opc:Field Name="Targets" TypeName="tns:BrowsePathTarget" LengthField="NoOfTargets" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="TranslateBrowsePathsToNodeIdsRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfBrowsePaths" TypeName="opc:Int32" />\r\n    <opc:Field Name="BrowsePaths" TypeName="tns:BrowsePath" LengthField="NoOfBrowsePaths" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="TranslateBrowsePathsToNodeIdsResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:BrowsePathResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RegisterNodesRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfNodesToRegister" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToRegister" TypeName="ua:NodeId" LengthField="NoOfNodesToRegister" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RegisterNodesResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfRegisteredNodeIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="RegisteredNodeIds" TypeName="ua:NodeId" LengthField="NoOfRegisteredNodeIds" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UnregisterNodesRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfNodesToUnregister" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToUnregister" TypeName="ua:NodeId" LengthField="NoOfNodesToUnregister" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UnregisterNodesResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:OpaqueType Name="Counter">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:OpaqueType Name="NumericRange">\r\n  </opc:OpaqueType>\r\n\r\n  <opc:StructuredType Name="EndpointConfiguration" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="OperationTimeout" TypeName="opc:Int32" />\r\n    <opc:Field Name="UseBinaryEncoding" TypeName="opc:Boolean" />\r\n    <opc:Field Name="MaxStringLength" TypeName="opc:Int32" />\r\n    <opc:Field Name="MaxByteStringLength" TypeName="opc:Int32" />\r\n    <opc:Field Name="MaxArrayLength" TypeName="opc:Int32" />\r\n    <opc:Field Name="MaxMessageSize" TypeName="opc:Int32" />\r\n    <opc:Field Name="MaxBufferSize" TypeName="opc:Int32" />\r\n    <opc:Field Name="ChannelLifetime" TypeName="opc:Int32" />\r\n    <opc:Field Name="SecurityTokenLifetime" TypeName="opc:Int32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="QueryDataDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RelativePath" TypeName="tns:RelativePath" />\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="IndexRange" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NodeTypeDescription" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="TypeDefinitionNode" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="IncludeSubTypes" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NoOfDataToReturn" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataToReturn" TypeName="tns:QueryDataDescription" LengthField="NoOfDataToReturn" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="FilterOperator" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Equals" Value="0" />\r\n    <opc:EnumeratedValue Name="IsNull" Value="1" />\r\n    <opc:EnumeratedValue Name="GreaterThan" Value="2" />\r\n    <opc:EnumeratedValue Name="LessThan" Value="3" />\r\n    <opc:EnumeratedValue Name="GreaterThanOrEqual" Value="4" />\r\n    <opc:EnumeratedValue Name="LessThanOrEqual" Value="5" />\r\n    <opc:EnumeratedValue Name="Like" Value="6" />\r\n    <opc:EnumeratedValue Name="Not" Value="7" />\r\n    <opc:EnumeratedValue Name="Between" Value="8" />\r\n    <opc:EnumeratedValue Name="InList" Value="9" />\r\n    <opc:EnumeratedValue Name="And" Value="10" />\r\n    <opc:EnumeratedValue Name="Or" Value="11" />\r\n    <opc:EnumeratedValue Name="Cast" Value="12" />\r\n    <opc:EnumeratedValue Name="InView" Value="13" />\r\n    <opc:EnumeratedValue Name="OfType" Value="14" />\r\n    <opc:EnumeratedValue Name="RelatedTo" Value="15" />\r\n    <opc:EnumeratedValue Name="BitwiseAnd" Value="16" />\r\n    <opc:EnumeratedValue Name="BitwiseOr" Value="17" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="QueryDataSet" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="TypeDefinitionNode" TypeName="ua:ExpandedNodeId" />\r\n    <opc:Field Name="NoOfValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="Values" TypeName="ua:Variant" LengthField="NoOfValues" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NodeReference" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ReferenceTypeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IsForward" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NoOfReferencedNodeIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReferencedNodeIds" TypeName="ua:NodeId" LengthField="NoOfReferencedNodeIds" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ContentFilterElement" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="FilterOperator" TypeName="tns:FilterOperator" />\r\n    <opc:Field Name="NoOfFilterOperands" TypeName="opc:Int32" />\r\n    <opc:Field Name="FilterOperands" TypeName="ua:ExtensionObject" LengthField="NoOfFilterOperands" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ContentFilter" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfElements" TypeName="opc:Int32" />\r\n    <opc:Field Name="Elements" TypeName="tns:ContentFilterElement" LengthField="NoOfElements" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="FilterOperand" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ElementOperand" BaseType="tns:FilterOperand">\r\n    <opc:Field Name="Index" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="LiteralOperand" BaseType="tns:FilterOperand">\r\n    <opc:Field Name="Value" TypeName="ua:Variant" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AttributeOperand" BaseType="tns:FilterOperand">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="Alias" TypeName="opc:String" />\r\n    <opc:Field Name="BrowsePath" TypeName="tns:RelativePath" />\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="IndexRange" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SimpleAttributeOperand" BaseType="tns:FilterOperand">\r\n    <opc:Field Name="TypeDefinitionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="NoOfBrowsePath" TypeName="opc:Int32" />\r\n    <opc:Field Name="BrowsePath" TypeName="ua:QualifiedName" LengthField="NoOfBrowsePath" />\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="IndexRange" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ContentFilterElementResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="NoOfOperandStatusCodes" TypeName="opc:Int32" />\r\n    <opc:Field Name="OperandStatusCodes" TypeName="ua:StatusCode" LengthField="NoOfOperandStatusCodes" />\r\n    <opc:Field Name="NoOfOperandDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="OperandDiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfOperandDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ContentFilterResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfElementResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="ElementResults" TypeName="tns:ContentFilterElementResult" LengthField="NoOfElementResults" />\r\n    <opc:Field Name="NoOfElementDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="ElementDiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfElementDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ParsingResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="NoOfDataStatusCodes" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataStatusCodes" TypeName="ua:StatusCode" LengthField="NoOfDataStatusCodes" />\r\n    <opc:Field Name="NoOfDataDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataDiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDataDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="QueryFirstRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="View" TypeName="tns:ViewDescription" />\r\n    <opc:Field Name="NoOfNodeTypes" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodeTypes" TypeName="tns:NodeTypeDescription" LengthField="NoOfNodeTypes" />\r\n    <opc:Field Name="Filter" TypeName="tns:ContentFilter" />\r\n    <opc:Field Name="MaxDataSetsToReturn" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MaxReferencesToReturn" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="QueryFirstResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfQueryDataSets" TypeName="opc:Int32" />\r\n    <opc:Field Name="QueryDataSets" TypeName="tns:QueryDataSet" LengthField="NoOfQueryDataSets" />\r\n    <opc:Field Name="ContinuationPoint" TypeName="opc:ByteString" />\r\n    <opc:Field Name="NoOfParsingResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="ParsingResults" TypeName="tns:ParsingResult" LengthField="NoOfParsingResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n    <opc:Field Name="FilterResult" TypeName="tns:ContentFilterResult" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="QueryNextRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="ReleaseContinuationPoint" TypeName="opc:Boolean" />\r\n    <opc:Field Name="ContinuationPoint" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="QueryNextResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfQueryDataSets" TypeName="opc:Int32" />\r\n    <opc:Field Name="QueryDataSets" TypeName="tns:QueryDataSet" LengthField="NoOfQueryDataSets" />\r\n    <opc:Field Name="RevisedContinuationPoint" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="TimestampsToReturn" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Source" Value="0" />\r\n    <opc:EnumeratedValue Name="Server" Value="1" />\r\n    <opc:EnumeratedValue Name="Both" Value="2" />\r\n    <opc:EnumeratedValue Name="Neither" Value="3" />\r\n    <opc:EnumeratedValue Name="Invalid" Value="4" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ReadValueId" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="IndexRange" TypeName="opc:String" />\r\n    <opc:Field Name="DataEncoding" TypeName="ua:QualifiedName" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReadRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="MaxAge" TypeName="opc:Double" />\r\n    <opc:Field Name="TimestampsToReturn" TypeName="tns:TimestampsToReturn" />\r\n    <opc:Field Name="NoOfNodesToRead" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToRead" TypeName="tns:ReadValueId" LengthField="NoOfNodesToRead" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReadResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:DataValue" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryReadValueId" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="IndexRange" TypeName="opc:String" />\r\n    <opc:Field Name="DataEncoding" TypeName="ua:QualifiedName" />\r\n    <opc:Field Name="ContinuationPoint" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryReadResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="ContinuationPoint" TypeName="opc:ByteString" />\r\n    <opc:Field Name="HistoryData" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryReadDetails" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReadEventDetails" BaseType="tns:HistoryReadDetails">\r\n    <opc:Field Name="NumValuesPerNode" TypeName="opc:UInt32" />\r\n    <opc:Field Name="StartTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="EndTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="Filter" TypeName="tns:EventFilter" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReadRawModifiedDetails" BaseType="tns:HistoryReadDetails">\r\n    <opc:Field Name="IsReadModified" TypeName="opc:Boolean" />\r\n    <opc:Field Name="StartTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="EndTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="NumValuesPerNode" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ReturnBounds" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReadProcessedDetails" BaseType="tns:HistoryReadDetails">\r\n    <opc:Field Name="StartTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="EndTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="ProcessingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="NoOfAggregateType" TypeName="opc:Int32" />\r\n    <opc:Field Name="AggregateType" TypeName="ua:NodeId" LengthField="NoOfAggregateType" />\r\n    <opc:Field Name="AggregateConfiguration" TypeName="tns:AggregateConfiguration" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReadAtTimeDetails" BaseType="tns:HistoryReadDetails">\r\n    <opc:Field Name="NoOfReqTimes" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReqTimes" TypeName="opc:DateTime" LengthField="NoOfReqTimes" />\r\n    <opc:Field Name="UseSimpleBounds" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ReadAnnotationDataDetails" BaseType="tns:HistoryReadDetails">\r\n    <opc:Field Name="NoOfReqTimes" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReqTimes" TypeName="opc:DateTime" LengthField="NoOfReqTimes" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryData" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfDataValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataValues" TypeName="ua:DataValue" LengthField="NoOfDataValues" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ModificationInfo" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ModificationTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="UpdateType" TypeName="tns:HistoryUpdateType" />\r\n    <opc:Field Name="UserName" TypeName="opc:String" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryModifiedData" BaseType="tns:HistoryData">\r\n    <opc:Field Name="NoOfDataValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="DataValues" TypeName="ua:DataValue" LengthField="NoOfDataValues" />\r\n    <opc:Field Name="NoOfModificationInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="ModificationInfos" TypeName="tns:ModificationInfo" LengthField="NoOfModificationInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryEvent" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfEvents" TypeName="opc:Int32" />\r\n    <opc:Field Name="Events" TypeName="tns:HistoryEventFieldList" LengthField="NoOfEvents" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryReadRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="HistoryReadDetails" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="TimestampsToReturn" TypeName="tns:TimestampsToReturn" />\r\n    <opc:Field Name="ReleaseContinuationPoints" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NoOfNodesToRead" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToRead" TypeName="tns:HistoryReadValueId" LengthField="NoOfNodesToRead" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryReadResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:HistoryReadResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="WriteValue" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="AttributeId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="IndexRange" TypeName="opc:String" />\r\n    <opc:Field Name="Value" TypeName="ua:DataValue" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="WriteRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfNodesToWrite" TypeName="opc:Int32" />\r\n    <opc:Field Name="NodesToWrite" TypeName="tns:WriteValue" LengthField="NoOfNodesToWrite" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="WriteResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryUpdateDetails" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="HistoryUpdateType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Insert" Value="1" />\r\n    <opc:EnumeratedValue Name="Replace" Value="2" />\r\n    <opc:EnumeratedValue Name="Update" Value="3" />\r\n    <opc:EnumeratedValue Name="Delete" Value="4" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="PerformUpdateType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Insert" Value="1" />\r\n    <opc:EnumeratedValue Name="Replace" Value="2" />\r\n    <opc:EnumeratedValue Name="Update" Value="3" />\r\n    <opc:EnumeratedValue Name="Remove" Value="4" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="UpdateDataDetails" BaseType="tns:HistoryUpdateDetails">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:HistoryUpdateDetails" />\r\n    <opc:Field Name="PerformInsertReplace" TypeName="tns:PerformUpdateType" />\r\n    <opc:Field Name="NoOfUpdateValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="UpdateValues" TypeName="ua:DataValue" LengthField="NoOfUpdateValues" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UpdateStructureDataDetails" BaseType="tns:HistoryUpdateDetails">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:HistoryUpdateDetails" />\r\n    <opc:Field Name="PerformInsertReplace" TypeName="tns:PerformUpdateType" />\r\n    <opc:Field Name="NoOfUpdateValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="UpdateValues" TypeName="ua:DataValue" LengthField="NoOfUpdateValues" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="UpdateEventDetails" BaseType="tns:HistoryUpdateDetails">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:HistoryUpdateDetails" />\r\n    <opc:Field Name="PerformInsertReplace" TypeName="tns:PerformUpdateType" />\r\n    <opc:Field Name="Filter" TypeName="tns:EventFilter" />\r\n    <opc:Field Name="NoOfEventData" TypeName="opc:Int32" />\r\n    <opc:Field Name="EventData" TypeName="tns:HistoryEventFieldList" LengthField="NoOfEventData" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteRawModifiedDetails" BaseType="tns:HistoryUpdateDetails">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:HistoryUpdateDetails" />\r\n    <opc:Field Name="IsDeleteModified" TypeName="opc:Boolean" />\r\n    <opc:Field Name="StartTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="EndTime" TypeName="opc:DateTime" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteAtTimeDetails" BaseType="tns:HistoryUpdateDetails">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:HistoryUpdateDetails" />\r\n    <opc:Field Name="NoOfReqTimes" TypeName="opc:Int32" />\r\n    <opc:Field Name="ReqTimes" TypeName="opc:DateTime" LengthField="NoOfReqTimes" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteEventDetails" BaseType="tns:HistoryUpdateDetails">\r\n    <opc:Field Name="NodeId" TypeName="ua:NodeId" SourceType="tns:HistoryUpdateDetails" />\r\n    <opc:Field Name="NoOfEventIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="EventIds" TypeName="opc:ByteString" LengthField="NoOfEventIds" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryUpdateResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="NoOfOperationResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="OperationResults" TypeName="ua:StatusCode" LengthField="NoOfOperationResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryUpdateRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfHistoryUpdateDetails" TypeName="opc:Int32" />\r\n    <opc:Field Name="HistoryUpdateDetails" TypeName="ua:ExtensionObject" LengthField="NoOfHistoryUpdateDetails" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryUpdateResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:HistoryUpdateResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CallMethodRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ObjectId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="MethodId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="NoOfInputArguments" TypeName="opc:Int32" />\r\n    <opc:Field Name="InputArguments" TypeName="ua:Variant" LengthField="NoOfInputArguments" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CallMethodResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="NoOfInputArgumentResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="InputArgumentResults" TypeName="ua:StatusCode" LengthField="NoOfInputArgumentResults" />\r\n    <opc:Field Name="NoOfInputArgumentDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="InputArgumentDiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfInputArgumentDiagnosticInfos" />\r\n    <opc:Field Name="NoOfOutputArguments" TypeName="opc:Int32" />\r\n    <opc:Field Name="OutputArguments" TypeName="ua:Variant" LengthField="NoOfOutputArguments" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CallRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfMethodsToCall" TypeName="opc:Int32" />\r\n    <opc:Field Name="MethodsToCall" TypeName="tns:CallMethodRequest" LengthField="NoOfMethodsToCall" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CallResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:CallMethodResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="MonitoringMode" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Disabled" Value="0" />\r\n    <opc:EnumeratedValue Name="Sampling" Value="1" />\r\n    <opc:EnumeratedValue Name="Reporting" Value="2" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="DataChangeTrigger" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Status" Value="0" />\r\n    <opc:EnumeratedValue Name="StatusValue" Value="1" />\r\n    <opc:EnumeratedValue Name="StatusValueTimestamp" Value="2" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="DeadbandType" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="Absolute" Value="1" />\r\n    <opc:EnumeratedValue Name="Percent" Value="2" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="MonitoringFilter" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataChangeFilter" BaseType="tns:MonitoringFilter">\r\n    <opc:Field Name="Trigger" TypeName="tns:DataChangeTrigger" />\r\n    <opc:Field Name="DeadbandType" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DeadbandValue" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EventFilter" BaseType="tns:MonitoringFilter">\r\n    <opc:Field Name="NoOfSelectClauses" TypeName="opc:Int32" />\r\n    <opc:Field Name="SelectClauses" TypeName="tns:SimpleAttributeOperand" LengthField="NoOfSelectClauses" />\r\n    <opc:Field Name="WhereClause" TypeName="tns:ContentFilter" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AggregateConfiguration" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="UseServerCapabilitiesDefaults" TypeName="opc:Boolean" />\r\n    <opc:Field Name="TreatUncertainAsBad" TypeName="opc:Boolean" />\r\n    <opc:Field Name="PercentDataBad" TypeName="opc:Byte" />\r\n    <opc:Field Name="PercentDataGood" TypeName="opc:Byte" />\r\n    <opc:Field Name="UseSlopedExtrapolation" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AggregateFilter" BaseType="tns:MonitoringFilter">\r\n    <opc:Field Name="StartTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="AggregateType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ProcessingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="AggregateConfiguration" TypeName="tns:AggregateConfiguration" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MonitoringFilterResult" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EventFilterResult" BaseType="tns:MonitoringFilterResult">\r\n    <opc:Field Name="NoOfSelectClauseResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="SelectClauseResults" TypeName="ua:StatusCode" LengthField="NoOfSelectClauseResults" />\r\n    <opc:Field Name="NoOfSelectClauseDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="SelectClauseDiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfSelectClauseDiagnosticInfos" />\r\n    <opc:Field Name="WhereClauseResult" TypeName="tns:ContentFilterResult" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AggregateFilterResult" BaseType="tns:MonitoringFilterResult">\r\n    <opc:Field Name="RevisedStartTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="RevisedProcessingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="RevisedAggregateConfiguration" TypeName="tns:AggregateConfiguration" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MonitoringParameters" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ClientHandle" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SamplingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="Filter" TypeName="ua:ExtensionObject" />\r\n    <opc:Field Name="QueueSize" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DiscardOldest" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MonitoredItemCreateRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ItemToMonitor" TypeName="tns:ReadValueId" />\r\n    <opc:Field Name="MonitoringMode" TypeName="tns:MonitoringMode" />\r\n    <opc:Field Name="RequestedParameters" TypeName="tns:MonitoringParameters" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MonitoredItemCreateResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="MonitoredItemId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RevisedSamplingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="RevisedQueueSize" TypeName="opc:UInt32" />\r\n    <opc:Field Name="FilterResult" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CreateMonitoredItemsRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TimestampsToReturn" TypeName="tns:TimestampsToReturn" />\r\n    <opc:Field Name="NoOfItemsToCreate" TypeName="opc:Int32" />\r\n    <opc:Field Name="ItemsToCreate" TypeName="tns:MonitoredItemCreateRequest" LengthField="NoOfItemsToCreate" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CreateMonitoredItemsResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:MonitoredItemCreateResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MonitoredItemModifyRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="MonitoredItemId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RequestedParameters" TypeName="tns:MonitoringParameters" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MonitoredItemModifyResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="RevisedSamplingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="RevisedQueueSize" TypeName="opc:UInt32" />\r\n    <opc:Field Name="FilterResult" TypeName="ua:ExtensionObject" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ModifyMonitoredItemsRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TimestampsToReturn" TypeName="tns:TimestampsToReturn" />\r\n    <opc:Field Name="NoOfItemsToModify" TypeName="opc:Int32" />\r\n    <opc:Field Name="ItemsToModify" TypeName="tns:MonitoredItemModifyRequest" LengthField="NoOfItemsToModify" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ModifyMonitoredItemsResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:MonitoredItemModifyResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SetMonitoringModeRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MonitoringMode" TypeName="tns:MonitoringMode" />\r\n    <opc:Field Name="NoOfMonitoredItemIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="MonitoredItemIds" TypeName="opc:UInt32" LengthField="NoOfMonitoredItemIds" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SetMonitoringModeResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SetTriggeringRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TriggeringItemId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfLinksToAdd" TypeName="opc:Int32" />\r\n    <opc:Field Name="LinksToAdd" TypeName="opc:UInt32" LengthField="NoOfLinksToAdd" />\r\n    <opc:Field Name="NoOfLinksToRemove" TypeName="opc:Int32" />\r\n    <opc:Field Name="LinksToRemove" TypeName="opc:UInt32" LengthField="NoOfLinksToRemove" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SetTriggeringResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfAddResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="AddResults" TypeName="ua:StatusCode" LengthField="NoOfAddResults" />\r\n    <opc:Field Name="NoOfAddDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="AddDiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfAddDiagnosticInfos" />\r\n    <opc:Field Name="NoOfRemoveResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="RemoveResults" TypeName="ua:StatusCode" LengthField="NoOfRemoveResults" />\r\n    <opc:Field Name="NoOfRemoveDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="RemoveDiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfRemoveDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteMonitoredItemsRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfMonitoredItemIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="MonitoredItemIds" TypeName="opc:UInt32" LengthField="NoOfMonitoredItemIds" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteMonitoredItemsResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CreateSubscriptionRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="RequestedPublishingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="RequestedLifetimeCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RequestedMaxKeepAliveCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MaxNotificationsPerPublish" TypeName="opc:UInt32" />\r\n    <opc:Field Name="PublishingEnabled" TypeName="opc:Boolean" />\r\n    <opc:Field Name="Priority" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="CreateSubscriptionResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RevisedPublishingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="RevisedLifetimeCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RevisedMaxKeepAliveCount" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ModifySubscriptionRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RequestedPublishingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="RequestedLifetimeCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RequestedMaxKeepAliveCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MaxNotificationsPerPublish" TypeName="opc:UInt32" />\r\n    <opc:Field Name="Priority" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ModifySubscriptionResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="RevisedPublishingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="RevisedLifetimeCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RevisedMaxKeepAliveCount" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SetPublishingModeRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="PublishingEnabled" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NoOfSubscriptionIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="SubscriptionIds" TypeName="opc:UInt32" LengthField="NoOfSubscriptionIds" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SetPublishingModeResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NotificationMessage" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SequenceNumber" TypeName="opc:UInt32" />\r\n    <opc:Field Name="PublishTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="NoOfNotificationData" TypeName="opc:Int32" />\r\n    <opc:Field Name="NotificationData" TypeName="ua:ExtensionObject" LengthField="NoOfNotificationData" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NotificationData" BaseType="ua:ExtensionObject">\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DataChangeNotification" BaseType="tns:NotificationData">\r\n    <opc:Field Name="NoOfMonitoredItems" TypeName="opc:Int32" />\r\n    <opc:Field Name="MonitoredItems" TypeName="tns:MonitoredItemNotification" LengthField="NoOfMonitoredItems" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="MonitoredItemNotification" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ClientHandle" TypeName="opc:UInt32" />\r\n    <opc:Field Name="Value" TypeName="ua:DataValue" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EventNotificationList" BaseType="tns:NotificationData">\r\n    <opc:Field Name="NoOfEvents" TypeName="opc:Int32" />\r\n    <opc:Field Name="Events" TypeName="tns:EventFieldList" LengthField="NoOfEvents" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EventFieldList" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ClientHandle" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfEventFields" TypeName="opc:Int32" />\r\n    <opc:Field Name="EventFields" TypeName="ua:Variant" LengthField="NoOfEventFields" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="HistoryEventFieldList" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfEventFields" TypeName="opc:Int32" />\r\n    <opc:Field Name="EventFields" TypeName="ua:Variant" LengthField="NoOfEventFields" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="StatusChangeNotification" BaseType="tns:NotificationData">\r\n    <opc:Field Name="Status" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="DiagnosticInfo" TypeName="ua:DiagnosticInfo" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SubscriptionAcknowledgement" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SequenceNumber" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PublishRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfSubscriptionAcknowledgements" TypeName="opc:Int32" />\r\n    <opc:Field Name="SubscriptionAcknowledgements" TypeName="tns:SubscriptionAcknowledgement" LengthField="NoOfSubscriptionAcknowledgements" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="PublishResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NoOfAvailableSequenceNumbers" TypeName="opc:Int32" />\r\n    <opc:Field Name="AvailableSequenceNumbers" TypeName="opc:UInt32" LengthField="NoOfAvailableSequenceNumbers" />\r\n    <opc:Field Name="MoreNotifications" TypeName="opc:Boolean" />\r\n    <opc:Field Name="NotificationMessage" TypeName="tns:NotificationMessage" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RepublishRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RetransmitSequenceNumber" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="RepublishResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NotificationMessage" TypeName="tns:NotificationMessage" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="TransferResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="NoOfAvailableSequenceNumbers" TypeName="opc:Int32" />\r\n    <opc:Field Name="AvailableSequenceNumbers" TypeName="opc:UInt32" LengthField="NoOfAvailableSequenceNumbers" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="TransferSubscriptionsRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfSubscriptionIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="SubscriptionIds" TypeName="opc:UInt32" LengthField="NoOfSubscriptionIds" />\r\n    <opc:Field Name="SendInitialValues" TypeName="opc:Boolean" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="TransferSubscriptionsResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="tns:TransferResult" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteSubscriptionsRequest" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="RequestHeader" TypeName="tns:RequestHeader" />\r\n    <opc:Field Name="NoOfSubscriptionIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="SubscriptionIds" TypeName="opc:UInt32" LengthField="NoOfSubscriptionIds" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DeleteSubscriptionsResponse" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ResponseHeader" TypeName="tns:ResponseHeader" />\r\n    <opc:Field Name="NoOfResults" TypeName="opc:Int32" />\r\n    <opc:Field Name="Results" TypeName="ua:StatusCode" LengthField="NoOfResults" />\r\n    <opc:Field Name="NoOfDiagnosticInfos" TypeName="opc:Int32" />\r\n    <opc:Field Name="DiagnosticInfos" TypeName="ua:DiagnosticInfo" LengthField="NoOfDiagnosticInfos" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="BuildInfo" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ProductUri" TypeName="opc:String" />\r\n    <opc:Field Name="ManufacturerName" TypeName="opc:String" />\r\n    <opc:Field Name="ProductName" TypeName="opc:String" />\r\n    <opc:Field Name="SoftwareVersion" TypeName="opc:String" />\r\n    <opc:Field Name="BuildNumber" TypeName="opc:String" />\r\n    <opc:Field Name="BuildDate" TypeName="opc:DateTime" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="RedundancySupport" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="None" Value="0" />\r\n    <opc:EnumeratedValue Name="Cold" Value="1" />\r\n    <opc:EnumeratedValue Name="Warm" Value="2" />\r\n    <opc:EnumeratedValue Name="Hot" Value="3" />\r\n    <opc:EnumeratedValue Name="Transparent" Value="4" />\r\n    <opc:EnumeratedValue Name="HotAndMirrored" Value="5" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:EnumeratedType Name="ServerState" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Running" Value="0" />\r\n    <opc:EnumeratedValue Name="Failed" Value="1" />\r\n    <opc:EnumeratedValue Name="NoConfiguration" Value="2" />\r\n    <opc:EnumeratedValue Name="Suspended" Value="3" />\r\n    <opc:EnumeratedValue Name="Shutdown" Value="4" />\r\n    <opc:EnumeratedValue Name="Test" Value="5" />\r\n    <opc:EnumeratedValue Name="CommunicationFault" Value="6" />\r\n    <opc:EnumeratedValue Name="Unknown" Value="7" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="RedundantServerDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ServerId" TypeName="opc:String" />\r\n    <opc:Field Name="ServiceLevel" TypeName="opc:Byte" />\r\n    <opc:Field Name="ServerState" TypeName="tns:ServerState" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EndpointUrlListDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NoOfEndpointUrlList" TypeName="opc:Int32" />\r\n    <opc:Field Name="EndpointUrlList" TypeName="opc:String" LengthField="NoOfEndpointUrlList" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="NetworkGroupDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ServerUri" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfNetworkPaths" TypeName="opc:Int32" />\r\n    <opc:Field Name="NetworkPaths" TypeName="tns:EndpointUrlListDataType" LengthField="NoOfNetworkPaths" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SamplingIntervalDiagnosticsDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SamplingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="MonitoredItemCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MaxMonitoredItemCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DisabledMonitoredItemCount" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ServerDiagnosticsSummaryDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="ServerViewCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CurrentSessionCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CumulatedSessionCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SecurityRejectedSessionCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RejectedSessionCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SessionTimeoutCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SessionAbortCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CurrentSubscriptionCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CumulatedSubscriptionCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="PublishingIntervalCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="SecurityRejectedRequestsCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RejectedRequestsCount" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ServerStatusDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StartTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="CurrentTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="State" TypeName="tns:ServerState" />\r\n    <opc:Field Name="BuildInfo" TypeName="tns:BuildInfo" />\r\n    <opc:Field Name="SecondsTillShutdown" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ShutdownReason" TypeName="ua:LocalizedText" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SessionDiagnosticsDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="SessionName" TypeName="opc:String" />\r\n    <opc:Field Name="ClientDescription" TypeName="tns:ApplicationDescription" />\r\n    <opc:Field Name="ServerUri" TypeName="opc:String" />\r\n    <opc:Field Name="EndpointUrl" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfLocaleIds" TypeName="opc:Int32" />\r\n    <opc:Field Name="LocaleIds" TypeName="opc:String" LengthField="NoOfLocaleIds" />\r\n    <opc:Field Name="ActualSessionTimeout" TypeName="opc:Double" />\r\n    <opc:Field Name="MaxResponseMessageSize" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ClientConnectionTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="ClientLastContactTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="CurrentSubscriptionsCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CurrentMonitoredItemsCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CurrentPublishRequestsInQueue" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TotalRequestCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="UnauthorizedRequestCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ReadCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="HistoryReadCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="WriteCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="HistoryUpdateCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="CallCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="CreateMonitoredItemsCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="ModifyMonitoredItemsCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="SetMonitoringModeCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="SetTriggeringCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="DeleteMonitoredItemsCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="CreateSubscriptionCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="ModifySubscriptionCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="SetPublishingModeCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="PublishCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="RepublishCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="TransferSubscriptionsCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="DeleteSubscriptionsCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="AddNodesCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="AddReferencesCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="DeleteNodesCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="DeleteReferencesCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="BrowseCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="BrowseNextCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="TranslateBrowsePathsToNodeIdsCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="QueryFirstCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="QueryNextCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="RegisterNodesCount" TypeName="tns:ServiceCounterDataType" />\r\n    <opc:Field Name="UnregisterNodesCount" TypeName="tns:ServiceCounterDataType" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SessionSecurityDiagnosticsDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="ClientUserIdOfSession" TypeName="opc:String" />\r\n    <opc:Field Name="NoOfClientUserIdHistory" TypeName="opc:Int32" />\r\n    <opc:Field Name="ClientUserIdHistory" TypeName="opc:String" LengthField="NoOfClientUserIdHistory" />\r\n    <opc:Field Name="AuthenticationMechanism" TypeName="opc:String" />\r\n    <opc:Field Name="Encoding" TypeName="opc:String" />\r\n    <opc:Field Name="TransportProtocol" TypeName="opc:String" />\r\n    <opc:Field Name="SecurityMode" TypeName="tns:MessageSecurityMode" />\r\n    <opc:Field Name="SecurityPolicyUri" TypeName="opc:String" />\r\n    <opc:Field Name="ClientCertificate" TypeName="opc:ByteString" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ServiceCounterDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="TotalCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="ErrorCount" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="StatusResult" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="StatusCode" TypeName="ua:StatusCode" />\r\n    <opc:Field Name="DiagnosticInfo" TypeName="ua:DiagnosticInfo" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SubscriptionDiagnosticsDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="SessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="SubscriptionId" TypeName="opc:UInt32" />\r\n    <opc:Field Name="Priority" TypeName="opc:Byte" />\r\n    <opc:Field Name="PublishingInterval" TypeName="opc:Double" />\r\n    <opc:Field Name="MaxKeepAliveCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MaxLifetimeCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MaxNotificationsPerPublish" TypeName="opc:UInt32" />\r\n    <opc:Field Name="PublishingEnabled" TypeName="opc:Boolean" />\r\n    <opc:Field Name="ModifyCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="EnableCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DisableCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RepublishRequestCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RepublishMessageRequestCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="RepublishMessageCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TransferRequestCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TransferredToAltClientCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="TransferredToSameClientCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="PublishRequestCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DataChangeNotificationsCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="EventNotificationsCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NotificationsCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="LatePublishRequestCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CurrentKeepAliveCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="CurrentLifetimeCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="UnacknowledgedMessageCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DiscardedMessageCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MonitoredItemCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="DisabledMonitoredItemCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="MonitoringQueueOverflowCount" TypeName="opc:UInt32" />\r\n    <opc:Field Name="NextSequenceNumber" TypeName="opc:UInt32" />\r\n    <opc:Field Name="EventQueueOverFlowCount" TypeName="opc:UInt32" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="ModelChangeStructureVerbMask" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="NodeAdded" Value="1" />\r\n    <opc:EnumeratedValue Name="NodeDeleted" Value="2" />\r\n    <opc:EnumeratedValue Name="ReferenceAdded" Value="4" />\r\n    <opc:EnumeratedValue Name="ReferenceDeleted" Value="8" />\r\n    <opc:EnumeratedValue Name="DataTypeChanged" Value="16" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ModelChangeStructureDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Affected" TypeName="ua:NodeId" />\r\n    <opc:Field Name="AffectedType" TypeName="ua:NodeId" />\r\n    <opc:Field Name="Verb" TypeName="opc:Byte" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="SemanticChangeStructureDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Affected" TypeName="ua:NodeId" />\r\n    <opc:Field Name="AffectedType" TypeName="ua:NodeId" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Range" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Low" TypeName="opc:Double" />\r\n    <opc:Field Name="High" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="EUInformation" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="NamespaceUri" TypeName="opc:String" />\r\n    <opc:Field Name="UnitId" TypeName="opc:Int32" />\r\n    <opc:Field Name="DisplayName" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="Description" TypeName="ua:LocalizedText" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="AxisScaleEnumeration" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="Linear" Value="0" />\r\n    <opc:EnumeratedValue Name="Log" Value="1" />\r\n    <opc:EnumeratedValue Name="Ln" Value="2" />\r\n  </opc:EnumeratedType>\r\n\r\n  <opc:StructuredType Name="ComplexNumberType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Real" TypeName="opc:Float" />\r\n    <opc:Field Name="Imaginary" TypeName="opc:Float" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="DoubleComplexNumberType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Real" TypeName="opc:Double" />\r\n    <opc:Field Name="Imaginary" TypeName="opc:Double" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="AxisInformation" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="EngineeringUnits" TypeName="tns:EUInformation" />\r\n    <opc:Field Name="EURange" TypeName="tns:Range" />\r\n    <opc:Field Name="Title" TypeName="ua:LocalizedText" />\r\n    <opc:Field Name="AxisScaleType" TypeName="tns:AxisScaleEnumeration" />\r\n    <opc:Field Name="NoOfAxisSteps" TypeName="opc:Int32" />\r\n    <opc:Field Name="AxisSteps" TypeName="opc:Double" LengthField="NoOfAxisSteps" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="XVType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="X" TypeName="opc:Double" />\r\n    <opc:Field Name="Value" TypeName="opc:Float" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ProgramDiagnosticDataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="CreateSessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="CreateClientName" TypeName="opc:String" />\r\n    <opc:Field Name="InvocationCreationTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="LastTransitionTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="LastMethodCall" TypeName="opc:String" />\r\n    <opc:Field Name="LastMethodSessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="NoOfLastMethodInputArguments" TypeName="opc:Int32" />\r\n    <opc:Field Name="LastMethodInputArguments" TypeName="tns:Argument" LengthField="NoOfLastMethodInputArguments" />\r\n    <opc:Field Name="NoOfLastMethodOutputArguments" TypeName="opc:Int32" />\r\n    <opc:Field Name="LastMethodOutputArguments" TypeName="tns:Argument" LengthField="NoOfLastMethodOutputArguments" />\r\n    <opc:Field Name="LastMethodCallTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="LastMethodReturnStatus" TypeName="tns:StatusResult" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="ProgramDiagnostic2DataType" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="CreateSessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="CreateClientName" TypeName="opc:String" />\r\n    <opc:Field Name="InvocationCreationTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="LastTransitionTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="LastMethodCall" TypeName="opc:String" />\r\n    <opc:Field Name="LastMethodSessionId" TypeName="ua:NodeId" />\r\n    <opc:Field Name="NoOfLastMethodInputArguments" TypeName="opc:Int32" />\r\n    <opc:Field Name="LastMethodInputArguments" TypeName="tns:Argument" LengthField="NoOfLastMethodInputArguments" />\r\n    <opc:Field Name="NoOfLastMethodOutputArguments" TypeName="opc:Int32" />\r\n    <opc:Field Name="LastMethodOutputArguments" TypeName="tns:Argument" LengthField="NoOfLastMethodOutputArguments" />\r\n    <opc:Field Name="NoOfLastMethodInputValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="LastMethodInputValues" TypeName="ua:Variant" LengthField="NoOfLastMethodInputValues" />\r\n    <opc:Field Name="NoOfLastMethodOutputValues" TypeName="opc:Int32" />\r\n    <opc:Field Name="LastMethodOutputValues" TypeName="ua:Variant" LengthField="NoOfLastMethodOutputValues" />\r\n    <opc:Field Name="LastMethodCallTime" TypeName="opc:DateTime" />\r\n    <opc:Field Name="LastMethodReturnStatus" TypeName="tns:StatusResult" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:StructuredType Name="Annotation" BaseType="ua:ExtensionObject">\r\n    <opc:Field Name="Message" TypeName="opc:String" />\r\n    <opc:Field Name="UserName" TypeName="opc:String" />\r\n    <opc:Field Name="AnnotationTime" TypeName="opc:DateTime" />\r\n  </opc:StructuredType>\r\n\r\n  <opc:EnumeratedType Name="ExceptionDeviationFormat" LengthInBits="32">\r\n    <opc:EnumeratedValue Name="AbsoluteValue" Value="0" />\r\n    <opc:EnumeratedValue Name="PercentOfValue" Value="1" />\r\n    <opc:EnumeratedValue Name="PercentOfRange" Value="2" />\r\n    <opc:EnumeratedValue Name="PercentOfEURange" Value="3" />\r\n    <opc:EnumeratedValue Name="Unknown" Value="4" />\r\n  </opc:EnumeratedType>\r\n\r\n</opc:TypeDictionary>', ua.VariantType.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7617, 0),
        BrowseName=QualifiedName("Opc.Ua", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(93, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(72, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7619, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15037, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14873, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17538, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17550, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15738, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23514, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12681, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15741, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15599, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15602, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15501, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15521, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14876, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15766, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15769, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14324, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15772, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15775, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15778, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15781, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15784, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15787, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21156, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15793, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15854, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15857, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15860, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21159, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21162, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21165, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15866, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15869, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15872, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15877, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15880, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15883, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15886, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21002, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15889, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21168, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15895, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15898, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15919, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15922, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15925, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15931, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17469, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15524, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15940, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15946, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16131, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18181, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18184, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18187, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7656, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14870, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12770, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8914, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12213, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7662, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7668, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7782, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12902, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12905, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7671, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7674, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7677, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7680, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7683, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12718, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12721, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7929, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7932, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7935, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7938, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7941, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7947, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8004, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8067, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8073, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8076, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7692, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8208, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11959, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11962, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8211, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8214, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8220, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8223, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8226, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7659, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8229, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8232, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8235, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8238, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8241, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12183, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12186, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12091, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12094, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8247, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15398, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8244, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(93, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(72, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('http://opcfoundation.org/UA/', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7619, 0),
        BrowseName=QualifiedName("NamespaceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(7619, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Deprecated"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        Value=ua.Variant(True, ua.VariantType.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15037, 0),
        BrowseName=QualifiedName("Deprecated", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeyValuePair"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('KeyValuePair', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14873, 0),
        BrowseName=QualifiedName("KeyValuePair", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AdditionalParametersType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AdditionalParametersType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17538, 0),
        BrowseName=QualifiedName("AdditionalParametersType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17538, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17538, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EphemeralKeyType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EphemeralKeyType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17550, 0),
        BrowseName=QualifiedName("EphemeralKeyType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EndpointType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15734, 0),
        BrowseName=QualifiedName("EndpointType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RationalNumber"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('RationalNumber', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18824, 0),
        BrowseName=QualifiedName("RationalNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18824, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Vector"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('Vector', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18827, 0),
        BrowseName=QualifiedName("Vector", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18827, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DVector"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ThreeDVector', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18830, 0),
        BrowseName=QualifiedName("3DVector", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18830, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CartesianCoordinates"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('CartesianCoordinates', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18833, 0),
        BrowseName=QualifiedName("CartesianCoordinates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18833, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DCartesianCoordinates"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ThreeDCartesianCoordinates', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18836, 0),
        BrowseName=QualifiedName("3DCartesianCoordinates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18836, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Orientation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('Orientation', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18839, 0),
        BrowseName=QualifiedName("Orientation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18839, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DOrientation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ThreeDOrientation', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18842, 0),
        BrowseName=QualifiedName("3DOrientation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18842, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Frame"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('Frame', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18845, 0),
        BrowseName=QualifiedName("Frame", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18845, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DFrame"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ThreeDFrame', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18848, 0),
        BrowseName=QualifiedName("3DFrame", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18848, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IdentityMappingRuleType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('IdentityMappingRuleType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15738, 0),
        BrowseName=QualifiedName("IdentityMappingRuleType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrencyUnitType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('CurrencyUnitType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23514, 0),
        BrowseName=QualifiedName("CurrencyUnitType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23514, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23514, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrustListDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('TrustListDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12681, 0),
        BrowseName=QualifiedName("TrustListDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeSchemaHeader"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataTypeSchemaHeader', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15741, 0),
        BrowseName=QualifiedName("DataTypeSchemaHeader", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataTypeDescription', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14855, 0),
        BrowseName=QualifiedName("DataTypeDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StructureDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('StructureDescription', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15599, 0),
        BrowseName=QualifiedName("StructureDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EnumDescription', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15602, 0),
        BrowseName=QualifiedName("EnumDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SimpleTypeDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SimpleTypeDescription', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15501, 0),
        BrowseName=QualifiedName("SimpleTypeDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15501, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15501, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UABinaryFileDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('UABinaryFileDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15521, 0),
        BrowseName=QualifiedName("UABinaryFileDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15521, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15521, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMetaDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataSetMetaDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14849, 0),
        BrowseName=QualifiedName("DataSetMetaDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FieldMetaData"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('FieldMetaData', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14852, 0),
        BrowseName=QualifiedName("FieldMetaData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfigurationVersionDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ConfigurationVersionDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14876, 0),
        BrowseName=QualifiedName("ConfigurationVersionDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedDataSetDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PublishedDataSetDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15766, 0),
        BrowseName=QualifiedName("PublishedDataSetDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15766, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedDataSetSourceDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PublishedDataSetSourceDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15769, 0),
        BrowseName=QualifiedName("PublishedDataSetSourceDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15769, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedVariableDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PublishedVariableDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14324, 0),
        BrowseName=QualifiedName("PublishedVariableDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedDataItemsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PublishedDataItemsDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15772, 0),
        BrowseName=QualifiedName("PublishedDataItemsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15772, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedEventsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PublishedEventsDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15775, 0),
        BrowseName=QualifiedName("PublishedEventsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15775, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataSetWriterDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15778, 0),
        BrowseName=QualifiedName("DataSetWriterDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15778, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataSetWriterTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15781, 0),
        BrowseName=QualifiedName("DataSetWriterTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15781, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataSetWriterMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15784, 0),
        BrowseName=QualifiedName("DataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15784, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PubSubGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PubSubGroupDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15787, 0),
        BrowseName=QualifiedName("PubSubGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15787, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('WriterGroupDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21156, 0),
        BrowseName=QualifiedName("WriterGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('WriterGroupTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15793, 0),
        BrowseName=QualifiedName("WriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15793, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15793, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('WriterGroupMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15854, 0),
        BrowseName=QualifiedName("WriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PubSubConnectionDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PubSubConnectionDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15857, 0),
        BrowseName=QualifiedName("PubSubConnectionDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConnectionTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ConnectionTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15860, 0),
        BrowseName=QualifiedName("ConnectionTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkAddressDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('NetworkAddressDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21159, 0),
        BrowseName=QualifiedName("NetworkAddressDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkAddressUrlDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('NetworkAddressUrlDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21162, 0),
        BrowseName=QualifiedName("NetworkAddressUrlDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReaderGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ReaderGroupDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21165, 0),
        BrowseName=QualifiedName("ReaderGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReaderGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ReaderGroupTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15866, 0),
        BrowseName=QualifiedName("ReaderGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReaderGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ReaderGroupMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15869, 0),
        BrowseName=QualifiedName("ReaderGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataSetReaderDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15872, 0),
        BrowseName=QualifiedName("DataSetReaderDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataSetReaderTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15877, 0),
        BrowseName=QualifiedName("DataSetReaderTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15877, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15877, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataSetReaderMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15880, 0),
        BrowseName=QualifiedName("DataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15880, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscribedDataSetDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SubscribedDataSetDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15883, 0),
        BrowseName=QualifiedName("SubscribedDataSetDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15883, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TargetVariablesDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('TargetVariablesDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15886, 0),
        BrowseName=QualifiedName("TargetVariablesDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15886, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FieldTargetDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('FieldTargetDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21002, 0),
        BrowseName=QualifiedName("FieldTargetDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21002, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21002, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscribedDataSetMirrorDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SubscribedDataSetMirrorDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15889, 0),
        BrowseName=QualifiedName("SubscribedDataSetMirrorDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15889, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PubSubConfigurationDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('PubSubConfigurationDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21168, 0),
        BrowseName=QualifiedName("PubSubConfigurationDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UadpWriterGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('UadpWriterGroupMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15895, 0),
        BrowseName=QualifiedName("UadpWriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UadpDataSetWriterMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('UadpDataSetWriterMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15898, 0),
        BrowseName=QualifiedName("UadpDataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UadpDataSetReaderMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('UadpDataSetReaderMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15919, 0),
        BrowseName=QualifiedName("UadpDataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15919, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15919, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("JsonWriterGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('JsonWriterGroupMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15922, 0),
        BrowseName=QualifiedName("JsonWriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15922, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("JsonDataSetWriterMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('JsonDataSetWriterMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15925, 0),
        BrowseName=QualifiedName("JsonDataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15925, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15925, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("JsonDataSetReaderMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('JsonDataSetReaderMessageDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15931, 0),
        BrowseName=QualifiedName("JsonDataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15931, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15931, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DatagramConnectionTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DatagramConnectionTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17469, 0),
        BrowseName=QualifiedName("DatagramConnectionTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17469, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17469, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DatagramWriterGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DatagramWriterGroupTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21171, 0),
        BrowseName=QualifiedName("DatagramWriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerConnectionTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('BrokerConnectionTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15524, 0),
        BrowseName=QualifiedName("BrokerConnectionTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerWriterGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('BrokerWriterGroupTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15940, 0),
        BrowseName=QualifiedName("BrokerWriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15940, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerDataSetWriterTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('BrokerDataSetWriterTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15943, 0),
        BrowseName=QualifiedName("BrokerDataSetWriterTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15943, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerDataSetReaderTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('BrokerDataSetReaderTransportDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15946, 0),
        BrowseName=QualifiedName("BrokerDataSetReaderTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15946, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15946, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AliasNameDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AliasNameDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23502, 0),
        BrowseName=QualifiedName("AliasNameDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RolePermissionType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('RolePermissionType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16131, 0),
        BrowseName=QualifiedName("RolePermissionType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeDefinition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DataTypeDefinition', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18178, 0),
        BrowseName=QualifiedName("DataTypeDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StructureField"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('StructureField', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18181, 0),
        BrowseName=QualifiedName("StructureField", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18181, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StructureDefinition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('StructureDefinition', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18184, 0),
        BrowseName=QualifiedName("StructureDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumDefinition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EnumDefinition', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18187, 0),
        BrowseName=QualifiedName("EnumDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Argument"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('Argument', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7650, 0),
        BrowseName=QualifiedName("Argument", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7650, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValueType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EnumValueType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7656, 0),
        BrowseName=QualifiedName("EnumValueType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7656, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumField"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EnumField', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14870, 0),
        BrowseName=QualifiedName("EnumField", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSet"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('OptionSet', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12767, 0),
        BrowseName=QualifiedName("OptionSet", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12767, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Union"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('Union', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12770, 0),
        BrowseName=QualifiedName("Union", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12770, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TimeZoneDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('TimeZoneDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8914, 0),
        BrowseName=QualifiedName("TimeZoneDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8914, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8914, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ApplicationDescription', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7665, 0),
        BrowseName=QualifiedName("ApplicationDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7665, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7665, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerOnNetwork"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ServerOnNetwork', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12213, 0),
        BrowseName=QualifiedName("ServerOnNetwork", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12213, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserTokenPolicy"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('UserTokenPolicy', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7662, 0),
        BrowseName=QualifiedName("UserTokenPolicy", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7662, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7662, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EndpointDescription', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7668, 0),
        BrowseName=QualifiedName("EndpointDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7668, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RegisteredServer"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('RegisteredServer', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7782, 0),
        BrowseName=QualifiedName("RegisteredServer", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7782, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiscoveryConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DiscoveryConfiguration', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12902, 0),
        BrowseName=QualifiedName("DiscoveryConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MdnsDiscoveryConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('MdnsDiscoveryConfiguration', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12905, 0),
        BrowseName=QualifiedName("MdnsDiscoveryConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12905, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SignedSoftwareCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SignedSoftwareCertificate', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7698, 0),
        BrowseName=QualifiedName("SignedSoftwareCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7698, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('UserIdentityToken', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7671, 0),
        BrowseName=QualifiedName("UserIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7671, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AnonymousIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AnonymousIdentityToken', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7674, 0),
        BrowseName=QualifiedName("AnonymousIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7674, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7674, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserNameIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('UserNameIdentityToken', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7677, 0),
        BrowseName=QualifiedName("UserNameIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("X509IdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('X509IdentityToken', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7680, 0),
        BrowseName=QualifiedName("X509IdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7680, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IssuedIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('IssuedIdentityToken', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7683, 0),
        BrowseName=QualifiedName("IssuedIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7683, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7683, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddNodesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AddNodesItem', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7728, 0),
        BrowseName=QualifiedName("AddNodesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddReferencesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AddReferencesItem', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7731, 0),
        BrowseName=QualifiedName("AddReferencesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7731, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteNodesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DeleteNodesItem', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7734, 0),
        BrowseName=QualifiedName("DeleteNodesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7734, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteReferencesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DeleteReferencesItem', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7737, 0),
        BrowseName=QualifiedName("DeleteReferencesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7737, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RelativePathElement"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('RelativePathElement', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12718, 0),
        BrowseName=QualifiedName("RelativePathElement", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12718, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RelativePath"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('RelativePath', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12721, 0),
        BrowseName=QualifiedName("RelativePath", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12721, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EndpointConfiguration', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7686, 0),
        BrowseName=QualifiedName("EndpointConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7686, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ContentFilterElement"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ContentFilterElement', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7929, 0),
        BrowseName=QualifiedName("ContentFilterElement", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7929, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ContentFilter"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ContentFilter', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7932, 0),
        BrowseName=QualifiedName("ContentFilter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7932, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FilterOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('FilterOperand', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7935, 0),
        BrowseName=QualifiedName("FilterOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7935, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ElementOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ElementOperand', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7938, 0),
        BrowseName=QualifiedName("ElementOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7938, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7938, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LiteralOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('LiteralOperand', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7941, 0),
        BrowseName=QualifiedName("LiteralOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7941, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7941, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AttributeOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AttributeOperand', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7944, 0),
        BrowseName=QualifiedName("AttributeOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7944, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7944, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SimpleAttributeOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SimpleAttributeOperand', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7947, 0),
        BrowseName=QualifiedName("SimpleAttributeOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7947, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryEvent"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('HistoryEvent', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8004, 0),
        BrowseName=QualifiedName("HistoryEvent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8004, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MonitoringFilter"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('MonitoringFilter', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8067, 0),
        BrowseName=QualifiedName("MonitoringFilter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventFilter"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EventFilter', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8073, 0),
        BrowseName=QualifiedName("EventFilter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AggregateConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AggregateConfiguration', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8076, 0),
        BrowseName=QualifiedName("AggregateConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryEventFieldList"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('HistoryEventFieldList', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8172, 0),
        BrowseName=QualifiedName("HistoryEventFieldList", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BuildInfo"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('BuildInfo', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7692, 0),
        BrowseName=QualifiedName("BuildInfo", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7692, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RedundantServerDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('RedundantServerDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8208, 0),
        BrowseName=QualifiedName("RedundantServerDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrlListDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EndpointUrlListDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11959, 0),
        BrowseName=QualifiedName("EndpointUrlListDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11959, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('NetworkGroupDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11962, 0),
        BrowseName=QualifiedName("NetworkGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11962, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SamplingIntervalDiagnosticsDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8211, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsSummaryDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ServerDiagnosticsSummaryDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8214, 0),
        BrowseName=QualifiedName("ServerDiagnosticsSummaryDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8214, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerStatusDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ServerStatusDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8217, 0),
        BrowseName=QualifiedName("ServerStatusDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8217, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SessionDiagnosticsDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8220, 0),
        BrowseName=QualifiedName("SessionDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8220, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SessionSecurityDiagnosticsDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8223, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8223, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServiceCounterDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ServiceCounterDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8226, 0),
        BrowseName=QualifiedName("ServiceCounterDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8226, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatusResult"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('StatusResult', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7659, 0),
        BrowseName=QualifiedName("StatusResult", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(7659, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SubscriptionDiagnosticsDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8229, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8229, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModelChangeStructureDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ModelChangeStructureDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8232, 0),
        BrowseName=QualifiedName("ModelChangeStructureDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8232, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SemanticChangeStructureDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('SemanticChangeStructureDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8235, 0),
        BrowseName=QualifiedName("SemanticChangeStructureDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8235, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8235, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Range"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('Range', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8238, 0),
        BrowseName=QualifiedName("Range", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8238, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EUInformation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('EUInformation', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8241, 0),
        BrowseName=QualifiedName("EUInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8241, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ComplexNumberType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ComplexNumberType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12183, 0),
        BrowseName=QualifiedName("ComplexNumberType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DoubleComplexNumberType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('DoubleComplexNumberType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12186, 0),
        BrowseName=QualifiedName("DoubleComplexNumberType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AxisInformation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('AxisInformation', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12091, 0),
        BrowseName=QualifiedName("AxisInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("XVType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('XVType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12094, 0),
        BrowseName=QualifiedName("XVType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProgramDiagnosticDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ProgramDiagnosticDataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8247, 0),
        BrowseName=QualifiedName("ProgramDiagnosticDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8247, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProgramDiagnostic2DataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('ProgramDiagnostic2DataType', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15398, 0),
        BrowseName=QualifiedName("ProgramDiagnostic2DataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15398, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Annotation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('Annotation', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8244, 0),
        BrowseName=QualifiedName("Annotation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(7617, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7617, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14802, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14533, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14533, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14802, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17541, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(16313, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17541, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16313, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(17541, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17542, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17541, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17553, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17548, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17548, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(17553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17554, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15949, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15528, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15528, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16024, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18851, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18806, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18806, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18860, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18851, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18852, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18807, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18807, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18863, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18853, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18808, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18808, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18866, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18853, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18854, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18809, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18809, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18869, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18855, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18810, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18810, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19049, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18856, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18811, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18811, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19052, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18856, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18857, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18812, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18812, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18858, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18813, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18813, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19058, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18859, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18814, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(18859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18814, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(18859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19061, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18859, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15728, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15634, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15634, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15730, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15728, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23520, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23498, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23520, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23498, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(23520, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23520, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12676, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12554, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12554, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12677, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12676, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17862, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17861, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17861, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17862, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15950, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15534, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15534, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14796, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14525, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14525, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14811, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14796, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15589, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15487, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15487, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15591, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15589, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15590, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15488, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15488, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15594, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15529, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15005, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15529, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15005, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15529, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15585, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15529, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15531, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15006, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15531, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15006, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15531, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15588, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15531, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14794, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14523, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14523, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14805, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14794, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14795, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14524, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14524, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14808, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14803, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14593, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14593, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14832, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14803, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15951, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15578, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15578, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15952, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15580, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16033, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15952, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14319, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14273, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14320, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14319, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15953, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15581, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15953, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15581, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15953, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16037, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15953, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15954, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15582, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16040, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15955, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15597, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15597, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16047, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15955, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15956, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15598, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16050, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15956, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15987, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15605, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15987, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15605, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15987, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16053, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15987, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15988, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15609, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15988, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15609, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15988, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15988, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21174, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15480, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21180, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15990, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15611, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15990, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15611, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15990, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16062, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15990, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15991, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15616, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16065, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15991, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15992, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15617, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15617, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16068, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15992, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15993, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15618, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15618, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16071, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15993, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21175, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15502, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21183, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21176, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15510, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15510, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21186, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21177, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15520, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15520, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15995, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15621, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15621, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16077, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15995, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15996, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15622, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15996, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16007, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15623, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15623, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16007, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16008, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15628, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15628, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16086, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16008, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16009, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15629, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16089, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16009, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16010, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15630, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15630, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16010, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16011, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15631, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15631, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16095, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16011, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14804, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14744, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14744, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16012, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15635, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16012, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21178, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15530, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15530, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21192, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16014, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15645, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15645, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16104, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16014, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16015, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15652, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15652, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16015, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16016, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15653, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16016, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16017, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15657, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15657, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16017, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16018, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15664, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15664, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16116, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16018, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16019, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15665, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16019, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17472, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17467, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17467, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(17472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17473, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21179, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15532, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15532, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(21179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21195, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15579, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15007, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15007, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15640, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16021, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15667, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15667, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16125, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16021, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16022, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15669, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15669, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16144, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16022, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16023, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15670, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15670, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16147, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16023, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23505, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23468, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23468, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(23505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16126, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(96, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(96, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(16126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16126, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14797, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(97, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14797, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(97, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14797, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18166, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14797, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14800, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(101, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14800, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14798, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(99, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(99, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14799, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(100, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14799, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(259, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(258, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(258, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11887, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11879, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11879, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11887, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11888, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11880, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11880, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(262, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(261, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(261, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(265, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(264, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(264, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(265, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(268, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(267, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(267, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(268, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(271, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(270, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(270, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(274, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(273, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(277, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(276, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(276, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(280, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(279, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(279, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(283, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(282, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(282, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(286, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(285, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(285, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(297, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(296, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(296, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8285, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(7616, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(7594, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(7616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7594, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(7616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8291, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(7616, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14801, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(102, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(14801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(14801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14826, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12757, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12755, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12759, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12757, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12758, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12756, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12756, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12762, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12758, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8913, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(8912, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(8913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8912, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(8913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8918, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8913, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(309, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(308, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(308, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8300, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(390, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(389, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(390, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(393, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(392, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(392, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(396, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(395, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(395, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(396, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15902, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15901, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15901, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15902, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21000, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(20999, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20999, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21000, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(421, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(420, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(421, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(424, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(423, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(424, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12195, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12189, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12201, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12196, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12190, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12190, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12197, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12191, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12191, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(305, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(304, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(305, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(304, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(305, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8297, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(305, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(313, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(312, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(313, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(312, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(313, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8303, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(313, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(427, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(426, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(426, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(427, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(430, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(429, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(430, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(430, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(433, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(432, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(432, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8417, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(433, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(436, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(435, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(435, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(436, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(439, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(438, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(439, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(438, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(439, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12892, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12890, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12890, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12893, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12891, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12893, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12199, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12193, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12193, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12200, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12194, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12194, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(442, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(441, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(441, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(442, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(445, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(444, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(445, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(444, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(445, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(448, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(447, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(447, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(448, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(451, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(450, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(450, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(451, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(454, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(453, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(453, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(454, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(345, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(344, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(345, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(344, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(345, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8333, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(345, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(457, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(456, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(457, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(460, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(459, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(459, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(460, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(463, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(462, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(463, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(317, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(316, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(317, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(316, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(317, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8306, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(317, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(320, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(319, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(320, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(319, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(320, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(320, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(323, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(322, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(322, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8312, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(326, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(325, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(325, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8315, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(939, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(938, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(938, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8318, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(939, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(466, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(465, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(466, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(469, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(468, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(469, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(468, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(469, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(472, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(471, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(471, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(472, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(475, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(474, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(475, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(474, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(475, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(478, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(477, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(478, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(481, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(480, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(481, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(350, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(350, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(350, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(353, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(352, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(352, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(356, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(355, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(355, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(359, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(358, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(358, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(362, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(361, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(362, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(361, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(362, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(365, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(364, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(364, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(368, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(367, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(367, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(371, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(370, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(370, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(374, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(373, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(373, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17608, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17606, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17609, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17607, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17609, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(377, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(376, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(376, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8363, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(484, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(483, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(483, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(484, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(487, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(486, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(487, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(490, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(489, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(489, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(490, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(380, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(379, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(379, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8366, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(493, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(492, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(492, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(493, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(496, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(495, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(495, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(496, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(383, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(382, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(382, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8369, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(383, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(499, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(498, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(498, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(499, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(502, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(501, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(501, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(502, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(386, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(385, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(385, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8372, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(386, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(505, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(504, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(504, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(505, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(508, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(507, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(507, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(512, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(511, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(511, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(515, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(514, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(515, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(514, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(515, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(519, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(518, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(518, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(519, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(523, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(522, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(523, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(523, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(526, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(525, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(526, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(525, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(526, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(529, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(528, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(529, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(528, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(529, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(532, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(531, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(532, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(531, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(532, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(535, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(534, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(534, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(535, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(538, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(537, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(538, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(537, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(538, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12712, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(538, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(541, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(540, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(541, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(540, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(541, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12715, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(541, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(544, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(543, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(544, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(543, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(544, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(547, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(546, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(546, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(550, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(549, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(549, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(550, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(553, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(552, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(552, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(553, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(556, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(555, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(556, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(559, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(558, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(558, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(559, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(562, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(561, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(561, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(562, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(565, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(564, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(565, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(568, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(567, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(568, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(332, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(331, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(331, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8321, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(571, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(570, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(570, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(571, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(574, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(573, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(574, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(578, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(577, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(578, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(577, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(578, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(581, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(580, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(581, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(584, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(583, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(583, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(584, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(587, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(586, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(586, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(587, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(590, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(589, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(589, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8570, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(590, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(593, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(592, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(592, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(593, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(596, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(595, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(595, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8576, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(596, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(599, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(598, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8579, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(599, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(602, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(601, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(601, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(602, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(605, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(604, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(604, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(605, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(608, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(607, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(608, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(611, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(610, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(610, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(611, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(614, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(613, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(613, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(614, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(617, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(616, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(617, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(620, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(619, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(619, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(620, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(623, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(622, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(623, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(627, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(626, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(626, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(627, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(630, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(629, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(630, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(633, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(632, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(632, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(633, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(636, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(635, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(636, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(639, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(638, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(638, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(642, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(642, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(645, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(644, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(644, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(645, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(648, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(647, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(647, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(648, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(651, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(650, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(651, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(654, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(653, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(654, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23506, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23497, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23497, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23506, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(657, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(656, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(656, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(657, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11218, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11216, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11216, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11218, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11219, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11217, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11219, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(660, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(659, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(660, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(659, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(660, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8639, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(660, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(663, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(662, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(663, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(662, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(663, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(666, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(665, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(666, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(669, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(668, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(669, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(668, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(669, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(672, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(671, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(672, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(671, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(672, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(675, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(674, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(675, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(674, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(675, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(678, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(678, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(681, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(680, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(680, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(681, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11296, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11295, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11295, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11296, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(684, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(683, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(683, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(684, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(687, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(686, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(687, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(690, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(689, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(689, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(690, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(693, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(692, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(692, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(693, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(696, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(695, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(696, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(695, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(696, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(699, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(698, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(699, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(702, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(701, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(701, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(705, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(704, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(704, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(705, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(708, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(707, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(707, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(711, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(710, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(714, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(713, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(713, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(720, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(719, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(719, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8702, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(720, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(723, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(722, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(722, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(723, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(726, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(725, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8708, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(949, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(948, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8711, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(729, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(728, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(729, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(732, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(731, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(732, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(735, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(734, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(735, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(738, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(737, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(738, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(741, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(740, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(740, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(741, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(744, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(743, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(743, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(744, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(747, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(746, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(747, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(750, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(749, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(749, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(750, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(753, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(752, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(752, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(753, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(756, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(755, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(756, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(759, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(758, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(758, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(762, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(761, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(761, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(765, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(764, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(764, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(765, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(768, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(767, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(768, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(771, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(770, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(770, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(771, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(774, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(773, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(773, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(774, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(777, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(776, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(776, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(777, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(780, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(779, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(779, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(780, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(783, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(782, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(782, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(783, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(786, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(785, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(785, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(786, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(789, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(788, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(788, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(789, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(792, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(791, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(791, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(792, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(795, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(794, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(794, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(795, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(798, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(797, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(797, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(798, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(801, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(800, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(800, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(801, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(804, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(803, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(804, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(946, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(945, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(946, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(945, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(946, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(810, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(809, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(810, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(809, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(810, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(807, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(806, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(806, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(915, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(914, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(914, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(915, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(918, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(917, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(917, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(921, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(920, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(921, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(920, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(921, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8807, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(921, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(819, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(818, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(818, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(819, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(822, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(821, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(822, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(825, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(824, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(825, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(828, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(827, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(828, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(831, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(830, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(831, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(834, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(833, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(834, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(837, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(836, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(837, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(840, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(839, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(840, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(843, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(842, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(846, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(845, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(849, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(848, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(339, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(338, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(338, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8327, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(854, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(853, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8843, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(854, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11949, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11943, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(11949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11951, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11949, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11950, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11944, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(11950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(11950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11954, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11950, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(857, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(856, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(856, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8846, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(857, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(860, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(859, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(859, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(863, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(862, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(862, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(866, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(865, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(865, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(869, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(868, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(868, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8858, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(872, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(871, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(871, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8861, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(872, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(300, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(299, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(299, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8294, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(875, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(874, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(874, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8864, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(875, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(878, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(877, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(877, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8867, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(878, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(898, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(897, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8870, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(898, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(885, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(884, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(884, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8873, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(885, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(888, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(887, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(887, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8876, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(888, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12173, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12171, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12174, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12172, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12081, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12079, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12082, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12080, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(12082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(12082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12086, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(895, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(894, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8882, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(895, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15401, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15396, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15396, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(15401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15401, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default XML"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(892, 0),
        BrowseName=QualifiedName("Default XML", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(891, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(39, 0),
        SourceNodeId=NumericNodeId(892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8879, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(892, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Opc.Ua"),
        DataType=ua.NodeId(ua.ObjectIds.ByteString),
        Value=ua.Variant(b'<xs:schema\r\n  xmlns:xs="http://www.w3.org/2001/XMLSchema"\r\n  xmlns:ua="http://opcfoundation.org/UA/2008/02/Types.xsd"\r\n  xmlns:tns="http://opcfoundation.org/UA/2008/02/Types.xsd"\r\n  targetNamespace="http://opcfoundation.org/UA/2008/02/Types.xsd"\r\n  elementFormDefault="qualified"\r\n>\r\n  <xs:element name="Boolean" nillable="true" type="xs:boolean" />\r\n\r\n  <xs:complexType name="ListOfBoolean">\r\n    <xs:sequence>\r\n      <xs:element name="Boolean" type="xs:boolean" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBoolean" type="tns:ListOfBoolean" nillable="true"></xs:element>\r\n\r\n  <xs:element name="SByte" nillable="true" type="xs:byte" />\r\n\r\n  <xs:complexType name="ListOfSByte">\r\n    <xs:sequence>\r\n      <xs:element name="SByte" type="xs:byte" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSByte" type="tns:ListOfSByte" nillable="true"></xs:element>\r\n\r\n  <xs:element name="Byte" nillable="true" type="xs:unsignedByte" />\r\n\r\n  <xs:complexType name="ListOfByte">\r\n    <xs:sequence>\r\n      <xs:element name="Byte" type="xs:unsignedByte" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfByte" type="tns:ListOfByte" nillable="true"></xs:element>\r\n\r\n  <xs:element name="Int16" nillable="true" type="xs:short" />\r\n\r\n  <xs:complexType name="ListOfInt16">\r\n    <xs:sequence>\r\n      <xs:element name="Int16" type="xs:short" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfInt16" type="tns:ListOfInt16" nillable="true"></xs:element>\r\n\r\n  <xs:element name="UInt16" nillable="true" type="xs:unsignedShort" />\r\n\r\n  <xs:complexType name="ListOfUInt16">\r\n    <xs:sequence>\r\n      <xs:element name="UInt16" type="xs:unsignedShort" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUInt16" type="tns:ListOfUInt16" nillable="true"></xs:element>\r\n\r\n  <xs:element name="Int32" nillable="true" type="xs:int" />\r\n\r\n  <xs:complexType name="ListOfInt32">\r\n    <xs:sequence>\r\n      <xs:element name="Int32" type="xs:int" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfInt32" type="tns:ListOfInt32" nillable="true"></xs:element>\r\n\r\n  <xs:element name="UInt32" nillable="true" type="xs:unsignedInt" />\r\n\r\n  <xs:complexType name="ListOfUInt32">\r\n    <xs:sequence>\r\n      <xs:element name="UInt32" type="xs:unsignedInt" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUInt32" type="tns:ListOfUInt32" nillable="true"></xs:element>\r\n\r\n  <xs:element name="Int64" nillable="true" type="xs:long" />\r\n\r\n  <xs:complexType name="ListOfInt64">\r\n    <xs:sequence>\r\n      <xs:element name="Int64" type="xs:long" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfInt64" type="tns:ListOfInt64" nillable="true"></xs:element>\r\n\r\n  <xs:element name="UInt64" nillable="true" type="xs:unsignedLong" />\r\n\r\n  <xs:complexType name="ListOfUInt64">\r\n    <xs:sequence>\r\n      <xs:element name="UInt64" type="xs:unsignedLong" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUInt64" type="tns:ListOfUInt64" nillable="true"></xs:element>\r\n\r\n  <xs:element name="Float" nillable="true" type="xs:float" />\r\n\r\n  <xs:complexType name="ListOfFloat">\r\n    <xs:sequence>\r\n      <xs:element name="Float" type="xs:float" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfFloat" type="tns:ListOfFloat" nillable="true"></xs:element>\r\n\r\n  <xs:element name="Double" nillable="true" type="xs:double" />\r\n\r\n  <xs:complexType name="ListOfDouble">\r\n    <xs:sequence>\r\n      <xs:element name="Double" type="xs:double" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDouble" type="tns:ListOfDouble" nillable="true"></xs:element>\r\n\r\n  <xs:element name="String" nillable="true" type="xs:string" />\r\n\r\n  <xs:complexType name="ListOfString">\r\n    <xs:sequence>\r\n      <xs:element name="String" type="xs:string" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfString" type="tns:ListOfString" nillable="true"></xs:element>\r\n\r\n  <xs:element name="DateTime" nillable="true" type="xs:dateTime" />\r\n\r\n  <xs:complexType name="ListOfDateTime">\r\n    <xs:sequence>\r\n      <xs:element name="DateTime" type="xs:dateTime" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDateTime" type="tns:ListOfDateTime" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Guid">\r\n    <xs:annotation>\r\n      <xs:appinfo>\r\n        <IsValueType xmlns="http://schemas.microsoft.com/2003/10/Serialization/">true</IsValueType>\r\n      </xs:appinfo>\r\n    </xs:annotation>\r\n    <xs:sequence>\r\n      <xs:element name="String" type="xs:string" minOccurs="0" maxOccurs="1" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Guid" type="tns:Guid" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ListOfGuid">\r\n    <xs:sequence>\r\n      <xs:element name="Guid" type="tns:Guid" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfGuid" type="tns:ListOfGuid" nillable="true"></xs:element>\r\n\r\n  <xs:element name="ByteString" nillable="true" type="xs:base64Binary" />\r\n\r\n  <xs:complexType name="ListOfByteString">\r\n    <xs:sequence>\r\n      <xs:element name="ByteString" type="xs:base64Binary" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfByteString" type="tns:ListOfByteString" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ListOfXmlElement">\r\n    <xs:sequence>\r\n      <xs:element name="XmlElement" minOccurs="0" maxOccurs="unbounded" nillable="true">\r\n        <xs:complexType>\r\n          <xs:sequence>\r\n            <xs:any minOccurs="0" processContents="lax"/>\r\n          </xs:sequence>\r\n        </xs:complexType>\r\n      </xs:element>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfXmlElement" type="tns:ListOfXmlElement" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="NodeId">\r\n    <xs:sequence>\r\n      <xs:element name="Identifier" type="xs:string" minOccurs="0" maxOccurs="1" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NodeId" type="tns:NodeId" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ListOfNodeId">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="tns:NodeId" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfNodeId" type="tns:ListOfNodeId" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ExpandedNodeId">\r\n    <xs:sequence>\r\n      <xs:element name="Identifier" type="xs:string" minOccurs="0" maxOccurs="1" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ExpandedNodeId" type="tns:ExpandedNodeId" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ListOfExpandedNodeId">\r\n    <xs:sequence>\r\n      <xs:element name="ExpandedNodeId" type="tns:ExpandedNodeId" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfExpandedNodeId" type="tns:ListOfExpandedNodeId" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="StatusCode">\r\n    <xs:annotation>\r\n      <xs:appinfo>\r\n        <IsValueType xmlns="http://schemas.microsoft.com/2003/10/Serialization/">true</IsValueType>\r\n      </xs:appinfo>\r\n    </xs:annotation>\r\n    <xs:sequence>\r\n      <xs:element name="Code" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="StatusCode" type="tns:StatusCode" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ListOfStatusCode">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="tns:StatusCode" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfStatusCode" type="tns:ListOfStatusCode" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DiagnosticInfo">\r\n    <xs:sequence>\r\n      <xs:element name="SymbolicId" type="xs:int" minOccurs="0" />\r\n      <xs:element name="NamespaceUri" type="xs:int" minOccurs="0" />\r\n      <xs:element name="Locale" type="xs:int" minOccurs="0" />\r\n      <xs:element name="LocalizedText" type="xs:int" minOccurs="0" />\r\n      <xs:element name="AdditionalInfo" type="xs:string" minOccurs="0" />\r\n      <xs:element name="InnerStatusCode" type="tns:StatusCode" minOccurs="0" />\r\n      <xs:element name="InnerDiagnosticInfo" type="tns:DiagnosticInfo" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DiagnosticInfo" type="tns:DiagnosticInfo" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ListOfDiagnosticInfo">\r\n    <xs:sequence>\r\n      <xs:element name="DiagnosticInfo" type="tns:DiagnosticInfo" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDiagnosticInfo" type="tns:ListOfDiagnosticInfo" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="LocalizedText">\r\n    <xs:sequence>\r\n      <xs:element name="Locale" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Text" type="xs:string" minOccurs="0"  nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="LocalizedText" type="tns:LocalizedText" nillable="true" />\r\n\r\n  <xs:complexType name="ListOfLocalizedText">\r\n    <xs:sequence>\r\n      <xs:element name="LocalizedText" type="tns:LocalizedText" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfLocalizedText" type="tns:ListOfLocalizedText" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="QualifiedName">\r\n    <xs:sequence>\r\n      <xs:element name="NamespaceIndex" type="xs:unsignedShort" minOccurs="0" />\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="QualifiedName" type="tns:QualifiedName" nillable="true" />\r\n\r\n  <xs:complexType name="ListOfQualifiedName">\r\n    <xs:sequence>\r\n      <xs:element name="QualifiedName" type="tns:QualifiedName" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfQualifiedName" type="tns:ListOfQualifiedName" nillable="true"></xs:element>\r\n\r\n  <!--\r\n    Some environments require a WSDL/XSD which explicitly defines all possible types.\r\n    The UA WSDL/XSD can be modified to support these environments by replacing the\r\n    definitions of the ExtensionObjectBody and VariantValue complex types with the\r\n    definitions in the comments shown here. Developers would then define subtypes\r\n    of the ExtensionObjectBody type which explicitly declare a choice between all of the\r\n    complex types used by the system. The ExampleExtensionObjectBody subtype is provides\r\n    a template based on a few common UA-defined complex types.\r\n    -->\r\n\r\n  <!--\r\n    <xs:complexType name="ExtensionObjectBody" />\r\n\r\n    <xs:complexType name="ExampleExtensionObjectBody">\r\n      <xs:complexContent>\r\n        <xs:extension base="tns:ExtensionObjectBody">\r\n          <xs:choice>\r\n            <xs:element name="Argument" type="tns:Argument" minOccurs="0" nillable="true" />\r\n            <xs:element name="UserIdentityToken" type="tns:UserIdentityToken" minOccurs="0" nillable="true" />\r\n            <xs:element name="UserNameIdentityToken" type="tns:UserNameIdentityToken" minOccurs="0" nillable="true" />\r\n          </xs:choice>\r\n        </xs:extension>\r\n      </xs:complexContent>\r\n    </xs:complexType>\r\n\r\n    <xs:complexType name="ExtensionObject">\r\n      <xs:sequence>\r\n        <xs:element name="TypeId" type="tns:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n        <xs:element name="Body" minOccurs="0" type="tns:ExtensionObjectBody" nillable="true" />\r\n      </xs:sequence>\r\n    </xs:complexType>\r\n    <xs:element name="ExtensionObject" type="tns:ExtensionObject" nillable="true" />\r\n    -->\r\n\r\n  <xs:complexType name="ExtensionObject">\r\n    <xs:sequence>\r\n      <xs:element name="TypeId" type="tns:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="Body" minOccurs="0" nillable="true">\r\n        <xs:complexType>\r\n          <xs:sequence>\r\n            <xs:any minOccurs="0" processContents="lax"/>\r\n          </xs:sequence>\r\n        </xs:complexType>\r\n      </xs:element>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ExtensionObject" type="tns:ExtensionObject" nillable="true" />\r\n\r\n  <xs:complexType name="ListOfExtensionObject">\r\n    <xs:sequence>\r\n      <xs:element name="ExtensionObject" type="tns:ExtensionObject" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfExtensionObject" type="tns:ListOfExtensionObject" nillable="true"></xs:element>\r\n\r\n  <!--\r\n    Some WSDL/XML compilers have issues with the xs:choice construct. For that reason\r\n    the default declaration of a Variant uses xs:any construct. The schema acutually\r\n    defined by the specification is provided by the Matrix and VariantValue complex types\r\n    shown in comments below. Application developers can replace the VariantValue declaration\r\n    with the specific declaration if they have a development environment that can handle\r\n    the xs:choice construct in a reasonable way.\r\n    -->\r\n\r\n  <!--\r\n    <xs:complexType name="Matrix">\r\n      <xs:sequence>\r\n        <xs:element name="Dimensions" type="tns:ListOfUInt32" minOccurs="0" nillable="true" />\r\n        <xs:element name="Value" minOccurs="0" nillable="true">\r\n          <xs:complexType mixed="false">\r\n            <xs:choice maxOccurs="unbounded">\r\n              <xs:element name="Boolean" type="xs:boolean" minOccurs="0" />\r\n              <xs:element name="SByte" type="xs:byte" minOccurs="0" />\r\n              <xs:element name="Byte" type="xs:unsignedByte" minOccurs="0" />\r\n              <xs:element name="Int16" type="xs:short" minOccurs="0" />\r\n              <xs:element name="UInt16" type="xs:unsignedShort" minOccurs="0" />\r\n              <xs:element name="Int32" type="xs:int" minOccurs="0" />\r\n              <xs:element name="UInt32" type="xs:unsignedInt" minOccurs="0" />\r\n              <xs:element name="Int64" type="xs:long" minOccurs="0" />\r\n              <xs:element name="UInt64" type="xs:unsignedLong" minOccurs="0" />\r\n              <xs:element name="Float" type="xs:float" minOccurs="0" />\r\n              <xs:element name="Double" type="xs:double" minOccurs="0" />\r\n              <xs:element name="String" type="xs:string" minOccurs="0" />\r\n              <xs:element name="DateTime" type="xs:dateTime" minOccurs="0" />\r\n              <xs:element name="Guid" type="tns:Guid" minOccurs="0" />\r\n              <xs:element name="ByteString" type="xs:base64Binary" minOccurs="0" />\r\n              <xs:element name="XmlElement" minOccurs="0" nillable="true">\r\n                <xs:complexType>\r\n                  <xs:sequence>\r\n                    <xs:any minOccurs="0" processContents="lax" />\r\n                  </xs:sequence>\r\n                </xs:complexType>\r\n              </xs:element>\r\n              <xs:element name="StatusCode" type="tns:StatusCode" minOccurs="0" />\r\n              <xs:element name="NodeId" type="tns:NodeId" minOccurs="0" />\r\n              <xs:element name="ExpandedNodeId" type="tns:ExpandedNodeId" minOccurs="0" />\r\n              <xs:element name="QualifiedName" type="tns:QualifiedName" minOccurs="0" />\r\n              <xs:element name="LocalizedText" type="tns:LocalizedText" minOccurs="0" />\r\n              <xs:element name="ExtensionObject" type="tns:ExtensionObject" minOccurs="0" />\r\n              <xs:element name="Variant" type="tns:Variant" minOccurs="0" />\r\n            </xs:choice>\r\n          </xs:complexType>\r\n        </xs:element>\r\n      </xs:sequence>\r\n    </xs:complexType>\r\n    <xs:element name="Matrix" type="tns:Matrix" nillable="true" />\r\n\r\n    <xs:complexType name="VariantValue">\r\n      <xs:choice>\r\n        <xs:element name="Boolean" type="xs:boolean" minOccurs="0" />\r\n        <xs:element name="SByte" type="xs:byte" minOccurs="0" />\r\n        <xs:element name="Byte" type="xs:unsignedByte" minOccurs="0" />\r\n        <xs:element name="Int16" type="xs:short" minOccurs="0" />\r\n        <xs:element name="UInt16" type="xs:unsignedShort" minOccurs="0" />\r\n        <xs:element name="Int32" type="xs:int" minOccurs="0" />\r\n        <xs:element name="UInt32" type="xs:unsignedInt" minOccurs="0" />\r\n        <xs:element name="Int64" type="xs:long" minOccurs="0" />\r\n        <xs:element name="UInt64" type="xs:unsignedLong" minOccurs="0" />\r\n        <xs:element name="Float" type="xs:float" minOccurs="0" />\r\n        <xs:element name="Double" type="xs:double" minOccurs="0" />\r\n        <xs:element name="String" type="xs:string" minOccurs="0" />\r\n        <xs:element name="DateTime" type="xs:dateTime" minOccurs="0" />\r\n        <xs:element name="Guid" type="tns:Guid" minOccurs="0" />\r\n        <xs:element name="ByteString" type="xs:base64Binary" minOccurs="0" />\r\n        <xs:element name="XmlElement" minOccurs="0" nillable="true">\r\n          <xs:complexType>\r\n            <xs:sequence>\r\n              <xs:any minOccurs="0" processContents="lax" />\r\n            </xs:sequence>\r\n          </xs:complexType>\r\n        </xs:element>\r\n        <xs:element name="StatusCode" type="tns:StatusCode" minOccurs="0" />\r\n        <xs:element name="NodeId" type="tns:NodeId" minOccurs="0" />\r\n        <xs:element name="ExpandedNodeId" type="tns:ExpandedNodeId" minOccurs="0" />\r\n        <xs:element name="QualifiedName" type="tns:QualifiedName" minOccurs="0" />\r\n        <xs:element name="LocalizedText" type="tns:LocalizedText" minOccurs="0" />\r\n        <xs:element name="ExtensionObject" type="tns:ExtensionObject" minOccurs="0" />\r\n        <xs:element name="ListOfBoolean" type="tns:ListOfBoolean" minOccurs="0" />\r\n        <xs:element name="ListOfSByte" type="tns:ListOfSByte" minOccurs="0" />\r\n        <xs:element name="ListOfByte" type="tns:ListOfByte" minOccurs="0" />\r\n        <xs:element name="ListOfInt16" type="tns:ListOfInt16" minOccurs="0" />\r\n        <xs:element name="ListOfUInt16" type="tns:ListOfUInt16" minOccurs="0" />\r\n        <xs:element name="ListOfInt32" type="tns:ListOfInt32" minOccurs="0" />\r\n        <xs:element name="ListOfUInt32" type="tns:ListOfUInt32" minOccurs="0" />\r\n        <xs:element name="ListOfInt64" type="tns:ListOfInt64" minOccurs="0" />\r\n        <xs:element name="ListOfUInt64" type="tns:ListOfUInt64" minOccurs="0" />\r\n        <xs:element name="ListOfFloat" type="tns:ListOfFloat" minOccurs="0" />\r\n        <xs:element name="ListOfDouble" type="tns:ListOfDouble" minOccurs="0" />\r\n        <xs:element name="ListOfString" type="tns:ListOfString" minOccurs="0" />\r\n        <xs:element name="ListOfDateTime" type="tns:ListOfDateTime" minOccurs="0" />\r\n        <xs:element name="ListOfGuid" type="tns:ListOfGuid" minOccurs="0" />\r\n        <xs:element name="ListOfByteString" type="tns:ListOfByteString" minOccurs="0" />\r\n        <xs:element name="ListOfXmlElement" type="tns:ListOfXmlElement" minOccurs="0" />\r\n        <xs:element name="ListOfStatusCode" type="tns:ListOfStatusCode" minOccurs="0" />\r\n        <xs:element name="ListOfNodeId" type="tns:ListOfNodeId" minOccurs="0" />\r\n        <xs:element name="ListOfExpandedNodeId" type="tns:ListOfExpandedNodeId" minOccurs="0" />\r\n        <xs:element name="ListOfQualifiedName" type="tns:ListOfQualifiedName" minOccurs="0" />\r\n        <xs:element name="ListOfLocalizedText" type="tns:ListOfLocalizedText" minOccurs="0" />\r\n        <xs:element name="ListOfExtensionObject" type="tns:ListOfExtensionObject" minOccurs="0" />\r\n        <xs:element name="ListOfVariant" type="tns:ListOfVariant" minOccurs="0" />\r\n        <xs:element name="Matrix" type="tns:Matrix" minOccurs="0" />\r\n      </xs:choice>\r\n    </xs:complexType>\r\n\r\n    <xs:complexType name="Variant">\r\n      <xs:sequence>\r\n        <xs:element name="Value" type="tns:VariantValue" minOccurs="0" nillable="true" />\r\n      </xs:sequence>\r\n    </xs:complexType>\r\n    <xs:element name="Variant" type="tns:Variant" nillable="true" />\r\n    -->\r\n\r\n  <xs:complexType name="Variant">\r\n    <xs:annotation>\r\n      <xs:appinfo>\r\n        <IsValueType xmlns="http://schemas.microsoft.com/2003/10/Serialization/">true</IsValueType>\r\n      </xs:appinfo>\r\n    </xs:annotation>\r\n    <xs:sequence>\r\n      <xs:element name="Value" minOccurs="0" nillable="true">\r\n        <xs:complexType>\r\n          <xs:sequence>\r\n            <xs:any minOccurs="0" processContents="lax" />\r\n          </xs:sequence>\r\n        </xs:complexType>\r\n      </xs:element>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Variant" type="tns:Variant" nillable="true" />\r\n\r\n  <xs:complexType name="ListOfVariant">\r\n    <xs:sequence>\r\n      <xs:element name="Variant" type="tns:Variant" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfVariant" type="tns:ListOfVariant" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataValue">\r\n    <xs:sequence>\r\n      <xs:element name="Value" type="tns:Variant" minOccurs="0" />\r\n      <xs:element name="StatusCode" type="tns:StatusCode" minOccurs="0" />\r\n      <xs:element name="SourceTimestamp" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="SourcePicoseconds" type="xs:unsignedShort" minOccurs="0" />\r\n      <xs:element name="ServerTimestamp" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="ServerPicoseconds" type="xs:unsignedShort" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataValue" type="tns:DataValue" nillable="true"/>\r\n\r\n  <xs:complexType name="ListOfDataValue">\r\n    <xs:sequence>\r\n      <xs:element name="DataValue" type="tns:DataValue" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataValue" type="tns:ListOfDataValue" nillable="true"></xs:element>\r\n\r\n  <xs:element name="InvokeServiceRequest" type="xs:base64Binary" nillable="true" />\r\n  <xs:element name="InvokeServiceResponse" type="xs:base64Binary" nillable="true" />\r\n\r\n  <xs:element name="ImageBMP" type="xs:base64Binary" />\r\n\r\n  <xs:element name="ImageGIF" type="xs:base64Binary" />\r\n\r\n  <xs:element name="ImageJPG" type="xs:base64Binary" />\r\n\r\n  <xs:element name="ImagePNG" type="xs:base64Binary" />\r\n\r\n  <xs:element name="AudioDataType" type="xs:base64Binary" />\r\n\r\n  <xs:element name="BitFieldMaskDataType" type="xs:unsignedLong" />\r\n\r\n  <xs:complexType name="KeyValuePair">\r\n    <xs:sequence>\r\n      <xs:element name="Key" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n      <xs:element name="Value" type="ua:Variant" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="KeyValuePair" type="tns:KeyValuePair" />\r\n\r\n  <xs:complexType name="ListOfKeyValuePair">\r\n    <xs:sequence>\r\n      <xs:element name="KeyValuePair" type="tns:KeyValuePair" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfKeyValuePair" type="tns:ListOfKeyValuePair" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="AdditionalParametersType">\r\n    <xs:sequence>\r\n      <xs:element name="Parameters" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AdditionalParametersType" type="tns:AdditionalParametersType" />\r\n\r\n  <xs:complexType name="EphemeralKeyType">\r\n    <xs:sequence>\r\n      <xs:element name="PublicKey" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="Signature" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EphemeralKeyType" type="tns:EphemeralKeyType" />\r\n\r\n  <xs:complexType name="EndpointType">\r\n    <xs:sequence>\r\n      <xs:element name="EndpointUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityMode" type="tns:MessageSecurityMode" minOccurs="0" />\r\n      <xs:element name="SecurityPolicyUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="TransportProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EndpointType" type="tns:EndpointType" />\r\n\r\n  <xs:complexType name="ListOfEndpointType">\r\n    <xs:sequence>\r\n      <xs:element name="EndpointType" type="tns:EndpointType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEndpointType" type="tns:ListOfEndpointType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="RationalNumber">\r\n    <xs:sequence>\r\n      <xs:element name="Numerator" type="xs:int" minOccurs="0" />\r\n      <xs:element name="Denominator" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RationalNumber" type="tns:RationalNumber" />\r\n\r\n  <xs:complexType name="ListOfRationalNumber">\r\n    <xs:sequence>\r\n      <xs:element name="RationalNumber" type="tns:RationalNumber" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfRationalNumber" type="tns:ListOfRationalNumber" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Vector">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Vector" type="tns:Vector" />\r\n\r\n  <xs:complexType name="ListOfVector">\r\n    <xs:sequence>\r\n      <xs:element name="Vector" type="tns:Vector" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfVector" type="tns:ListOfVector" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ThreeDVector">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:Vector">\r\n        <xs:sequence>\r\n          <xs:element name="X" type="xs:double" minOccurs="0" />\r\n          <xs:element name="Y" type="xs:double" minOccurs="0" />\r\n          <xs:element name="Z" type="xs:double" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ThreeDVector" type="tns:ThreeDVector" />\r\n\r\n  <xs:complexType name="ListOfThreeDVector">\r\n    <xs:sequence>\r\n      <xs:element name="ThreeDVector" type="tns:ThreeDVector" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfThreeDVector" type="tns:ListOfThreeDVector" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="CartesianCoordinates">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CartesianCoordinates" type="tns:CartesianCoordinates" />\r\n\r\n  <xs:complexType name="ListOfCartesianCoordinates">\r\n    <xs:sequence>\r\n      <xs:element name="CartesianCoordinates" type="tns:CartesianCoordinates" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfCartesianCoordinates" type="tns:ListOfCartesianCoordinates" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ThreeDCartesianCoordinates">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:CartesianCoordinates">\r\n        <xs:sequence>\r\n          <xs:element name="X" type="xs:double" minOccurs="0" />\r\n          <xs:element name="Y" type="xs:double" minOccurs="0" />\r\n          <xs:element name="Z" type="xs:double" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ThreeDCartesianCoordinates" type="tns:ThreeDCartesianCoordinates" />\r\n\r\n  <xs:complexType name="ListOfThreeDCartesianCoordinates">\r\n    <xs:sequence>\r\n      <xs:element name="ThreeDCartesianCoordinates" type="tns:ThreeDCartesianCoordinates" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfThreeDCartesianCoordinates" type="tns:ListOfThreeDCartesianCoordinates" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Orientation">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Orientation" type="tns:Orientation" />\r\n\r\n  <xs:complexType name="ListOfOrientation">\r\n    <xs:sequence>\r\n      <xs:element name="Orientation" type="tns:Orientation" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfOrientation" type="tns:ListOfOrientation" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ThreeDOrientation">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:Orientation">\r\n        <xs:sequence>\r\n          <xs:element name="A" type="xs:double" minOccurs="0" />\r\n          <xs:element name="B" type="xs:double" minOccurs="0" />\r\n          <xs:element name="C" type="xs:double" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ThreeDOrientation" type="tns:ThreeDOrientation" />\r\n\r\n  <xs:complexType name="ListOfThreeDOrientation">\r\n    <xs:sequence>\r\n      <xs:element name="ThreeDOrientation" type="tns:ThreeDOrientation" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfThreeDOrientation" type="tns:ListOfThreeDOrientation" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Frame">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Frame" type="tns:Frame" />\r\n\r\n  <xs:complexType name="ListOfFrame">\r\n    <xs:sequence>\r\n      <xs:element name="Frame" type="tns:Frame" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfFrame" type="tns:ListOfFrame" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ThreeDFrame">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:Frame">\r\n        <xs:sequence>\r\n          <xs:element name="CartesianCoordinates" type="tns:ThreeDCartesianCoordinates" minOccurs="0" nillable="true" />\r\n          <xs:element name="Orientation" type="tns:ThreeDOrientation" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ThreeDFrame" type="tns:ThreeDFrame" />\r\n\r\n  <xs:complexType name="ListOfThreeDFrame">\r\n    <xs:sequence>\r\n      <xs:element name="ThreeDFrame" type="tns:ThreeDFrame" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfThreeDFrame" type="tns:ListOfThreeDFrame" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="OpenFileMode">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Read_1" />\r\n      <xs:enumeration value="Write_2" />\r\n      <xs:enumeration value="EraseExisting_4" />\r\n      <xs:enumeration value="Append_8" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="OpenFileMode" type="tns:OpenFileMode" />\r\n\r\n  <xs:complexType name="ListOfOpenFileMode">\r\n    <xs:sequence>\r\n      <xs:element name="OpenFileMode" type="tns:OpenFileMode" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfOpenFileMode" type="tns:ListOfOpenFileMode" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="IdentityCriteriaType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="UserName_1" />\r\n      <xs:enumeration value="Thumbprint_2" />\r\n      <xs:enumeration value="Role_3" />\r\n      <xs:enumeration value="GroupId_4" />\r\n      <xs:enumeration value="Anonymous_5" />\r\n      <xs:enumeration value="AuthenticatedUser_6" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="IdentityCriteriaType" type="tns:IdentityCriteriaType" />\r\n\r\n  <xs:complexType name="ListOfIdentityCriteriaType">\r\n    <xs:sequence>\r\n      <xs:element name="IdentityCriteriaType" type="tns:IdentityCriteriaType" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfIdentityCriteriaType" type="tns:ListOfIdentityCriteriaType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="IdentityMappingRuleType">\r\n    <xs:sequence>\r\n      <xs:element name="CriteriaType" type="tns:IdentityCriteriaType" minOccurs="0" />\r\n      <xs:element name="Criteria" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="IdentityMappingRuleType" type="tns:IdentityMappingRuleType" />\r\n\r\n  <xs:complexType name="ListOfIdentityMappingRuleType">\r\n    <xs:sequence>\r\n      <xs:element name="IdentityMappingRuleType" type="tns:IdentityMappingRuleType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfIdentityMappingRuleType" type="tns:ListOfIdentityMappingRuleType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="CurrencyUnitType">\r\n    <xs:sequence>\r\n      <xs:element name="NumericCode" type="xs:short" minOccurs="0" />\r\n      <xs:element name="Exponent" type="xs:byte" minOccurs="0" />\r\n      <xs:element name="AlphabeticCode" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Currency" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CurrencyUnitType" type="tns:CurrencyUnitType" />\r\n\r\n  <xs:complexType name="ListOfCurrencyUnitType">\r\n    <xs:sequence>\r\n      <xs:element name="CurrencyUnitType" type="tns:CurrencyUnitType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfCurrencyUnitType" type="tns:ListOfCurrencyUnitType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="TrustListMasks">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="None_0" />\r\n      <xs:enumeration value="TrustedCertificates_1" />\r\n      <xs:enumeration value="TrustedCrls_2" />\r\n      <xs:enumeration value="IssuerCertificates_4" />\r\n      <xs:enumeration value="IssuerCrls_8" />\r\n      <xs:enumeration value="All_15" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="TrustListMasks" type="tns:TrustListMasks" />\r\n\r\n  <xs:complexType name="TrustListDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SpecifiedLists" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TrustedCertificates" type="ua:ListOfByteString" minOccurs="0" nillable="true" />\r\n      <xs:element name="TrustedCrls" type="ua:ListOfByteString" minOccurs="0" nillable="true" />\r\n      <xs:element name="IssuerCertificates" type="ua:ListOfByteString" minOccurs="0" nillable="true" />\r\n      <xs:element name="IssuerCrls" type="ua:ListOfByteString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="TrustListDataType" type="tns:TrustListDataType" />\r\n\r\n  <xs:complexType name="ListOfTrustListDataType">\r\n    <xs:sequence>\r\n      <xs:element name="TrustListDataType" type="tns:TrustListDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfTrustListDataType" type="tns:ListOfTrustListDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DecimalDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Scale" type="xs:short" minOccurs="0" />\r\n      <xs:element name="Value" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DecimalDataType" type="tns:DecimalDataType" />\r\n\r\n  <xs:complexType name="DataTypeSchemaHeader">\r\n    <xs:sequence>\r\n      <xs:element name="Namespaces" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="StructureDataTypes" type="tns:ListOfStructureDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="EnumDataTypes" type="tns:ListOfEnumDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="SimpleDataTypes" type="tns:ListOfSimpleTypeDescription" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataTypeSchemaHeader" type="tns:DataTypeSchemaHeader" />\r\n\r\n  <xs:complexType name="ListOfDataTypeSchemaHeader">\r\n    <xs:sequence>\r\n      <xs:element name="DataTypeSchemaHeader" type="tns:DataTypeSchemaHeader" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataTypeSchemaHeader" type="tns:ListOfDataTypeSchemaHeader" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataTypeDescription">\r\n    <xs:sequence>\r\n      <xs:element name="DataTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="Name" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataTypeDescription" type="tns:DataTypeDescription" />\r\n\r\n  <xs:complexType name="ListOfDataTypeDescription">\r\n    <xs:sequence>\r\n      <xs:element name="DataTypeDescription" type="tns:DataTypeDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataTypeDescription" type="tns:ListOfDataTypeDescription" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="StructureDescription">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataTypeDescription">\r\n        <xs:sequence>\r\n          <xs:element name="StructureDefinition" type="tns:StructureDefinition" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="StructureDescription" type="tns:StructureDescription" />\r\n\r\n  <xs:complexType name="ListOfStructureDescription">\r\n    <xs:sequence>\r\n      <xs:element name="StructureDescription" type="tns:StructureDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfStructureDescription" type="tns:ListOfStructureDescription" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="EnumDescription">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataTypeDescription">\r\n        <xs:sequence>\r\n          <xs:element name="EnumDefinition" type="tns:EnumDefinition" minOccurs="0" nillable="true" />\r\n          <xs:element name="BuiltInType" type="xs:unsignedByte" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="EnumDescription" type="tns:EnumDescription" />\r\n\r\n  <xs:complexType name="ListOfEnumDescription">\r\n    <xs:sequence>\r\n      <xs:element name="EnumDescription" type="tns:EnumDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEnumDescription" type="tns:ListOfEnumDescription" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="SimpleTypeDescription">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataTypeDescription">\r\n        <xs:sequence>\r\n          <xs:element name="BaseDataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="BuiltInType" type="xs:unsignedByte" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="SimpleTypeDescription" type="tns:SimpleTypeDescription" />\r\n\r\n  <xs:complexType name="ListOfSimpleTypeDescription">\r\n    <xs:sequence>\r\n      <xs:element name="SimpleTypeDescription" type="tns:SimpleTypeDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSimpleTypeDescription" type="tns:ListOfSimpleTypeDescription" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="UABinaryFileDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataTypeSchemaHeader">\r\n        <xs:sequence>\r\n          <xs:element name="SchemaLocation" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="FileHeader" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n          <xs:element name="Body" type="ua:Variant" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UABinaryFileDataType" type="tns:UABinaryFileDataType" />\r\n\r\n  <xs:complexType name="ListOfUABinaryFileDataType">\r\n    <xs:sequence>\r\n      <xs:element name="UABinaryFileDataType" type="tns:UABinaryFileDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUABinaryFileDataType" type="tns:ListOfUABinaryFileDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="PubSubState">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Disabled_0" />\r\n      <xs:enumeration value="Paused_1" />\r\n      <xs:enumeration value="Operational_2" />\r\n      <xs:enumeration value="Error_3" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="PubSubState" type="tns:PubSubState" />\r\n\r\n  <xs:complexType name="ListOfPubSubState">\r\n    <xs:sequence>\r\n      <xs:element name="PubSubState" type="tns:PubSubState" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPubSubState" type="tns:ListOfPubSubState" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataSetMetaDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataTypeSchemaHeader">\r\n        <xs:sequence>\r\n          <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n          <xs:element name="Fields" type="tns:ListOfFieldMetaData" minOccurs="0" nillable="true" />\r\n          <xs:element name="DataSetClassId" type="ua:Guid" minOccurs="0" />\r\n          <xs:element name="ConfigurationVersion" type="tns:ConfigurationVersionDataType" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DataSetMetaDataType" type="tns:DataSetMetaDataType" />\r\n\r\n  <xs:complexType name="ListOfDataSetMetaDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetMetaDataType" type="tns:DataSetMetaDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetMetaDataType" type="tns:ListOfDataSetMetaDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="FieldMetaData">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="FieldFlags" type="tns:DataSetFieldFlags" minOccurs="0" />\r\n      <xs:element name="BuiltInType" type="xs:unsignedByte" minOccurs="0" />\r\n      <xs:element name="DataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ValueRank" type="xs:int" minOccurs="0" />\r\n      <xs:element name="ArrayDimensions" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n      <xs:element name="MaxStringLength" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DataSetFieldId" type="ua:Guid" minOccurs="0" />\r\n      <xs:element name="Properties" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="FieldMetaData" type="tns:FieldMetaData" />\r\n\r\n  <xs:complexType name="ListOfFieldMetaData">\r\n    <xs:sequence>\r\n      <xs:element name="FieldMetaData" type="tns:FieldMetaData" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfFieldMetaData" type="tns:ListOfFieldMetaData" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="DataSetFieldFlags">\r\n    <xs:restriction base="xs:unsignedShort">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="DataSetFieldFlags" type="tns:DataSetFieldFlags" />\r\n\r\n  <xs:complexType name="ConfigurationVersionDataType">\r\n    <xs:sequence>\r\n      <xs:element name="MajorVersion" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MinorVersion" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ConfigurationVersionDataType" type="tns:ConfigurationVersionDataType" />\r\n\r\n  <xs:complexType name="ListOfConfigurationVersionDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ConfigurationVersionDataType" type="tns:ConfigurationVersionDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfConfigurationVersionDataType" type="tns:ListOfConfigurationVersionDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PublishedDataSetDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataSetFolder" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataSetMetaData" type="tns:DataSetMetaDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="ExtensionFields" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataSetSource" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PublishedDataSetDataType" type="tns:PublishedDataSetDataType" />\r\n\r\n  <xs:complexType name="ListOfPublishedDataSetDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PublishedDataSetDataType" type="tns:PublishedDataSetDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPublishedDataSetDataType" type="tns:ListOfPublishedDataSetDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PublishedDataSetSourceDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PublishedDataSetSourceDataType" type="tns:PublishedDataSetSourceDataType" />\r\n\r\n  <xs:complexType name="ListOfPublishedDataSetSourceDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PublishedDataSetSourceDataType" type="tns:PublishedDataSetSourceDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPublishedDataSetSourceDataType" type="tns:ListOfPublishedDataSetSourceDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PublishedVariableDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PublishedVariable" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SamplingIntervalHint" type="xs:double" minOccurs="0" />\r\n      <xs:element name="DeadbandType" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DeadbandValue" type="xs:double" minOccurs="0" />\r\n      <xs:element name="IndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubstituteValue" type="ua:Variant" minOccurs="0" />\r\n      <xs:element name="MetaDataProperties" type="ua:ListOfQualifiedName" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PublishedVariableDataType" type="tns:PublishedVariableDataType" />\r\n\r\n  <xs:complexType name="ListOfPublishedVariableDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PublishedVariableDataType" type="tns:PublishedVariableDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPublishedVariableDataType" type="tns:ListOfPublishedVariableDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PublishedDataItemsDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:PublishedDataSetSourceDataType">\r\n        <xs:sequence>\r\n          <xs:element name="PublishedData" type="tns:ListOfPublishedVariableDataType" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="PublishedDataItemsDataType" type="tns:PublishedDataItemsDataType" />\r\n\r\n  <xs:complexType name="ListOfPublishedDataItemsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PublishedDataItemsDataType" type="tns:PublishedDataItemsDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPublishedDataItemsDataType" type="tns:ListOfPublishedDataItemsDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PublishedEventsDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:PublishedDataSetSourceDataType">\r\n        <xs:sequence>\r\n          <xs:element name="EventNotifier" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="SelectedFields" type="tns:ListOfSimpleAttributeOperand" minOccurs="0" nillable="true" />\r\n          <xs:element name="Filter" type="tns:ContentFilter" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="PublishedEventsDataType" type="tns:PublishedEventsDataType" />\r\n\r\n  <xs:complexType name="ListOfPublishedEventsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PublishedEventsDataType" type="tns:PublishedEventsDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPublishedEventsDataType" type="tns:ListOfPublishedEventsDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="DataSetFieldContentMask">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="DataSetFieldContentMask" type="tns:DataSetFieldContentMask" />\r\n\r\n  <xs:complexType name="ListOfDataSetFieldContentMask">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetFieldContentMask" type="tns:DataSetFieldContentMask" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetFieldContentMask" type="tns:ListOfDataSetFieldContentMask" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataSetWriterDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Enabled" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="DataSetWriterId" type="xs:unsignedShort" minOccurs="0" />\r\n      <xs:element name="DataSetFieldContentMask" type="tns:DataSetFieldContentMask" minOccurs="0" />\r\n      <xs:element name="KeyFrameCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DataSetName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataSetWriterProperties" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n      <xs:element name="TransportSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="MessageSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataSetWriterDataType" type="tns:DataSetWriterDataType" />\r\n\r\n  <xs:complexType name="ListOfDataSetWriterDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetWriterDataType" type="tns:DataSetWriterDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetWriterDataType" type="tns:ListOfDataSetWriterDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataSetWriterTransportDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataSetWriterTransportDataType" type="tns:DataSetWriterTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfDataSetWriterTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetWriterTransportDataType" type="tns:DataSetWriterTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetWriterTransportDataType" type="tns:ListOfDataSetWriterTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataSetWriterMessageDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataSetWriterMessageDataType" type="tns:DataSetWriterMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfDataSetWriterMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetWriterMessageDataType" type="tns:DataSetWriterMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetWriterMessageDataType" type="tns:ListOfDataSetWriterMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PubSubGroupDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Enabled" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="SecurityMode" type="tns:MessageSecurityMode" minOccurs="0" />\r\n      <xs:element name="SecurityGroupId" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityKeyServices" type="tns:ListOfEndpointDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="MaxNetworkMessageSize" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="GroupProperties" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PubSubGroupDataType" type="tns:PubSubGroupDataType" />\r\n\r\n  <xs:complexType name="ListOfPubSubGroupDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PubSubGroupDataType" type="tns:PubSubGroupDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPubSubGroupDataType" type="tns:ListOfPubSubGroupDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="WriterGroupDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:PubSubGroupDataType">\r\n        <xs:sequence>\r\n          <xs:element name="WriterGroupId" type="xs:unsignedShort" minOccurs="0" />\r\n          <xs:element name="PublishingInterval" type="xs:double" minOccurs="0" />\r\n          <xs:element name="KeepAliveTime" type="xs:double" minOccurs="0" />\r\n          <xs:element name="Priority" type="xs:unsignedByte" minOccurs="0" />\r\n          <xs:element name="LocaleIds" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n          <xs:element name="HeaderLayoutUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="TransportSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n          <xs:element name="MessageSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n          <xs:element name="DataSetWriters" type="tns:ListOfDataSetWriterDataType" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="WriterGroupDataType" type="tns:WriterGroupDataType" />\r\n\r\n  <xs:complexType name="ListOfWriterGroupDataType">\r\n    <xs:sequence>\r\n      <xs:element name="WriterGroupDataType" type="tns:WriterGroupDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfWriterGroupDataType" type="tns:ListOfWriterGroupDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="WriterGroupTransportDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="WriterGroupTransportDataType" type="tns:WriterGroupTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfWriterGroupTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="WriterGroupTransportDataType" type="tns:WriterGroupTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfWriterGroupTransportDataType" type="tns:ListOfWriterGroupTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="WriterGroupMessageDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="WriterGroupMessageDataType" type="tns:WriterGroupMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfWriterGroupMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="WriterGroupMessageDataType" type="tns:WriterGroupMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfWriterGroupMessageDataType" type="tns:ListOfWriterGroupMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PubSubConnectionDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Enabled" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="PublisherId" type="ua:Variant" minOccurs="0" />\r\n      <xs:element name="TransportProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Address" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="ConnectionProperties" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n      <xs:element name="TransportSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="WriterGroups" type="tns:ListOfWriterGroupDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReaderGroups" type="tns:ListOfReaderGroupDataType" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PubSubConnectionDataType" type="tns:PubSubConnectionDataType" />\r\n\r\n  <xs:complexType name="ListOfPubSubConnectionDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PubSubConnectionDataType" type="tns:PubSubConnectionDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPubSubConnectionDataType" type="tns:ListOfPubSubConnectionDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ConnectionTransportDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ConnectionTransportDataType" type="tns:ConnectionTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfConnectionTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ConnectionTransportDataType" type="tns:ConnectionTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfConnectionTransportDataType" type="tns:ListOfConnectionTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="NetworkAddressDataType">\r\n    <xs:sequence>\r\n      <xs:element name="NetworkInterface" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NetworkAddressDataType" type="tns:NetworkAddressDataType" />\r\n\r\n  <xs:complexType name="ListOfNetworkAddressDataType">\r\n    <xs:sequence>\r\n      <xs:element name="NetworkAddressDataType" type="tns:NetworkAddressDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfNetworkAddressDataType" type="tns:ListOfNetworkAddressDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="NetworkAddressUrlDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NetworkAddressDataType">\r\n        <xs:sequence>\r\n          <xs:element name="Url" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="NetworkAddressUrlDataType" type="tns:NetworkAddressUrlDataType" />\r\n\r\n  <xs:complexType name="ListOfNetworkAddressUrlDataType">\r\n    <xs:sequence>\r\n      <xs:element name="NetworkAddressUrlDataType" type="tns:NetworkAddressUrlDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfNetworkAddressUrlDataType" type="tns:ListOfNetworkAddressUrlDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ReaderGroupDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:PubSubGroupDataType">\r\n        <xs:sequence>\r\n          <xs:element name="TransportSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n          <xs:element name="MessageSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n          <xs:element name="DataSetReaders" type="tns:ListOfDataSetReaderDataType" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReaderGroupDataType" type="tns:ReaderGroupDataType" />\r\n\r\n  <xs:complexType name="ListOfReaderGroupDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ReaderGroupDataType" type="tns:ReaderGroupDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfReaderGroupDataType" type="tns:ListOfReaderGroupDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ReaderGroupTransportDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ReaderGroupTransportDataType" type="tns:ReaderGroupTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfReaderGroupTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ReaderGroupTransportDataType" type="tns:ReaderGroupTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfReaderGroupTransportDataType" type="tns:ListOfReaderGroupTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ReaderGroupMessageDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ReaderGroupMessageDataType" type="tns:ReaderGroupMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfReaderGroupMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ReaderGroupMessageDataType" type="tns:ReaderGroupMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfReaderGroupMessageDataType" type="tns:ListOfReaderGroupMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataSetReaderDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Enabled" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="PublisherId" type="ua:Variant" minOccurs="0" />\r\n      <xs:element name="WriterGroupId" type="xs:unsignedShort" minOccurs="0" />\r\n      <xs:element name="DataSetWriterId" type="xs:unsignedShort" minOccurs="0" />\r\n      <xs:element name="DataSetMetaData" type="tns:DataSetMetaDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataSetFieldContentMask" type="tns:DataSetFieldContentMask" minOccurs="0" />\r\n      <xs:element name="MessageReceiveTimeout" type="xs:double" minOccurs="0" />\r\n      <xs:element name="KeyFrameCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="HeaderLayoutUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityMode" type="tns:MessageSecurityMode" minOccurs="0" />\r\n      <xs:element name="SecurityGroupId" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityKeyServices" type="tns:ListOfEndpointDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataSetReaderProperties" type="tns:ListOfKeyValuePair" minOccurs="0" nillable="true" />\r\n      <xs:element name="TransportSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="MessageSettings" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscribedDataSet" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataSetReaderDataType" type="tns:DataSetReaderDataType" />\r\n\r\n  <xs:complexType name="ListOfDataSetReaderDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetReaderDataType" type="tns:DataSetReaderDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetReaderDataType" type="tns:ListOfDataSetReaderDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataSetReaderTransportDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataSetReaderTransportDataType" type="tns:DataSetReaderTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfDataSetReaderTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetReaderTransportDataType" type="tns:DataSetReaderTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetReaderTransportDataType" type="tns:ListOfDataSetReaderTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataSetReaderMessageDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataSetReaderMessageDataType" type="tns:DataSetReaderMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfDataSetReaderMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetReaderMessageDataType" type="tns:DataSetReaderMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetReaderMessageDataType" type="tns:ListOfDataSetReaderMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="SubscribedDataSetDataType">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SubscribedDataSetDataType" type="tns:SubscribedDataSetDataType" />\r\n\r\n  <xs:complexType name="ListOfSubscribedDataSetDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SubscribedDataSetDataType" type="tns:SubscribedDataSetDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSubscribedDataSetDataType" type="tns:ListOfSubscribedDataSetDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="TargetVariablesDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:SubscribedDataSetDataType">\r\n        <xs:sequence>\r\n          <xs:element name="TargetVariables" type="tns:ListOfFieldTargetDataType" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="TargetVariablesDataType" type="tns:TargetVariablesDataType" />\r\n\r\n  <xs:complexType name="ListOfTargetVariablesDataType">\r\n    <xs:sequence>\r\n      <xs:element name="TargetVariablesDataType" type="tns:TargetVariablesDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfTargetVariablesDataType" type="tns:ListOfTargetVariablesDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="FieldTargetDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetFieldId" type="ua:Guid" minOccurs="0" />\r\n      <xs:element name="ReceiverIndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="TargetNodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="WriteIndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="OverrideValueHandling" type="tns:OverrideValueHandling" minOccurs="0" />\r\n      <xs:element name="OverrideValue" type="ua:Variant" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="FieldTargetDataType" type="tns:FieldTargetDataType" />\r\n\r\n  <xs:complexType name="ListOfFieldTargetDataType">\r\n    <xs:sequence>\r\n      <xs:element name="FieldTargetDataType" type="tns:FieldTargetDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfFieldTargetDataType" type="tns:ListOfFieldTargetDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="OverrideValueHandling">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Disabled_0" />\r\n      <xs:enumeration value="LastUsableValue_1" />\r\n      <xs:enumeration value="OverrideValue_2" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="OverrideValueHandling" type="tns:OverrideValueHandling" />\r\n\r\n  <xs:complexType name="ListOfOverrideValueHandling">\r\n    <xs:sequence>\r\n      <xs:element name="OverrideValueHandling" type="tns:OverrideValueHandling" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfOverrideValueHandling" type="tns:ListOfOverrideValueHandling" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="SubscribedDataSetMirrorDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:SubscribedDataSetDataType">\r\n        <xs:sequence>\r\n          <xs:element name="ParentNodeName" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="RolePermissions" type="tns:ListOfRolePermissionType" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="SubscribedDataSetMirrorDataType" type="tns:SubscribedDataSetMirrorDataType" />\r\n\r\n  <xs:complexType name="ListOfSubscribedDataSetMirrorDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SubscribedDataSetMirrorDataType" type="tns:SubscribedDataSetMirrorDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSubscribedDataSetMirrorDataType" type="tns:ListOfSubscribedDataSetMirrorDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PubSubConfigurationDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PublishedDataSets" type="tns:ListOfPublishedDataSetDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="Connections" type="tns:ListOfPubSubConnectionDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="Enabled" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PubSubConfigurationDataType" type="tns:PubSubConfigurationDataType" />\r\n\r\n  <xs:complexType name="ListOfPubSubConfigurationDataType">\r\n    <xs:sequence>\r\n      <xs:element name="PubSubConfigurationDataType" type="tns:PubSubConfigurationDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPubSubConfigurationDataType" type="tns:ListOfPubSubConfigurationDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="DataSetOrderingType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Undefined_0" />\r\n      <xs:enumeration value="AscendingWriterId_1" />\r\n      <xs:enumeration value="AscendingWriterIdSingle_2" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="DataSetOrderingType" type="tns:DataSetOrderingType" />\r\n\r\n  <xs:complexType name="ListOfDataSetOrderingType">\r\n    <xs:sequence>\r\n      <xs:element name="DataSetOrderingType" type="tns:DataSetOrderingType" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataSetOrderingType" type="tns:ListOfDataSetOrderingType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="UadpNetworkMessageContentMask">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="UadpNetworkMessageContentMask" type="tns:UadpNetworkMessageContentMask" />\r\n\r\n  <xs:complexType name="ListOfUadpNetworkMessageContentMask">\r\n    <xs:sequence>\r\n      <xs:element name="UadpNetworkMessageContentMask" type="tns:UadpNetworkMessageContentMask" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUadpNetworkMessageContentMask" type="tns:ListOfUadpNetworkMessageContentMask" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="UadpWriterGroupMessageDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:WriterGroupMessageDataType">\r\n        <xs:sequence>\r\n          <xs:element name="GroupVersion" type="xs:unsignedInt" minOccurs="0" />\r\n          <xs:element name="DataSetOrdering" type="tns:DataSetOrderingType" minOccurs="0" />\r\n          <xs:element name="NetworkMessageContentMask" type="tns:UadpNetworkMessageContentMask" minOccurs="0" />\r\n          <xs:element name="SamplingOffset" type="xs:double" minOccurs="0" />\r\n          <xs:element name="PublishingOffset" type="ua:ListOfDouble" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UadpWriterGroupMessageDataType" type="tns:UadpWriterGroupMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfUadpWriterGroupMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="UadpWriterGroupMessageDataType" type="tns:UadpWriterGroupMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUadpWriterGroupMessageDataType" type="tns:ListOfUadpWriterGroupMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="UadpDataSetMessageContentMask">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="UadpDataSetMessageContentMask" type="tns:UadpDataSetMessageContentMask" />\r\n\r\n  <xs:complexType name="ListOfUadpDataSetMessageContentMask">\r\n    <xs:sequence>\r\n      <xs:element name="UadpDataSetMessageContentMask" type="tns:UadpDataSetMessageContentMask" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUadpDataSetMessageContentMask" type="tns:ListOfUadpDataSetMessageContentMask" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="UadpDataSetWriterMessageDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataSetWriterMessageDataType">\r\n        <xs:sequence>\r\n          <xs:element name="DataSetMessageContentMask" type="tns:UadpDataSetMessageContentMask" minOccurs="0" />\r\n          <xs:element name="ConfiguredSize" type="xs:unsignedShort" minOccurs="0" />\r\n          <xs:element name="NetworkMessageNumber" type="xs:unsignedShort" minOccurs="0" />\r\n          <xs:element name="DataSetOffset" type="xs:unsignedShort" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UadpDataSetWriterMessageDataType" type="tns:UadpDataSetWriterMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfUadpDataSetWriterMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="UadpDataSetWriterMessageDataType" type="tns:UadpDataSetWriterMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUadpDataSetWriterMessageDataType" type="tns:ListOfUadpDataSetWriterMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="UadpDataSetReaderMessageDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataSetReaderMessageDataType">\r\n        <xs:sequence>\r\n          <xs:element name="GroupVersion" type="xs:unsignedInt" minOccurs="0" />\r\n          <xs:element name="NetworkMessageNumber" type="xs:unsignedShort" minOccurs="0" />\r\n          <xs:element name="DataSetOffset" type="xs:unsignedShort" minOccurs="0" />\r\n          <xs:element name="DataSetClassId" type="ua:Guid" minOccurs="0" />\r\n          <xs:element name="NetworkMessageContentMask" type="tns:UadpNetworkMessageContentMask" minOccurs="0" />\r\n          <xs:element name="DataSetMessageContentMask" type="tns:UadpDataSetMessageContentMask" minOccurs="0" />\r\n          <xs:element name="PublishingInterval" type="xs:double" minOccurs="0" />\r\n          <xs:element name="ReceiveOffset" type="xs:double" minOccurs="0" />\r\n          <xs:element name="ProcessingOffset" type="xs:double" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UadpDataSetReaderMessageDataType" type="tns:UadpDataSetReaderMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfUadpDataSetReaderMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="UadpDataSetReaderMessageDataType" type="tns:UadpDataSetReaderMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUadpDataSetReaderMessageDataType" type="tns:ListOfUadpDataSetReaderMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="JsonNetworkMessageContentMask">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="JsonNetworkMessageContentMask" type="tns:JsonNetworkMessageContentMask" />\r\n\r\n  <xs:complexType name="ListOfJsonNetworkMessageContentMask">\r\n    <xs:sequence>\r\n      <xs:element name="JsonNetworkMessageContentMask" type="tns:JsonNetworkMessageContentMask" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfJsonNetworkMessageContentMask" type="tns:ListOfJsonNetworkMessageContentMask" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="JsonWriterGroupMessageDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:WriterGroupMessageDataType">\r\n        <xs:sequence>\r\n          <xs:element name="NetworkMessageContentMask" type="tns:JsonNetworkMessageContentMask" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="JsonWriterGroupMessageDataType" type="tns:JsonWriterGroupMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfJsonWriterGroupMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="JsonWriterGroupMessageDataType" type="tns:JsonWriterGroupMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfJsonWriterGroupMessageDataType" type="tns:ListOfJsonWriterGroupMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="JsonDataSetMessageContentMask">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="JsonDataSetMessageContentMask" type="tns:JsonDataSetMessageContentMask" />\r\n\r\n  <xs:complexType name="ListOfJsonDataSetMessageContentMask">\r\n    <xs:sequence>\r\n      <xs:element name="JsonDataSetMessageContentMask" type="tns:JsonDataSetMessageContentMask" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfJsonDataSetMessageContentMask" type="tns:ListOfJsonDataSetMessageContentMask" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="JsonDataSetWriterMessageDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataSetWriterMessageDataType">\r\n        <xs:sequence>\r\n          <xs:element name="DataSetMessageContentMask" type="tns:JsonDataSetMessageContentMask" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="JsonDataSetWriterMessageDataType" type="tns:JsonDataSetWriterMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfJsonDataSetWriterMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="JsonDataSetWriterMessageDataType" type="tns:JsonDataSetWriterMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfJsonDataSetWriterMessageDataType" type="tns:ListOfJsonDataSetWriterMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="JsonDataSetReaderMessageDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataSetReaderMessageDataType">\r\n        <xs:sequence>\r\n          <xs:element name="NetworkMessageContentMask" type="tns:JsonNetworkMessageContentMask" minOccurs="0" />\r\n          <xs:element name="DataSetMessageContentMask" type="tns:JsonDataSetMessageContentMask" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="JsonDataSetReaderMessageDataType" type="tns:JsonDataSetReaderMessageDataType" />\r\n\r\n  <xs:complexType name="ListOfJsonDataSetReaderMessageDataType">\r\n    <xs:sequence>\r\n      <xs:element name="JsonDataSetReaderMessageDataType" type="tns:JsonDataSetReaderMessageDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfJsonDataSetReaderMessageDataType" type="tns:ListOfJsonDataSetReaderMessageDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DatagramConnectionTransportDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:ConnectionTransportDataType">\r\n        <xs:sequence>\r\n          <xs:element name="DiscoveryAddress" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DatagramConnectionTransportDataType" type="tns:DatagramConnectionTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfDatagramConnectionTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DatagramConnectionTransportDataType" type="tns:DatagramConnectionTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDatagramConnectionTransportDataType" type="tns:ListOfDatagramConnectionTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DatagramWriterGroupTransportDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:WriterGroupTransportDataType">\r\n        <xs:sequence>\r\n          <xs:element name="MessageRepeatCount" type="xs:unsignedByte" minOccurs="0" />\r\n          <xs:element name="MessageRepeatDelay" type="xs:double" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DatagramWriterGroupTransportDataType" type="tns:DatagramWriterGroupTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfDatagramWriterGroupTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="DatagramWriterGroupTransportDataType" type="tns:DatagramWriterGroupTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDatagramWriterGroupTransportDataType" type="tns:ListOfDatagramWriterGroupTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="BrokerConnectionTransportDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:ConnectionTransportDataType">\r\n        <xs:sequence>\r\n          <xs:element name="ResourceUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="AuthenticationProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="BrokerConnectionTransportDataType" type="tns:BrokerConnectionTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfBrokerConnectionTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="BrokerConnectionTransportDataType" type="tns:BrokerConnectionTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrokerConnectionTransportDataType" type="tns:ListOfBrokerConnectionTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="BrokerTransportQualityOfService">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="NotSpecified_0" />\r\n      <xs:enumeration value="BestEffort_1" />\r\n      <xs:enumeration value="AtLeastOnce_2" />\r\n      <xs:enumeration value="AtMostOnce_3" />\r\n      <xs:enumeration value="ExactlyOnce_4" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="BrokerTransportQualityOfService" type="tns:BrokerTransportQualityOfService" />\r\n\r\n  <xs:complexType name="ListOfBrokerTransportQualityOfService">\r\n    <xs:sequence>\r\n      <xs:element name="BrokerTransportQualityOfService" type="tns:BrokerTransportQualityOfService" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrokerTransportQualityOfService" type="tns:ListOfBrokerTransportQualityOfService" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="BrokerWriterGroupTransportDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:WriterGroupTransportDataType">\r\n        <xs:sequence>\r\n          <xs:element name="QueueName" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="ResourceUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="AuthenticationProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="RequestedDeliveryGuarantee" type="tns:BrokerTransportQualityOfService" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="BrokerWriterGroupTransportDataType" type="tns:BrokerWriterGroupTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfBrokerWriterGroupTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="BrokerWriterGroupTransportDataType" type="tns:BrokerWriterGroupTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrokerWriterGroupTransportDataType" type="tns:ListOfBrokerWriterGroupTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="BrokerDataSetWriterTransportDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataSetWriterTransportDataType">\r\n        <xs:sequence>\r\n          <xs:element name="QueueName" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="ResourceUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="AuthenticationProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="RequestedDeliveryGuarantee" type="tns:BrokerTransportQualityOfService" minOccurs="0" />\r\n          <xs:element name="MetaDataQueueName" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="MetaDataUpdateTime" type="xs:double" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="BrokerDataSetWriterTransportDataType" type="tns:BrokerDataSetWriterTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfBrokerDataSetWriterTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="BrokerDataSetWriterTransportDataType" type="tns:BrokerDataSetWriterTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrokerDataSetWriterTransportDataType" type="tns:ListOfBrokerDataSetWriterTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="BrokerDataSetReaderTransportDataType">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataSetReaderTransportDataType">\r\n        <xs:sequence>\r\n          <xs:element name="QueueName" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="ResourceUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="AuthenticationProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="RequestedDeliveryGuarantee" type="tns:BrokerTransportQualityOfService" minOccurs="0" />\r\n          <xs:element name="MetaDataQueueName" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="BrokerDataSetReaderTransportDataType" type="tns:BrokerDataSetReaderTransportDataType" />\r\n\r\n  <xs:complexType name="ListOfBrokerDataSetReaderTransportDataType">\r\n    <xs:sequence>\r\n      <xs:element name="BrokerDataSetReaderTransportDataType" type="tns:BrokerDataSetReaderTransportDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrokerDataSetReaderTransportDataType" type="tns:ListOfBrokerDataSetReaderTransportDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="DiagnosticsLevel">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Basic_0" />\r\n      <xs:enumeration value="Advanced_1" />\r\n      <xs:enumeration value="Info_2" />\r\n      <xs:enumeration value="Log_3" />\r\n      <xs:enumeration value="Debug_4" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="DiagnosticsLevel" type="tns:DiagnosticsLevel" />\r\n\r\n  <xs:complexType name="ListOfDiagnosticsLevel">\r\n    <xs:sequence>\r\n      <xs:element name="DiagnosticsLevel" type="tns:DiagnosticsLevel" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDiagnosticsLevel" type="tns:ListOfDiagnosticsLevel" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="PubSubDiagnosticsCounterClassification">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Information_0" />\r\n      <xs:enumeration value="Error_1" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="PubSubDiagnosticsCounterClassification" type="tns:PubSubDiagnosticsCounterClassification" />\r\n\r\n  <xs:complexType name="ListOfPubSubDiagnosticsCounterClassification">\r\n    <xs:sequence>\r\n      <xs:element name="PubSubDiagnosticsCounterClassification" type="tns:PubSubDiagnosticsCounterClassification" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfPubSubDiagnosticsCounterClassification" type="tns:ListOfPubSubDiagnosticsCounterClassification" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="AliasNameDataType">\r\n    <xs:sequence>\r\n      <xs:element name="AliasName" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReferencedNodes" type="ua:ListOfExpandedNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AliasNameDataType" type="tns:AliasNameDataType" />\r\n\r\n  <xs:complexType name="ListOfAliasNameDataType">\r\n    <xs:sequence>\r\n      <xs:element name="AliasNameDataType" type="tns:AliasNameDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfAliasNameDataType" type="tns:ListOfAliasNameDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="IdType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Numeric_0" />\r\n      <xs:enumeration value="String_1" />\r\n      <xs:enumeration value="Guid_2" />\r\n      <xs:enumeration value="Opaque_3" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="IdType" type="tns:IdType" />\r\n\r\n  <xs:complexType name="ListOfIdType">\r\n    <xs:sequence>\r\n      <xs:element name="IdType" type="tns:IdType" minOccurs="0" maxOccurs="unbounded" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfIdType" type="tns:ListOfIdType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="NodeClass">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Unspecified_0" />\r\n      <xs:enumeration value="Object_1" />\r\n      <xs:enumeration value="Variable_2" />\r\n      <xs:enumeration value="Method_4" />\r\n      <xs:enumeration value="ObjectType_8" />\r\n      <xs:enumeration value="VariableType_16" />\r\n      <xs:enumeration value="ReferenceType_32" />\r\n      <xs:enumeration value="DataType_64" />\r\n      <xs:enumeration value="View_128" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="NodeClass" type="tns:NodeClass" />\r\n\r\n  <xs:simpleType  name="PermissionType">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="PermissionType" type="tns:PermissionType" />\r\n\r\n  <xs:simpleType  name="AccessLevelType">\r\n    <xs:restriction base="xs:unsignedByte">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="AccessLevelType" type="tns:AccessLevelType" />\r\n\r\n  <xs:simpleType  name="AccessLevelExType">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="AccessLevelExType" type="tns:AccessLevelExType" />\r\n\r\n  <xs:simpleType  name="EventNotifierType">\r\n    <xs:restriction base="xs:unsignedByte">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="EventNotifierType" type="tns:EventNotifierType" />\r\n\r\n  <xs:complexType name="RolePermissionType">\r\n    <xs:sequence>\r\n      <xs:element name="RoleId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="Permissions" type="tns:PermissionType" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RolePermissionType" type="tns:RolePermissionType" />\r\n\r\n  <xs:complexType name="ListOfRolePermissionType">\r\n    <xs:sequence>\r\n      <xs:element name="RolePermissionType" type="tns:RolePermissionType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfRolePermissionType" type="tns:ListOfRolePermissionType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DataTypeDefinition">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DataTypeDefinition" type="tns:DataTypeDefinition" />\r\n\r\n  <xs:complexType name="ListOfDataTypeDefinition">\r\n    <xs:sequence>\r\n      <xs:element name="DataTypeDefinition" type="tns:DataTypeDefinition" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDataTypeDefinition" type="tns:ListOfDataTypeDefinition" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="StructureType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Structure_0" />\r\n      <xs:enumeration value="StructureWithOptionalFields_1" />\r\n      <xs:enumeration value="Union_2" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="StructureType" type="tns:StructureType" />\r\n\r\n  <xs:complexType name="StructureField">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ValueRank" type="xs:int" minOccurs="0" />\r\n      <xs:element name="ArrayDimensions" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n      <xs:element name="MaxStringLength" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="IsOptional" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="StructureField" type="tns:StructureField" />\r\n\r\n  <xs:complexType name="ListOfStructureField">\r\n    <xs:sequence>\r\n      <xs:element name="StructureField" type="tns:StructureField" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfStructureField" type="tns:ListOfStructureField" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="StructureDefinition">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataTypeDefinition">\r\n        <xs:sequence>\r\n          <xs:element name="DefaultEncodingId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="BaseDataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="StructureType" type="tns:StructureType" minOccurs="0" />\r\n          <xs:element name="Fields" type="tns:ListOfStructureField" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="StructureDefinition" type="tns:StructureDefinition" />\r\n\r\n  <xs:complexType name="ListOfStructureDefinition">\r\n    <xs:sequence>\r\n      <xs:element name="StructureDefinition" type="tns:StructureDefinition" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfStructureDefinition" type="tns:ListOfStructureDefinition" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="EnumDefinition">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DataTypeDefinition">\r\n        <xs:sequence>\r\n          <xs:element name="Fields" type="tns:ListOfEnumField" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="EnumDefinition" type="tns:EnumDefinition" />\r\n\r\n  <xs:complexType name="ListOfEnumDefinition">\r\n    <xs:sequence>\r\n      <xs:element name="EnumDefinition" type="tns:EnumDefinition" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEnumDefinition" type="tns:ListOfEnumDefinition" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Node">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodeClass" type="tns:NodeClass" minOccurs="0" />\r\n      <xs:element name="BrowseName" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n      <xs:element name="DisplayName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="WriteMask" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="UserWriteMask" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RolePermissions" type="tns:ListOfRolePermissionType" minOccurs="0" nillable="true" />\r\n      <xs:element name="UserRolePermissions" type="tns:ListOfRolePermissionType" minOccurs="0" nillable="true" />\r\n      <xs:element name="AccessRestrictions" type="xs:unsignedShort" minOccurs="0" />\r\n      <xs:element name="References" type="tns:ListOfReferenceNode" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Node" type="tns:Node" />\r\n\r\n  <xs:complexType name="ListOfNode">\r\n    <xs:sequence>\r\n      <xs:element name="Node" type="tns:Node" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfNode" type="tns:ListOfNode" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="InstanceNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:Node">\r\n        <xs:sequence>\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="InstanceNode" type="tns:InstanceNode" />\r\n\r\n  <xs:complexType name="TypeNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:Node">\r\n        <xs:sequence>\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="TypeNode" type="tns:TypeNode" />\r\n\r\n  <xs:complexType name="ObjectNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:InstanceNode">\r\n        <xs:sequence>\r\n          <xs:element name="EventNotifier" type="xs:unsignedByte" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ObjectNode" type="tns:ObjectNode" />\r\n\r\n  <xs:complexType name="ObjectTypeNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:TypeNode">\r\n        <xs:sequence>\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ObjectTypeNode" type="tns:ObjectTypeNode" />\r\n\r\n  <xs:complexType name="VariableNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:InstanceNode">\r\n        <xs:sequence>\r\n          <xs:element name="Value" type="ua:Variant" minOccurs="0" />\r\n          <xs:element name="DataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="ValueRank" type="xs:int" minOccurs="0" />\r\n          <xs:element name="ArrayDimensions" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n          <xs:element name="AccessLevel" type="xs:unsignedByte" minOccurs="0" />\r\n          <xs:element name="UserAccessLevel" type="xs:unsignedByte" minOccurs="0" />\r\n          <xs:element name="MinimumSamplingInterval" type="xs:double" minOccurs="0" />\r\n          <xs:element name="Historizing" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="AccessLevelEx" type="xs:unsignedInt" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="VariableNode" type="tns:VariableNode" />\r\n\r\n  <xs:complexType name="VariableTypeNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:TypeNode">\r\n        <xs:sequence>\r\n          <xs:element name="Value" type="ua:Variant" minOccurs="0" />\r\n          <xs:element name="DataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="ValueRank" type="xs:int" minOccurs="0" />\r\n          <xs:element name="ArrayDimensions" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="VariableTypeNode" type="tns:VariableTypeNode" />\r\n\r\n  <xs:complexType name="ReferenceTypeNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:TypeNode">\r\n        <xs:sequence>\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="Symmetric" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="InverseName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReferenceTypeNode" type="tns:ReferenceTypeNode" />\r\n\r\n  <xs:complexType name="MethodNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:InstanceNode">\r\n        <xs:sequence>\r\n          <xs:element name="Executable" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="UserExecutable" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="MethodNode" type="tns:MethodNode" />\r\n\r\n  <xs:complexType name="ViewNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:InstanceNode">\r\n        <xs:sequence>\r\n          <xs:element name="ContainsNoLoops" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="EventNotifier" type="xs:unsignedByte" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ViewNode" type="tns:ViewNode" />\r\n\r\n  <xs:complexType name="DataTypeNode">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:TypeNode">\r\n        <xs:sequence>\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="DataTypeDefinition" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DataTypeNode" type="tns:DataTypeNode" />\r\n\r\n  <xs:complexType name="ReferenceNode">\r\n    <xs:sequence>\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IsInverse" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="TargetId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ReferenceNode" type="tns:ReferenceNode" />\r\n\r\n  <xs:complexType name="ListOfReferenceNode">\r\n    <xs:sequence>\r\n      <xs:element name="ReferenceNode" type="tns:ReferenceNode" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfReferenceNode" type="tns:ListOfReferenceNode" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Argument">\r\n    <xs:sequence>\r\n      <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ValueRank" type="xs:int" minOccurs="0" />\r\n      <xs:element name="ArrayDimensions" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n      <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Argument" type="tns:Argument" />\r\n\r\n  <xs:complexType name="ListOfArgument">\r\n    <xs:sequence>\r\n      <xs:element name="Argument" type="tns:Argument" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfArgument" type="tns:ListOfArgument" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="EnumValueType">\r\n    <xs:sequence>\r\n      <xs:element name="Value" type="xs:long" minOccurs="0" />\r\n      <xs:element name="DisplayName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EnumValueType" type="tns:EnumValueType" />\r\n\r\n  <xs:complexType name="ListOfEnumValueType">\r\n    <xs:sequence>\r\n      <xs:element name="EnumValueType" type="tns:EnumValueType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEnumValueType" type="tns:ListOfEnumValueType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="EnumField">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:EnumValueType">\r\n        <xs:sequence>\r\n          <xs:element name="Name" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="EnumField" type="tns:EnumField" />\r\n\r\n  <xs:complexType name="ListOfEnumField">\r\n    <xs:sequence>\r\n      <xs:element name="EnumField" type="tns:EnumField" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEnumField" type="tns:ListOfEnumField" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="OptionSet">\r\n    <xs:sequence>\r\n      <xs:element name="Value" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="ValidBits" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="OptionSet" type="tns:OptionSet" />\r\n\r\n  <xs:complexType name="ListOfOptionSet">\r\n    <xs:sequence>\r\n      <xs:element name="OptionSet" type="tns:OptionSet" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfOptionSet" type="tns:ListOfOptionSet" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Union">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Union" type="tns:Union" />\r\n\r\n  <xs:complexType name="ListOfUnion">\r\n    <xs:sequence>\r\n      <xs:element name="Union" type="tns:Union" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUnion" type="tns:ListOfUnion" nillable="true"></xs:element>\r\n\r\n  <xs:element name="NormalizedString" type="xs:string" />\r\n\r\n  <xs:element name="DecimalString" type="xs:string" />\r\n\r\n  <xs:element name="DurationString" type="xs:string" />\r\n\r\n  <xs:element name="TimeString" type="xs:string" />\r\n\r\n  <xs:element name="DateString" type="xs:string" />\r\n\r\n  <xs:element name="Duration" type="xs:double" />\r\n\r\n  <xs:element name="UtcTime" type="xs:dateTime" />\r\n\r\n  <xs:element name="LocaleId" type="xs:string" />\r\n\r\n  <xs:complexType name="TimeZoneDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Offset" type="xs:short" minOccurs="0" />\r\n      <xs:element name="DaylightSavingInOffset" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="TimeZoneDataType" type="tns:TimeZoneDataType" />\r\n\r\n  <xs:complexType name="ListOfTimeZoneDataType">\r\n    <xs:sequence>\r\n      <xs:element name="TimeZoneDataType" type="tns:TimeZoneDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfTimeZoneDataType" type="tns:ListOfTimeZoneDataType" nillable="true"></xs:element>\r\n\r\n  <xs:element name="Index" type="xs:unsignedInt" />\r\n\r\n  <xs:element name="IntegerId" type="xs:unsignedInt" />\r\n\r\n  <xs:simpleType  name="ApplicationType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Server_0" />\r\n      <xs:enumeration value="Client_1" />\r\n      <xs:enumeration value="ClientAndServer_2" />\r\n      <xs:enumeration value="DiscoveryServer_3" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="ApplicationType" type="tns:ApplicationType" />\r\n\r\n  <xs:complexType name="ApplicationDescription">\r\n    <xs:sequence>\r\n      <xs:element name="ApplicationUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ProductUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ApplicationName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="ApplicationType" type="tns:ApplicationType" minOccurs="0" />\r\n      <xs:element name="GatewayServerUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiscoveryProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiscoveryUrls" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ApplicationDescription" type="tns:ApplicationDescription" />\r\n\r\n  <xs:complexType name="ListOfApplicationDescription">\r\n    <xs:sequence>\r\n      <xs:element name="ApplicationDescription" type="tns:ApplicationDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfApplicationDescription" type="tns:ListOfApplicationDescription" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="RequestHeader">\r\n    <xs:sequence>\r\n      <xs:element name="AuthenticationToken" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="Timestamp" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="RequestHandle" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ReturnDiagnostics" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="AuditEntryId" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="TimeoutHint" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="AdditionalHeader" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RequestHeader" type="tns:RequestHeader" />\r\n\r\n  <xs:complexType name="ResponseHeader">\r\n    <xs:sequence>\r\n      <xs:element name="Timestamp" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="RequestHandle" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ServiceResult" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="ServiceDiagnostics" type="ua:DiagnosticInfo" minOccurs="0" nillable="true" />\r\n      <xs:element name="StringTable" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="AdditionalHeader" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ResponseHeader" type="tns:ResponseHeader" />\r\n\r\n  <xs:element name="VersionTime" type="xs:unsignedInt" />\r\n\r\n  <xs:complexType name="ServiceFault">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ServiceFault" type="tns:ServiceFault" />\r\n\r\n  <xs:complexType name="SessionlessInvokeRequestType">\r\n    <xs:sequence>\r\n      <xs:element name="UrisVersion" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="NamespaceUris" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerUris" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="LocaleIds" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServiceId" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SessionlessInvokeRequestType" type="tns:SessionlessInvokeRequestType" />\r\n\r\n  <xs:complexType name="SessionlessInvokeResponseType">\r\n    <xs:sequence>\r\n      <xs:element name="NamespaceUris" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerUris" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServiceId" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SessionlessInvokeResponseType" type="tns:SessionlessInvokeResponseType" />\r\n\r\n  <xs:complexType name="FindServersRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="EndpointUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="LocaleIds" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerUris" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="FindServersRequest" type="tns:FindServersRequest" />\r\n\r\n  <xs:complexType name="FindServersResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Servers" type="tns:ListOfApplicationDescription" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="FindServersResponse" type="tns:FindServersResponse" />\r\n\r\n  <xs:complexType name="ServerOnNetwork">\r\n    <xs:sequence>\r\n      <xs:element name="RecordId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ServerName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiscoveryUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerCapabilities" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ServerOnNetwork" type="tns:ServerOnNetwork" />\r\n\r\n  <xs:complexType name="ListOfServerOnNetwork">\r\n    <xs:sequence>\r\n      <xs:element name="ServerOnNetwork" type="tns:ServerOnNetwork" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfServerOnNetwork" type="tns:ListOfServerOnNetwork" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="FindServersOnNetworkRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="StartingRecordId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MaxRecordsToReturn" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ServerCapabilityFilter" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="FindServersOnNetworkRequest" type="tns:FindServersOnNetworkRequest" />\r\n\r\n  <xs:complexType name="FindServersOnNetworkResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastCounterResetTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="Servers" type="tns:ListOfServerOnNetwork" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="FindServersOnNetworkResponse" type="tns:FindServersOnNetworkResponse" />\r\n\r\n  <xs:element name="ApplicationInstanceCertificate" type="xs:base64Binary" />\r\n\r\n  <xs:simpleType  name="MessageSecurityMode">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Invalid_0" />\r\n      <xs:enumeration value="None_1" />\r\n      <xs:enumeration value="Sign_2" />\r\n      <xs:enumeration value="SignAndEncrypt_3" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="MessageSecurityMode" type="tns:MessageSecurityMode" />\r\n\r\n  <xs:simpleType  name="UserTokenType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Anonymous_0" />\r\n      <xs:enumeration value="UserName_1" />\r\n      <xs:enumeration value="Certificate_2" />\r\n      <xs:enumeration value="IssuedToken_3" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="UserTokenType" type="tns:UserTokenType" />\r\n\r\n  <xs:complexType name="UserTokenPolicy">\r\n    <xs:sequence>\r\n      <xs:element name="PolicyId" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="TokenType" type="tns:UserTokenType" minOccurs="0" />\r\n      <xs:element name="IssuedTokenType" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="IssuerEndpointUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityPolicyUri" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="UserTokenPolicy" type="tns:UserTokenPolicy" />\r\n\r\n  <xs:complexType name="ListOfUserTokenPolicy">\r\n    <xs:sequence>\r\n      <xs:element name="UserTokenPolicy" type="tns:UserTokenPolicy" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfUserTokenPolicy" type="tns:ListOfUserTokenPolicy" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="EndpointDescription">\r\n    <xs:sequence>\r\n      <xs:element name="EndpointUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Server" type="tns:ApplicationDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerCertificate" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityMode" type="tns:MessageSecurityMode" minOccurs="0" />\r\n      <xs:element name="SecurityPolicyUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="UserIdentityTokens" type="tns:ListOfUserTokenPolicy" minOccurs="0" nillable="true" />\r\n      <xs:element name="TransportProfileUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityLevel" type="xs:unsignedByte" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EndpointDescription" type="tns:EndpointDescription" />\r\n\r\n  <xs:complexType name="ListOfEndpointDescription">\r\n    <xs:sequence>\r\n      <xs:element name="EndpointDescription" type="tns:EndpointDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEndpointDescription" type="tns:ListOfEndpointDescription" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="GetEndpointsRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="EndpointUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="LocaleIds" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="ProfileUris" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="GetEndpointsRequest" type="tns:GetEndpointsRequest" />\r\n\r\n  <xs:complexType name="GetEndpointsResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Endpoints" type="tns:ListOfEndpointDescription" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="GetEndpointsResponse" type="tns:GetEndpointsResponse" />\r\n\r\n  <xs:complexType name="RegisteredServer">\r\n    <xs:sequence>\r\n      <xs:element name="ServerUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ProductUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerNames" type="ua:ListOfLocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerType" type="tns:ApplicationType" minOccurs="0" />\r\n      <xs:element name="GatewayServerUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiscoveryUrls" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="SemaphoreFilePath" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="IsOnline" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RegisteredServer" type="tns:RegisteredServer" />\r\n\r\n  <xs:complexType name="ListOfRegisteredServer">\r\n    <xs:sequence>\r\n      <xs:element name="RegisteredServer" type="tns:RegisteredServer" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfRegisteredServer" type="tns:ListOfRegisteredServer" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="RegisterServerRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Server" type="tns:RegisteredServer" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RegisterServerRequest" type="tns:RegisterServerRequest" />\r\n\r\n  <xs:complexType name="RegisterServerResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RegisterServerResponse" type="tns:RegisterServerResponse" />\r\n\r\n  <xs:complexType name="DiscoveryConfiguration">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DiscoveryConfiguration" type="tns:DiscoveryConfiguration" />\r\n\r\n  <xs:complexType name="MdnsDiscoveryConfiguration">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:DiscoveryConfiguration">\r\n        <xs:sequence>\r\n          <xs:element name="MdnsServerName" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="ServerCapabilities" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="MdnsDiscoveryConfiguration" type="tns:MdnsDiscoveryConfiguration" />\r\n\r\n  <xs:complexType name="RegisterServer2Request">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Server" type="tns:RegisteredServer" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiscoveryConfiguration" type="ua:ListOfExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RegisterServer2Request" type="tns:RegisterServer2Request" />\r\n\r\n  <xs:complexType name="RegisterServer2Response">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ConfigurationResults" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RegisterServer2Response" type="tns:RegisterServer2Response" />\r\n\r\n  <xs:simpleType  name="SecurityTokenRequestType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Issue_0" />\r\n      <xs:enumeration value="Renew_1" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="SecurityTokenRequestType" type="tns:SecurityTokenRequestType" />\r\n\r\n  <xs:complexType name="ChannelSecurityToken">\r\n    <xs:sequence>\r\n      <xs:element name="ChannelId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TokenId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CreatedAt" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="RevisedLifetime" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ChannelSecurityToken" type="tns:ChannelSecurityToken" />\r\n\r\n  <xs:complexType name="OpenSecureChannelRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientProtocolVersion" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RequestType" type="tns:SecurityTokenRequestType" minOccurs="0" />\r\n      <xs:element name="SecurityMode" type="tns:MessageSecurityMode" minOccurs="0" />\r\n      <xs:element name="ClientNonce" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="RequestedLifetime" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="OpenSecureChannelRequest" type="tns:OpenSecureChannelRequest" />\r\n\r\n  <xs:complexType name="OpenSecureChannelResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerProtocolVersion" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SecurityToken" type="tns:ChannelSecurityToken" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerNonce" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="OpenSecureChannelResponse" type="tns:OpenSecureChannelResponse" />\r\n\r\n  <xs:complexType name="CloseSecureChannelRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CloseSecureChannelRequest" type="tns:CloseSecureChannelRequest" />\r\n\r\n  <xs:complexType name="CloseSecureChannelResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CloseSecureChannelResponse" type="tns:CloseSecureChannelResponse" />\r\n\r\n  <xs:complexType name="SignedSoftwareCertificate">\r\n    <xs:sequence>\r\n      <xs:element name="CertificateData" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="Signature" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SignedSoftwareCertificate" type="tns:SignedSoftwareCertificate" />\r\n\r\n  <xs:complexType name="ListOfSignedSoftwareCertificate">\r\n    <xs:sequence>\r\n      <xs:element name="SignedSoftwareCertificate" type="tns:SignedSoftwareCertificate" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSignedSoftwareCertificate" type="tns:ListOfSignedSoftwareCertificate" nillable="true"></xs:element>\r\n\r\n  <xs:element name="SessionAuthenticationToken" type="ua:NodeId" />\r\n\r\n  <xs:complexType name="SignatureData">\r\n    <xs:sequence>\r\n      <xs:element name="Algorithm" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Signature" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SignatureData" type="tns:SignatureData" />\r\n\r\n  <xs:complexType name="CreateSessionRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientDescription" type="tns:ApplicationDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="EndpointUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SessionName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientNonce" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientCertificate" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="RequestedSessionTimeout" type="xs:double" minOccurs="0" />\r\n      <xs:element name="MaxResponseMessageSize" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CreateSessionRequest" type="tns:CreateSessionRequest" />\r\n\r\n  <xs:complexType name="CreateSessionResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="AuthenticationToken" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="RevisedSessionTimeout" type="xs:double" minOccurs="0" />\r\n      <xs:element name="ServerNonce" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerCertificate" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerEndpoints" type="tns:ListOfEndpointDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerSoftwareCertificates" type="tns:ListOfSignedSoftwareCertificate" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerSignature" type="tns:SignatureData" minOccurs="0" nillable="true" />\r\n      <xs:element name="MaxRequestMessageSize" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CreateSessionResponse" type="tns:CreateSessionResponse" />\r\n\r\n  <xs:complexType name="UserIdentityToken">\r\n    <xs:sequence>\r\n      <xs:element name="PolicyId" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="UserIdentityToken" type="tns:UserIdentityToken" />\r\n\r\n  <xs:complexType name="AnonymousIdentityToken">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:UserIdentityToken">\r\n        <xs:sequence>\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="AnonymousIdentityToken" type="tns:AnonymousIdentityToken" />\r\n\r\n  <xs:complexType name="UserNameIdentityToken">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:UserIdentityToken">\r\n        <xs:sequence>\r\n          <xs:element name="UserName" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="Password" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n          <xs:element name="EncryptionAlgorithm" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UserNameIdentityToken" type="tns:UserNameIdentityToken" />\r\n\r\n  <xs:complexType name="X509IdentityToken">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:UserIdentityToken">\r\n        <xs:sequence>\r\n          <xs:element name="CertificateData" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="X509IdentityToken" type="tns:X509IdentityToken" />\r\n\r\n  <xs:complexType name="IssuedIdentityToken">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:UserIdentityToken">\r\n        <xs:sequence>\r\n          <xs:element name="TokenData" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n          <xs:element name="EncryptionAlgorithm" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="IssuedIdentityToken" type="tns:IssuedIdentityToken" />\r\n\r\n  <xs:element name="RsaEncryptedSecret" type="ua:Variant" />\r\n\r\n  <xs:element name="EccEncryptedSecret" type="ua:Variant" />\r\n\r\n  <xs:complexType name="ActivateSessionRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientSignature" type="tns:SignatureData" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientSoftwareCertificates" type="tns:ListOfSignedSoftwareCertificate" minOccurs="0" nillable="true" />\r\n      <xs:element name="LocaleIds" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="UserIdentityToken" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="UserTokenSignature" type="tns:SignatureData" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ActivateSessionRequest" type="tns:ActivateSessionRequest" />\r\n\r\n  <xs:complexType name="ActivateSessionResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerNonce" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ActivateSessionResponse" type="tns:ActivateSessionResponse" />\r\n\r\n  <xs:complexType name="CloseSessionRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="DeleteSubscriptions" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CloseSessionRequest" type="tns:CloseSessionRequest" />\r\n\r\n  <xs:complexType name="CloseSessionResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CloseSessionResponse" type="tns:CloseSessionResponse" />\r\n\r\n  <xs:complexType name="CancelRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="RequestHandle" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CancelRequest" type="tns:CancelRequest" />\r\n\r\n  <xs:complexType name="CancelResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="CancelCount" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CancelResponse" type="tns:CancelResponse" />\r\n\r\n  <xs:simpleType  name="NodeAttributesMask">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="None_0" />\r\n      <xs:enumeration value="AccessLevel_1" />\r\n      <xs:enumeration value="ArrayDimensions_2" />\r\n      <xs:enumeration value="BrowseName_4" />\r\n      <xs:enumeration value="ContainsNoLoops_8" />\r\n      <xs:enumeration value="DataType_16" />\r\n      <xs:enumeration value="Description_32" />\r\n      <xs:enumeration value="DisplayName_64" />\r\n      <xs:enumeration value="EventNotifier_128" />\r\n      <xs:enumeration value="Executable_256" />\r\n      <xs:enumeration value="Historizing_512" />\r\n      <xs:enumeration value="InverseName_1024" />\r\n      <xs:enumeration value="IsAbstract_2048" />\r\n      <xs:enumeration value="MinimumSamplingInterval_4096" />\r\n      <xs:enumeration value="NodeClass_8192" />\r\n      <xs:enumeration value="NodeId_16384" />\r\n      <xs:enumeration value="Symmetric_32768" />\r\n      <xs:enumeration value="UserAccessLevel_65536" />\r\n      <xs:enumeration value="UserExecutable_131072" />\r\n      <xs:enumeration value="UserWriteMask_262144" />\r\n      <xs:enumeration value="ValueRank_524288" />\r\n      <xs:enumeration value="WriteMask_1048576" />\r\n      <xs:enumeration value="Value_2097152" />\r\n      <xs:enumeration value="DataTypeDefinition_4194304" />\r\n      <xs:enumeration value="RolePermissions_8388608" />\r\n      <xs:enumeration value="AccessRestrictions_16777216" />\r\n      <xs:enumeration value="All_33554431" />\r\n      <xs:enumeration value="BaseNode_26501220" />\r\n      <xs:enumeration value="Object_26501348" />\r\n      <xs:enumeration value="ObjectType_26503268" />\r\n      <xs:enumeration value="Variable_26571383" />\r\n      <xs:enumeration value="VariableType_28600438" />\r\n      <xs:enumeration value="Method_26632548" />\r\n      <xs:enumeration value="ReferenceType_26537060" />\r\n      <xs:enumeration value="View_26501356" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="NodeAttributesMask" type="tns:NodeAttributesMask" />\r\n\r\n  <xs:complexType name="NodeAttributes">\r\n    <xs:sequence>\r\n      <xs:element name="SpecifiedAttributes" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DisplayName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="WriteMask" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="UserWriteMask" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NodeAttributes" type="tns:NodeAttributes" />\r\n\r\n  <xs:complexType name="ObjectAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="EventNotifier" type="xs:unsignedByte" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ObjectAttributes" type="tns:ObjectAttributes" />\r\n\r\n  <xs:complexType name="VariableAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="Value" type="ua:Variant" minOccurs="0" />\r\n          <xs:element name="DataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="ValueRank" type="xs:int" minOccurs="0" />\r\n          <xs:element name="ArrayDimensions" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n          <xs:element name="AccessLevel" type="xs:unsignedByte" minOccurs="0" />\r\n          <xs:element name="UserAccessLevel" type="xs:unsignedByte" minOccurs="0" />\r\n          <xs:element name="MinimumSamplingInterval" type="xs:double" minOccurs="0" />\r\n          <xs:element name="Historizing" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="VariableAttributes" type="tns:VariableAttributes" />\r\n\r\n  <xs:complexType name="MethodAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="Executable" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="UserExecutable" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="MethodAttributes" type="tns:MethodAttributes" />\r\n\r\n  <xs:complexType name="ObjectTypeAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ObjectTypeAttributes" type="tns:ObjectTypeAttributes" />\r\n\r\n  <xs:complexType name="VariableTypeAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="Value" type="ua:Variant" minOccurs="0" />\r\n          <xs:element name="DataType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="ValueRank" type="xs:int" minOccurs="0" />\r\n          <xs:element name="ArrayDimensions" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="VariableTypeAttributes" type="tns:VariableTypeAttributes" />\r\n\r\n  <xs:complexType name="ReferenceTypeAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="Symmetric" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="InverseName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReferenceTypeAttributes" type="tns:ReferenceTypeAttributes" />\r\n\r\n  <xs:complexType name="DataTypeAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="IsAbstract" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DataTypeAttributes" type="tns:DataTypeAttributes" />\r\n\r\n  <xs:complexType name="ViewAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="ContainsNoLoops" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="EventNotifier" type="xs:unsignedByte" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ViewAttributes" type="tns:ViewAttributes" />\r\n\r\n  <xs:complexType name="GenericAttributeValue">\r\n    <xs:sequence>\r\n      <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="Value" type="ua:Variant" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="GenericAttributeValue" type="tns:GenericAttributeValue" />\r\n\r\n  <xs:complexType name="ListOfGenericAttributeValue">\r\n    <xs:sequence>\r\n      <xs:element name="GenericAttributeValue" type="tns:GenericAttributeValue" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfGenericAttributeValue" type="tns:ListOfGenericAttributeValue" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="GenericAttributes">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NodeAttributes">\r\n        <xs:sequence>\r\n          <xs:element name="AttributeValues" type="tns:ListOfGenericAttributeValue" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="GenericAttributes" type="tns:GenericAttributes" />\r\n\r\n  <xs:complexType name="AddNodesItem">\r\n    <xs:sequence>\r\n      <xs:element name="ParentNodeId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="RequestedNewNodeId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="BrowseName" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodeClass" type="tns:NodeClass" minOccurs="0" />\r\n      <xs:element name="NodeAttributes" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="TypeDefinition" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AddNodesItem" type="tns:AddNodesItem" />\r\n\r\n  <xs:complexType name="ListOfAddNodesItem">\r\n    <xs:sequence>\r\n      <xs:element name="AddNodesItem" type="tns:AddNodesItem" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfAddNodesItem" type="tns:ListOfAddNodesItem" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="AddNodesResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="AddedNodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AddNodesResult" type="tns:AddNodesResult" />\r\n\r\n  <xs:complexType name="ListOfAddNodesResult">\r\n    <xs:sequence>\r\n      <xs:element name="AddNodesResult" type="tns:AddNodesResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfAddNodesResult" type="tns:ListOfAddNodesResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="AddNodesRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodesToAdd" type="tns:ListOfAddNodesItem" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AddNodesRequest" type="tns:AddNodesRequest" />\r\n\r\n  <xs:complexType name="AddNodesResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfAddNodesResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AddNodesResponse" type="tns:AddNodesResponse" />\r\n\r\n  <xs:complexType name="AddReferencesItem">\r\n    <xs:sequence>\r\n      <xs:element name="SourceNodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IsForward" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="TargetServerUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="TargetNodeId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="TargetNodeClass" type="tns:NodeClass" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AddReferencesItem" type="tns:AddReferencesItem" />\r\n\r\n  <xs:complexType name="ListOfAddReferencesItem">\r\n    <xs:sequence>\r\n      <xs:element name="AddReferencesItem" type="tns:AddReferencesItem" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfAddReferencesItem" type="tns:ListOfAddReferencesItem" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="AddReferencesRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReferencesToAdd" type="tns:ListOfAddReferencesItem" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AddReferencesRequest" type="tns:AddReferencesRequest" />\r\n\r\n  <xs:complexType name="AddReferencesResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AddReferencesResponse" type="tns:AddReferencesResponse" />\r\n\r\n  <xs:complexType name="DeleteNodesItem">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="DeleteTargetReferences" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteNodesItem" type="tns:DeleteNodesItem" />\r\n\r\n  <xs:complexType name="ListOfDeleteNodesItem">\r\n    <xs:sequence>\r\n      <xs:element name="DeleteNodesItem" type="tns:DeleteNodesItem" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDeleteNodesItem" type="tns:ListOfDeleteNodesItem" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DeleteNodesRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodesToDelete" type="tns:ListOfDeleteNodesItem" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteNodesRequest" type="tns:DeleteNodesRequest" />\r\n\r\n  <xs:complexType name="DeleteNodesResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteNodesResponse" type="tns:DeleteNodesResponse" />\r\n\r\n  <xs:complexType name="DeleteReferencesItem">\r\n    <xs:sequence>\r\n      <xs:element name="SourceNodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IsForward" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="TargetNodeId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="DeleteBidirectional" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteReferencesItem" type="tns:DeleteReferencesItem" />\r\n\r\n  <xs:complexType name="ListOfDeleteReferencesItem">\r\n    <xs:sequence>\r\n      <xs:element name="DeleteReferencesItem" type="tns:DeleteReferencesItem" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfDeleteReferencesItem" type="tns:ListOfDeleteReferencesItem" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="DeleteReferencesRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReferencesToDelete" type="tns:ListOfDeleteReferencesItem" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteReferencesRequest" type="tns:DeleteReferencesRequest" />\r\n\r\n  <xs:complexType name="DeleteReferencesResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteReferencesResponse" type="tns:DeleteReferencesResponse" />\r\n\r\n  <xs:simpleType  name="AttributeWriteMask">\r\n    <xs:restriction base="xs:unsignedInt">\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="AttributeWriteMask" type="tns:AttributeWriteMask" />\r\n\r\n  <xs:simpleType  name="BrowseDirection">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Forward_0" />\r\n      <xs:enumeration value="Inverse_1" />\r\n      <xs:enumeration value="Both_2" />\r\n      <xs:enumeration value="Invalid_3" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="BrowseDirection" type="tns:BrowseDirection" />\r\n\r\n  <xs:complexType name="ViewDescription">\r\n    <xs:sequence>\r\n      <xs:element name="ViewId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="Timestamp" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="ViewVersion" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ViewDescription" type="tns:ViewDescription" />\r\n\r\n  <xs:complexType name="BrowseDescription">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="BrowseDirection" type="tns:BrowseDirection" minOccurs="0" />\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IncludeSubtypes" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="NodeClassMask" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ResultMask" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowseDescription" type="tns:BrowseDescription" />\r\n\r\n  <xs:complexType name="ListOfBrowseDescription">\r\n    <xs:sequence>\r\n      <xs:element name="BrowseDescription" type="tns:BrowseDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrowseDescription" type="tns:ListOfBrowseDescription" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="BrowseResultMask">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="None_0" />\r\n      <xs:enumeration value="ReferenceTypeId_1" />\r\n      <xs:enumeration value="IsForward_2" />\r\n      <xs:enumeration value="NodeClass_4" />\r\n      <xs:enumeration value="BrowseName_8" />\r\n      <xs:enumeration value="DisplayName_16" />\r\n      <xs:enumeration value="TypeDefinition_32" />\r\n      <xs:enumeration value="All_63" />\r\n      <xs:enumeration value="ReferenceTypeInfo_3" />\r\n      <xs:enumeration value="TargetInfo_60" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="BrowseResultMask" type="tns:BrowseResultMask" />\r\n\r\n  <xs:complexType name="ReferenceDescription">\r\n    <xs:sequence>\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IsForward" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="NodeId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="BrowseName" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n      <xs:element name="DisplayName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodeClass" type="tns:NodeClass" minOccurs="0" />\r\n      <xs:element name="TypeDefinition" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ReferenceDescription" type="tns:ReferenceDescription" />\r\n\r\n  <xs:complexType name="ListOfReferenceDescription">\r\n    <xs:sequence>\r\n      <xs:element name="ReferenceDescription" type="tns:ReferenceDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfReferenceDescription" type="tns:ListOfReferenceDescription" nillable="true"></xs:element>\r\n\r\n  <xs:element name="ContinuationPoint" type="xs:base64Binary" />\r\n\r\n  <xs:complexType name="BrowseResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="ContinuationPoint" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="References" type="tns:ListOfReferenceDescription" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowseResult" type="tns:BrowseResult" />\r\n\r\n  <xs:complexType name="ListOfBrowseResult">\r\n    <xs:sequence>\r\n      <xs:element name="BrowseResult" type="tns:BrowseResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrowseResult" type="tns:ListOfBrowseResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="BrowseRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="View" type="tns:ViewDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="RequestedMaxReferencesPerNode" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="NodesToBrowse" type="tns:ListOfBrowseDescription" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowseRequest" type="tns:BrowseRequest" />\r\n\r\n  <xs:complexType name="BrowseResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfBrowseResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowseResponse" type="tns:BrowseResponse" />\r\n\r\n  <xs:complexType name="BrowseNextRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReleaseContinuationPoints" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="ContinuationPoints" type="ua:ListOfByteString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowseNextRequest" type="tns:BrowseNextRequest" />\r\n\r\n  <xs:complexType name="BrowseNextResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfBrowseResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowseNextResponse" type="tns:BrowseNextResponse" />\r\n\r\n  <xs:complexType name="RelativePathElement">\r\n    <xs:sequence>\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IsInverse" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="IncludeSubtypes" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="TargetName" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RelativePathElement" type="tns:RelativePathElement" />\r\n\r\n  <xs:complexType name="ListOfRelativePathElement">\r\n    <xs:sequence>\r\n      <xs:element name="RelativePathElement" type="tns:RelativePathElement" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfRelativePathElement" type="tns:ListOfRelativePathElement" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="RelativePath">\r\n    <xs:sequence>\r\n      <xs:element name="Elements" type="tns:ListOfRelativePathElement" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RelativePath" type="tns:RelativePath" />\r\n\r\n  <xs:complexType name="BrowsePath">\r\n    <xs:sequence>\r\n      <xs:element name="StartingNode" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="RelativePath" type="tns:RelativePath" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowsePath" type="tns:BrowsePath" />\r\n\r\n  <xs:complexType name="ListOfBrowsePath">\r\n    <xs:sequence>\r\n      <xs:element name="BrowsePath" type="tns:BrowsePath" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrowsePath" type="tns:ListOfBrowsePath" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="BrowsePathTarget">\r\n    <xs:sequence>\r\n      <xs:element name="TargetId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="RemainingPathIndex" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowsePathTarget" type="tns:BrowsePathTarget" />\r\n\r\n  <xs:complexType name="ListOfBrowsePathTarget">\r\n    <xs:sequence>\r\n      <xs:element name="BrowsePathTarget" type="tns:BrowsePathTarget" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrowsePathTarget" type="tns:ListOfBrowsePathTarget" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="BrowsePathResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="Targets" type="tns:ListOfBrowsePathTarget" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BrowsePathResult" type="tns:BrowsePathResult" />\r\n\r\n  <xs:complexType name="ListOfBrowsePathResult">\r\n    <xs:sequence>\r\n      <xs:element name="BrowsePathResult" type="tns:BrowsePathResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfBrowsePathResult" type="tns:ListOfBrowsePathResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="TranslateBrowsePathsToNodeIdsRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="BrowsePaths" type="tns:ListOfBrowsePath" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="TranslateBrowsePathsToNodeIdsRequest" type="tns:TranslateBrowsePathsToNodeIdsRequest" />\r\n\r\n  <xs:complexType name="TranslateBrowsePathsToNodeIdsResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfBrowsePathResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="TranslateBrowsePathsToNodeIdsResponse" type="tns:TranslateBrowsePathsToNodeIdsResponse" />\r\n\r\n  <xs:complexType name="RegisterNodesRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodesToRegister" type="ua:ListOfNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RegisterNodesRequest" type="tns:RegisterNodesRequest" />\r\n\r\n  <xs:complexType name="RegisterNodesResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="RegisteredNodeIds" type="ua:ListOfNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RegisterNodesResponse" type="tns:RegisterNodesResponse" />\r\n\r\n  <xs:complexType name="UnregisterNodesRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodesToUnregister" type="ua:ListOfNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="UnregisterNodesRequest" type="tns:UnregisterNodesRequest" />\r\n\r\n  <xs:complexType name="UnregisterNodesResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="UnregisterNodesResponse" type="tns:UnregisterNodesResponse" />\r\n\r\n  <xs:element name="Counter" type="xs:unsignedInt" />\r\n\r\n  <xs:element name="NumericRange" type="xs:string" />\r\n\r\n  <xs:complexType name="EndpointConfiguration">\r\n    <xs:sequence>\r\n      <xs:element name="OperationTimeout" type="xs:int" minOccurs="0" />\r\n      <xs:element name="UseBinaryEncoding" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="MaxStringLength" type="xs:int" minOccurs="0" />\r\n      <xs:element name="MaxByteStringLength" type="xs:int" minOccurs="0" />\r\n      <xs:element name="MaxArrayLength" type="xs:int" minOccurs="0" />\r\n      <xs:element name="MaxMessageSize" type="xs:int" minOccurs="0" />\r\n      <xs:element name="MaxBufferSize" type="xs:int" minOccurs="0" />\r\n      <xs:element name="ChannelLifetime" type="xs:int" minOccurs="0" />\r\n      <xs:element name="SecurityTokenLifetime" type="xs:int" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EndpointConfiguration" type="tns:EndpointConfiguration" />\r\n\r\n  <xs:complexType name="ListOfEndpointConfiguration">\r\n    <xs:sequence>\r\n      <xs:element name="EndpointConfiguration" type="tns:EndpointConfiguration" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEndpointConfiguration" type="tns:ListOfEndpointConfiguration" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="QueryDataDescription">\r\n    <xs:sequence>\r\n      <xs:element name="RelativePath" type="tns:RelativePath" minOccurs="0" nillable="true" />\r\n      <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="IndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="QueryDataDescription" type="tns:QueryDataDescription" />\r\n\r\n  <xs:complexType name="ListOfQueryDataDescription">\r\n    <xs:sequence>\r\n      <xs:element name="QueryDataDescription" type="tns:QueryDataDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfQueryDataDescription" type="tns:ListOfQueryDataDescription" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="NodeTypeDescription">\r\n    <xs:sequence>\r\n      <xs:element name="TypeDefinitionNode" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IncludeSubTypes" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="DataToReturn" type="tns:ListOfQueryDataDescription" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NodeTypeDescription" type="tns:NodeTypeDescription" />\r\n\r\n  <xs:complexType name="ListOfNodeTypeDescription">\r\n    <xs:sequence>\r\n      <xs:element name="NodeTypeDescription" type="tns:NodeTypeDescription" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfNodeTypeDescription" type="tns:ListOfNodeTypeDescription" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="FilterOperator">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Equals_0" />\r\n      <xs:enumeration value="IsNull_1" />\r\n      <xs:enumeration value="GreaterThan_2" />\r\n      <xs:enumeration value="LessThan_3" />\r\n      <xs:enumeration value="GreaterThanOrEqual_4" />\r\n      <xs:enumeration value="LessThanOrEqual_5" />\r\n      <xs:enumeration value="Like_6" />\r\n      <xs:enumeration value="Not_7" />\r\n      <xs:enumeration value="Between_8" />\r\n      <xs:enumeration value="InList_9" />\r\n      <xs:enumeration value="And_10" />\r\n      <xs:enumeration value="Or_11" />\r\n      <xs:enumeration value="Cast_12" />\r\n      <xs:enumeration value="InView_13" />\r\n      <xs:enumeration value="OfType_14" />\r\n      <xs:enumeration value="RelatedTo_15" />\r\n      <xs:enumeration value="BitwiseAnd_16" />\r\n      <xs:enumeration value="BitwiseOr_17" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="FilterOperator" type="tns:FilterOperator" />\r\n\r\n  <xs:complexType name="QueryDataSet">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="TypeDefinitionNode" type="ua:ExpandedNodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="Values" type="ua:ListOfVariant" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="QueryDataSet" type="tns:QueryDataSet" />\r\n\r\n  <xs:complexType name="ListOfQueryDataSet">\r\n    <xs:sequence>\r\n      <xs:element name="QueryDataSet" type="tns:QueryDataSet" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfQueryDataSet" type="tns:ListOfQueryDataSet" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="NodeReference">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReferenceTypeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IsForward" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="ReferencedNodeIds" type="ua:ListOfNodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NodeReference" type="tns:NodeReference" />\r\n\r\n  <xs:complexType name="ListOfNodeReference">\r\n    <xs:sequence>\r\n      <xs:element name="NodeReference" type="tns:NodeReference" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfNodeReference" type="tns:ListOfNodeReference" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ContentFilterElement">\r\n    <xs:sequence>\r\n      <xs:element name="FilterOperator" type="tns:FilterOperator" minOccurs="0" />\r\n      <xs:element name="FilterOperands" type="ua:ListOfExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ContentFilterElement" type="tns:ContentFilterElement" />\r\n\r\n  <xs:complexType name="ListOfContentFilterElement">\r\n    <xs:sequence>\r\n      <xs:element name="ContentFilterElement" type="tns:ContentFilterElement" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfContentFilterElement" type="tns:ListOfContentFilterElement" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ContentFilter">\r\n    <xs:sequence>\r\n      <xs:element name="Elements" type="tns:ListOfContentFilterElement" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ContentFilter" type="tns:ContentFilter" />\r\n\r\n  <xs:complexType name="ListOfContentFilter">\r\n    <xs:sequence>\r\n      <xs:element name="ContentFilter" type="tns:ContentFilter" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfContentFilter" type="tns:ListOfContentFilter" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="FilterOperand">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="FilterOperand" type="tns:FilterOperand" />\r\n\r\n  <xs:complexType name="ElementOperand">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:FilterOperand">\r\n        <xs:sequence>\r\n          <xs:element name="Index" type="xs:unsignedInt" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ElementOperand" type="tns:ElementOperand" />\r\n\r\n  <xs:complexType name="LiteralOperand">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:FilterOperand">\r\n        <xs:sequence>\r\n          <xs:element name="Value" type="ua:Variant" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="LiteralOperand" type="tns:LiteralOperand" />\r\n\r\n  <xs:complexType name="AttributeOperand">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:FilterOperand">\r\n        <xs:sequence>\r\n          <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="Alias" type="xs:string" minOccurs="0" nillable="true" />\r\n          <xs:element name="BrowsePath" type="tns:RelativePath" minOccurs="0" nillable="true" />\r\n          <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n          <xs:element name="IndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="AttributeOperand" type="tns:AttributeOperand" />\r\n\r\n  <xs:complexType name="SimpleAttributeOperand">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:FilterOperand">\r\n        <xs:sequence>\r\n          <xs:element name="TypeDefinitionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="BrowsePath" type="ua:ListOfQualifiedName" minOccurs="0" nillable="true" />\r\n          <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n          <xs:element name="IndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="SimpleAttributeOperand" type="tns:SimpleAttributeOperand" />\r\n\r\n  <xs:complexType name="ListOfSimpleAttributeOperand">\r\n    <xs:sequence>\r\n      <xs:element name="SimpleAttributeOperand" type="tns:SimpleAttributeOperand" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSimpleAttributeOperand" type="tns:ListOfSimpleAttributeOperand" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ContentFilterElementResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="OperandStatusCodes" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="OperandDiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ContentFilterElementResult" type="tns:ContentFilterElementResult" />\r\n\r\n  <xs:complexType name="ListOfContentFilterElementResult">\r\n    <xs:sequence>\r\n      <xs:element name="ContentFilterElementResult" type="tns:ContentFilterElementResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfContentFilterElementResult" type="tns:ListOfContentFilterElementResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ContentFilterResult">\r\n    <xs:sequence>\r\n      <xs:element name="ElementResults" type="tns:ListOfContentFilterElementResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="ElementDiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ContentFilterResult" type="tns:ContentFilterResult" />\r\n\r\n  <xs:complexType name="ParsingResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="DataStatusCodes" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataDiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ParsingResult" type="tns:ParsingResult" />\r\n\r\n  <xs:complexType name="ListOfParsingResult">\r\n    <xs:sequence>\r\n      <xs:element name="ParsingResult" type="tns:ParsingResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfParsingResult" type="tns:ListOfParsingResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="QueryFirstRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="View" type="tns:ViewDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodeTypes" type="tns:ListOfNodeTypeDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="Filter" type="tns:ContentFilter" minOccurs="0" nillable="true" />\r\n      <xs:element name="MaxDataSetsToReturn" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MaxReferencesToReturn" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="QueryFirstRequest" type="tns:QueryFirstRequest" />\r\n\r\n  <xs:complexType name="QueryFirstResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="QueryDataSets" type="tns:ListOfQueryDataSet" minOccurs="0" nillable="true" />\r\n      <xs:element name="ContinuationPoint" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="ParsingResults" type="tns:ListOfParsingResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n      <xs:element name="FilterResult" type="tns:ContentFilterResult" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="QueryFirstResponse" type="tns:QueryFirstResponse" />\r\n\r\n  <xs:complexType name="QueryNextRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="ReleaseContinuationPoint" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="ContinuationPoint" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="QueryNextRequest" type="tns:QueryNextRequest" />\r\n\r\n  <xs:complexType name="QueryNextResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="QueryDataSets" type="tns:ListOfQueryDataSet" minOccurs="0" nillable="true" />\r\n      <xs:element name="RevisedContinuationPoint" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="QueryNextResponse" type="tns:QueryNextResponse" />\r\n\r\n  <xs:simpleType  name="TimestampsToReturn">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Source_0" />\r\n      <xs:enumeration value="Server_1" />\r\n      <xs:enumeration value="Both_2" />\r\n      <xs:enumeration value="Neither_3" />\r\n      <xs:enumeration value="Invalid_4" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="TimestampsToReturn" type="tns:TimestampsToReturn" />\r\n\r\n  <xs:complexType name="ReadValueId">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="IndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataEncoding" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ReadValueId" type="tns:ReadValueId" />\r\n\r\n  <xs:complexType name="ListOfReadValueId">\r\n    <xs:sequence>\r\n      <xs:element name="ReadValueId" type="tns:ReadValueId" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfReadValueId" type="tns:ListOfReadValueId" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ReadRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="MaxAge" type="xs:double" minOccurs="0" />\r\n      <xs:element name="TimestampsToReturn" type="tns:TimestampsToReturn" minOccurs="0" />\r\n      <xs:element name="NodesToRead" type="tns:ListOfReadValueId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ReadRequest" type="tns:ReadRequest" />\r\n\r\n  <xs:complexType name="ReadResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfDataValue" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ReadResponse" type="tns:ReadResponse" />\r\n\r\n  <xs:complexType name="HistoryReadValueId">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="IndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="DataEncoding" type="ua:QualifiedName" minOccurs="0" nillable="true" />\r\n      <xs:element name="ContinuationPoint" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryReadValueId" type="tns:HistoryReadValueId" />\r\n\r\n  <xs:complexType name="ListOfHistoryReadValueId">\r\n    <xs:sequence>\r\n      <xs:element name="HistoryReadValueId" type="tns:HistoryReadValueId" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfHistoryReadValueId" type="tns:ListOfHistoryReadValueId" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="HistoryReadResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="ContinuationPoint" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n      <xs:element name="HistoryData" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryReadResult" type="tns:HistoryReadResult" />\r\n\r\n  <xs:complexType name="ListOfHistoryReadResult">\r\n    <xs:sequence>\r\n      <xs:element name="HistoryReadResult" type="tns:HistoryReadResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfHistoryReadResult" type="tns:ListOfHistoryReadResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="HistoryReadDetails">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryReadDetails" type="tns:HistoryReadDetails" />\r\n\r\n  <xs:complexType name="ReadEventDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryReadDetails">\r\n        <xs:sequence>\r\n          <xs:element name="NumValuesPerNode" type="xs:unsignedInt" minOccurs="0" />\r\n          <xs:element name="StartTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="EndTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="Filter" type="tns:EventFilter" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReadEventDetails" type="tns:ReadEventDetails" />\r\n\r\n  <xs:complexType name="ReadRawModifiedDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryReadDetails">\r\n        <xs:sequence>\r\n          <xs:element name="IsReadModified" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="StartTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="EndTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="NumValuesPerNode" type="xs:unsignedInt" minOccurs="0" />\r\n          <xs:element name="ReturnBounds" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReadRawModifiedDetails" type="tns:ReadRawModifiedDetails" />\r\n\r\n  <xs:complexType name="ReadProcessedDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryReadDetails">\r\n        <xs:sequence>\r\n          <xs:element name="StartTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="EndTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="ProcessingInterval" type="xs:double" minOccurs="0" />\r\n          <xs:element name="AggregateType" type="ua:ListOfNodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="AggregateConfiguration" type="tns:AggregateConfiguration" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReadProcessedDetails" type="tns:ReadProcessedDetails" />\r\n\r\n  <xs:complexType name="ReadAtTimeDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryReadDetails">\r\n        <xs:sequence>\r\n          <xs:element name="ReqTimes" type="ua:ListOfDateTime" minOccurs="0" nillable="true" />\r\n          <xs:element name="UseSimpleBounds" type="xs:boolean" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReadAtTimeDetails" type="tns:ReadAtTimeDetails" />\r\n\r\n  <xs:complexType name="ReadAnnotationDataDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryReadDetails">\r\n        <xs:sequence>\r\n          <xs:element name="ReqTimes" type="ua:ListOfDateTime" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="ReadAnnotationDataDetails" type="tns:ReadAnnotationDataDetails" />\r\n\r\n  <xs:complexType name="HistoryData">\r\n    <xs:sequence>\r\n      <xs:element name="DataValues" type="ua:ListOfDataValue" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryData" type="tns:HistoryData" />\r\n\r\n  <xs:complexType name="ModificationInfo">\r\n    <xs:sequence>\r\n      <xs:element name="ModificationTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="UpdateType" type="tns:HistoryUpdateType" minOccurs="0" />\r\n      <xs:element name="UserName" type="xs:string" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ModificationInfo" type="tns:ModificationInfo" />\r\n\r\n  <xs:complexType name="ListOfModificationInfo">\r\n    <xs:sequence>\r\n      <xs:element name="ModificationInfo" type="tns:ModificationInfo" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfModificationInfo" type="tns:ListOfModificationInfo" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="HistoryModifiedData">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryData">\r\n        <xs:sequence>\r\n          <xs:element name="ModificationInfos" type="tns:ListOfModificationInfo" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryModifiedData" type="tns:HistoryModifiedData" />\r\n\r\n  <xs:complexType name="HistoryEvent">\r\n    <xs:sequence>\r\n      <xs:element name="Events" type="tns:ListOfHistoryEventFieldList" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryEvent" type="tns:HistoryEvent" />\r\n\r\n  <xs:complexType name="HistoryReadRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="HistoryReadDetails" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="TimestampsToReturn" type="tns:TimestampsToReturn" minOccurs="0" />\r\n      <xs:element name="ReleaseContinuationPoints" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="NodesToRead" type="tns:ListOfHistoryReadValueId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryReadRequest" type="tns:HistoryReadRequest" />\r\n\r\n  <xs:complexType name="HistoryReadResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfHistoryReadResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryReadResponse" type="tns:HistoryReadResponse" />\r\n\r\n  <xs:complexType name="WriteValue">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="AttributeId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="IndexRange" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Value" type="ua:DataValue" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="WriteValue" type="tns:WriteValue" />\r\n\r\n  <xs:complexType name="ListOfWriteValue">\r\n    <xs:sequence>\r\n      <xs:element name="WriteValue" type="tns:WriteValue" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfWriteValue" type="tns:ListOfWriteValue" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="WriteRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="NodesToWrite" type="tns:ListOfWriteValue" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="WriteRequest" type="tns:WriteRequest" />\r\n\r\n  <xs:complexType name="WriteResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="WriteResponse" type="tns:WriteResponse" />\r\n\r\n  <xs:complexType name="HistoryUpdateDetails">\r\n    <xs:sequence>\r\n      <xs:element name="NodeId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryUpdateDetails" type="tns:HistoryUpdateDetails" />\r\n\r\n  <xs:simpleType  name="HistoryUpdateType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Insert_1" />\r\n      <xs:enumeration value="Replace_2" />\r\n      <xs:enumeration value="Update_3" />\r\n      <xs:enumeration value="Delete_4" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="HistoryUpdateType" type="tns:HistoryUpdateType" />\r\n\r\n  <xs:simpleType  name="PerformUpdateType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Insert_1" />\r\n      <xs:enumeration value="Replace_2" />\r\n      <xs:enumeration value="Update_3" />\r\n      <xs:enumeration value="Remove_4" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="PerformUpdateType" type="tns:PerformUpdateType" />\r\n\r\n  <xs:complexType name="UpdateDataDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryUpdateDetails">\r\n        <xs:sequence>\r\n          <xs:element name="PerformInsertReplace" type="tns:PerformUpdateType" minOccurs="0" />\r\n          <xs:element name="UpdateValues" type="ua:ListOfDataValue" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UpdateDataDetails" type="tns:UpdateDataDetails" />\r\n\r\n  <xs:complexType name="UpdateStructureDataDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryUpdateDetails">\r\n        <xs:sequence>\r\n          <xs:element name="PerformInsertReplace" type="tns:PerformUpdateType" minOccurs="0" />\r\n          <xs:element name="UpdateValues" type="ua:ListOfDataValue" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UpdateStructureDataDetails" type="tns:UpdateStructureDataDetails" />\r\n\r\n  <xs:complexType name="UpdateEventDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryUpdateDetails">\r\n        <xs:sequence>\r\n          <xs:element name="PerformInsertReplace" type="tns:PerformUpdateType" minOccurs="0" />\r\n          <xs:element name="Filter" type="tns:EventFilter" minOccurs="0" nillable="true" />\r\n          <xs:element name="EventData" type="tns:ListOfHistoryEventFieldList" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="UpdateEventDetails" type="tns:UpdateEventDetails" />\r\n\r\n  <xs:complexType name="DeleteRawModifiedDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryUpdateDetails">\r\n        <xs:sequence>\r\n          <xs:element name="IsDeleteModified" type="xs:boolean" minOccurs="0" />\r\n          <xs:element name="StartTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="EndTime" type="xs:dateTime" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteRawModifiedDetails" type="tns:DeleteRawModifiedDetails" />\r\n\r\n  <xs:complexType name="DeleteAtTimeDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryUpdateDetails">\r\n        <xs:sequence>\r\n          <xs:element name="ReqTimes" type="ua:ListOfDateTime" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteAtTimeDetails" type="tns:DeleteAtTimeDetails" />\r\n\r\n  <xs:complexType name="DeleteEventDetails">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:HistoryUpdateDetails">\r\n        <xs:sequence>\r\n          <xs:element name="EventIds" type="ua:ListOfByteString" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteEventDetails" type="tns:DeleteEventDetails" />\r\n\r\n  <xs:complexType name="HistoryUpdateResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="OperationResults" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryUpdateResult" type="tns:HistoryUpdateResult" />\r\n\r\n  <xs:complexType name="ListOfHistoryUpdateResult">\r\n    <xs:sequence>\r\n      <xs:element name="HistoryUpdateResult" type="tns:HistoryUpdateResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfHistoryUpdateResult" type="tns:ListOfHistoryUpdateResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="HistoryUpdateRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="HistoryUpdateDetails" type="ua:ListOfExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryUpdateRequest" type="tns:HistoryUpdateRequest" />\r\n\r\n  <xs:complexType name="HistoryUpdateResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfHistoryUpdateResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryUpdateResponse" type="tns:HistoryUpdateResponse" />\r\n\r\n  <xs:complexType name="CallMethodRequest">\r\n    <xs:sequence>\r\n      <xs:element name="ObjectId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="MethodId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="InputArguments" type="ua:ListOfVariant" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CallMethodRequest" type="tns:CallMethodRequest" />\r\n\r\n  <xs:complexType name="ListOfCallMethodRequest">\r\n    <xs:sequence>\r\n      <xs:element name="CallMethodRequest" type="tns:CallMethodRequest" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfCallMethodRequest" type="tns:ListOfCallMethodRequest" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="CallMethodResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="InputArgumentResults" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="InputArgumentDiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n      <xs:element name="OutputArguments" type="ua:ListOfVariant" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CallMethodResult" type="tns:CallMethodResult" />\r\n\r\n  <xs:complexType name="ListOfCallMethodResult">\r\n    <xs:sequence>\r\n      <xs:element name="CallMethodResult" type="tns:CallMethodResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfCallMethodResult" type="tns:ListOfCallMethodResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="CallRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="MethodsToCall" type="tns:ListOfCallMethodRequest" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CallRequest" type="tns:CallRequest" />\r\n\r\n  <xs:complexType name="CallResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfCallMethodResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CallResponse" type="tns:CallResponse" />\r\n\r\n  <xs:simpleType  name="MonitoringMode">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Disabled_0" />\r\n      <xs:enumeration value="Sampling_1" />\r\n      <xs:enumeration value="Reporting_2" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="MonitoringMode" type="tns:MonitoringMode" />\r\n\r\n  <xs:simpleType  name="DataChangeTrigger">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Status_0" />\r\n      <xs:enumeration value="StatusValue_1" />\r\n      <xs:enumeration value="StatusValueTimestamp_2" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="DataChangeTrigger" type="tns:DataChangeTrigger" />\r\n\r\n  <xs:simpleType  name="DeadbandType">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="None_0" />\r\n      <xs:enumeration value="Absolute_1" />\r\n      <xs:enumeration value="Percent_2" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="DeadbandType" type="tns:DeadbandType" />\r\n\r\n  <xs:complexType name="MonitoringFilter">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoringFilter" type="tns:MonitoringFilter" />\r\n\r\n  <xs:complexType name="DataChangeFilter">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:MonitoringFilter">\r\n        <xs:sequence>\r\n          <xs:element name="Trigger" type="tns:DataChangeTrigger" minOccurs="0" />\r\n          <xs:element name="DeadbandType" type="xs:unsignedInt" minOccurs="0" />\r\n          <xs:element name="DeadbandValue" type="xs:double" minOccurs="0" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DataChangeFilter" type="tns:DataChangeFilter" />\r\n\r\n  <xs:complexType name="EventFilter">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:MonitoringFilter">\r\n        <xs:sequence>\r\n          <xs:element name="SelectClauses" type="tns:ListOfSimpleAttributeOperand" minOccurs="0" nillable="true" />\r\n          <xs:element name="WhereClause" type="tns:ContentFilter" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="EventFilter" type="tns:EventFilter" />\r\n\r\n  <xs:complexType name="AggregateConfiguration">\r\n    <xs:sequence>\r\n      <xs:element name="UseServerCapabilitiesDefaults" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="TreatUncertainAsBad" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="PercentDataBad" type="xs:unsignedByte" minOccurs="0" />\r\n      <xs:element name="PercentDataGood" type="xs:unsignedByte" minOccurs="0" />\r\n      <xs:element name="UseSlopedExtrapolation" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AggregateConfiguration" type="tns:AggregateConfiguration" />\r\n\r\n  <xs:complexType name="AggregateFilter">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:MonitoringFilter">\r\n        <xs:sequence>\r\n          <xs:element name="StartTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="AggregateType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n          <xs:element name="ProcessingInterval" type="xs:double" minOccurs="0" />\r\n          <xs:element name="AggregateConfiguration" type="tns:AggregateConfiguration" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="AggregateFilter" type="tns:AggregateFilter" />\r\n\r\n  <xs:complexType name="MonitoringFilterResult">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoringFilterResult" type="tns:MonitoringFilterResult" />\r\n\r\n  <xs:complexType name="EventFilterResult">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:MonitoringFilterResult">\r\n        <xs:sequence>\r\n          <xs:element name="SelectClauseResults" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n          <xs:element name="SelectClauseDiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n          <xs:element name="WhereClauseResult" type="tns:ContentFilterResult" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="EventFilterResult" type="tns:EventFilterResult" />\r\n\r\n  <xs:complexType name="AggregateFilterResult">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:MonitoringFilterResult">\r\n        <xs:sequence>\r\n          <xs:element name="RevisedStartTime" type="xs:dateTime" minOccurs="0" />\r\n          <xs:element name="RevisedProcessingInterval" type="xs:double" minOccurs="0" />\r\n          <xs:element name="RevisedAggregateConfiguration" type="tns:AggregateConfiguration" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="AggregateFilterResult" type="tns:AggregateFilterResult" />\r\n\r\n  <xs:complexType name="MonitoringParameters">\r\n    <xs:sequence>\r\n      <xs:element name="ClientHandle" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SamplingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="Filter" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n      <xs:element name="QueueSize" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DiscardOldest" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoringParameters" type="tns:MonitoringParameters" />\r\n\r\n  <xs:complexType name="MonitoredItemCreateRequest">\r\n    <xs:sequence>\r\n      <xs:element name="ItemToMonitor" type="tns:ReadValueId" minOccurs="0" nillable="true" />\r\n      <xs:element name="MonitoringMode" type="tns:MonitoringMode" minOccurs="0" />\r\n      <xs:element name="RequestedParameters" type="tns:MonitoringParameters" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoredItemCreateRequest" type="tns:MonitoredItemCreateRequest" />\r\n\r\n  <xs:complexType name="ListOfMonitoredItemCreateRequest">\r\n    <xs:sequence>\r\n      <xs:element name="MonitoredItemCreateRequest" type="tns:MonitoredItemCreateRequest" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfMonitoredItemCreateRequest" type="tns:ListOfMonitoredItemCreateRequest" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="MonitoredItemCreateResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="MonitoredItemId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RevisedSamplingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="RevisedQueueSize" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="FilterResult" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoredItemCreateResult" type="tns:MonitoredItemCreateResult" />\r\n\r\n  <xs:complexType name="ListOfMonitoredItemCreateResult">\r\n    <xs:sequence>\r\n      <xs:element name="MonitoredItemCreateResult" type="tns:MonitoredItemCreateResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfMonitoredItemCreateResult" type="tns:ListOfMonitoredItemCreateResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="CreateMonitoredItemsRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TimestampsToReturn" type="tns:TimestampsToReturn" minOccurs="0" />\r\n      <xs:element name="ItemsToCreate" type="tns:ListOfMonitoredItemCreateRequest" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CreateMonitoredItemsRequest" type="tns:CreateMonitoredItemsRequest" />\r\n\r\n  <xs:complexType name="CreateMonitoredItemsResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfMonitoredItemCreateResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CreateMonitoredItemsResponse" type="tns:CreateMonitoredItemsResponse" />\r\n\r\n  <xs:complexType name="MonitoredItemModifyRequest">\r\n    <xs:sequence>\r\n      <xs:element name="MonitoredItemId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RequestedParameters" type="tns:MonitoringParameters" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoredItemModifyRequest" type="tns:MonitoredItemModifyRequest" />\r\n\r\n  <xs:complexType name="ListOfMonitoredItemModifyRequest">\r\n    <xs:sequence>\r\n      <xs:element name="MonitoredItemModifyRequest" type="tns:MonitoredItemModifyRequest" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfMonitoredItemModifyRequest" type="tns:ListOfMonitoredItemModifyRequest" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="MonitoredItemModifyResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="RevisedSamplingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="RevisedQueueSize" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="FilterResult" type="ua:ExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoredItemModifyResult" type="tns:MonitoredItemModifyResult" />\r\n\r\n  <xs:complexType name="ListOfMonitoredItemModifyResult">\r\n    <xs:sequence>\r\n      <xs:element name="MonitoredItemModifyResult" type="tns:MonitoredItemModifyResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfMonitoredItemModifyResult" type="tns:ListOfMonitoredItemModifyResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ModifyMonitoredItemsRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TimestampsToReturn" type="tns:TimestampsToReturn" minOccurs="0" />\r\n      <xs:element name="ItemsToModify" type="tns:ListOfMonitoredItemModifyRequest" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ModifyMonitoredItemsRequest" type="tns:ModifyMonitoredItemsRequest" />\r\n\r\n  <xs:complexType name="ModifyMonitoredItemsResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfMonitoredItemModifyResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ModifyMonitoredItemsResponse" type="tns:ModifyMonitoredItemsResponse" />\r\n\r\n  <xs:complexType name="SetMonitoringModeRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MonitoringMode" type="tns:MonitoringMode" minOccurs="0" />\r\n      <xs:element name="MonitoredItemIds" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SetMonitoringModeRequest" type="tns:SetMonitoringModeRequest" />\r\n\r\n  <xs:complexType name="SetMonitoringModeResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SetMonitoringModeResponse" type="tns:SetMonitoringModeResponse" />\r\n\r\n  <xs:complexType name="SetTriggeringRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TriggeringItemId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="LinksToAdd" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n      <xs:element name="LinksToRemove" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SetTriggeringRequest" type="tns:SetTriggeringRequest" />\r\n\r\n  <xs:complexType name="SetTriggeringResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="AddResults" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="AddDiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n      <xs:element name="RemoveResults" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="RemoveDiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SetTriggeringResponse" type="tns:SetTriggeringResponse" />\r\n\r\n  <xs:complexType name="DeleteMonitoredItemsRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MonitoredItemIds" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteMonitoredItemsRequest" type="tns:DeleteMonitoredItemsRequest" />\r\n\r\n  <xs:complexType name="DeleteMonitoredItemsResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteMonitoredItemsResponse" type="tns:DeleteMonitoredItemsResponse" />\r\n\r\n  <xs:complexType name="CreateSubscriptionRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="RequestedPublishingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="RequestedLifetimeCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RequestedMaxKeepAliveCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MaxNotificationsPerPublish" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="PublishingEnabled" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="Priority" type="xs:unsignedByte" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CreateSubscriptionRequest" type="tns:CreateSubscriptionRequest" />\r\n\r\n  <xs:complexType name="CreateSubscriptionResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RevisedPublishingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="RevisedLifetimeCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RevisedMaxKeepAliveCount" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="CreateSubscriptionResponse" type="tns:CreateSubscriptionResponse" />\r\n\r\n  <xs:complexType name="ModifySubscriptionRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RequestedPublishingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="RequestedLifetimeCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RequestedMaxKeepAliveCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MaxNotificationsPerPublish" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="Priority" type="xs:unsignedByte" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ModifySubscriptionRequest" type="tns:ModifySubscriptionRequest" />\r\n\r\n  <xs:complexType name="ModifySubscriptionResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="RevisedPublishingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="RevisedLifetimeCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RevisedMaxKeepAliveCount" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ModifySubscriptionResponse" type="tns:ModifySubscriptionResponse" />\r\n\r\n  <xs:complexType name="SetPublishingModeRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="PublishingEnabled" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="SubscriptionIds" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SetPublishingModeRequest" type="tns:SetPublishingModeRequest" />\r\n\r\n  <xs:complexType name="SetPublishingModeResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SetPublishingModeResponse" type="tns:SetPublishingModeResponse" />\r\n\r\n  <xs:complexType name="NotificationMessage">\r\n    <xs:sequence>\r\n      <xs:element name="SequenceNumber" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="PublishTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="NotificationData" type="ua:ListOfExtensionObject" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NotificationMessage" type="tns:NotificationMessage" />\r\n\r\n  <xs:complexType name="NotificationData">\r\n    <xs:sequence>\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NotificationData" type="tns:NotificationData" />\r\n\r\n  <xs:complexType name="DataChangeNotification">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NotificationData">\r\n        <xs:sequence>\r\n          <xs:element name="MonitoredItems" type="tns:ListOfMonitoredItemNotification" minOccurs="0" nillable="true" />\r\n          <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="DataChangeNotification" type="tns:DataChangeNotification" />\r\n\r\n  <xs:complexType name="MonitoredItemNotification">\r\n    <xs:sequence>\r\n      <xs:element name="ClientHandle" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="Value" type="ua:DataValue" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="MonitoredItemNotification" type="tns:MonitoredItemNotification" />\r\n\r\n  <xs:complexType name="ListOfMonitoredItemNotification">\r\n    <xs:sequence>\r\n      <xs:element name="MonitoredItemNotification" type="tns:MonitoredItemNotification" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfMonitoredItemNotification" type="tns:ListOfMonitoredItemNotification" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="EventNotificationList">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NotificationData">\r\n        <xs:sequence>\r\n          <xs:element name="Events" type="tns:ListOfEventFieldList" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="EventNotificationList" type="tns:EventNotificationList" />\r\n\r\n  <xs:complexType name="EventFieldList">\r\n    <xs:sequence>\r\n      <xs:element name="ClientHandle" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="EventFields" type="ua:ListOfVariant" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EventFieldList" type="tns:EventFieldList" />\r\n\r\n  <xs:complexType name="ListOfEventFieldList">\r\n    <xs:sequence>\r\n      <xs:element name="EventFieldList" type="tns:EventFieldList" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEventFieldList" type="tns:ListOfEventFieldList" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="HistoryEventFieldList">\r\n    <xs:sequence>\r\n      <xs:element name="EventFields" type="ua:ListOfVariant" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="HistoryEventFieldList" type="tns:HistoryEventFieldList" />\r\n\r\n  <xs:complexType name="ListOfHistoryEventFieldList">\r\n    <xs:sequence>\r\n      <xs:element name="HistoryEventFieldList" type="tns:HistoryEventFieldList" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfHistoryEventFieldList" type="tns:ListOfHistoryEventFieldList" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="StatusChangeNotification">\r\n    <xs:complexContent mixed="false">\r\n      <xs:extension base="tns:NotificationData">\r\n        <xs:sequence>\r\n          <xs:element name="Status" type="ua:StatusCode" minOccurs="0" />\r\n          <xs:element name="DiagnosticInfo" type="ua:DiagnosticInfo" minOccurs="0" nillable="true" />\r\n        </xs:sequence>\r\n      </xs:extension>\r\n    </xs:complexContent>\r\n  </xs:complexType>\r\n  <xs:element name="StatusChangeNotification" type="tns:StatusChangeNotification" />\r\n\r\n  <xs:complexType name="SubscriptionAcknowledgement">\r\n    <xs:sequence>\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SequenceNumber" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SubscriptionAcknowledgement" type="tns:SubscriptionAcknowledgement" />\r\n\r\n  <xs:complexType name="ListOfSubscriptionAcknowledgement">\r\n    <xs:sequence>\r\n      <xs:element name="SubscriptionAcknowledgement" type="tns:SubscriptionAcknowledgement" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSubscriptionAcknowledgement" type="tns:ListOfSubscriptionAcknowledgement" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="PublishRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionAcknowledgements" type="tns:ListOfSubscriptionAcknowledgement" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PublishRequest" type="tns:PublishRequest" />\r\n\r\n  <xs:complexType name="PublishResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="AvailableSequenceNumbers" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n      <xs:element name="MoreNotifications" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="NotificationMessage" type="tns:NotificationMessage" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="PublishResponse" type="tns:PublishResponse" />\r\n\r\n  <xs:complexType name="RepublishRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RetransmitSequenceNumber" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RepublishRequest" type="tns:RepublishRequest" />\r\n\r\n  <xs:complexType name="RepublishResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="NotificationMessage" type="tns:NotificationMessage" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RepublishResponse" type="tns:RepublishResponse" />\r\n\r\n  <xs:complexType name="TransferResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="AvailableSequenceNumbers" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="TransferResult" type="tns:TransferResult" />\r\n\r\n  <xs:complexType name="ListOfTransferResult">\r\n    <xs:sequence>\r\n      <xs:element name="TransferResult" type="tns:TransferResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfTransferResult" type="tns:ListOfTransferResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="TransferSubscriptionsRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionIds" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n      <xs:element name="SendInitialValues" type="xs:boolean" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="TransferSubscriptionsRequest" type="tns:TransferSubscriptionsRequest" />\r\n\r\n  <xs:complexType name="TransferSubscriptionsResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="tns:ListOfTransferResult" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="TransferSubscriptionsResponse" type="tns:TransferSubscriptionsResponse" />\r\n\r\n  <xs:complexType name="DeleteSubscriptionsRequest">\r\n    <xs:sequence>\r\n      <xs:element name="RequestHeader" type="tns:RequestHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionIds" type="ua:ListOfUInt32" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteSubscriptionsRequest" type="tns:DeleteSubscriptionsRequest" />\r\n\r\n  <xs:complexType name="DeleteSubscriptionsResponse">\r\n    <xs:sequence>\r\n      <xs:element name="ResponseHeader" type="tns:ResponseHeader" minOccurs="0" nillable="true" />\r\n      <xs:element name="Results" type="ua:ListOfStatusCode" minOccurs="0" nillable="true" />\r\n      <xs:element name="DiagnosticInfos" type="ua:ListOfDiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DeleteSubscriptionsResponse" type="tns:DeleteSubscriptionsResponse" />\r\n\r\n  <xs:complexType name="BuildInfo">\r\n    <xs:sequence>\r\n      <xs:element name="ProductUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ManufacturerName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ProductName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SoftwareVersion" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="BuildNumber" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="BuildDate" type="xs:dateTime" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="BuildInfo" type="tns:BuildInfo" />\r\n\r\n  <xs:simpleType  name="RedundancySupport">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="None_0" />\r\n      <xs:enumeration value="Cold_1" />\r\n      <xs:enumeration value="Warm_2" />\r\n      <xs:enumeration value="Hot_3" />\r\n      <xs:enumeration value="Transparent_4" />\r\n      <xs:enumeration value="HotAndMirrored_5" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="RedundancySupport" type="tns:RedundancySupport" />\r\n\r\n  <xs:simpleType  name="ServerState">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Running_0" />\r\n      <xs:enumeration value="Failed_1" />\r\n      <xs:enumeration value="NoConfiguration_2" />\r\n      <xs:enumeration value="Suspended_3" />\r\n      <xs:enumeration value="Shutdown_4" />\r\n      <xs:enumeration value="Test_5" />\r\n      <xs:enumeration value="CommunicationFault_6" />\r\n      <xs:enumeration value="Unknown_7" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="ServerState" type="tns:ServerState" />\r\n\r\n  <xs:complexType name="RedundantServerDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ServerId" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServiceLevel" type="xs:unsignedByte" minOccurs="0" />\r\n      <xs:element name="ServerState" type="tns:ServerState" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="RedundantServerDataType" type="tns:RedundantServerDataType" />\r\n\r\n  <xs:complexType name="ListOfRedundantServerDataType">\r\n    <xs:sequence>\r\n      <xs:element name="RedundantServerDataType" type="tns:RedundantServerDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfRedundantServerDataType" type="tns:ListOfRedundantServerDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="EndpointUrlListDataType">\r\n    <xs:sequence>\r\n      <xs:element name="EndpointUrlList" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EndpointUrlListDataType" type="tns:EndpointUrlListDataType" />\r\n\r\n  <xs:complexType name="ListOfEndpointUrlListDataType">\r\n    <xs:sequence>\r\n      <xs:element name="EndpointUrlListDataType" type="tns:EndpointUrlListDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfEndpointUrlListDataType" type="tns:ListOfEndpointUrlListDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="NetworkGroupDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ServerUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="NetworkPaths" type="tns:ListOfEndpointUrlListDataType" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="NetworkGroupDataType" type="tns:NetworkGroupDataType" />\r\n\r\n  <xs:complexType name="ListOfNetworkGroupDataType">\r\n    <xs:sequence>\r\n      <xs:element name="NetworkGroupDataType" type="tns:NetworkGroupDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfNetworkGroupDataType" type="tns:ListOfNetworkGroupDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="SamplingIntervalDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SamplingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="MonitoredItemCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MaxMonitoredItemCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DisabledMonitoredItemCount" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SamplingIntervalDiagnosticsDataType" type="tns:SamplingIntervalDiagnosticsDataType" />\r\n\r\n  <xs:complexType name="ListOfSamplingIntervalDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SamplingIntervalDiagnosticsDataType" type="tns:SamplingIntervalDiagnosticsDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSamplingIntervalDiagnosticsDataType" type="tns:ListOfSamplingIntervalDiagnosticsDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ServerDiagnosticsSummaryDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ServerViewCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CurrentSessionCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CumulatedSessionCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SecurityRejectedSessionCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RejectedSessionCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SessionTimeoutCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SessionAbortCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CurrentSubscriptionCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CumulatedSubscriptionCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="PublishingIntervalCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="SecurityRejectedRequestsCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RejectedRequestsCount" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ServerDiagnosticsSummaryDataType" type="tns:ServerDiagnosticsSummaryDataType" />\r\n\r\n  <xs:complexType name="ServerStatusDataType">\r\n    <xs:sequence>\r\n      <xs:element name="StartTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="CurrentTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="State" type="tns:ServerState" minOccurs="0" />\r\n      <xs:element name="BuildInfo" type="tns:BuildInfo" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecondsTillShutdown" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ShutdownReason" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ServerStatusDataType" type="tns:ServerStatusDataType" />\r\n\r\n  <xs:complexType name="SessionDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="SessionName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientDescription" type="tns:ApplicationDescription" minOccurs="0" nillable="true" />\r\n      <xs:element name="ServerUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="EndpointUrl" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="LocaleIds" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="ActualSessionTimeout" type="xs:double" minOccurs="0" />\r\n      <xs:element name="MaxResponseMessageSize" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ClientConnectionTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="ClientLastContactTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="CurrentSubscriptionsCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CurrentMonitoredItemsCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CurrentPublishRequestsInQueue" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TotalRequestCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="UnauthorizedRequestCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ReadCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="HistoryReadCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="WriteCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="HistoryUpdateCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="CallCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="CreateMonitoredItemsCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="ModifyMonitoredItemsCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="SetMonitoringModeCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="SetTriggeringCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="DeleteMonitoredItemsCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="CreateSubscriptionCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="ModifySubscriptionCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="SetPublishingModeCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="PublishCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="RepublishCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="TransferSubscriptionsCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="DeleteSubscriptionsCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="AddNodesCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="AddReferencesCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="DeleteNodesCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="DeleteReferencesCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="BrowseCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="BrowseNextCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="TranslateBrowsePathsToNodeIdsCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="QueryFirstCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="QueryNextCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="RegisterNodesCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n      <xs:element name="UnregisterNodesCount" type="tns:ServiceCounterDataType" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SessionDiagnosticsDataType" type="tns:SessionDiagnosticsDataType" />\r\n\r\n  <xs:complexType name="ListOfSessionDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SessionDiagnosticsDataType" type="tns:SessionDiagnosticsDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSessionDiagnosticsDataType" type="tns:ListOfSessionDiagnosticsDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="SessionSecurityDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientUserIdOfSession" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientUserIdHistory" type="ua:ListOfString" minOccurs="0" nillable="true" />\r\n      <xs:element name="AuthenticationMechanism" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="Encoding" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="TransportProtocol" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="SecurityMode" type="tns:MessageSecurityMode" minOccurs="0" />\r\n      <xs:element name="SecurityPolicyUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="ClientCertificate" type="xs:base64Binary" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SessionSecurityDiagnosticsDataType" type="tns:SessionSecurityDiagnosticsDataType" />\r\n\r\n  <xs:complexType name="ListOfSessionSecurityDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SessionSecurityDiagnosticsDataType" type="tns:SessionSecurityDiagnosticsDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSessionSecurityDiagnosticsDataType" type="tns:ListOfSessionSecurityDiagnosticsDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="ServiceCounterDataType">\r\n    <xs:sequence>\r\n      <xs:element name="TotalCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="ErrorCount" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ServiceCounterDataType" type="tns:ServiceCounterDataType" />\r\n\r\n  <xs:complexType name="StatusResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusCode" type="ua:StatusCode" minOccurs="0" />\r\n      <xs:element name="DiagnosticInfo" type="ua:DiagnosticInfo" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="StatusResult" type="tns:StatusResult" />\r\n\r\n  <xs:complexType name="ListOfStatusResult">\r\n    <xs:sequence>\r\n      <xs:element name="StatusResult" type="tns:StatusResult" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfStatusResult" type="tns:ListOfStatusResult" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="SubscriptionDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="SubscriptionId" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="Priority" type="xs:unsignedByte" minOccurs="0" />\r\n      <xs:element name="PublishingInterval" type="xs:double" minOccurs="0" />\r\n      <xs:element name="MaxKeepAliveCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MaxLifetimeCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MaxNotificationsPerPublish" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="PublishingEnabled" type="xs:boolean" minOccurs="0" />\r\n      <xs:element name="ModifyCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="EnableCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DisableCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RepublishRequestCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RepublishMessageRequestCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="RepublishMessageCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TransferRequestCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TransferredToAltClientCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="TransferredToSameClientCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="PublishRequestCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DataChangeNotificationsCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="EventNotificationsCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="NotificationsCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="LatePublishRequestCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CurrentKeepAliveCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="CurrentLifetimeCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="UnacknowledgedMessageCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DiscardedMessageCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MonitoredItemCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="DisabledMonitoredItemCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="MonitoringQueueOverflowCount" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="NextSequenceNumber" type="xs:unsignedInt" minOccurs="0" />\r\n      <xs:element name="EventQueueOverFlowCount" type="xs:unsignedInt" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SubscriptionDiagnosticsDataType" type="tns:SubscriptionDiagnosticsDataType" />\r\n\r\n  <xs:complexType name="ListOfSubscriptionDiagnosticsDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SubscriptionDiagnosticsDataType" type="tns:SubscriptionDiagnosticsDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSubscriptionDiagnosticsDataType" type="tns:ListOfSubscriptionDiagnosticsDataType" nillable="true"></xs:element>\r\n\r\n  <xs:simpleType  name="ModelChangeStructureVerbMask">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="NodeAdded_1" />\r\n      <xs:enumeration value="NodeDeleted_2" />\r\n      <xs:enumeration value="ReferenceAdded_4" />\r\n      <xs:enumeration value="ReferenceDeleted_8" />\r\n      <xs:enumeration value="DataTypeChanged_16" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="ModelChangeStructureVerbMask" type="tns:ModelChangeStructureVerbMask" />\r\n\r\n  <xs:complexType name="ModelChangeStructureDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Affected" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="AffectedType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="Verb" type="xs:unsignedByte" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ModelChangeStructureDataType" type="tns:ModelChangeStructureDataType" />\r\n\r\n  <xs:complexType name="ListOfModelChangeStructureDataType">\r\n    <xs:sequence>\r\n      <xs:element name="ModelChangeStructureDataType" type="tns:ModelChangeStructureDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfModelChangeStructureDataType" type="tns:ListOfModelChangeStructureDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="SemanticChangeStructureDataType">\r\n    <xs:sequence>\r\n      <xs:element name="Affected" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="AffectedType" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="SemanticChangeStructureDataType" type="tns:SemanticChangeStructureDataType" />\r\n\r\n  <xs:complexType name="ListOfSemanticChangeStructureDataType">\r\n    <xs:sequence>\r\n      <xs:element name="SemanticChangeStructureDataType" type="tns:SemanticChangeStructureDataType" minOccurs="0" maxOccurs="unbounded" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ListOfSemanticChangeStructureDataType" type="tns:ListOfSemanticChangeStructureDataType" nillable="true"></xs:element>\r\n\r\n  <xs:complexType name="Range">\r\n    <xs:sequence>\r\n      <xs:element name="Low" type="xs:double" minOccurs="0" />\r\n      <xs:element name="High" type="xs:double" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Range" type="tns:Range" />\r\n\r\n  <xs:complexType name="EUInformation">\r\n    <xs:sequence>\r\n      <xs:element name="NamespaceUri" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="UnitId" type="xs:int" minOccurs="0" />\r\n      <xs:element name="DisplayName" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="Description" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="EUInformation" type="tns:EUInformation" />\r\n\r\n  <xs:simpleType  name="AxisScaleEnumeration">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="Linear_0" />\r\n      <xs:enumeration value="Log_1" />\r\n      <xs:enumeration value="Ln_2" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="AxisScaleEnumeration" type="tns:AxisScaleEnumeration" />\r\n\r\n  <xs:complexType name="ComplexNumberType">\r\n    <xs:sequence>\r\n      <xs:element name="Real" type="xs:float" minOccurs="0" />\r\n      <xs:element name="Imaginary" type="xs:float" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ComplexNumberType" type="tns:ComplexNumberType" />\r\n\r\n  <xs:complexType name="DoubleComplexNumberType">\r\n    <xs:sequence>\r\n      <xs:element name="Real" type="xs:double" minOccurs="0" />\r\n      <xs:element name="Imaginary" type="xs:double" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="DoubleComplexNumberType" type="tns:DoubleComplexNumberType" />\r\n\r\n  <xs:complexType name="AxisInformation">\r\n    <xs:sequence>\r\n      <xs:element name="EngineeringUnits" type="tns:EUInformation" minOccurs="0" nillable="true" />\r\n      <xs:element name="EURange" type="tns:Range" minOccurs="0" nillable="true" />\r\n      <xs:element name="Title" type="ua:LocalizedText" minOccurs="0" nillable="true" />\r\n      <xs:element name="AxisScaleType" type="tns:AxisScaleEnumeration" minOccurs="0" />\r\n      <xs:element name="AxisSteps" type="ua:ListOfDouble" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="AxisInformation" type="tns:AxisInformation" />\r\n\r\n  <xs:complexType name="XVType">\r\n    <xs:sequence>\r\n      <xs:element name="X" type="xs:double" minOccurs="0" />\r\n      <xs:element name="Value" type="xs:float" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="XVType" type="tns:XVType" />\r\n\r\n  <xs:complexType name="ProgramDiagnosticDataType">\r\n    <xs:sequence>\r\n      <xs:element name="CreateSessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="CreateClientName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="InvocationCreationTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="LastTransitionTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="LastMethodCall" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodSessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodInputArguments" type="tns:ListOfArgument" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodOutputArguments" type="tns:ListOfArgument" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodCallTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="LastMethodReturnStatus" type="tns:StatusResult" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ProgramDiagnosticDataType" type="tns:ProgramDiagnosticDataType" />\r\n\r\n  <xs:complexType name="ProgramDiagnostic2DataType">\r\n    <xs:sequence>\r\n      <xs:element name="CreateSessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="CreateClientName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="InvocationCreationTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="LastTransitionTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="LastMethodCall" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodSessionId" type="ua:NodeId" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodInputArguments" type="tns:ListOfArgument" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodOutputArguments" type="tns:ListOfArgument" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodInputValues" type="ua:ListOfVariant" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodOutputValues" type="ua:ListOfVariant" minOccurs="0" nillable="true" />\r\n      <xs:element name="LastMethodCallTime" type="xs:dateTime" minOccurs="0" />\r\n      <xs:element name="LastMethodReturnStatus" type="tns:StatusResult" minOccurs="0" nillable="true" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="ProgramDiagnostic2DataType" type="tns:ProgramDiagnostic2DataType" />\r\n\r\n  <xs:complexType name="Annotation">\r\n    <xs:sequence>\r\n      <xs:element name="Message" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="UserName" type="xs:string" minOccurs="0" nillable="true" />\r\n      <xs:element name="AnnotationTime" type="xs:dateTime" minOccurs="0" />\r\n    </xs:sequence>\r\n  </xs:complexType>\r\n  <xs:element name="Annotation" type="tns:Annotation" />\r\n\r\n  <xs:simpleType  name="ExceptionDeviationFormat">\r\n    <xs:restriction base="xs:string">\r\n      <xs:enumeration value="AbsoluteValue_0" />\r\n      <xs:enumeration value="PercentOfValue_1" />\r\n      <xs:enumeration value="PercentOfRange_2" />\r\n      <xs:enumeration value="PercentOfEURange_3" />\r\n      <xs:enumeration value="Unknown_4" />\r\n    </xs:restriction>\r\n  </xs:simpleType>\r\n  <xs:element name="ExceptionDeviationFormat" type="tns:ExceptionDeviationFormat" />\r\n\r\n</xs:schema>', ua.VariantType.ByteString),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8252, 0),
        BrowseName=QualifiedName("Opc.Ua", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(92, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(72, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8254, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15039, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14829, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17542, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17554, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16024, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18860, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18863, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18866, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18869, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19049, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19052, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19055, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19058, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(19061, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15730, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12677, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16027, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14811, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15591, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15594, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15585, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15588, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14805, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14808, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14832, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16030, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16033, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14320, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16037, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16040, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16047, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16050, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16053, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16056, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21180, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16062, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16065, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16068, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16071, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21183, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21186, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16077, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16086, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16089, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16092, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16095, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14835, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16098, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21192, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16104, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16107, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16110, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16113, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16116, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16119, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17473, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(21195, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15640, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16125, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16144, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16147, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23508, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16127, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18166, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18169, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8285, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8291, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14826, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12759, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12762, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8918, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8300, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12201, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8297, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8303, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8417, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8333, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8306, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8309, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8312, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8315, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8318, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8363, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8366, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8369, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8372, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12712, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12715, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8321, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8570, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8576, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8579, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8639, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8702, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8708, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8711, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8807, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8327, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8843, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11951, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11954, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8846, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8849, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8852, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8855, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8858, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8861, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8294, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8864, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8867, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8870, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8873, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8876, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12175, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12178, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12083, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12086, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8882, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15402, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8879, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(92, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(72, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NamespaceUri"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant('http://opcfoundation.org/UA/2008/02/Types.xsd', ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8254, 0),
        BrowseName=QualifiedName("NamespaceUri", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(8254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Deprecated"),
        DataType=ua.NodeId(ua.ObjectIds.Boolean),
        Value=ua.Variant(True, ua.VariantType.Boolean),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15039, 0),
        BrowseName=QualifiedName("Deprecated", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(46, 0),
        TypeDefinition=NumericNodeId(68, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15039, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(68, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(46, 0),
        SourceNodeId=NumericNodeId(15039, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("KeyValuePair"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='KeyValuePair']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14829, 0),
        BrowseName=QualifiedName("KeyValuePair", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14829, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AdditionalParametersType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AdditionalParametersType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17542, 0),
        BrowseName=QualifiedName("AdditionalParametersType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17542, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EphemeralKeyType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EphemeralKeyType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17554, 0),
        BrowseName=QualifiedName("EphemeralKeyType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17554, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EndpointType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16024, 0),
        BrowseName=QualifiedName("EndpointType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16024, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RationalNumber"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='RationalNumber']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18860, 0),
        BrowseName=QualifiedName("RationalNumber", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18860, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Vector"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='Vector']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18863, 0),
        BrowseName=QualifiedName("Vector", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18863, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DVector"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ThreeDVector']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18866, 0),
        BrowseName=QualifiedName("3DVector", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18866, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CartesianCoordinates"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='CartesianCoordinates']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18869, 0),
        BrowseName=QualifiedName("CartesianCoordinates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18869, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DCartesianCoordinates"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ThreeDCartesianCoordinates']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19049, 0),
        BrowseName=QualifiedName("3DCartesianCoordinates", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19049, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19049, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Orientation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='Orientation']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19052, 0),
        BrowseName=QualifiedName("Orientation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19052, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DOrientation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ThreeDOrientation']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19055, 0),
        BrowseName=QualifiedName("3DOrientation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19055, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Frame"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='Frame']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19058, 0),
        BrowseName=QualifiedName("Frame", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("3DFrame"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ThreeDFrame']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19061, 0),
        BrowseName=QualifiedName("3DFrame", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(19061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IdentityMappingRuleType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='IdentityMappingRuleType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15730, 0),
        BrowseName=QualifiedName("IdentityMappingRuleType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15730, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("CurrencyUnitType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='CurrencyUnitType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23522, 0),
        BrowseName=QualifiedName("CurrencyUnitType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23522, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TrustListDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='TrustListDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12677, 0),
        BrowseName=QualifiedName("TrustListDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12677, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeSchemaHeader"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataTypeSchemaHeader']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16027, 0),
        BrowseName=QualifiedName("DataTypeSchemaHeader", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16027, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataTypeDescription']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14811, 0),
        BrowseName=QualifiedName("DataTypeDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14811, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StructureDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='StructureDescription']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15591, 0),
        BrowseName=QualifiedName("StructureDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15591, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EnumDescription']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15594, 0),
        BrowseName=QualifiedName("EnumDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15594, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SimpleTypeDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SimpleTypeDescription']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15585, 0),
        BrowseName=QualifiedName("SimpleTypeDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15585, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UABinaryFileDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='UABinaryFileDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15588, 0),
        BrowseName=QualifiedName("UABinaryFileDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15588, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetMetaDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataSetMetaDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14805, 0),
        BrowseName=QualifiedName("DataSetMetaDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14805, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FieldMetaData"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='FieldMetaData']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14808, 0),
        BrowseName=QualifiedName("FieldMetaData", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14808, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConfigurationVersionDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ConfigurationVersionDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14832, 0),
        BrowseName=QualifiedName("ConfigurationVersionDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14832, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedDataSetDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PublishedDataSetDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16030, 0),
        BrowseName=QualifiedName("PublishedDataSetDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16030, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedDataSetSourceDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PublishedDataSetSourceDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16033, 0),
        BrowseName=QualifiedName("PublishedDataSetSourceDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16033, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedVariableDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PublishedVariableDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14320, 0),
        BrowseName=QualifiedName("PublishedVariableDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14320, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14320, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedDataItemsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PublishedDataItemsDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16037, 0),
        BrowseName=QualifiedName("PublishedDataItemsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16037, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PublishedEventsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PublishedEventsDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16040, 0),
        BrowseName=QualifiedName("PublishedEventsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16040, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataSetWriterDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16047, 0),
        BrowseName=QualifiedName("DataSetWriterDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16047, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataSetWriterTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16050, 0),
        BrowseName=QualifiedName("DataSetWriterTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetWriterMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataSetWriterMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16053, 0),
        BrowseName=QualifiedName("DataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16053, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PubSubGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PubSubGroupDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16056, 0),
        BrowseName=QualifiedName("PubSubGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16056, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='WriterGroupDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21180, 0),
        BrowseName=QualifiedName("WriterGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='WriterGroupTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16062, 0),
        BrowseName=QualifiedName("WriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("WriterGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='WriterGroupMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16065, 0),
        BrowseName=QualifiedName("WriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PubSubConnectionDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PubSubConnectionDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16068, 0),
        BrowseName=QualifiedName("PubSubConnectionDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ConnectionTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ConnectionTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16071, 0),
        BrowseName=QualifiedName("ConnectionTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkAddressDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='NetworkAddressDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21183, 0),
        BrowseName=QualifiedName("NetworkAddressDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkAddressUrlDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='NetworkAddressUrlDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21186, 0),
        BrowseName=QualifiedName("NetworkAddressUrlDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReaderGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ReaderGroupDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21189, 0),
        BrowseName=QualifiedName("ReaderGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReaderGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ReaderGroupTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16077, 0),
        BrowseName=QualifiedName("ReaderGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ReaderGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ReaderGroupMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16080, 0),
        BrowseName=QualifiedName("ReaderGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataSetReaderDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16083, 0),
        BrowseName=QualifiedName("DataSetReaderDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataSetReaderTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16086, 0),
        BrowseName=QualifiedName("DataSetReaderTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataSetReaderMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataSetReaderMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16089, 0),
        BrowseName=QualifiedName("DataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscribedDataSetDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SubscribedDataSetDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16092, 0),
        BrowseName=QualifiedName("SubscribedDataSetDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TargetVariablesDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='TargetVariablesDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16095, 0),
        BrowseName=QualifiedName("TargetVariablesDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FieldTargetDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='FieldTargetDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14835, 0),
        BrowseName=QualifiedName("FieldTargetDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14835, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscribedDataSetMirrorDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SubscribedDataSetMirrorDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16098, 0),
        BrowseName=QualifiedName("SubscribedDataSetMirrorDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("PubSubConfigurationDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='PubSubConfigurationDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21192, 0),
        BrowseName=QualifiedName("PubSubConfigurationDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UadpWriterGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='UadpWriterGroupMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16104, 0),
        BrowseName=QualifiedName("UadpWriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UadpDataSetWriterMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='UadpDataSetWriterMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16107, 0),
        BrowseName=QualifiedName("UadpDataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UadpDataSetReaderMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='UadpDataSetReaderMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16110, 0),
        BrowseName=QualifiedName("UadpDataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16110, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("JsonWriterGroupMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='JsonWriterGroupMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16113, 0),
        BrowseName=QualifiedName("JsonWriterGroupMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16113, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("JsonDataSetWriterMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='JsonDataSetWriterMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16116, 0),
        BrowseName=QualifiedName("JsonDataSetWriterMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16116, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("JsonDataSetReaderMessageDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='JsonDataSetReaderMessageDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16119, 0),
        BrowseName=QualifiedName("JsonDataSetReaderMessageDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16119, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DatagramConnectionTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DatagramConnectionTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17473, 0),
        BrowseName=QualifiedName("DatagramConnectionTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17473, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(17473, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DatagramWriterGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DatagramWriterGroupTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21195, 0),
        BrowseName=QualifiedName("DatagramWriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(21195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerConnectionTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='BrokerConnectionTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15640, 0),
        BrowseName=QualifiedName("BrokerConnectionTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15640, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerWriterGroupTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='BrokerWriterGroupTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16125, 0),
        BrowseName=QualifiedName("BrokerWriterGroupTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16125, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerDataSetWriterTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='BrokerDataSetWriterTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16144, 0),
        BrowseName=QualifiedName("BrokerDataSetWriterTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BrokerDataSetReaderTransportDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='BrokerDataSetReaderTransportDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16147, 0),
        BrowseName=QualifiedName("BrokerDataSetReaderTransportDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AliasNameDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AliasNameDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23508, 0),
        BrowseName=QualifiedName("AliasNameDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(23508, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RolePermissionType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='RolePermissionType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16127, 0),
        BrowseName=QualifiedName("RolePermissionType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(16127, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DataTypeDefinition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DataTypeDefinition']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18166, 0),
        BrowseName=QualifiedName("DataTypeDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StructureField"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='StructureField']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18169, 0),
        BrowseName=QualifiedName("StructureField", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StructureDefinition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='StructureDefinition']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18172, 0),
        BrowseName=QualifiedName("StructureDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumDefinition"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EnumDefinition']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(18175, 0),
        BrowseName=QualifiedName("EnumDefinition", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(18175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(18175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Argument"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='Argument']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8285, 0),
        BrowseName=QualifiedName("Argument", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumValueType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EnumValueType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8291, 0),
        BrowseName=QualifiedName("EnumValueType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8291, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8291, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EnumField"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EnumField']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(14826, 0),
        BrowseName=QualifiedName("EnumField", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(14826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(14826, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("OptionSet"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='OptionSet']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12759, 0),
        BrowseName=QualifiedName("OptionSet", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12759, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Union"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='Union']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12762, 0),
        BrowseName=QualifiedName("Union", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12762, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("TimeZoneDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='TimeZoneDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8918, 0),
        BrowseName=QualifiedName("TimeZoneDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8918, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ApplicationDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ApplicationDescription']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8300, 0),
        BrowseName=QualifiedName("ApplicationDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8300, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerOnNetwork"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ServerOnNetwork']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12201, 0),
        BrowseName=QualifiedName("ServerOnNetwork", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserTokenPolicy"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='UserTokenPolicy']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8297, 0),
        BrowseName=QualifiedName("UserTokenPolicy", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8297, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointDescription"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EndpointDescription']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8303, 0),
        BrowseName=QualifiedName("EndpointDescription", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8303, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RegisteredServer"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='RegisteredServer']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8417, 0),
        BrowseName=QualifiedName("RegisteredServer", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8417, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DiscoveryConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DiscoveryConfiguration']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12894, 0),
        BrowseName=QualifiedName("DiscoveryConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12894, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MdnsDiscoveryConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='MdnsDiscoveryConfiguration']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12897, 0),
        BrowseName=QualifiedName("MdnsDiscoveryConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12897, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SignedSoftwareCertificate"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SignedSoftwareCertificate']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8333, 0),
        BrowseName=QualifiedName("SignedSoftwareCertificate", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='UserIdentityToken']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8306, 0),
        BrowseName=QualifiedName("UserIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8306, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AnonymousIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AnonymousIdentityToken']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8309, 0),
        BrowseName=QualifiedName("AnonymousIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8309, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("UserNameIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='UserNameIdentityToken']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8312, 0),
        BrowseName=QualifiedName("UserNameIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("X509IdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='X509IdentityToken']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8315, 0),
        BrowseName=QualifiedName("X509IdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("IssuedIdentityToken"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='IssuedIdentityToken']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8318, 0),
        BrowseName=QualifiedName("IssuedIdentityToken", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8318, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddNodesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AddNodesItem']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8363, 0),
        BrowseName=QualifiedName("AddNodesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AddReferencesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AddReferencesItem']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8366, 0),
        BrowseName=QualifiedName("AddReferencesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteNodesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DeleteNodesItem']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8369, 0),
        BrowseName=QualifiedName("DeleteNodesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DeleteReferencesItem"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DeleteReferencesItem']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8372, 0),
        BrowseName=QualifiedName("DeleteReferencesItem", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RelativePathElement"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='RelativePathElement']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12712, 0),
        BrowseName=QualifiedName("RelativePathElement", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12712, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RelativePath"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='RelativePath']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12715, 0),
        BrowseName=QualifiedName("RelativePath", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12715, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EndpointConfiguration']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8321, 0),
        BrowseName=QualifiedName("EndpointConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ContentFilterElement"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ContentFilterElement']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8564, 0),
        BrowseName=QualifiedName("ContentFilterElement", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8564, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ContentFilter"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ContentFilter']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8567, 0),
        BrowseName=QualifiedName("ContentFilter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8567, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("FilterOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='FilterOperand']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8570, 0),
        BrowseName=QualifiedName("FilterOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8570, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ElementOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ElementOperand']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8573, 0),
        BrowseName=QualifiedName("ElementOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8573, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("LiteralOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='LiteralOperand']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8576, 0),
        BrowseName=QualifiedName("LiteralOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8576, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AttributeOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AttributeOperand']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8579, 0),
        BrowseName=QualifiedName("AttributeOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8579, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SimpleAttributeOperand"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SimpleAttributeOperand']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8582, 0),
        BrowseName=QualifiedName("SimpleAttributeOperand", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8582, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryEvent"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='HistoryEvent']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8639, 0),
        BrowseName=QualifiedName("HistoryEvent", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8639, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("MonitoringFilter"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='MonitoringFilter']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8702, 0),
        BrowseName=QualifiedName("MonitoringFilter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8702, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EventFilter"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EventFilter']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8708, 0),
        BrowseName=QualifiedName("EventFilter", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8708, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AggregateConfiguration"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AggregateConfiguration']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8711, 0),
        BrowseName=QualifiedName("AggregateConfiguration", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8711, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("HistoryEventFieldList"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='HistoryEventFieldList']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8807, 0),
        BrowseName=QualifiedName("HistoryEventFieldList", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8807, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("BuildInfo"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='BuildInfo']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8327, 0),
        BrowseName=QualifiedName("BuildInfo", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("RedundantServerDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='RedundantServerDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8843, 0),
        BrowseName=QualifiedName("RedundantServerDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8843, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EndpointUrlListDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EndpointUrlListDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11951, 0),
        BrowseName=QualifiedName("EndpointUrlListDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11951, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("NetworkGroupDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='NetworkGroupDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(11954, 0),
        BrowseName=QualifiedName("NetworkGroupDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(11954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(11954, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SamplingIntervalDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SamplingIntervalDiagnosticsDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8846, 0),
        BrowseName=QualifiedName("SamplingIntervalDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8846, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerDiagnosticsSummaryDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ServerDiagnosticsSummaryDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8849, 0),
        BrowseName=QualifiedName("ServerDiagnosticsSummaryDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8849, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServerStatusDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ServerStatusDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8852, 0),
        BrowseName=QualifiedName("ServerStatusDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8852, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SessionDiagnosticsDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8855, 0),
        BrowseName=QualifiedName("SessionDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8855, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SessionSecurityDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SessionSecurityDiagnosticsDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8858, 0),
        BrowseName=QualifiedName("SessionSecurityDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8858, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ServiceCounterDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ServiceCounterDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8861, 0),
        BrowseName=QualifiedName("ServiceCounterDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8861, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("StatusResult"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='StatusResult']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8294, 0),
        BrowseName=QualifiedName("StatusResult", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SubscriptionDiagnosticsDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SubscriptionDiagnosticsDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8864, 0),
        BrowseName=QualifiedName("SubscriptionDiagnosticsDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8864, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ModelChangeStructureDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ModelChangeStructureDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8867, 0),
        BrowseName=QualifiedName("ModelChangeStructureDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8867, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("SemanticChangeStructureDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='SemanticChangeStructureDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8870, 0),
        BrowseName=QualifiedName("SemanticChangeStructureDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8870, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Range"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='Range']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8873, 0),
        BrowseName=QualifiedName("Range", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8873, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("EUInformation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='EUInformation']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8876, 0),
        BrowseName=QualifiedName("EUInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8876, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ComplexNumberType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ComplexNumberType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12175, 0),
        BrowseName=QualifiedName("ComplexNumberType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("DoubleComplexNumberType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='DoubleComplexNumberType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12178, 0),
        BrowseName=QualifiedName("DoubleComplexNumberType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12178, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("AxisInformation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='AxisInformation']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12083, 0),
        BrowseName=QualifiedName("AxisInformation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("XVType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='XVType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(12086, 0),
        BrowseName=QualifiedName("XVType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(12086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(12086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProgramDiagnosticDataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ProgramDiagnosticDataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8882, 0),
        BrowseName=QualifiedName("ProgramDiagnosticDataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8882, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("ProgramDiagnostic2DataType"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='ProgramDiagnostic2DataType']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15402, 0),
        BrowseName=QualifiedName("ProgramDiagnostic2DataType", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(15402, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.VariableAttributes(
        DisplayName=LocalizedText("Annotation"),
        DataType=ua.NodeId(ua.ObjectIds.String),
        Value=ua.Variant("//xs:element[@name='Annotation']", ua.VariantType.String),
        ValueRank=-1,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(8879, 0),
        BrowseName=QualifiedName("Annotation", 0),
        NodeClass_=NodeClass.Variable,
        ParentNodeId=NumericNodeId(8252, 0),
        ReferenceTypeId=NumericNodeId(47, 0),
        TypeDefinition=NumericNodeId(69, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(8879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(69, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(47, 0),
        SourceNodeId=NumericNodeId(8879, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8252, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15041, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14533, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14533, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15041, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17547, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(16313, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(16313, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17547, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17557, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17548, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17548, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17557, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16150, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15528, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15528, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19064, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18806, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18806, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19064, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19065, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18807, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18807, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19066, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18808, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18808, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19067, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18809, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18809, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19068, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18810, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18810, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19069, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18811, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18811, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19070, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18812, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18812, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19071, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18813, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18813, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(19072, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(18814, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(19072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(18814, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(19072, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15042, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15634, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15634, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15042, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23528, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23498, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23528, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23498, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23528, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15044, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12554, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15044, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12554, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15044, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15045, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17861, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17861, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15045, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16151, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15534, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15534, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15057, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14525, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14525, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15057, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15058, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15487, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15487, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15058, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15059, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15488, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15488, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15059, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15700, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15005, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15005, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15700, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15714, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15006, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15006, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15714, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15050, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14523, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14523, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15050, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15051, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14524, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14524, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15051, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15049, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14593, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15049, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14593, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15049, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16152, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15578, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15578, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16153, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15580, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15060, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14273, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15060, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16154, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15581, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15581, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16154, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16155, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15582, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15582, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16155, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16156, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15597, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15597, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16156, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16157, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15598, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16158, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15605, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15605, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16159, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15609, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15609, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21198, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15480, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16161, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15611, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15611, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16280, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15616, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16281, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15617, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15617, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16282, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15618, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15618, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21199, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15502, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15502, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21200, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15510, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15510, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21201, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15520, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15520, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16284, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15621, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15621, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16285, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15622, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16286, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15623, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15623, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16287, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15628, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15628, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16288, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15629, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16308, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15630, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15630, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16308, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16310, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15631, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15631, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16310, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15061, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(14744, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(14744, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15061, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16311, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15635, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16311, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21202, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15530, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15530, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16323, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15645, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15645, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16391, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15652, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15652, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16391, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16392, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15653, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16392, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16393, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15657, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15657, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16393, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16394, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15664, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15664, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16394, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16404, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15665, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16404, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(17476, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17467, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(17476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17467, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(17476, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(21203, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15532, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(21203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15532, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(21203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15726, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15007, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15007, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15726, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16524, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15667, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15667, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16524, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16525, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15669, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16525, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15669, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16525, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(16526, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15670, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(16526, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15670, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(16526, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23511, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23468, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23468, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23511, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15062, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(96, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(96, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15062, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15063, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(97, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(97, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15063, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15065, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(101, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(101, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15065, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15066, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(99, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(99, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15066, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15067, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(100, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(100, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15067, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15068, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(258, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(258, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15068, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15069, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11879, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11879, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15069, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15070, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11880, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11880, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15070, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15071, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(261, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(261, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15071, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15073, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(264, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(264, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15073, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15074, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(267, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(267, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15074, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15075, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(270, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(270, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15075, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15076, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(273, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(273, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15076, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15077, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(276, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(276, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15077, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15078, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(279, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(279, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15078, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15079, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(282, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(282, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15079, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15080, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(285, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(285, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15080, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15081, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(296, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(296, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15081, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15082, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(7594, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(7594, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15082, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15083, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(102, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(102, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15083, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15084, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12755, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15084, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15085, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12756, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12756, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15085, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15086, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(8912, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(8912, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15086, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15087, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(308, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(308, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15087, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15088, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(389, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(389, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15088, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15089, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(392, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(392, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15089, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15090, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(395, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(395, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15090, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15091, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15901, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15901, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15091, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15092, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(20999, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(20999, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15092, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15093, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(420, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(420, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15093, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15094, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(423, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(423, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15094, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15095, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12189, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12189, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15095, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15096, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12190, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12190, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15096, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15097, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12191, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12191, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15097, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15098, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(304, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(304, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15098, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15099, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(312, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(312, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15099, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15100, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(426, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(426, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15100, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15101, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(429, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(429, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15101, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15102, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(432, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(432, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15102, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15103, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(435, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(435, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15103, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15104, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(438, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(438, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15104, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15105, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12890, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12890, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15105, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15106, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12891, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15106, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15107, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12193, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12193, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15107, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15130, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12194, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12194, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15130, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15131, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(441, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(441, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15131, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15132, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(444, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(444, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15132, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15133, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(447, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(447, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15133, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15134, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(450, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(450, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15134, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15135, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(453, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(453, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15135, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15136, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(344, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(344, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15136, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15137, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(456, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(456, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15137, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15138, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(459, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(459, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15138, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15139, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(462, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(462, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15139, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15140, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(316, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(316, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15140, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15141, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(319, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(319, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15141, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15142, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(322, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(322, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15142, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15143, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(325, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(325, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15143, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15144, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(938, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(938, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15144, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15145, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(465, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(465, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15145, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15146, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(468, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(468, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15146, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15147, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(471, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(471, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15147, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15148, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(474, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(474, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15148, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15149, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(477, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(477, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15149, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15150, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(480, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(480, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15150, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15151, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(349, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(349, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15151, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15152, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(352, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(352, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15152, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15153, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(355, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(355, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15153, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15157, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(358, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(358, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15157, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15158, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(361, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(361, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15158, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15159, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(364, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(364, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15159, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15160, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(367, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(367, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15160, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15161, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(370, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(370, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15161, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15162, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(373, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(373, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15162, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15163, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17606, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17606, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15163, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15164, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(17607, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(17607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15164, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15165, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(376, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(376, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15165, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15166, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(483, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(483, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15166, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15167, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(486, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(486, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15167, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15168, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(489, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(489, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15168, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15169, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(379, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(379, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15169, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15170, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(492, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(492, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15170, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15171, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(495, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(495, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15171, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15172, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(382, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(382, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15172, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15173, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(498, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(498, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15173, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15174, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(501, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(501, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15174, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15175, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(385, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(385, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15175, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15176, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(504, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(504, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15176, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15177, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(507, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(507, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15177, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15179, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(511, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(511, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15179, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15180, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(514, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(514, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15180, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15182, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(518, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(518, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15182, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15183, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(522, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(522, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15183, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15184, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(525, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(525, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15184, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15185, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(528, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(528, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15185, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15186, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(531, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(531, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15186, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15187, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(534, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(534, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15187, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15188, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(537, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(537, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15188, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15189, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(540, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(540, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15189, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15190, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(543, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(543, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15190, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15191, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(546, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(546, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15191, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15192, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(549, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(549, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15192, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15193, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(552, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(552, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15193, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15194, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(555, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15194, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(555, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15194, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15195, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(558, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(558, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15195, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15196, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(561, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(561, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15196, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15197, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(564, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(564, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15197, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15198, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(567, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(567, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15198, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15199, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(331, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(331, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15199, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15200, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(570, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(570, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15200, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15201, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(573, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(573, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15201, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15202, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(577, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(577, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15202, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15203, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(580, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(580, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15203, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15204, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(583, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(583, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15204, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15205, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(586, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(586, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15205, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15206, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(589, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(589, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15206, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15207, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(592, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(592, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15207, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15208, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(595, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(595, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15208, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15209, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(598, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(598, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15209, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15210, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(601, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15210, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(601, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15210, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15211, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(604, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(604, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15211, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15228, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(607, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15228, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(607, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15228, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15236, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(610, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15236, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(610, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15236, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15244, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(613, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(613, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15244, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15252, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(616, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(616, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15252, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15254, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(619, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(619, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15254, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15255, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(622, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(622, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15255, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15256, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(626, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(626, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15256, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15257, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(629, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(629, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15257, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15258, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(632, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(632, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15258, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15259, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(635, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(635, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15259, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15260, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(638, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(638, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15260, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15261, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(641, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(641, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15261, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15262, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(644, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(644, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15262, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15263, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(647, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(647, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15263, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15264, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(650, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(650, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15264, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15269, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(653, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(653, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15269, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(23512, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(23497, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(23512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(23497, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(23512, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15270, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(656, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(656, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15270, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15271, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11216, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11216, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15271, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15272, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11217, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11217, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15272, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15273, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(659, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(659, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15273, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15274, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(662, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(662, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15274, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15275, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(665, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(665, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15275, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15276, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(668, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(668, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15276, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15277, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(671, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(671, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15277, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15278, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(674, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(674, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15278, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15279, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(677, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(677, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15279, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15280, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(680, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(680, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15280, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15281, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11295, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11295, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15281, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15282, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(683, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(683, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15282, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15283, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(686, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(686, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15283, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15284, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(689, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(689, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15284, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15285, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(692, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(692, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15285, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15286, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(695, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(695, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15286, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15287, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(698, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(698, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15287, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15288, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(701, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(701, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15288, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15289, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(704, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(704, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15289, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15290, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(707, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15290, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(707, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15290, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15291, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(710, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15291, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(710, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15291, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15292, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(713, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15292, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(713, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15292, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15293, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(719, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15293, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(719, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15293, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15294, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(722, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(722, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15294, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15295, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(725, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(725, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15295, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15304, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(948, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15304, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(948, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15304, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15312, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(728, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(728, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15312, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15313, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(731, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15313, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(731, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15313, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15314, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(734, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(734, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15314, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15315, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(737, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(737, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15315, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15320, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(740, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15320, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(740, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15320, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15321, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(743, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(743, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15321, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15322, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(746, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(746, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15322, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15323, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(749, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(749, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15323, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15324, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(752, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(752, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15324, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15325, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(755, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(755, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15325, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15326, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(758, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(758, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15326, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15327, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(761, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(761, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15327, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15328, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(764, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15328, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(764, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15328, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15329, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(767, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(767, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15329, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15331, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(770, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(770, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15331, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15332, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(773, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(773, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15332, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15333, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(776, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(776, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15333, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15335, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(779, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(779, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15335, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15336, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(782, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(782, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15336, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15337, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(785, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(785, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15337, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15338, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(788, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(788, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15338, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15339, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(791, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(791, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15339, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15340, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(794, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(794, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15340, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15341, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(797, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(797, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15341, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15342, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(800, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15342, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(800, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15342, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15343, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(803, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15343, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(803, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15343, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15344, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(945, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15344, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(945, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15344, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15345, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(809, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15345, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(809, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15345, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15346, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(806, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15346, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(806, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15346, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15347, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(914, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15347, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(914, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15347, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15348, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(917, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(917, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15348, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15349, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(920, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15349, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(920, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15349, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15350, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(818, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15350, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(818, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15350, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15351, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(821, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(821, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15351, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15352, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(824, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15352, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(824, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15352, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15353, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(827, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(827, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15353, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15354, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(830, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(830, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15354, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15355, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(833, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(833, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15355, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15356, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(836, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(836, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15356, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15357, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(839, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(839, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15357, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15358, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(842, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(842, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15358, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15359, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(845, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(845, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15359, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15360, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(848, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(848, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15360, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15361, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(338, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15361, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(338, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15361, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15362, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(853, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15362, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(853, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15362, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15363, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11943, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11943, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15363, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15364, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(11944, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(11944, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15364, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15365, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(856, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(856, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15365, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15366, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(859, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(859, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15366, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15367, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(862, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(862, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15367, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15368, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(865, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(865, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15368, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15369, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(868, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(868, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15369, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15370, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(871, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(871, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15370, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15371, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(299, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(299, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15371, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15372, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(874, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(874, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15372, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15373, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(877, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(877, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15373, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15374, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(897, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(897, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15374, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15375, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(884, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(884, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15375, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15376, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(887, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(887, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15376, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15377, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12171, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12171, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15377, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15378, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12172, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12172, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15378, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15379, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12079, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12079, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15379, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15380, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(12080, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(12080, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15380, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15381, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(894, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(894, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15381, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15405, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(15396, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(15396, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15405, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)

    attrs = ua.ObjectAttributes(
        DisplayName=LocalizedText("Default JSON"),
        EventNotifier=0,
        )
    node = ua.AddNodesItem(
        RequestedNewNodeId=NumericNodeId(15382, 0),
        BrowseName=QualifiedName("Default JSON", 0),
        NodeClass_=NodeClass.Object,
        ParentNodeId=NumericNodeId(891, 0),
        ReferenceTypeId=NumericNodeId(38, 0),
        TypeDefinition=NumericNodeId(76, 0),
        NodeAttributes=attrs,
    )
    server.add_nodes([node])
    refs = []
    ref = ua.AddReferencesItem(
        IsForward=False,
        ReferenceTypeId=NumericNodeId(38, 0),
        SourceNodeId=NumericNodeId(15382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(891, 0),
        )
    refs.append(ref)
    ref = ua.AddReferencesItem(
        IsForward=True,
        ReferenceTypeId=NumericNodeId(40, 0),
        SourceNodeId=NumericNodeId(15382, 0),
        TargetNodeClass=NodeClass.DataType,
        TargetNodeId=NumericNodeId(76, 0),
        )
    refs.append(ref)
    server.add_references(refs)
